/************************************************************************/
/*									*/
/*  Manipulate iten properties in a document.				*/
/*									*/
/************************************************************************/

#   include	"config.h"

#   include	<stdlib.h>
#   include	<string.h>
#   include	<stdio.h>

#   include	<debugon.h>

#   include	<appUnit.h>

#   include	"docBuf.h"

/************************************************************************/
/*									*/
/*  Change paragraph properties and tell what has been changed.		*/
/*									*/
/************************************************************************/

int docUpdParaProperties(	unsigned int *			pPpChgPask,
				BufferItem *			paraBi,
				unsigned int			ppUpdMask,
				const ParagraphProperties *	ppNew )
    {
    unsigned int		ppChgMask= 0;

    ParagraphProperties *	pp= &(paraBi->biParaProperties);

    if  ( ppUpdMask & PPupdFIRST_INDENT )
	{
	if  ( pp->ppFirstIndentTwips != ppNew->ppFirstIndentTwips )
	    {
	    pp->ppFirstIndentTwips= ppNew->ppFirstIndentTwips;
	    ppChgMask |= PPupdFIRST_INDENT;
	    }
	}

    if  ( ppUpdMask & PPupdLEFT_INDENT )
	{
	if  ( pp->ppLeftIndentTwips != ppNew->ppLeftIndentTwips )
	    {
	    pp->ppLeftIndentTwips= ppNew->ppLeftIndentTwips;
	    ppChgMask |= PPupdLEFT_INDENT;
	    }
	}

    if  ( ppUpdMask & PPupdRIGHT_INDENT )
	{
	if  ( pp->ppRightIndentTwips != ppNew->ppRightIndentTwips )
	    {
	    pp->ppRightIndentTwips= ppNew->ppRightIndentTwips;
	    ppChgMask |= PPupdRIGHT_INDENT;
	    }
	}

    if  ( ppUpdMask & PPupdTAB_STOPS )
	{
	if  ( pp->ppTabCount > 0 || ppNew->ppTabCount > 0 )
	    {
	    if  ( pp->ppTabCount != ppNew->ppTabCount )
		{ ppChgMask |= PPupdTAB_STOPS;	}

	    if  ( ppNew->ppTabCount == 0 )
		{ pp->ppTabCount= 0;	}
	    else{
		TabStop *			ts;
		TabStop *			ots= pp->ppTabStops;
		TabStop *			nts= ppNew->ppTabStops;
		int				tab;

		ts= (TabStop *)malloc( ppNew->ppTabCount* sizeof(TabStop) );
		if  ( ! ts )
		    { LXDEB(ppNew->ppTabCount,ts); return -1;	}

		for ( tab= 0; tab < ppNew->ppTabCount; tab++ )
		    {
		    if  ( pp->ppTabCount == ppNew->ppTabCount )
			{
			if  ( nts[tab].tsTwips != ots[tab].tsTwips	||
			      nts[tab].tsKind != ots[tab].tsKind	||
			      nts[tab].tsLeader != ots[tab].tsLeader	)
			    {
			    ppChgMask |= PPupdTAB_STOPS;
			    ts[tab]= nts[tab];
			    }
			else{ ts[tab]= ots[tab];	}
			}
		    else{ ts[tab]= nts[tab];		}
		    }

		pp->ppTabCount= ppNew->ppTabCount;
		if  ( pp->ppTabStops )
		    { free( pp->ppTabStops );	}

		pp->ppTabStops= ts;
		}
	    }
	}

    if  ( ppUpdMask & PPupdALIGNMENT )
	{
	if  ( pp->ppAlignment != ppNew->ppAlignment )
	    {
	    pp->ppAlignment= ppNew->ppAlignment;
	    ppChgMask |= PPupdALIGNMENT;
	    }
	}

    if  ( ppUpdMask & PPupdSPACE_BEFORE )
	{
	if  ( pp->ppSpaceBeforeTwips != ppNew->ppSpaceBeforeTwips )
	    {
	    pp->ppSpaceBeforeTwips= ppNew->ppSpaceBeforeTwips;
	    ppChgMask |= PPupdSPACE_BEFORE;
	    }
	}

    if  ( ppUpdMask & PPupdSPACE_AFTER )
	{
	if  ( pp->ppSpaceAfterTwips != ppNew->ppSpaceAfterTwips )
	    {
	    pp->ppSpaceAfterTwips= ppNew->ppSpaceAfterTwips;
	    ppChgMask |= PPupdSPACE_AFTER;
	    }
	}

    if  ( ppUpdMask & PPupdNEWPAGE )
	{
	if  ( pp->ppStartsOnNewPage != ppNew->ppStartsOnNewPage )
	    {
	    pp->ppStartsOnNewPage= ppNew->ppStartsOnNewPage;
	    ppChgMask |= PPupdNEWPAGE;
	    }
	}

    if  ( ppUpdMask & PPupdLINE_SPACING )
	{
	if  ( pp->ppLineSpacingTwips != ppNew->ppLineSpacingTwips )
	    {
	    pp->ppLineSpacingTwips= ppNew->ppLineSpacingTwips;
	    ppChgMask |= PPupdLINE_SPACING;
	    }
	if  ( pp->ppLineSpacingIsMultiple != ppNew->ppLineSpacingIsMultiple )
	    {
	    pp->ppLineSpacingIsMultiple= ppNew->ppLineSpacingIsMultiple;
	    ppChgMask |= PPupdLINE_SPACING;
	    }
	}

    if  ( ppUpdMask & PPupdTOP_BORDER )
	{
	if  ( ! docEqualBorder( &(pp->ppTopBorder),
						&(ppNew->ppTopBorder ) ) )
	    {
	    docCopyBorderProperties( &(pp->ppTopBorder),
						&(ppNew->ppTopBorder) );
	    ppChgMask |= PPupdTOP_BORDER;
	    }
	}

    if  ( ppUpdMask & PPupdBOTTOM_BORDER )
	{
	if  ( ! docEqualBorder( &(pp->ppBottomBorder),
						&(ppNew->ppBottomBorder ) ) )
	    {
	    docCopyBorderProperties( &(pp->ppBottomBorder),
						&(ppNew->ppBottomBorder) );
	    ppChgMask |= PPupdBOTTOM_BORDER;
	    }
	}

    *pPpChgPask= ppChgMask; return 0;
    }

/************************************************************************/
/*									*/
/*  Change section properties and tell what has been changed.		*/
/*									*/
/************************************************************************/

int docUpdSectProperties(	unsigned int *			pSpChgMask,
				BufferItem *			sectBi,
				unsigned int			spUpdMask,
				const SectionProperties *	spNew )
    {
    SectionProperties *	sp= &(sectBi->biSectProperties);
    unsigned int	spChgMask= SPupdNONE;

    if  ( spUpdMask & SPupdPAGE_WIDTH )
	{
	if  ( sp->spPageWideTwips != spNew->spPageWideTwips )
	    {
	    sp->spPageWideTwips= spNew->spPageWideTwips;
	    spChgMask |= SPupdPAGE_WIDTH;
	    }
	}

    if  ( spUpdMask & SPupdPAGE_HEIGHT )
	{
	if  ( sp->spPageHighTwips != spNew->spPageHighTwips )
	    {
	    sp->spPageHighTwips= spNew->spPageHighTwips;
	    spChgMask |= SPupdPAGE_HEIGHT;
	    }
	}

    if  ( spUpdMask & SPupdLEFT_MARGIN )
	{
	if  ( sp->spLeftMarginTwips != spNew->spLeftMarginTwips )
	    {
	    sp->spLeftMarginTwips= spNew->spLeftMarginTwips;
	    spChgMask |= SPupdLEFT_MARGIN;
	    }
	}

    if  ( spUpdMask & SPupdRIGHT_MARGIN )
	{
	if  ( sp->spRightMarginTwips != spNew->spRightMarginTwips )
	    {
	    sp->spRightMarginTwips= spNew->spRightMarginTwips;
	    spChgMask |= SPupdRIGHT_MARGIN;
	    }
	}

    if  ( spUpdMask & SPupdTOP_MARGIN )
	{
	if  ( sp->spTopMarginTwips != spNew->spTopMarginTwips )
	    {
	    sp->spTopMarginTwips= spNew->spTopMarginTwips;
	    spChgMask |= SPupdTOP_MARGIN;
	    }
	}

    if  ( spUpdMask & SPupdBOTTOM_MARGIN )
	{
	if  ( sp->spBottomMarginTwips != spNew->spBottomMarginTwips )
	    {
	    sp->spBottomMarginTwips= spNew->spBottomMarginTwips;
	    spChgMask |= SPupdBOTTOM_MARGIN;
	    }
	}

    if  ( spUpdMask & SPupdHEADER_POSITION )
	{
	if  ( sp->spHeaderYTwips != spNew->spHeaderYTwips )
	    {
	    sp->spHeaderYTwips= spNew->spHeaderYTwips;
	    spChgMask |= SPupdHEADER_POSITION;
	    }
	}

    if  ( spUpdMask & SPupdFOOTER_POSITION )
	{
	if  ( sp->spFooterYTwips != spNew->spFooterYTwips )
	    {
	    sp->spFooterYTwips= spNew->spFooterYTwips;
	    spChgMask |= SPupdFOOTER_POSITION;
	    }
	}

    if  ( spUpdMask & SPupdNUMBER_STYLE )
	{
	if  ( sp->spPageNumberStyle != spNew->spPageNumberStyle )
	    {
	    sp->spPageNumberStyle= spNew->spPageNumberStyle;
	    spChgMask |= SPupdNUMBER_STYLE;
	    }
	}

    if  ( spUpdMask & SPupdBREAK_KIND )
	{
	if  ( sp->spBreakKind != spNew->spBreakKind )
	    {
	    sp->spBreakKind= spNew->spBreakKind;
	    spChgMask |= SPupdBREAK_KIND;
	    }
	}

    *pSpChgMask= spChgMask; return 0;
    }
