/************************************************************************/
/*									*/
/*  Postscript printing and formatting related data structures.		*/
/*									*/
/************************************************************************/

#   include	<psFont.h>
#   include	"docBuf.h"
#   include	"docDraw.h"
#   include	"appPs.h"

typedef struct ParticuleData
    {
    int			pdX0;
    int			pdWidth;
    int			pdVisibleWidth;
    int			pdTabNumber;
    int			pdTabKind;
    int			pdTabPosition;
    const AfmFontInfo *	pdAfi;

    int			pdVisiblePixels;
    int			pdWhiteUnits;
    int			pdCorrectBy;
    } ParticuleData;

#   define LINEDISTFAC	7

/************************************************************************/
/*									*/
/*  Pass through information for a layout job.				*/
/*									*/
/************************************************************************/

typedef int (*LAYOUT_SCREEN_LINE)( TextLine *			tl,
				const BufferItem *		bi,
				int				part,
				int				accepted,
				AppDrawingData *		add,
				const FormattingFrame *		ffPixels,
				TextParticule *			tp,
				ParticuleData *			pd,
				int				y );

typedef int (*START_PARAGRAPH)( FormattingFrame *		ff,
				int				bottom,
				BufferItem *			bi,
				AppDrawingData *		add,
				const DocumentProperties *	dp );

typedef int (*START_ROW)(	BufferItem *			rowBi,
				AppDrawingData *		add,
				const DocumentProperties *	dp );

typedef struct ScreenLayout
    {
    START_ROW			slStartRow;
    START_PARAGRAPH		slStartParagraph;
    LAYOUT_SCREEN_LINE		slLayoutLine;
    } ScreenLayout;

typedef struct LayoutJob
    {
    LayoutPosition		ljPosition;
    LayoutPosition		ljBottom;

    DocumentRectangle *		ljChangedRectanglePixels;

    int				ljInHeaderFooter;
    ScreenLayout		ljLayoutScreen;
    ScreenLayout		ljPlaceScreen;
    } LayoutJob;

/************************************************************************/
/*									*/
/*  Routine declarations.						*/
/*									*/
/************************************************************************/

extern int docPsLineBox(	TextLine *			tl,
				int *				pPageBreak,
				const BufferItem *		bi,
				int				part,
				int				tabInterval,
				const DocumentFontList *	dfl,
				AppPhysicalFontList *		apfl,
				const TextParticule *		tp,
				ParticuleData *			pd,
				const FormattingFrame *		ff );

extern AfmFontInfo * docPsPrintGetAfi(	const AppPhysicalFontList *	apfl,
					int				physf );

extern int docPsClaimParticuleData(	const BufferItem *	bi,
					ParticuleData **	pData );

extern int docPsParagraphLineExtents(	const AppPhysicalFontList *	apfl,
					BufferItem *			bi );

extern int docPsPrintGetItemFonts(	const BufferItem *		bi,
					DocumentFontList *		dfl,
					const AppPhysicalFontList *	apfl,
					PostScriptFont **		pFntLst,
					int *				pCount );

extern int docPsPrintDocument(	FILE *			f,
			const char *			title,
			const char *			applicationName,
			AppDrawingData *		add,
			BufferDocument *		bd,
			int				nup,
			int				horizontal,
			int				firstPage,
			int				lastPage,
			const DocumentGeometry *	dgPrinter,
			LAYOUT_HEADER			layoutHeader,
			LAYOUT_FOOTER			layoutFooter );

extern int docPsLayoutItemAndParents(	BufferItem *			bi,
					AppDrawingData *		add,
					const DocumentProperties *	dp,
					LayoutJob *			lj,
					int			andParents );

extern int docPsPlaceItem(	BufferItem *			bi,
				AppDrawingData *		add,
				const DocumentProperties *	dp,
				LayoutJob *			lj );

extern int docPsAdjustParaLayout( BufferItem *			bi,
				int				line,
				int				stroffShift,
				int				upto,
				AppDrawingData *		add,
				const DocumentProperties *	dp,
				LayoutJob *			lj );

extern void docPsStartPageLayout( int				pageNumber,
				AppDrawingData *		add,
				const SectionProperties *	sp,
				LayoutPosition *		lpTop );

extern int docPsListObjectFonts(	const InsertedObject *	io,
					const char *		afmDirectory,
					const char *		prefix,
					PostScriptFont **	pFontList,
					int *			pCount );

