/************************************************************************/
/*									*/
/*  Editor functionality, indepentent of the application.		*/
/*									*/
/************************************************************************/

#   include	"config.h"

#   include	<stddef.h>
#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<ctype.h>

#   include	"tedApp.h"

#   include	<X11/Xatom.h>

#   include	<debugon.h>

/************************************************************************/
/*									*/
/*  Change the attributes of the selection.				*/
/*									*/
/*  1)  Adjust the start line for recalculating the layout of the first	*/
/*	paragraph in the selection.					*/
/*  2)  If necessary, split the first particule to change the text	*/
/*	attributes of the second half.					*/
/*  3)  For all paragraphs in the selection upto the last one..		*/
/*  4)  Change the text attributes.					*/
/*									*/
/************************************************************************/

static int tedChangeParaProperties( unsigned int *		pPpChgMask,
				    unsigned int *		pTaChgMask,
				    BufferItem *		paraBi,
				    int				partFrom,
				    int				partUpto,
				    AppDrawingData *		add,
				    DocumentFontList *		dfl,
				    unsigned int		taUpdMask,
				    TextAttribute		taNew,
				    unsigned int		ppUpdMask,
				    const ParagraphProperties *	ppNew )
    {
    unsigned int	ppChgMask= 0;
    unsigned int	taChgMask= 0;

    if  ( taUpdMask )
	{
	TextParticule *		tp= paraBi->biParaParticules+ partFrom;
	int			part;

	tp= paraBi->biParaParticules+ partFrom;
	for ( part= partFrom; part < partUpto; tp++, part++ )
	    {
	    taChgMask |= docAttributeDifference(
				    tp->tpTextAttribute, taNew, taUpdMask );

	    if  ( tedChangeParticuleAttribute( add, tp, dfl,
						    taUpdMask, taNew ) )
		{ XDEB(taUpdMask); return -1;	}
	    }
	}

    if  ( ppUpdMask )
	{
	if  ( docUpdParaProperties( &ppChgMask, paraBi, ppUpdMask, ppNew ) )
	    { XDEB(ppUpdMask); return -1;	}

	if  ( ppChgMask )
	    { tedParaScreenGeometry( paraBi, add );	}
	}

    *pTaChgMask= taChgMask;
    *pPpChgMask= ppChgMask;
    return 0;
    }

int tedChangeSelectionProperties( EditDocument *		ed,
				unsigned int			taUpdMask,
				TextAttribute			taNew,
				unsigned int			ppUpdMask,
				const ParagraphProperties *	ppNew,
				unsigned int			spUpdMask,
				const SectionProperties *	spNew )
    {
    AppDrawingData *		add= &(ed->edDrawingData);
    int				oldBackY1= add->addBackRect.drY1;

    TedDocument *		td= (TedDocument *)ed->edPrivateData;
    const BufferSelection *	bs= &(td->tdSelection);
    BufferDocument *		bd= td->tdDocument;
    DocumentProperties *	dp= &(bd->bdProperties);
    DocumentFontList *		dfl= &(dp->dpFontList);

    int				part;

    TextParticule *		tp;

    BufferItem *		paraBi= bs->bsBegin.bpBi;
    BufferItem *		sectBi= (BufferItem *)0;

    DocumentRectangle		drChanged;
    
    int				firstParticuleSplit= 0;

    unsigned int		selTaChgMask= 0;
    unsigned int		selPpChgMask= 0;
    unsigned int		selSpChgMask= 0;
    unsigned int		paraTaChgMask;
    unsigned int		paraPpChgMask;
    unsigned int		sectSpChgMask;

    if  ( ! paraBi )
	{ XDEB(paraBi); return -1; }

    if  ( tedHasIBarSelection( td ) )
	{ taUpdMask= 0;	}

    part= bs->bsBegin.bpParticule;

    drChanged= td->tdSelectedRectangle;
    drChanged.drX0= add->addBackRect.drX0;
    drChanged.drX1= add->addBackRect.drX1;

    /*  1  */
    tp= paraBi->biParaParticules+ part;

    /*  2  */
    if  ( (unsigned)bs->bsBegin.bpStroff > tp->tpStroff			&&
	  docAttributeDifference( tp->tpTextAttribute, taNew,
							    taUpdMask )	)
	{
	int		stroff= bs->bsBegin.bpStroff;

	if  ( (unsigned)stroff < tp->tpStroff+ tp->tpStrlen )
	    {
	    TextParticule *	newTp;

	    newTp= docCopyParticule( paraBi, part+ 1, stroff,
			tp->tpStroff+ tp->tpStrlen- stroff, DOCkindTEXT, tp );
	    if  ( ! newTp )
		{ XDEB(newTp); return -1;	}

	    tp= paraBi->biParaParticules+ part;
	    tp->tpStrlen= stroff- tp->tpStroff;
	    tp= newTp; part++;

	    if  ( bs->bsEnd.bpBi == paraBi )
		{ firstParticuleSplit= 1;	}
	    }
	else{ tp++; part++; }
	}

    sectSpChgMask= 0;

    sectBi= paraBi->biParent->biParent->biParent;
    if  ( spUpdMask )
	{
	if  ( docUpdSectProperties( &sectSpChgMask, sectBi, spUpdMask, spNew ) )
	    { XDEB(spUpdMask); return -1;	}

	selSpChgMask |= sectSpChgMask;
	}

    /*  3  */
    while( paraBi != bs->bsEnd.bpBi )
	{
	int	col;
	int	partUpto= paraBi->biParaParticuleCount;

	col= paraBi->biParent->biNumberInParent;

	if  ( bs->bsCol0 < 0					||
	      bs->bsCol1 < 0					||
	      ( col >= bs->bsCol0 && col <= bs->bsCol1 )	)
	    {
	    /*  4  */
	    if  ( tedChangeParaProperties( &paraPpChgMask, &paraTaChgMask,
				    paraBi, part, partUpto, add, dfl,
				    taUpdMask, taNew, ppUpdMask, ppNew ) )
		{ XXDEB(taUpdMask,ppUpdMask);	}

	    if  ( sectBi != paraBi->biParent->biParent->biParent )
		{
		if  ( sectSpChgMask )
		    { tedApplyItemFormat( sectBi, dp, add, &drChanged ); }

		sectBi= paraBi->biParent->biParent->biParent;
		if  ( spUpdMask )
		    {
		    if  ( docUpdSectProperties( &sectSpChgMask,
						    sectBi, spUpdMask, spNew ) )
			{ XDEB(spUpdMask); return -1;	}

		    selSpChgMask |= sectSpChgMask;
		    }
		}

	    if  ( paraPpChgMask || paraTaChgMask )
		{
		tedApplyItemFormat( paraBi, dp, add, &drChanged );

		if  ( paraPpChgMask && paraBi == bs->bsBegin.bpBi )
		    { tedDocAdaptHorizontalRuler( ed, paraBi, 1 );	}

		selTaChgMask |= paraTaChgMask;
		selPpChgMask |= paraPpChgMask;
		}
	    }

	paraBi= docNextParagraph( paraBi );
	if  ( ! paraBi )
	    { XDEB(paraBi); return -1;	}

	part= 0; firstParticuleSplit= 0;
	}

    /*  4  */
    if  ( tedChangeParaProperties( &paraPpChgMask, &paraTaChgMask, paraBi,
			part+ firstParticuleSplit,
			bs->bsEnd.bpParticule+ firstParticuleSplit, add, dfl,
			taUpdMask, taNew, ppUpdMask, ppNew ) )
	{ XXDEB(taUpdMask,ppUpdMask);	}

    part= bs->bsEnd.bpParticule+ firstParticuleSplit;
    tp= paraBi->biParaParticules+ part;

    if  ( sectBi != paraBi->biParent->biParent->biParent )
	{
	if  ( sectSpChgMask )
	    { tedApplyItemFormat( sectBi, dp, add, &drChanged ); }

	sectBi= paraBi->biParent->biParent->biParent;
	if  ( spUpdMask )
	    {
	    if  ( docUpdSectProperties( &sectSpChgMask,
						    sectBi, spUpdMask, spNew ) )
		{ XDEB(spUpdMask); return -1;	}

	    selSpChgMask |= sectSpChgMask;
	    }
	}

    if  ( sectSpChgMask )
	{ tedApplyItemFormat( sectBi, dp, add, &drChanged ); }

    if  ( part < paraBi->biParaParticuleCount				&&
	  docAttributeDifference( tp->tpTextAttribute, taNew,
							    taUpdMask )	)
	{
	int	strend= tp->tpStroff+ tp->tpStrlen;

	if  ( bs->bsEnd.bpStroff < strend )
	    {
	    TextParticule *	newTp;

	    newTp= docCopyParticule( paraBi, part+ 1, bs->bsEnd.bpStroff,
		    strend- bs->bsEnd.bpStroff,
		    DOCkindTEXT, tp );

	    if  ( ! newTp )
		{ XDEB(newTp); return -1;	}

	    tp= paraBi->biParaParticules+ part;
	    tp->tpStrlen= bs->bsEnd.bpStroff- tp->tpStroff;
	    }

	paraTaChgMask |= docAttributeDifference( tp->tpTextAttribute,
							taNew, taUpdMask );

	/*  2  */
	if  ( tedChangeParticuleAttribute( add, tp, dfl, taUpdMask, taNew ) )
	    { LDEB(1);	}
	}

    if  ( paraPpChgMask || paraTaChgMask )
	{
	tedApplyItemFormat( paraBi, dp, add, &drChanged );

	if  ( paraPpChgMask && paraBi == bs->bsBegin.bpBi )
	    { tedDocAdaptHorizontalRuler( ed, paraBi, 1 );	}

	selPpChgMask |= paraPpChgMask;
	selTaChgMask |= paraTaChgMask;
	}

    if  ( selPpChgMask || selTaChgMask || selSpChgMask )
	{
	int				scrolledX= 0;
	int				scrolledY= 0;

	docFindLineAndParticule( td->tdSelection.bsBegin.bpBi,
		    td->tdSelection.bsBegin.bpStroff,
		    &(td->tdSelection.bsBegin), 1 );
	docFindLineAndParticule( td->tdSelection.bsEnd.bpBi,
		    td->tdSelection.bsEnd.bpStroff,
		    &(td->tdSelection.bsEnd), 0 );

	tedCalculateSelectedLines( td );
	tedSelectionCoordinates( &(td->tdSelection), add );

	tedScrollToSelection( ed, &scrolledX, &scrolledY );

	tedExposeRectangle( ed, &drChanged, scrolledX, scrolledY );

	tedSelectionRectangle( &(td->tdSelectedRectangle),
					     add, &(td->tdSelection) );

	if  ( add->addBackRect.drY1 != oldBackY1 )
	    {
	    appDocSetScrollbarValues( ed );
	    appSetShellConstraints( ed );
	    }

	tedAdaptToolsToSelection( ed );

	appDocumentChanged( ed, 1 );
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Set an I Bar selection and cause the I Bar to be drawn.		*/
/*									*/
/*  1)  As an extra service, check consistency of fields		*/
/*	administration.							*/
/*									*/
/************************************************************************/

static int tedEditSetIBarSelection(	EditDocument *	ed,
					BufferItem *	bi,
					int *		pScrolledX,
					int *		pScrolledY,
					int		stroff )
    {
    TedDocument *	td= (TedDocument *)ed->edPrivateData;

    const int		andExposeRuler= 1;
    BufferPosition *	bp= &(td->tdSelection.bsBegin);

    td->tdVisibleSelectionCopied= 0;

    if  ( td->tdObjectSelected )
	{
	XUnmapWindow( XtDisplay( ed->edDocumentWidget ), td->tdObjectWindow );
	td->tdObjectSelected= 0;
	}

    if  ( docFindLineAndParticule( bi, stroff, bp, 1 ) )
	{ LDEB(stroff); return -1;	}

    /*  1  */
    if  ( bp->bpParticule > 0 )
	{
	TextParticule *	tp= bi->biParaParticules+ bp->bpParticule;

	if  ( tp[-1].tpKind == DOCkindFIELDSTART	&&
	      ! tp->tpTextAttribute.taInField		)
	    { LDEB(1); docListItem(0,bi); }
	}

    tedSetSelectedPosition( ed, td, bp, pScrolledX, pScrolledY );

    tedAdaptToolsToPosition( ed, andExposeRuler );

    return 0;
    }

int tedEditFinishIBarSelection(	EditDocument *			ed,
				BufferItem *			bi,
				const DocumentRectangle *	drChanged,
				int				oldBackY1,
				int				stroff )
    {
    AppDrawingData *		add= &(ed->edDrawingData);

    int				scrolledX= 0;
    int				scrolledY= 0;

    if  ( bi )
	{ tedEditSetIBarSelection( ed, bi, &scrolledX, &scrolledY, stroff ); }
    else{ XDEB(bi);	}

    if  ( add->addBackRect.drY1 != oldBackY1 )
	{
	appDocSetScrollbarValues( ed );
	appSetShellConstraints( ed );
	}

    tedExposeRectangle( ed, drChanged, scrolledX, scrolledY );

    return 0;
    }

/************************************************************************/
/*									*/
/*  Delete some buffer items but first close the objects in them.	*/
/*									*/
/************************************************************************/

void tedDeleteItems(	BufferDocument *	bd,
			BufferItem *		bi,
			int			first,
			int			count,
			Display *		display )
    {
    int		i;
    int		f= first;

    for ( i= 0; i < count; f++, i++ )
	{
	docCloseItemObjects( bd, bi->biGroupChildren[f],
					(void *)display, tedCloseObject );
	}

    docDeleteItems( bd, bi, first, count );
    }

/************************************************************************/
/*									*/
/*  Replace the selection in the document with new text.		*/
/*									*/
/*  b)  Replace the tail of the beginning paragraph with the new text.	*/
/*  c)  Merge the two paragraphs.					*/
/*  d)  Erase all paragraphs after the beginning of the selection	*/
/*	to, including the end.						*/
/*  e)  Update the paragraph buffer and the particule administration.	*/
/*									*/
/*  B)  The first particule of the line was split, also reformat the	*/
/*	previous line.							*/
/*	If paragraphs were merged, redraw the whole neighbourhood.	*/
/*  C)  - Recalculate the geometry from the start of the selection to	*/
/*	  the end of the paragraph.					*/
/*	- Redraw the affected part of the paragraph.			*/
/*  6)  Update the notion of current particule and current line.	*/
/*  7)  Redraw the I-Bar.						*/
/*									*/
/************************************************************************/

/*  B  */
static void tedAdjustRedrawBegin( TedDocument *			td,
				const DocumentRectangle *	drBack,
				DocumentRectangle *		drChanged,
				int *				pLine,
				int				part )
    {
    TextLine *	tl= td->tdSelection.bsBegin.bpBi->biParaLines+ *pLine;

    if  ( *pLine == 0 )
	{
	if  ( td->tdSelection.bsBegin.bpBi != td->tdSelection.bsEnd.bpBi )
	    {
	    drChanged->drX0= drBack->drX0;
	    drChanged->drX1= drBack->drX1;

	    if  ( drChanged->drY0 > tl->tlTopPosition.lpYPixels )
		{ drChanged->drY0=  tl->tlTopPosition.lpYPixels;	}
	    }

	return;
	}

    if  ( part == tl->tlFirstParticule					||
	  td->tdSelection.bsBegin.bpBi != td->tdSelection.bsEnd.bpBi	)
	{
	(*pLine)--; tl--;

	drChanged->drX0= drBack->drX0;
	drChanged->drX1= drBack->drX1;

	if  ( drChanged->drY0 > tl->tlTopPosition.lpYPixels )
	    { drChanged->drY0=  tl->tlTopPosition.lpYPixels;	}
	}

    return;
    }

void tedReplaceSelection(	EditDocument *		ed,
				const unsigned char *	addedText,
				int			addedLength )
    {
    TedDocument *		td= (TedDocument *)ed->edPrivateData;
    BufferDocument *		bd= td->tdDocument;
    DocumentProperties *	dp= &(bd->bdProperties);

    Widget			w= ed->edDocumentWidget;
    Window			win= XtWindow( w );
    GC				gc= ed->edGc;
	
    int				stroffShift= 0;
    int				partShift= 0;

    BufferItem *		bi= td->tdSelection.bsBegin.bpBi;
    int				paragraphsMerged;
	
    int				line;
    int				part;

    AppDrawingData *		add= &(ed->edDrawingData);

    int				oldBackY1= add->addBackRect.drY1;

    const DocumentRectangle *	drBack= &(add->addBackRect);
    DocumentRectangle		drChanged;
    int				stroff;

    TextLine *			tl;
    int				upto;

    stroff= td->tdSelection.bsBegin.bpStroff;
    part= td->tdSelection.bsBegin.bpParticule;
    line= td->tdSelection.bsBegin.bpLine;

    tl= bi->biParaLines+ line;
    upto= tl->tlStroff+ tl->tlStrlen;

    drChanged= td->tdSelectedRectangle;
    drChanged.drX1= drBack->drX1;

    paragraphsMerged= td->tdSelection.bsEnd.bpBi != bi;

    /*  b,c,d,e  */
    if  ( docReplaceSelection( bd, &(td->tdSelection),
		    &partShift, &stroffShift,
		    addedText, addedLength, td->tdCurrentTextAttribute,
		    (void *)add->addDisplay, tedCloseObject ) )
	{ LDEB(addedLength); return;	}

    /*  B  */
    tedAdjustRedrawBegin( td, drBack, &drChanged, &line, part );

    /*  C  */
    if  ( tedAdjustParagraphLayout( bi, line, stroffShift,
				upto+ addedLength, add, dp, &drChanged ) )
	{ LDEB(line); }

    if  ( paragraphsMerged )
	{
	BufferItem *	nextBi= docNextParagraph( bi );

	if  ( nextBi && tedLayoutItem( nextBi, dp, add, &drChanged ) )
	    { LDEB(1);	}
	}

    if  ( tedOpenItemObjects( bi, &ed->edColors, add, win, gc ) )
	{ LDEB(1);	}

    /*  6,7  */
    tedEditFinishIBarSelection( ed, bi, &drChanged, oldBackY1,
							stroff+ addedLength );

    return;
    }

/************************************************************************/
/*									*/
/*  Merge the selection into one paragraph.				*/
/*									*/
/************************************************************************/

static int tedIsIndentationParticule(	const BufferItem *	bi,
					const TextParticule *	tp )
    {
    const unsigned char *	s= bi->biParaString+ tp->tpStroff;
    int				i;

    if  ( tp->tpKind == DOCkindTAB )
	{ return 1;	}

    if  ( tp->tpKind != DOCkindTEXT )
	{ return 0;	}

    for ( i= 0; i < tp->tpStrlen; i++ )
	{
	if  ( *(s++) != ' ' )
	    { return 0;	}
	}

    return 1;
    }

void tedMergeParagraphsInSelection(	EditDocument *	ed )
    {
    TedDocument *		td= (TedDocument *)ed->edPrivateData;
    BufferDocument *		bd= td->tdDocument;
    DocumentProperties *	dp= &(bd->bdProperties);
    AppDrawingData *		add= &(ed->edDrawingData);

    BufferItem *		biTo;
    BufferItem *		biFrom;

    int				oldBackY1= add->addBackRect.drY1;

    const DocumentRectangle *	drBack= &(add->addBackRect);
    DocumentRectangle		drChanged;
    
    int				endMoved= 0;

    drChanged= td->tdSelectedRectangle;
    drChanged.drX1= drBack->drX1;

    if  ( td->tdSelection.bsBegin.bpBi == td->tdSelection.bsEnd.bpBi )
	{ LDEB(1); return; }

    biTo= td->tdSelection.bsBegin.bpBi;
    biFrom= docNextParagraph( biTo );
    if  ( ! biTo )
	{ XDEB(biTo); return;	}

    endMoved= biTo->biParaStrlen;

    for (;;)
	{
	int	partFrom= 0;

	int	particulesInserted= 0;
	int	charactersCopied= 0;


	while( partFrom < biFrom->biParaParticuleCount- 1		&&
	       tedIsIndentationParticule( biFrom,
				biFrom->biParaParticules+ partFrom )	)
	    { partFrom++; }

	if  ( partFrom < biFrom->biParaParticuleCount )
	    {
	    int		toCount= biTo->biParaParticuleCount;

	    if  ( toCount > 0						&&
		  biTo->biParaParticules[toCount-1].tpKind ==
							DOCkindTEXT	&&
		  biTo->biParaString[biTo->biParaStrlen-1] != ' '	)
		{
		if  ( docInflateTextString( biTo, 1 ) )
		    { LDEB(biTo->biParaStrlen); return; }

		biTo->biParaString[biTo->biParaStrlen++]= ' ';
		biTo->biParaString[biTo->biParaStrlen  ]= '\0';
		biTo->biParaParticules[toCount- 1].tpStrlen++;
		endMoved++;
		}

	    if  ( docCopyParticules( biTo, biFrom,
			biTo->biParaParticuleCount, partFrom,
			biFrom->biParaParticuleCount- partFrom,
			&particulesInserted, &charactersCopied, (char *)0 ) )
		{ LDEB(biFrom->biParaParticuleCount); return;	}

	    }

	if  ( biFrom == td->tdSelection.bsEnd.bpBi )
	    {
	    endMoved -= biFrom->biParaParticules[partFrom].tpStroff;
	    break;
	    }

	endMoved += charactersCopied;

	biFrom= docNextParagraph( biFrom );
	if  ( ! biFrom )
	    { XDEB(biFrom); return;	}
	}

    biTo= td->tdSelection.bsBegin.bpBi;
    biFrom= docNextParagraph( biTo );
    if  ( ! biTo )
	{ XDEB(biTo); return;	}

    for (;;)
	{
	BufferItem *	nextBi= docNextParagraph( biFrom );

	docDeleteItems( bd, biFrom->biParent, biFrom->biNumberInParent, 1 );

	if  ( biFrom == td->tdSelection.bsEnd.bpBi )
	    { break;	}

	if  ( ! nextBi )
	    { XDEB(nextBi); return;	}
	biFrom= nextBi;
	}

    if  ( tedLayoutItem( biTo, dp, add, &drChanged ) )
	{ LDEB(1);	}

    tedExposeRectangle( ed, &drChanged, /*scrolledX,Y*/ 0,0 );

    td->tdSelection.bsEnd.bpBi= biTo;
    td->tdSelection.bsEnd.bpStroff += endMoved;

    docFindLineAndParticule( td->tdSelection.bsEnd.bpBi,
		td->tdSelection.bsEnd.bpStroff, &(td->tdSelection.bsEnd), 0 );

    tedCalculateSelectedLines( td );
    tedSelectionCoordinates( &(td->tdSelection), add );

    tedSelectionRectangle( &(td->tdSelectedRectangle),
					     add, &(td->tdSelection) );

    if  ( add->addBackRect.drY1 != oldBackY1 )
	{
	appDocSetScrollbarValues( ed );
	appSetShellConstraints( ed );
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Replace the selection in a document with another document.		*/
/*  ( Used with 'paste', 'undo', 'redo'. )				*/
/*									*/
/*  B)  The first particule of the line was split, probably, part fits	*/
/*	on the previous line. Reformat from the previous particule.	*/
/*	If paragraphs were merged, redraw the whole neighbourhood.	*/
/*									*/
/*  1)  Replace the selection of the target with the text of those	*/
/*	particules at the beginning of the source that have the same	*/
/*	attributes.							*/
/*  2)  Insert the rest of the first paragraph of the source into the	*/
/*	target.								*/
/*  4)  If the insertion consists of more than one paragraph, split the	*/
/*	target paragraph.						*/
/*  5)  Insert the last particule of the insertion as text.		*/
/*									*/
/*  z)  Copy all paragraphs between the first and last (exclusive) of	*/
/*	the source to the target.					*/
/*									*/
/************************************************************************/

static int tedIncludeParagraphs(	const char *		filename,
					int *			pEndedInTable,
					int			startWithTable,
					BufferItem *		biFrom,
					BufferItem *		biTo,
					BufferPosition *	bpEndFrom,
					BufferDocument *	bdTo,
					AppDrawingData *	add,
					DocumentRectangle *	drChanged,
					AppColors *		ac,
					Window			win,
					GC			gc )
    {
    DocumentProperties *	dpTo= &(bdTo->bdProperties);

    BufferItem *		rowBi= (BufferItem *)0;
    int				inTable= 0;

    if  ( ! startWithTable )
	{
	biFrom= docNextParagraph( biFrom );
	if  ( ! biFrom )
	    { XDEB(biFrom); return -1;	}

	if  ( biFrom == bpEndFrom->bpBi )
	    { return 0; }
	}

    for (;;)
	{
	BufferItem *	insBi;

	if  ( biFrom->biParaInTable )
	    {
	    BufferPosition	bp;

	    BufferItem *	biRowTo= biTo->biParent->biParent;
	    BufferItem *	biRowFrom= biFrom->biParent->biParent;

	    inTable= 1;

	    if  ( biTo->biNumberInParent !=
					biTo->biParent->biGroupChildCount -1 )
		{
		if  ( docSplitGroupItem( &insBi,
				biTo->biParent, biTo->biNumberInParent+ 1 ) )
		    { LDEB(1); return -1;	}

		}

	    if  ( biTo->biParent->biNumberInParent ==
					    biRowTo->biGroupChildCount -1 )
		{
		tedLayoutItem( biTo->biParent, dpTo, add, drChanged );
		tedLayoutItem( insBi, dpTo, add, drChanged );
		}
	    else{
		if  ( docSplitGroupItem( &insBi,
			    biRowTo, biTo->biParent->biNumberInParent+ 1 ) )
		    { LDEB(1); return -1;	}

		tedLayoutItem( biRowTo, dpTo, add, drChanged );
		tedLayoutItem( insBi, dpTo, add, drChanged );

		biRowTo= insBi;
		}

	    rowBi= docCopyRowItem( bdTo, biRowTo->biParent,
			biRowTo->biNumberInParent+ 1, biRowFrom, filename );

	    if  ( ! rowBi )
		{ XDEB(rowBi); return -1;	}

	    if  ( tedLayoutItem( rowBi, dpTo, add, drChanged ) )
		{ LDEB(1); return -1;	}

	    if  ( tedOpenItemObjects( rowBi, ac, add, win, gc ) )
		{ LDEB(1); return -1;	}

	    if  ( docLastPosition( biRowFrom, &bp ) )
		{ LDEB(1); break;	}

	    if  ( biFrom == bpEndFrom->bpBi )
		{ break;	}

	    biFrom= docNextParagraph( bp.bpBi );
	    if  ( ! biFrom )
		{ break;	}

	    if  ( biFrom == bpEndFrom->bpBi && ! biFrom->biParaInTable )
		{ break;	}

	    if  ( docLastPosition( rowBi, &bp ) )
		{ LDEB(1); break;	}

	    biTo= bp.bpBi;
	    }
	else{
	    if  ( rowBi )
		{
		insBi= docInsertItem( rowBi->biParent,
				    rowBi->biNumberInParent+ 1, DOClevROW );
		if  ( ! insBi )
		    { XDEB(insBi); return -1;	}

		insBi= docInsertItem( insBi, 0, DOClevCELL );
		if  ( ! insBi )
		    { XDEB(insBi); return -1;	}

		insBi= docCopyParaItem( bdTo, insBi, 0, biFrom, filename );
		}
	    else{
		insBi= docCopyParaItem( bdTo, biTo->biParent,
				biTo->biNumberInParent+ 1, biFrom, filename );
		}

	    inTable= 0;

	    if  ( ! insBi )
		{ XDEB(insBi); return -1;	}

	    if  ( tedLayoutItem( insBi, dpTo, add, drChanged ) )
		{ LDEB(1); return -1;	}

	    if  ( tedOpenItemObjects( insBi, ac, add, win, gc ) )
		{ LDEB(1); return -1;	}

	    biFrom= docNextParagraph( biFrom );
	    if  ( ! biFrom )
		{ XDEB(biFrom); return -1;	}

	    if  ( biFrom == bpEndFrom->bpBi )
		{ break;	}

	    biTo= insBi;
	    rowBi= (BufferItem *)0;
	    }
	}

    *pEndedInTable= inTable; return 0;
    }

/************************************************************************/
/*									*/
/*  Include the tail of a paragraph in another paragraph.		*/
/*									*/
/*  It is known by the caller that at the beginning if the series of	*/
/*  particules that is to be included, no particule merging is		*/
/*  possible/necessary.							*/
/*									*/
/*  1)  Inclusion at the end of the paragraph... No particule merging.	*/
/*  2)  If the last particule of the source, and the subsequent		*/
/*	particule of the target have the same attributes, let the	*/
/*	text replacement routine decide whether to merge them or not.	*/
/*  3)  Copy those particules that are to be copied 'as is.'		*/
/*  4)  Insert the last paricule as text if this is desirable.		*/
/*									*/
/************************************************************************/

static int tedIncludeTail(	BufferDocument *	bdTo,
				BufferItem *            biTo,
				const BufferItem *	biFrom,
				int			partTo,
				int			partFrom,
				int *			pPartShift,
				int *			pStroffShift,
				Display *		display,
				const char *		refFileName )
    {
    int			copyLastAsText;
    int			particulesCopied;

    TextParticule *	tpTo;
    TextParticule *	tpFrom;

    int			particulesInserted= 0;
    int			charactersCopied= 0;

    /*  1  */
    if  ( biTo->biParaStrlen == 0		||
	  partTo == biTo->biParaParticuleCount	)
	{ copyLastAsText= 0;	}
    else{
	tpTo= biTo->biParaParticules+ partTo;
	tpFrom= biFrom->biParaParticules+ biFrom->biParaParticuleCount- 1;

	/*  2  */
	if  ( tpFrom->tpKind == DOCkindTEXT				&&
	      tpTo->tpKind == DOCkindTEXT				&&
	      docEqualTextAttributes( &(tpTo->tpTextAttribute),
					&(tpFrom->tpTextAttribute) )	)
	    { copyLastAsText= 1;	}
	else{ copyLastAsText= 0;	}
	}

    /*  3  */
    particulesCopied= biFrom->biParaParticuleCount- partFrom- copyLastAsText;
    if  ( particulesCopied > 0 )
	{
	if  ( docCopyParticules( biTo, biFrom, partTo, partFrom,
				    particulesCopied, &particulesInserted,
				    &charactersCopied, refFileName ) )
	    { LDEB(1); return -1;	}
	}
    else{
	if  ( particulesCopied < 0 )
	    { LDEB(particulesCopied); particulesCopied= 0;	}
	}

    /*  4  */
    if  ( copyLastAsText > 0 )
	{
	tpTo= biTo->biParaParticules+ partTo+ particulesCopied;

	tpFrom= biFrom->biParaParticules+
				biFrom->biParaParticuleCount- copyLastAsText;

	if  ( docParaReplaceText( bdTo, biTo, partTo+ particulesCopied,
		    tpTo->tpStroff,
		    &particulesInserted, &charactersCopied, tpTo->tpStroff,
		    biFrom->biParaString+ tpFrom->tpStroff,
		    tpFrom->tpStrlen,
		    tpFrom->tpTextAttribute,
		    (void *)display, tedCloseObject ) )
	    { LDEB(1); return -1;	}
	}

    *pPartShift += particulesInserted;
    *pStroffShift += charactersCopied;

    return 0;
    }

static int tedIncludeHead(	BufferItem *		biTo,
				const BufferItem *	biFrom,
				const int		part,
				const int		stroff,
				TedDocument *		tdTo,
				int *			pPartShift,
				int *			pStroffShift,
				int *			pLine,
				int *			pCharactersCopied,
				DocumentRectangle *	drChanged,
				const AppDrawingData *	add,
				const char *		refFileName )
    {
    const DocumentRectangle *	drBack= &(add->addBackRect);

    TextParticule *		tpTo;
    TextParticule *		tpFrom;

    int				partTo= part;
    unsigned int		stroffTo= stroff;

    int				charactersCopied= 0;
    int				particulesCopied;
    int				particuleSplit= 0;

    BufferDocument *		bdTo= tdTo->tdDocument;

    tpTo= biTo->biParaParticules+ tdTo->tdSelection.bsBegin.bpParticule;
    tpFrom= biFrom->biParaParticules;

    if  ( biTo->biParaStrlen == 0 )
	{
	int		toInTable= biTo->biParaInTable;

	if  ( docCopyParagraphRuler( &biTo->biParaProperties,
					    &biFrom->biParaProperties )	)
	    { LDEB(1); }

	biTo->biParaInTable= toInTable;
	tedParaScreenGeometry( biTo, add );
	}

    /*  B  */
    tedAdjustRedrawBegin( tdTo, drBack, drChanged, pLine, partTo );

    /*  1  */
    particulesCopied= 0;
    while( particulesCopied < biFrom->biParaParticuleCount		&&
	   tpTo->tpKind == DOCkindTEXT					&&
	   tpFrom[particulesCopied].tpKind == DOCkindTEXT		&&
	   docEqualTextAttributes( &(tpTo->tpTextAttribute),
			&(tpFrom[particulesCopied].tpTextAttribute) )	)
	{ particulesCopied++; }

    if  ( particulesCopied > 0 )
	{
	charactersCopied= tpFrom[particulesCopied-1].tpStroff+
				    tpFrom[particulesCopied-1].tpStrlen;
			    
	if  ( docReplaceSelection( bdTo, &(tdTo->tdSelection),
		    pPartShift, pStroffShift,
		    biFrom->biParaString+ tpFrom->tpStroff, charactersCopied,
		    tpFrom[particulesCopied-1].tpTextAttribute,
		    (void *)add->addDisplay, tedCloseObject ) )
	    { LDEB(tpFrom->tpStrlen); return -1;	}
	}
    else{
	if  ( ! tedHasIBarSelection( tdTo )				&&
	      docReplaceSelection( bdTo, &(tdTo->tdSelection),
		    pPartShift, pStroffShift,
		    (unsigned char *)0, 0, tdTo->tdCurrentTextAttribute,
		    (void *)add->addDisplay, tedCloseObject )		)
	    { LDEB(0); return -1;	}
	}

    stroffTo += charactersCopied;

    if  ( stroffTo >= biTo->biParaStrlen )
	{ partTo= biTo->biParaParticuleCount;	}
    else{
	partTo= 0; tpTo= biTo->biParaParticules;
	while( tpTo->tpStroff+ tpTo->tpStrlen < stroffTo )
	    { partTo++; tpTo++;	}

	if  ( tpTo->tpStroff != stroffTo				&&
	      tpTo->tpStroff+ tpTo->tpStrlen != stroffTo		)
	    { particuleSplit= 1;	}

	if  ( particulesCopied < biFrom->biParaParticuleCount	&&
	      particuleSplit					)
	    {
	    if  ( ! docCopyParticule( biTo, partTo+ 1, stroffTo,
		    tpTo->tpStroff+ tpTo->tpStrlen- stroffTo,
							DOCkindTEXT, tpTo ) )
		{ LDEB(1); return -1;	}

	    tpTo= biTo->biParaParticules+ partTo;
	    tpTo->tpStrlen= stroffTo- tpTo->tpStroff;

	    (*pPartShift)++; partTo++;
	    }
	}

    if  ( particulesCopied < biFrom->biParaParticuleCount )
	{
	int	partFrom= particulesCopied;

	if  ( tedIncludeTail( bdTo, biTo, biFrom, partTo, partFrom, pPartShift,
				pStroffShift, add->addDisplay, refFileName ) )
	    { LLDEB(partTo,partFrom); return -1;	}

	charactersCopied= biFrom->biParaStrlen;
	}

    *pCharactersCopied= charactersCopied;
    return 0;
    }

static int tedIncludeDocument(	Widget			w,
				EditDocument *		edTo,
				BufferDocument *	bdFrom )
    {
    AppDrawingData *		add= &(edTo->edDrawingData);

    TedDocument *		tdTo= (TedDocument *)edTo->edPrivateData;
    BufferDocument *		bdTo= tdTo->tdDocument;
    DocumentProperties *	dpTo= &(bdTo->bdProperties);

    Display *			display= XtDisplay( w );
    Window			win= XtWindow( w );
    GC				gc= edTo->edGc;
	
    int				stroffShift= 0;
    int				partShift= 0;
    int				endedInTable= 0;

    BufferPosition		bpBeginFrom;
    BufferPosition		bpEndFrom;

    BufferItem *		biFrom;

    BufferItem *		biTo;

    int				line;
    int				part;
    int				stroff;
    int				stroffEnd;

    int				charactersCopied= 0;

    int				oldBackY1= add->addBackRect.drY1;

    const DocumentRectangle *	drBack= &(add->addBackRect);
    DocumentRectangle		drChanged;

    if  ( docFirstPosition( &(bdFrom->bdItem), &bpBeginFrom ) )
	{ LDEB(1); return -1;	}
    if  ( docLastPosition( &(bdFrom->bdItem), &bpEndFrom ) )
	{ LDEB(1); return -1;	}

    biFrom= bpBeginFrom.bpBi;
    biTo= tdTo->tdSelection.bsBegin.bpBi;

    stroff= tdTo->tdSelection.bsBegin.bpStroff;
    part= tdTo->tdSelection.bsBegin.bpParticule;
    line= tdTo->tdSelection.bsBegin.bpLine;

    drChanged= tdTo->tdSelectedRectangle;
    drChanged.drX0= drBack->drX0;
    drChanged.drX1= drBack->drX1;

    if  ( ! biFrom->biParaInTable )
	{
	if  ( tedIncludeHead( biTo, biFrom, part, stroff, tdTo,
		    &partShift, &stroffShift, &line,
		    &charactersCopied, &drChanged, add, edTo->edFilename ) )
	    { LDEB(1); return -1;	}
	}

    if  ( bpEndFrom.bpBi != biFrom )
	{
	BufferItem *	newBi;

	const int	particulesCopied= 0;
	TextParticule *	tpFrom;

	/*  4  */
	if  ( docSplitParaItem( bdTo, &newBi, biTo, stroff+ charactersCopied ) )
	    { LLDEB(part+ partShift,stroff+ charactersCopied); return -1; }

	/*  C  */
	tedLayoutItem( biTo, dpTo, add, &drChanged );

	if  ( tedOpenItemObjects( biTo, &edTo->edColors, add, win, gc ) )
	    { LDEB(1); return -1;	}

	if  ( tedIncludeParagraphs( edTo->edFilename, &endedInTable,
				    biFrom->biParaInTable,
				    biFrom, biTo, &bpEndFrom,
				    bdTo, add, &drChanged,
				    &edTo->edColors, win, gc ) )
	    { LDEB(1); return -1;	}

	part= 0; stroff= 0; partShift= 0; stroffShift= 0;

	biFrom= bpEndFrom.bpBi;
	tpFrom= biFrom->biParaParticules;

	if  ( ! endedInTable && biFrom->biParaStrlen > 0 )
	    {
	    int		partTo= part;
	    int		partFrom= particulesCopied;
	    int		wasEmpty= newBi->biParaStrlen == 0;
	    int		savedInTable= newBi->biParaInTable;

	    if  ( tedIncludeTail( bdTo, newBi, biFrom, partTo, partFrom,
			&partShift, &stroffShift, display, edTo->edFilename ) )
		{ LLDEB(partTo,partFrom); return -1;	}

	    if  ( wasEmpty &&
		    docCopyParagraphProperties( &newBi->biParaProperties,
						&biFrom->biParaProperties ) )
		{ LDEB(wasEmpty);	}

	    newBi->biParaInTable= savedInTable;
	    }

	if  ( tedLayoutItem( newBi, dpTo, add, &drChanged ) )
	    { LDEB(1); return -1;	}

	if  ( tedOpenItemObjects( newBi, &edTo->edColors, add, win, gc ) )
	    { LDEB(1); return -1;	}

	biTo= newBi;
	stroffEnd= 0+ stroffShift;
	}
    else{
	stroffEnd= tdTo->tdSelection.bsBegin.bpStroff+ charactersCopied;

	/*  C  */
	if  ( tedAdjustParagraphLayout( biTo, line, stroffShift,
					stroffEnd, add, dpTo, &drChanged ) )
	    { LDEB(1); }

	if  ( tedOpenItemObjects( biTo, &edTo->edColors, add, win, gc ) )
	    { LDEB(1); return -1;	}
	}

    tedEditFinishIBarSelection( edTo, biTo, &drChanged, oldBackY1, stroffEnd );

    return 0;
    }

int tedIncludePlainDocument(	Widget			w,
				EditDocument *		ed,
				BufferDocument *	bdFrom )
    {
    TedDocument *		td= (TedDocument *)ed->edPrivateData;
    BufferPosition		bpBeginTo;
    BufferPosition		bpBeginFrom;
    BufferItem *		bi;

    TextAttribute		ta;

    bpBeginTo= td->tdSelection.bsBegin;
    if  ( ! bpBeginTo.bpBi				||
	  bpBeginTo.bpBi->biParaParticuleCount == 0	)
	{ LDEB(1); return -1;	}

    ta= bpBeginTo.bpBi->biParaParticules[bpBeginTo.bpParticule].tpTextAttribute;

    if  ( docFirstPosition( &(bdFrom->bdItem), &bpBeginFrom ) )
	{ LDEB(1); return -1;	}
    bi= bpBeginFrom.bpBi;

    for (;;)
	{
	TextParticule *		tp;
	int			part;

	tp= bi->biParaParticules;

	for ( part= 0; part < bi->biParaParticuleCount; part++, tp++ )
	    { tp->tpTextAttribute= ta; }

	bi= docNextParagraph( bi );
	if  ( ! bi )
	    { break;	}
	}

    return tedIncludeDocument( w, ed, bdFrom );
    }

int tedIncludeRtfDocument(	Widget			w,
				EditDocument *		ed,
				BufferDocument *	bdFrom )
    {
    TedDocument *		tdTo= (TedDocument *)ed->edPrivateData;
    BufferDocument *		bdTo= tdTo->tdDocument;
    DocumentProperties *	dpTo= &(bdTo->bdProperties);

    BufferPosition		bpBeginFrom;
    BufferItem *		bi;

    int *			fontMap= (int *)0;
    int *			fontUsed= (int *)0;

    int				fromFont;
    int				fromFontCount;
    DocumentFont *		fromFonts;

    int				toFont;
    int				toFontCount;
    DocumentFont *		toFonts;

    if  ( bdFrom->bdItem.biDocContainsTables )
	{
	if  ( tdTo->tdSelection.bsCol0 >= 0 )
	    { return 0;	}
	if  ( tdTo->tdSelection.bsBegin.bpBi->biParaInTable )
	    { return 0;	}
	if  ( tdTo->tdSelection.bsEnd.bpBi->biParaInTable )
	    { return 0;	}
	}

    fromFontCount= bdFrom->bdProperties.dpFontList.dflCount;
    fromFonts= bdFrom->bdProperties.dpFontList.dflFonts;

    toFontCount= dpTo->dpFontList.dflCount;
    toFonts= dpTo->dpFontList.dflFonts;

    toFonts= (DocumentFont *)realloc( toFonts,
		    ( toFontCount+ fromFontCount )* sizeof(DocumentFont) );
    if  ( ! toFonts )
	{ LXDEB(toFontCount+ fromFontCount,toFonts); return -1;	}
    dpTo->dpFontList.dflFonts= toFonts;

    fontMap= (int *)malloc( fromFontCount* sizeof( int ) );
    if  ( ! fontMap )
	{ LXDEB(fromFontCount,fontMap); return -1; }

    fontUsed= (int *)malloc( fromFontCount* sizeof( int ) );
    if  ( ! fontUsed )
	{ LXDEB(fromFontCount,fontUsed); return -1; }

    for ( fromFont= 0; fromFont < fromFontCount; fromFont++ )
	{ fontUsed[fromFont]= 0;	}

    if  ( docFirstPosition( &(bdFrom->bdItem), &bpBeginFrom ) )
	{ LDEB(1); return -1;	}
    bi= bpBeginFrom.bpBi;

    for (;;)
	{
	TextParticule *		tp;
	int			part;

	tp= bi->biParaParticules;

	for ( part= 0; part < bi->biParaParticuleCount; part++, tp++ )
	    { fontUsed[tp->tpTextAttribute.taFontNumber]= 1; }

	bi= docNextParagraph( bi );
	if  ( ! bi )
	    { break;	}
	}

    for ( fromFont= 0; fromFont < fromFontCount; fromFont++ )
	{
	if  ( ! fontUsed[fromFont] )
	    { continue;	}

	for ( toFont= 0; toFont < toFontCount; toFont++ )
	    {
	    if  ( toFonts[toFont].dfDocFamilyNumber < 0 )
		{ continue;	} 

	    if  ( ! toFonts[toFont].dfFamilyStyle )
		{ XDEB(toFonts[toFont].dfFamilyStyle); continue;	}

	    if  ( ! strcmp( fromFonts[fromFont].dfFamilyStyle,
					toFonts[toFont].dfFamilyStyle )	&&
		  ! strcmp( fromFonts[fromFont].dfName,
					toFonts[toFont].dfName )	)
		{ break;	}
	    }

	if  ( toFont >= toFontCount )
	    {
	    DocumentFont *	df;

	    df= docInsertFont( &(dpTo->dpFontList),
				    -1,
				    fromFonts[fromFont].dfFamilyStyle,
				    fromFonts[fromFont].dfName );

	    if  ( ! df )
		{ XDEB(df); return -1;	}

	    toFont= df->dfDocFamilyNumber;

	    toFonts= dpTo->dpFontList.dflFonts;
	    toFontCount= dpTo->dpFontList.dflCount;
	    }

	fontMap[fromFont]= toFont;
	}

    if  ( docFirstPosition( &(bdFrom->bdItem), &bpBeginFrom ) )
	{ LDEB(1); return -1;	}
    bi= bpBeginFrom.bpBi;

    for (;;)
	{
	TextParticule *		tp;
	int			part;

	tp= bi->biParaParticules;

	for ( part= 0; part < bi->biParaParticuleCount; part++, tp++ )
	    {
	    if  ( tp->tpTextAttribute.taFontNumber < 0			||
		  tp->tpTextAttribute.taFontNumber >= fromFontCount	)
		{
		LDEB(part);
		LLDEB(tp->tpTextAttribute.taFontNumber,fromFontCount);
		tp->tpTextAttribute.taFontNumber= 0;
		}
	    tp->tpTextAttribute.taFontNumber=
				    fontMap[tp->tpTextAttribute.taFontNumber];
	    }

	bi= docNextParagraph( bi );
	if  ( ! bi )
	    { break;	}
	}

    if  ( fontMap )
	{ free( fontMap );	}
    if  ( fontUsed )
	{ free( fontUsed );	}

    return tedIncludeDocument( w, ed, bdFrom );
    }

/************************************************************************/
/*									*/
/*  Split a paragraph.							*/
/*									*/
/*  1)  Split in the buffer administration.				*/
/*  2)  Recalculate the geometry from the start of the selection to the	*/
/*	end of the paragraph.						*/
/*  3)  Redraw the affected part of the paragraph.			*/
/*  4)  Layout the new paragraph.					*/
/*									*/
/************************************************************************/

int tedSplitParaContents(	BufferItem **		pNewBi,
				BufferItem *		bi,
				DocumentRectangle *	drChanged,
				EditDocument *		ed,
				int			splitLevel,
				int			onNewPage )
    {
    TedDocument *		td= (TedDocument *)ed->edPrivateData;
    BufferDocument *		bd= td->tdDocument;
    DocumentProperties *	dp= &(bd->bdProperties);

    BufferItem *		newBi;

    AppDrawingData *		add= &(ed->edDrawingData);

    Widget			w= ed->edDocumentWidget;
    Window			win= XtWindow( w );
    GC				gc= ed->edGc;

    int				n;
    int				level;

    /*  1  */
    if  ( docSplitParaItem( bd, &newBi, bi, td->tdSelection.bsBegin.bpStroff ) )
	{ LDEB(td->tdSelection.bsBegin.bpStroff); return -1;	}

    newBi->biParaStartsOnNewPage= ! newBi->biParaInTable && onNewPage;
    
    tedParaScreenGeometry( newBi, add );

    /*  3  */
    *drChanged= td->tdSelectedRectangle;
    drChanged->drX0= add->addBackRect.drX0;
    drChanged->drX1= add->addBackRect.drX1; 

    if  ( bi->biParaBottomBorder.bpIsSet )
	{ bi->biParaBottomBorder.bpIsSet= 0;	}
    if  ( bi->biParaSpaceAfterTwips > 0 )
	{ bi->biParaSpaceAfterTwips= 0;		}

    if  ( newBi->biParaTopBorder.bpIsSet )
	{ newBi->biParaTopBorder.bpIsSet= 0;	}
    if  ( newBi->biParaSpaceBeforeTwips > 0 )
	{ newBi->biParaSpaceBeforeTwips= 0;	}

    /*  2  */
    if  ( tedLayoutItem( bi, dp, add, drChanged ) < 0 )
	{ LDEB(1); return -1;	}

    /*  4  */
    if  ( tedLayoutItem( newBi, dp, add, drChanged ) < 0 )
	{ LDEB(1); return -1;	}

    if  ( tedOpenItemObjects( newBi, &ed->edColors, add, win, gc ) )
	{ LDEB(1); return -1;	}

    level= DOClevPARA;
    n= newBi->biNumberInParent;
    while( level > splitLevel )
	{
	BufferItem *	splitBi;

	if  ( docSplitGroupItem( &splitBi, bi->biParent, n ) )
	    { LDEB(1); return -1;	}

	tedLayoutItem( bi->biParent, dp, add, drChanged );
	tedLayoutItem( splitBi, dp, add, drChanged );

	n= splitBi->biNumberInParent+ 1;
	bi= splitBi; level--;
	}

    *pNewBi= newBi; return 0;
    }

void tedSplitParagraph(		EditDocument *		ed,
				int			onNewPage )
    {
    AppDrawingData *	add= &(ed->edDrawingData);
    TedDocument *	td= (TedDocument *)ed->edPrivateData;
    BufferItem *	bi= td->tdSelection.bsBegin.bpBi;
    BufferItem *	newBi;

    int			oldBackY1= add->addBackRect.drY1;

    DocumentRectangle	drChanged;

    if  ( tedSplitParaContents( &newBi, bi, &drChanged,
						ed, DOClevPARA, onNewPage ) )
	{ LDEB(1); return;	}

    /*  7,8  */
    tedEditFinishIBarSelection( ed, newBi, &drChanged, oldBackY1, 0 );

    return;
    }

/************************************************************************/
/*									*/
/*  Insert a 'Tab'.							*/
/*									*/
/************************************************************************/

void tedSetTab(		Widget			w,
			EditDocument *		ed )
    {
    AppDrawingData *		add= &(ed->edDrawingData);
    TedDocument *		td= (TedDocument *)ed->edPrivateData;
    BufferDocument *		bd= td->tdDocument;
    DocumentProperties *	dp= &(bd->bdProperties);
    TextParticule *		tp;

    int				stroffShift;
    int				partShift;

    BufferItem *		bi= td->tdSelection.bsBegin.bpBi;
    int				oldX0;

    int				line;
    int				part;

    int				oldBackY1= add->addBackRect.drY1;

    DocumentRectangle		drChanged;
    const DocumentRectangle *	drBack= &(add->addBackRect);

    part= td->tdSelection.bsBegin.bpParticule;
    line= td->tdSelection.bsBegin.bpLine;

    /*  1  */
    oldX0= bi->biParaParticules[part].tpX0;

    tp= docParaSpecialParticule( bi, DOCkindTAB,
				    part, td->tdSelection.bsEnd.bpStroff,
				    &partShift, &stroffShift );
    if  ( ! tp )
	{ LXDEB(part,tp); return;	}

    drChanged= td->tdSelectedRectangle;
    drChanged.drX1= drBack->drX1;

    /*  3,4,5  */
    if  ( tedAdjustParagraphLayout( bi, line, stroffShift,
				td->tdSelection.bsBegin.bpStroff+ stroffShift,
					    add, dp, &drChanged ) )
	{ LDEB(line); }

    /*  6,7  */
    tedEditFinishIBarSelection( ed, bi, &drChanged, oldBackY1,
			    td->tdSelection.bsBegin.bpStroff+ stroffShift );

    return;
    }

/************************************************************************/
/*									*/
/*  Insert an 'Object'.							*/
/*									*/
/************************************************************************/

int tedInsertNewObject(		Widget			w,
				InsertedObject *	io,
				EditDocument *		ed )
    {
    AppDrawingData *		add= &(ed->edDrawingData);
    TedDocument *		td= (TedDocument *)ed->edPrivateData;
    BufferDocument *		bd= td->tdDocument;
    DocumentProperties *	dp= &(bd->bdProperties);
    TextParticule *		tp;

    int				stroffShift;
    int				partShift;

    BufferItem *		bi= td->tdSelection.bsBegin.bpBi;
    int				oldX0;

    int				line;
    int				part;

    int				oldBackY1= add->addBackRect.drY1;

    DocumentRectangle		drChanged;
    const DocumentRectangle *	drBack= &(add->addBackRect);

    part= td->tdSelection.bsBegin.bpParticule;
    line= td->tdSelection.bsBegin.bpLine;

    /*  1  */
    oldX0= bi->biParaParticules[part].tpX0;

    tp= docParaSpecialParticule( bi, DOCkindOBJECT,
				    part, td->tdSelection.bsEnd.bpStroff,
				    &partShift, &stroffShift );
    if  ( ! tp )
	{ LXDEB(part,tp); return -1;	}

    tp->tpObjectNumber= bi->biParaObjectCount++;

    drChanged= td->tdSelectedRectangle;
    drChanged.drX1= drBack->drX1;

    /*  3,4,5  */
    if  ( tedAdjustParagraphLayout( bi, line, stroffShift,
				td->tdSelection.bsBegin.bpStroff+ stroffShift,
				add, dp, &drChanged ) )
	{ LDEB(line); }

    /*  6,7  */
    tedEditFinishIBarSelection( ed, bi, &drChanged, oldBackY1,
			    td->tdSelection.bsBegin.bpStroff+ stroffShift );

    tp->tpPhysicalFont= 0;

    return 0;
    }

/************************************************************************/
/*									*/
/*  An object has been resized: Redo layout.				*/
/*									*/
/************************************************************************/
int tedResizeObject(		EditDocument *		ed,
				BufferItem *		bi,
				int			line,
				TextParticule *		tp,
				InsertedObject *	io,
				int			newWidth,
				int			newHeight,
				int			ox,
				int			oy )
    {
    TedDocument *		td= (TedDocument *)ed->edPrivateData;
    BufferDocument *		bd= td->tdDocument;
    DocumentProperties *	dp= &(bd->bdProperties);
    AppDrawingData *		add= &(ed->edDrawingData);

    Widget			w= ed->edDocumentWidget;
    Window			win= XtWindow( w );
    GC				gc= ed->edGc;

    int				stroffShift= 0;

    int				oldBackY1= add->addBackRect.drY1;
    DocumentRectangle		drChanged;

    drChanged= td->tdSelectedRectangle;
    drChanged.drX0= add->addBackRect.drX0;
    drChanged.drX1= add->addBackRect.drX1;

    io->ioPixelsWide= newWidth;
    io->ioPixelsHigh= newHeight;
    io->ioScaleX= ( 100.0* io->ioPixelsWide )/
	    TWIPStoPIXELS( add->addMagnifiedPixelsPerTwip, io->ioTwipsWide );
    io->ioScaleY= ( 100.0* io->ioPixelsHigh )/
	    TWIPStoPIXELS( add->addMagnifiedPixelsPerTwip, io->ioTwipsHigh );

    tp->tpPixelsWide= newWidth;

    if  ( tedReopenObject( bd, bi, tp, &ed->edColors,
					    &(dp->dpFontList), add, win, gc ) )
	{ LDEB(1); return -1;	}

    if  ( tedAdjustParagraphLayout( bi, line, stroffShift,
				td->tdSelection.bsBegin.bpStroff+ stroffShift,
				add, dp, &drChanged ) )
	{ LDEB(line); }

    tedExposeRectangle( ed, &drChanged, /*scrolledX,Y*/ 0,0 );

    tedCalculateSelectedLines( td );
    tedSelectionCoordinates( &(td->tdSelection), add );

    tedSelectionRectangle( &(td->tdSelectedRectangle),
					     add, &(td->tdSelection) );

    if  ( add->addBackRect.drY1 != oldBackY1 )
	{
	appDocSetScrollbarValues( ed );
	appSetShellConstraints( ed );
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Apply a change to the layout of a paragraph.			*/
/*									*/
/************************************************************************/

void tedEditApplyItemFormat(	EditDocument *	ed,
				BufferItem *	bi )
    {
    TedDocument *		td= (TedDocument *)ed->edPrivateData;
    BufferDocument *		bd= td->tdDocument;
    DocumentProperties *	dp= &(bd->bdProperties);
    AppDrawingData *		add= &(ed->edDrawingData);

    DocumentRectangle		drChanged;

    drChanged.drX0= add->addBackRect.drX0;
    drChanged.drX1= add->addBackRect.drX1;
    drChanged.drY0= bi->biTopPosition.lpYPixels;
    drChanged.drY1= bi->biBelowPosition.lpYPixels- 1;

    if  ( drChanged.drY0 > 0 )
	{ drChanged.drY0--;	}

    tedApplyItemFormat( bi, dp, add, &drChanged );

    tedExposeRectangle( ed, &drChanged, /*scrolledX,Y*/ 0,0 );

    tedCalculateSelectedLines( td );
    tedSelectionCoordinates( &(td->tdSelection), add );

    tedSelectionRectangle( &(td->tdSelectedRectangle),
						 add, &(td->tdSelection) );

    return;
    }

/************************************************************************/
/*									*/
/*  Replace the selection with something new.				*/
/*									*/
/*  1)  No selection... Just refuse.					*/
/*  2)  Replace nothing with nothing: Do nothing.			*/
/*									*/
/************************************************************************/

void tedAppReplaceSelection(	EditDocument *		ed,
				const unsigned char *	word,
				int			len	)
    {
    TedDocument *	td= (TedDocument *)ed->edPrivateData;

    /*  1  */
    if  ( ! td->tdSelection.bsBegin.bpBi )
	{ XDEB(td->tdSelection.bsBegin.bpBi); return; }

    /*  2  */
    if  ( tedHasIBarSelection( td ) && len == 0	)
	{ return;	}

    tedReplaceSelection( ed, word, len );

    appDocumentChanged( ed, 1 );

    return;
    }

/************************************************************************/
/*									*/
/*  Delete the current paragraph from a document.			*/
/*									*/
/************************************************************************/

static int tedEmptyParagraph(	BufferItem *		paraBi,
				BufferDocument *	bd,
				Display *		display )
    {
    int			part;
    TextParticule *	tp= paraBi->biParaParticules;
    TextAttribute	ta= tp->tpTextAttribute;

    paraBi->biParaStrlen= 0;

    for ( part= 0; part < paraBi->biParaParticuleCount; tp++, part++ )
	{
	tedCloseObject( bd, paraBi, tp, (void *)display );
	docCleanParticuleObject( paraBi, tp );
	}

    docDeleteParticules( paraBi, 0, paraBi->biParaParticuleCount );
    docDeleteLines( paraBi, 0, paraBi->biParaLineCount );

    if  ( ! docInsertTextParticule( paraBi, 0, 0, 0, DOCkindTEXT, ta ) )
	{ LDEB(1); return -1;	}

    return 0;
    }

int tedDeleteCurrentParagraph(	EditApplication *	ea )
    {
    EditDocument *		ed= ea->eaCurrentDocument;
    TedDocument *		td= (TedDocument *)ed->edPrivateData;
    BufferDocument *		bd= td->tdDocument;
    DocumentProperties *	dp= &(bd->bdProperties);
    BufferItem *		paraBi= td->tdSelection.bsBegin.bpBi;
    AppDrawingData *		add= &(ed->edDrawingData);

    BufferPosition		bpNew;
    BufferItem *		parentBi;

    int				oldBackY1= add->addBackRect.drY1;
    DocumentRectangle		drChanged;

    drChanged.drX0= add->addBackRect.drX0;
    drChanged.drX1= add->addBackRect.drX1;
    drChanged.drY0= paraBi->biTopPosition.lpYPixels- 1;
    drChanged.drY1= paraBi->biBelowPosition.lpYPixels- 1;

    docInitPosition( &bpNew );
    if  ( docLastPosition( paraBi, &bpNew ) 	||
	  docNextPosition( &bpNew )		)
	{
	const int	lastOne= 0;

	docInitPosition( &bpNew );
	if  ( docFirstPosition( paraBi, &bpNew )	||
	      docPrevPosition( &bpNew, lastOne )	)
	    { docInitPosition( &bpNew ); }
	}

    parentBi= paraBi->biParent;
    if  ( parentBi->biGroupChildCount < 2 )
	{
	if  ( tedEmptyParagraph( paraBi, bd,
					XtDisplay( ed->edDocumentWidget ) ) )
	    { LDEB(1); return -1;	}

	if  ( docFirstPosition( paraBi, &bpNew ) )
	    { LDEB(1); return -1;	}

	tedLayoutItem( paraBi, dp, add, &drChanged );

	bpNew.bpBi= paraBi; bpNew.bpStroff= 0;
	}
    else{
	tedDeleteItems( bd, parentBi, paraBi->biNumberInParent, 1,
					XtDisplay( ed->edDocumentWidget ) );
	tedLayoutItem( parentBi, dp, add, &drChanged );
	}

    appDocumentChanged( ed, 1 );

    tedEditFinishIBarSelection( ed, bpNew.bpBi, &drChanged, oldBackY1,
							    bpNew.bpStroff );

    return 0;
    }

/************************************************************************/
/*									*/
/*  Make sure that a Section contains an empty paragraph.		*/
/*									*/
/************************************************************************/

static int tedSectionParagraph(		BufferItem **			pParaBi,
					BufferItem *			sectBi,
					const ParagraphProperties *	pp,
					TextAttribute			ta )
    {
    BufferItem *	bi;

    bi= docInsertItem( sectBi, 0, DOClevROW );
    if  ( ! bi )
	{ XDEB(bi); return -1;	}

    bi= docInsertItem( bi, 0, DOClevCELL );
    if  ( ! bi )
	{ XDEB(bi); return -1;	}

    bi= docInsertItem( bi, 0, DOClevPARA );
    if  ( ! bi )
	{ XDEB(bi); return -1;	}

    if  ( docCopyParagraphProperties( &(bi->biParaProperties), pp ) )
	{ LDEB(1);	}

    if  ( ! docInsertTextParticule( bi, 0, 0, 0, DOCkindTEXT, ta ) )
	{ LDEB(1); return -1;	}

    *pParaBi= bi; return 0;
    }

/************************************************************************/
/*									*/
/*  Delete the current section.						*/
/*									*/
/************************************************************************/

int tedDeleteCurrentSection(	EditApplication *	ea )
    {
    EditDocument *		ed= ea->eaCurrentDocument;
    TedDocument *		td= (TedDocument *)ed->edPrivateData;
    BufferDocument *		bd= td->tdDocument;
    DocumentProperties *	dp= &(bd->bdProperties);
    BufferItem *		paraBi= td->tdSelection.bsBegin.bpBi;
    BufferItem *		sectBi;
    AppDrawingData *		add= &(ed->edDrawingData);

    BufferPosition		bpNew;
    BufferItem *		parentBi;

    int				oldBackY1= add->addBackRect.drY1;
    DocumentRectangle		drChanged;

    sectBi= paraBi->biParent->biParent->biParent;

    drChanged.drX0= add->addBackRect.drX0;
    drChanged.drX1= add->addBackRect.drX1;
    drChanged.drY0= sectBi->biTopPosition.lpYPixels- 1;
    drChanged.drY1= sectBi->biBelowPosition.lpYPixels- 1;

    docInitPosition( &bpNew );
    if  ( docLastPosition( sectBi, &bpNew ) 	||
	  docNextPosition( &bpNew )		)
	{
	const int	lastOne= 0;

	docInitPosition( &bpNew );
	if  ( docFirstPosition( sectBi, &bpNew )	||
	      docPrevPosition( &bpNew, lastOne )	)
	    { docInitPosition( &bpNew ); }
	}

    parentBi= sectBi->biParent;
    if  ( parentBi->biGroupChildCount < 2 )
	{
	ParagraphProperties	pp;
	TextParticule *		tp= paraBi->biParaParticules;
	TextAttribute		ta= tp->tpTextAttribute;

	docInitParagraphProperties( &pp );

	if  ( docCopyParagraphProperties( &pp, &(paraBi->biParaProperties) ) )
	    { LDEB(1); return -1;	}

	tedDeleteItems( bd, sectBi, 0, sectBi->biGroupChildCount,
					XtDisplay( ed->edDocumentWidget ) );

	if  ( tedSectionParagraph( &paraBi, sectBi, &pp, ta ) )
	    { LDEB(1); return -1;	}

	docCleanParagraphProperties( &pp );

	bpNew.bpBi= paraBi; bpNew.bpStroff= 0;
	}
    else{
	tedDeleteItems( bd, parentBi, paraBi->biNumberInParent, 1,
					XtDisplay( ed->edDocumentWidget ) );
	}

    tedLayoutItem( parentBi, dp, add, &drChanged );

    appDocumentChanged( ed, 1 );

    tedEditFinishIBarSelection( ed, bpNew.bpBi, &drChanged, oldBackY1,
							    bpNew.bpStroff );

    return 0;
    }

/************************************************************************/
/*									*/
/*  Insert a paragraph Before/After the selection, depending on the	*/
/*  value of the 'after' argument.					*/
/*									*/
/************************************************************************/

int tedInsertParagraph(	EditApplication *	ea,
			int			after )
    {
    EditDocument *		ed= ea->eaCurrentDocument;
    TedDocument *		td= (TedDocument *)ed->edPrivateData;
    BufferDocument *		bd= td->tdDocument;
    DocumentProperties *	dp= &(bd->bdProperties);
    AppDrawingData *		add= &(ed->edDrawingData);

    BufferItem *		paraBi;
    BufferItem *		newBi;

    TextAttribute		ta;
    BufferItem *		parentBi;
    int				pos;

    int				oldBackY1= add->addBackRect.drY1;
    DocumentRectangle		drChanged;

    if  ( after )
	{
	paraBi= td->tdSelection.bsEnd.bpBi;
	if  ( ! paraBi )
	    { XDEB(paraBi); return -1;	}

	pos= paraBi->biNumberInParent+ 1;
	ta= paraBi->biParaParticules[
			    paraBi->biParaParticuleCount-1].tpTextAttribute;
	}
    else{
	paraBi= td->tdSelection.bsBegin.bpBi;
	if  ( ! paraBi )
	    { XDEB(paraBi); return -1;	}

	pos= paraBi->biNumberInParent;
	ta= paraBi->biParaParticules[0].tpTextAttribute;
	}

    drChanged.drX0= add->addBackRect.drX0;
    drChanged.drX1= add->addBackRect.drX1;
    drChanged.drY0= paraBi->biTopPosition.lpYPixels- 1;
    drChanged.drY1= paraBi->biBelowPosition.lpYPixels- 1;

    parentBi= paraBi->biParent;

    newBi= docInsertItem( parentBi, pos, DOClevPARA );
    if  ( ! newBi )
	{ XDEB(newBi); return -1;	}

    if  ( docCopyParagraphProperties( &(newBi->biParaProperties),
						&(paraBi->biParaProperties) ) )
	{ LDEB(1);	}

    if  ( ! docInsertTextParticule( newBi, 0, 0, 0, DOCkindTEXT, ta ) )
	{ LDEB(1); return -1;	}

    tedLayoutItem( paraBi, dp, add, &drChanged );

    tedLayoutItem( newBi, dp, add, &drChanged );

    appDocumentChanged( ed, 1 );

    tedEditFinishIBarSelection( ed, newBi, &drChanged, oldBackY1, 0 );

    return 0;
    }

int tedInsertSection(	EditApplication *	ea,
			int			after )
    {
    EditDocument *		ed= ea->eaCurrentDocument;
    TedDocument *		td= (TedDocument *)ed->edPrivateData;
    BufferDocument *		bd= td->tdDocument;
    DocumentProperties *	dp= &(bd->bdProperties);
    AppDrawingData *		add= &(ed->edDrawingData);

    BufferItem *		paraBi;
    BufferItem *		sectBi;
    BufferItem *		newBi;
    BufferItem *		bi;

    TextAttribute		ta;
    BufferItem *		parentBi;
    int				pos;

    int				oldBackY1= add->addBackRect.drY1;
    DocumentRectangle		drChanged;

    if  ( after )
	{
	paraBi= td->tdSelection.bsEnd.bpBi;
	if  ( ! paraBi )
	    { XDEB(paraBi); return -1;	}
	sectBi= paraBi->biParent->biParent->biParent;

	pos= sectBi->biNumberInParent+ 1;
	ta= paraBi->biParaParticules[
			    paraBi->biParaParticuleCount-1].tpTextAttribute;
	}
    else{
	paraBi= td->tdSelection.bsBegin.bpBi;
	if  ( ! paraBi )
	    { XDEB(paraBi); return -1;	}
	sectBi= paraBi->biParent->biParent->biParent;

	pos= sectBi->biNumberInParent;
	ta= paraBi->biParaParticules[0].tpTextAttribute;
	}

    drChanged.drX0= add->addBackRect.drX0;
    drChanged.drX1= add->addBackRect.drX1;
    drChanged.drY0= sectBi->biTopPosition.lpYPixels- 1;
    drChanged.drY1= sectBi->biBelowPosition.lpYPixels- 1;

    parentBi= sectBi->biParent;

    newBi= docInsertItem( parentBi, pos, DOClevSECT );
    if  ( ! newBi )
	{ XDEB(newBi); return -1;	}

    if  ( docCopySectionProperties( &(newBi->biSectProperties),
						&(sectBi->biSectProperties) ) )
	{ LDEB(1);	}

    if  ( tedSectionParagraph( &bi, newBi, &(paraBi->biParaProperties), ta ) )
	{ LDEB(1); return -1;	}

    tedLayoutItem( sectBi, dp, add, &drChanged );

    tedLayoutItem( newBi, dp, add, &drChanged );

    appDocumentChanged( ed, 1 );

    tedEditFinishIBarSelection( ed, bi, &drChanged, oldBackY1, 0 );

    return 0;
    }

/************************************************************************/
/*									*/
/*  Insert a line break in the document.				*/
/*									*/
/************************************************************************/

void tedDocInsertLineBreak(	Widget		option,
				XtPointer	voided,
				XtPointer	voidpbcs )
    {
    EditDocument *		ed= (EditDocument *)voided;
    AppDrawingData *		add= &(ed->edDrawingData);
    TedDocument *		td= (TedDocument *)ed->edPrivateData;
    BufferDocument *		bd= td->tdDocument;
    DocumentProperties *	dp= &(bd->bdProperties);
    TextParticule *		tp;

    int				stroffShift;
    int				partShift;

    BufferItem *		bi= td->tdSelection.bsBegin.bpBi;
    int				oldX0;

    int				line;
    int				part;

    int				oldBackY1= add->addBackRect.drY1;

    DocumentRectangle		drChanged;
    const DocumentRectangle *	drBack= &(add->addBackRect);

    part= td->tdSelection.bsBegin.bpParticule;
    line= td->tdSelection.bsBegin.bpLine;

    /*  1  */
    oldX0= bi->biParaParticules[part].tpX0;

    tp= docParaSpecialParticule( bi, DOCkindLINEBREAK,
				    part, td->tdSelection.bsEnd.bpStroff,
				    &partShift, &stroffShift );
    if  ( ! tp )
	{ LXDEB(part,tp); return;	}

    drChanged= td->tdSelectedRectangle;
    drChanged.drX0= drBack->drX0;
    drChanged.drX1= drBack->drX1;

    appDocumentChanged( ed, 1 );

    /*  3,4,5  */
    if  ( tedLayoutItem( bi, dp, add, &drChanged ) )
	{ LDEB(1); }

    /*  6,7  */
    tedEditFinishIBarSelection( ed, bi, &drChanged, oldBackY1,
			    td->tdSelection.bsBegin.bpStroff+ stroffShift );

    return;
    }

/************************************************************************/
/*									*/
/*  Insert a page break in the document.				*/
/*									*/
/************************************************************************/

void tedDocInsertPageBreak(	Widget		option,
				XtPointer	voided,
				XtPointer	voidpbcs )
    {
    EditDocument *	ed= (EditDocument *)voided;
    const int		onNewPage= 1;

    tedSplitParagraph( ed, onNewPage );
    return;
    }


/************************************************************************/
/*									*/
/*  Insert a section break in the document.				*/
/*									*/
/************************************************************************/

void tedDocInsertSectBreak(	Widget		option,
				XtPointer	voided,
				XtPointer	voidpbcs )
    {
    EditDocument *	ed= (EditDocument *)voided;
    const int		onNewPage= 0;

    tedSplitParagraph( ed, onNewPage );
    return;
    }

