/************************************************************************/
/*									*/
/*  A Format Tool.							*/
/*									*/
/************************************************************************/

#   include	"config.h"

#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<debugon.h>

#   include	<X11/Xatom.h>
#   include	<X11/IntrinsicP.h>
#   include	<Xm/Text.h>
#   include	<Xm/Form.h>
#   include	<Xm/ToggleB.h>
#   include	<Xm/Frame.h>
#   include	<Xm/ScrolledW.h>
#   include	<Xm/Label.h>
#   include	<Xm/RowColumn.h>
#   include	<Xm/PanedW.h>

#   include	<Xm/MwmUtil.h>
#   include	<Xm/Protocols.h>

#   include	<appUtil.h>
#   include	<appGeoString.h>
#   include	<appUnit.h>

#   include	"tedApp.h"
#   include	"tedFormatTool.h"

void *	TED_Inspector;

/************************************************************************/
/*									*/
/*  Resources for the format tool.					*/
/*									*/
/************************************************************************/

typedef struct TedFormatToolResources
    {
    InspectorSubjectResources	tttrSubjectResources[TEDtsiCOUNT];

    TablePageResources		tttrTablePageResources;
    RowPageResources		tttrRowPageResources;
    ColumnPageResources		tttrColumnPageResources;
    ParagraphPageResources	tttrParaPageResources;
    SectionPageResources	tttrSectPageResources;
    } TedFormatToolResources;

# define xx(x)	x,x

static XtResource TED_TedFormatToolResourceTable[]=
    {
	/****************************************************************/
	/*  'Table' page.						*/
	/****************************************************************/
	{ xx("tableToolTable"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiTABLE].isrSubjectName),
		    XtRString, "Table" },
	{ xx("tableToolChangeTable"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiTABLE].isrApplyToSubject),
		    XtRString, "Apply to Table" },
	{ xx("tableToolRevert"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiTABLE].isrRevert),
		    XtRString, "Revert" },
	/*  no Next Table */
	/*  no Prev Table */
	{ xx("tableToolSelectTable"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiTABLE].isrSelectButtonText),
		    XtRString, "Select Table" },
	{ xx("tableToolDeleteTable"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiTABLE].isrDeleteButtonText),
		    XtRString, "Delete Table" },
	/*  no Insert Table */
	/*  no Append Table */
	/**/
	{ xx("tableToolCellMargin"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrTablePageResources.tprCellMargin),
		    XtRString, "Cell Margin" },
	{ xx("tableToolLeftIndent"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrTablePageResources.tprLeftIndent),
		    XtRString, "Left Margin" },


	/****************************************************************/
	/*  'Row' page.							*/
	/****************************************************************/
	{ xx("tableToolRow"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiROW].isrSubjectName),
		    XtRString, "Row" },
	{ xx("tableToolChangeRow"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiROW].isrApplyToSubject),
		    XtRString, "Apply to Row" },
	{ xx("tableToolRevert"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiROW].isrRevert),
		    XtRString, "Revert" },
	{ xx("tableToolNextRow"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiROW].isrNextButtonText),
		    XtRString, "Next Row" },
	{ xx("tableToolPreviousRow"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiROW].isrPrevButtonText),
		    XtRString, "Previous Row" },
	{ xx("tableToolSelectRow"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiROW].isrSelectButtonText),
		    XtRString, "Select Row" },
	{ xx("tableToolDeleteRow"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiROW].isrDeleteButtonText),
		    XtRString, "Delete Row" },
	{ xx("tableToolInsertRow"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiROW].isrInsertButtonText),
		    XtRString, "Insert Row" },
	{ xx("tableToolAppendRow"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiROW].isrAppendButtonText),
		    XtRString, "Append Row" },
	/**/
	{ xx("tableToolHeightFree"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrRowPageResources.rprHeightFree),
		    XtRString, "Height Free" },
	{ xx("tableToolHeightExactly"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrRowPageResources.rprHeightExactly),
		    XtRString, "Height Exactly" },
	{ xx("tableToolHeightAtLeast"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrRowPageResources.rprHeightAtLeast),
		    XtRString, "Height at Least" },
	/**/
	{ xx("tableToolCellMargin"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrRowPageResources.rprCellMargin),
		    XtRString, "Cell Margin" },
	{ xx("tableToolLeftIndent"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrRowPageResources.rprLeftIndent),
		    XtRString, "Left Margin" },
	/**/
	{ xx("tableToolRowTopBorder"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrRowPageResources.rprTopBorder),
		    XtRString, "Top Border" },
	{ xx("tableToolRowBottomBorder"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrRowPageResources.rprBottomBorder),
		    XtRString, "Bottom Border" },
	/**/
	{ xx("formatToolRowKeepOnPage"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrRowPageResources.rprKeepOnPage),
		    XtRString, "Keep on One Page" },
	{ xx("formatToolRowIsTableHeader"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrRowPageResources.rprIsTableHeader),
		    XtRString, "Is Table Header" },


	/****************************************************************/
	/*  'Column' page.						*/
	/****************************************************************/
	{ xx("tableToolColumn"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiCOLUMN].isrSubjectName),
		    XtRString, "Column" },
	{ xx("tableToolChangeColumn"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiCOLUMN].isrApplyToSubject),
		    XtRString, "Apply to Column" },
	{ xx("tableToolRevert"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiCOLUMN].isrRevert),
		    XtRString, "Revert" },
	{ xx("tableToolNextColumn"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiCOLUMN].isrNextButtonText),
		    XtRString, "Next Column" },
	{ xx("tableToolPreviousColumn"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiCOLUMN].isrPrevButtonText),
		    XtRString, "Previous Column" },
	{ xx("tableToolSelectColumn"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiCOLUMN].isrSelectButtonText),
		    XtRString, "Select Column" },
	{ xx("tableToolDeleteColumn"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiCOLUMN].isrDeleteButtonText),
		    XtRString, "Delete Column" },
	{ xx("tableToolInsertColumn"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiCOLUMN].isrInsertButtonText),
		    XtRString, "Insert Column" },
	{ xx("tableToolAppendColumn"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiCOLUMN].isrAppendButtonText),
		    XtRString, "Append Column" },
	/**/
	{ xx("tableToolWidth"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrColumnPageResources.cprWidth),
		    XtRString, "Width" },
	/**/
	{ xx("tableToolColumnLeftBorder"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrColumnPageResources.cprLeftBorder),
		    XtRString, "Left Border" },
	{ xx("tableToolColumnRightBorder"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrColumnPageResources.cprRightBorder),
		    XtRString, "Right Border" },


	/****************************************************************/
	/*  'Paragraph' page.						*/
	/****************************************************************/
	{ xx("formatToolPara"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiPARA].isrSubjectName),
		    XtRString, "Paragraph" },
	{ xx("formatToolChangePara"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiPARA].isrApplyToSubject),
		    XtRString, "Apply to Paragraph" },
	{ xx("tableToolRevert"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiPARA].isrRevert),
		    XtRString, "Revert" },
	{ xx("formatToolNextPara"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiPARA].isrNextButtonText),
		    XtRString, "Next Paragraph" },
	{ xx("formatToolPreviousPara"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiPARA].isrPrevButtonText),
		    XtRString, "Previous Paragraph" },
	{ xx("formatToolSelectPara"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiPARA].isrSelectButtonText),
		    XtRString, "Select Paragraph" },
	{ xx("formatToolDeletePara"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiPARA].isrDeleteButtonText),
		    XtRString, "Delete Paragraph" },
	{ xx("formatToolInsertPara"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiPARA].isrInsertButtonText),
		    XtRString, "Insert Paragraph" },
	{ xx("formatToolAppendPara"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiPARA].isrAppendButtonText),
		    XtRString, "Append Paragraph" },
	/**/
	{ xx("formatToolParaFirstIndent"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrParaPageResources.pprParaFirstIndent),
		    XtRString, "First Line Indent" },
	{ xx("formatToolParaLeftIndent"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrParaPageResources.pprParaLeftIndent),
		    XtRString, "Left Indent" },
	{ xx("formatToolParaRightIndent"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrParaPageResources.pprParaRightIndent),
		    XtRString, "Right Indent" },
	/**/
	{ xx("formatToolParaSpaceAbove"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrParaPageResources.pprParaSpaceAbove),
		    XtRString, "Space Above" },
	{ xx("formatToolParaSpaceBelow"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrParaPageResources.pprParaSpaceBelow),
		    XtRString, "Space Below" },
	/**/
	{ xx("formatToolParaTopBorder"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrParaPageResources.pprTopBorder),
		    XtRString, "Top Border" },
	{ xx("formatToolParaBottomBorder"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrParaPageResources.pprBottomBorder),
		    XtRString, "Bottom Border" },
	/**/
	{ xx("formatToolParaOnNewPage"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrParaPageResources.pprOnNewPage),
		    XtRString, "Start on New Page" },
	{ xx("formatToolParaOnOnePage"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrParaPageResources.pprOnOnePage),
		    XtRString, "Keep on One Page" },
	/**/
	{ xx("formatToolLineDistFree"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrParaPageResources.pprParaLineDistFree),
		    XtRString, "Spacing Free" },
	{ xx("formatToolLineDistExactly"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrParaPageResources.pprParaLineDistExactly),
		    XtRString, "Spacing Exactly" },
	{ xx("formatToolLineDistAtLeast"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrParaPageResources.pprParaLineDistAtLeast),
		    XtRString, "Spacing at Least" },
	/**/
	{ xx("formatToolParaAlignment"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrParaPageResources.pprParaAlignment),
		XtRString, "Alignment" },
	{ xx("formatToolParaAlignLeft"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrParaPageResources.pprParaAlignMenuTexts[DOCiaLEFT]),
		XtRString, "Left" },
	{ xx("formatToolParaAlignRight"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrParaPageResources.pprParaAlignMenuTexts[DOCiaRIGHT]),
		XtRString, "Right" },
	{ xx("formatToolParaAlignCentered"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrParaPageResources.pprParaAlignMenuTexts[DOCiaCENTERED]),
		XtRString, "Centered" },
	{ xx("formatToolParaAlignJustified"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrParaPageResources.pprParaAlignMenuTexts[DOCiaJUSTIFIED]),
		XtRString, "Justified" },

	/****************************************************************/
	/*  'Section' page.						*/
	/****************************************************************/
	{ xx("formatToolSect"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiSECT].isrSubjectName),
		    XtRString, "Section" },
	{ xx("formatToolChangeSect"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiSECT].isrApplyToSubject),
		    XtRString, "Apply to Section" },
	{ xx("tableToolRevert"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiSECT].isrRevert),
		    XtRString, "Revert" },
	{ xx("formatToolNextSect"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiSECT].isrNextButtonText),
		    XtRString, "Next Section" },
	{ xx("formatToolPreviousSect"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiSECT].isrPrevButtonText),
		    XtRString, "Previous Section" },
	{ xx("formatToolSelectSect"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiSECT].isrSelectButtonText),
		    XtRString, "Select Section" },
	{ xx("formatToolDeleteSect"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiSECT].isrDeleteButtonText),
		    XtRString, "Delete Section" },
	{ xx("formatToolInsertSect"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiSECT].isrInsertButtonText),
		    XtRString, "Insert Section" },
	{ xx("formatToolAppendSect"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSubjectResources[TEDtsiSECT].isrAppendButtonText),
		    XtRString, "Append Section" },
	/**/
	{ xx("formatToolSectPageWide"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSectPageResources.sprPageWide),
		    XtRString, "Page Width" },
	{ xx("formatToolSectPageHigh"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSectPageResources.sprPageHigh),
		    XtRString, "Page Height" },
	/**/
	{ xx("formatToolSectLeftMargin"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSectPageResources.sprLeftMargin),
		    XtRString, "Left Margin" },
	{ xx("formatToolSectRightMargin"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSectPageResources.sprRightMargin),
		    XtRString, "Right Margin" },
	{ xx("formatToolSectTopMargin"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSectPageResources.sprTopMargin),
		    XtRString, "Top Margin" },
	{ xx("formatToolSectBottomMargin"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSectPageResources.sprBottomMargin),
		    XtRString, "Bottom Margin" },
	{ xx("formatToolSectHeaderY"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSectPageResources.sprHeaderY),
		    XtRString, "Header Position" },
	{ xx("formatToolSectFooterY"), XtRString, sizeof(char *),
		    offsetof(TedFormatToolResources,
		    tttrSectPageResources.sprFooterY),
		    XtRString, "Footer Position" },
	/**/
	{ xx("formatToolSectPageNumberStyle"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrSectPageResources.sprPageNumberStyle),
		XtRString, "Page Number Style" },
	{ xx("formatToolSectNumberStyleDec"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrSectPageResources.sprNumberStyleMenuTexts[DOCpgnDEC]),
		XtRString, "1, 2, 3" },
	{ xx("formatToolSectNumberStyleUcrm"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrSectPageResources.sprNumberStyleMenuTexts[DOCpgnUCRM]),
		XtRString, "I, II, III" },
	{ xx("formatToolSectNumberStyleLcrm"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrSectPageResources.sprNumberStyleMenuTexts[DOCpgnLCRM]),
		XtRString, "i, ii, iii" },
	{ xx("formatToolSectNumberStyleUcltr"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrSectPageResources.sprNumberStyleMenuTexts[DOCpgnUCLTR]),
		XtRString, "A, B, C" },
	{ xx("formatToolSectNumberStyleLcltr"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrSectPageResources.sprNumberStyleMenuTexts[DOCpgnLCLTR]),
		XtRString, "a, b, c" },
	/**/
	{ xx("formatToolSectBreakKind"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrSectPageResources.sprBreakKind),
		XtRString, "Begins" },
	{ xx("formatToolSectBreakKindNone"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrSectPageResources.sprBreakKindMenuTexts[DOCsbkNONE]),
		XtRString, "Below Previous" },
	{ xx("formatToolSectBreakKindCol"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrSectPageResources.sprBreakKindMenuTexts[DOCsbkCOL]),
		XtRString, "In Next Column" },
	{ xx("formatToolSectBreakKindPage"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrSectPageResources.sprBreakKindMenuTexts[DOCsbkPAGE]),
		XtRString, "On New Page" },
	{ xx("formatToolSectBreakKindEven"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrSectPageResources.sprBreakKindMenuTexts[DOCsbkEVEN]),
		XtRString, "On Even Page" },
	{ xx("formatToolSectBreakKindOdd"), XtRString, sizeof(char *),
		offsetof(TedFormatToolResources,
		tttrSectPageResources.sprBreakKindMenuTexts[DOCsbkODD]),
		XtRString, "On Odd Page" },

    };

/************************************************************************/
/*									*/
/*  Update a 'HasBorder' toggle.					*/
/*									*/
/************************************************************************/

void tedFormatReflectBorder(	Widget				toggle,
				const BorderProperties *	bp )
    {
    int		set= False;

    if  ( bp->bpWidthTwips > 0	&&
	  bp->bpIsSet		)
	{ set= True;	}

    XmToggleButtonSetState( toggle, set, False );

    return;
    }

/************************************************************************/
/*									*/
/*  Refresh a height chooser.						*/
/*									*/
/************************************************************************/

void tedFormatRefreshHeightChooser(	HeightChooser *	hc,
					int		height )
    {
    char		scratch[50];

    if  ( height == 0 )
	{
	appEnableText( hc->hcText, 0 );

	XtVaSetValues( hc->hcMenu,
			XmNmenuHistory,		hc->hcFreeItem,
			NULL );

	XmTextSetString( hc->hcText, "" );
	}
    else{
	appEnableText( hc->hcText, 1 );

	if  ( height > 0 )
	    {
	    XtVaSetValues( hc->hcMenu,
			XmNmenuHistory,		hc->hcAtLeastItem,
			NULL );

	    appGeoLengthToString( scratch,  height, UNITtyPOINTS );
	    }
	else{
	    XtVaSetValues( hc->hcMenu,
			XmNmenuHistory,		hc->hcExactlyItem,
			NULL );

	    appGeoLengthToString( scratch, -height, UNITtyPOINTS );
	    }

	XmTextSetString( hc->hcText, scratch );
	}
    }

int tedFormatToolGetHeight(	int *		pHeight,
				HeightChooser *	hc )
    {
    Widget		item= (Widget)0;

    int			sign;

    int			height;
    int			changed;
    const int		requirePositive= 1;

    XtVaGetValues( hc->hcMenu,
			    XmNmenuHistory,		&item,
			    NULL );

    if  ( item == hc->hcFreeItem )
	{ sign= height= 0;	}
    else{
	if  ( item == hc->hcAtLeastItem )
	    { sign= 1;	}
	else{
	    if  ( item == hc->hcExactlyItem )
		{ sign= -1;			}
	    else{ XDEB(item); return -1;	}
	    }
	}

    height= *pHeight;
    height *= sign;

    if  ( sign != 0 )
	{
	if  ( appGetLengthFromTextWidget( hc->hcText,
			&height, &changed, UNITtyPOINTS, requirePositive ) )
	    { return -1; }

	height *= sign;
	}

    *pHeight= height;

    return 0;
    }

int tedFormatToolGetGapWidth(		const RowProperties *	rp,
					int *			pValue,
					Widget			w )
    {
    int				value;
    int				changed;

    int				minWidth;
    const CellProperties *	cp;
    int				col;

    value= rp->rpHalfGapWidthTwips;
    if  ( appGetLengthFromTextWidget( w, &value, &changed, UNITtyPOINTS, 1 ) )
	{ return -1;	}

    cp= rp->rpCells;
    minWidth= cp->cpRightBoundaryTwips- rp->rpLeftIndentTwips;

    for ( col= 1; col < rp->rpCellCount; cp++, col++ )
	{
	int	cellWidth;

	cellWidth= cp[1].cpRightBoundaryTwips- cp[0].cpRightBoundaryTwips;

	if  ( minWidth > cellWidth )
	    { minWidth=  cellWidth;	}
	}

    if  ( 2* value >= minWidth )
	{
	char		scratch[50];

	value= minWidth/3;

	appGeoLengthToString( scratch, value, UNITtyPOINTS );
	XmTextSetString( w, scratch );
	appRefuseTextValue( w );
	return 1;
	}

    return 0;
    }

int tedFormatToolGetRowLeftIndent(	const RowProperties *	rp,
					const int		pageLeftMargin,
					int *			pValue,
					Widget			w )
    {
    int				value;
    int				changed;

    char			scratch[50];

    const int			requirePositive= 0;

    value= rp->rpLeftIndentTwips;
    if  ( appGetLengthFromTextWidget( w, &value, &changed,
					    UNITtyPOINTS, requirePositive ) )
	{ return -1;	}

    if  ( value < -pageLeftMargin )
	{
	value= -pageLeftMargin+ 10;

	appGeoLengthToString( scratch, value, UNITtyPOINTS );
	XmTextSetString( w, scratch );
	appRefuseTextValue( w );
	return 1;
	}

    if  ( value+ 2* rp->rpHalfGapWidthTwips >=
					rp->rpCells[0].cpRightBoundaryTwips )
	{
	value= rp->rpCells[0].cpRightBoundaryTwips- 3* rp->rpHalfGapWidthTwips;

	if  ( value < -pageLeftMargin )
	    { appRefuseTextValue( w ); return -1;	}

	appGeoLengthToString( scratch, value, UNITtyPOINTS );
	XmTextSetString( w, scratch );
	appRefuseTextValue( w );
	return 1;
	}

    *pValue= value; return 0;
    }

/************************************************************************/
/*									*/
/*  One of the border on/off toggles has been activated by a user.	*/
/*									*/
/************************************************************************/

void tedFormatToggleBorder(	BorderProperties *	bp,
				int			set )
    {
    if  ( set )
	{
	bp->bpIsSet= 1;

	if  ( bp->bpWidthTwips == 0 )
	    { bp->bpWidthTwips= 15;	}
	}
    else{ bp->bpIsSet= 0;		}
    }

/************************************************************************/
/*									*/
/*  Make the form with the two buttons.					*/
/*									*/
/************************************************************************/

void tedFormatHeightChosen(	Widget			w,
				HeightChooser *		hc,
				int			defaultValue )
    {
    char		scratch[50];
    int			val;

    if  ( w == hc->hcFreeItem )
	{
	appEnableText( hc->hcText, 0 );

	XmTextSetString( hc->hcText, "" );

	return;
	}

    if  ( w == hc->hcAtLeastItem )
	{
	appEnableText( hc->hcText, 1 );

	val= defaultValue;

	if  ( val < 0 )
	    { val= -val;	}

	appGeoLengthToString( scratch, val, UNITtyPOINTS );

	XmTextSetString( hc->hcText, scratch );

	return;
	}

    if  ( w == hc->hcExactlyItem )
	{
	appEnableText( hc->hcText, 1 );

	val= defaultValue;

	appGeoLengthToString( scratch, val, UNITtyPOINTS );

	XmTextSetString( hc->hcText, scratch );

	return;
	}

    LDEB(1);
    }

void tedFormatMakeHeightRow(	Widget *		pRow,
				void *			through,
				Widget			parent,
				HeightChooser *		hc,
				XtCallbackProc		callback )
{
    Arg		al[20];
    int		ac= 0;

    Widget	row;
    Widget	menuForm;
    Widget	pulldown;
    Widget	menu;
    Widget	text;

    row= appMakeButtonRow( parent, 2 );

    /**************/
    ac= 0;
    XtSetArg( al[ac], XmNtopAttachment,		XmATTACH_FORM ); ac++;

    XtSetArg( al[ac], XmNleftAttachment,	XmATTACH_POSITION ); ac++;
    XtSetArg( al[ac], XmNleftPosition,		0 ); ac++;

    XtSetArg( al[ac], XmNrightAttachment,	XmATTACH_POSITION ); ac++;
    XtSetArg( al[ac], XmNrightPosition,		1 ); ac++;

    menuForm= XmCreateForm( row, WIDGET_NAME, al, ac );

    appMakePulldownList( &pulldown, &menu, menuForm );

    XtVaSetValues( pulldown,
			XmNresizeWidth,		True,
			NULL );

    XtVaSetValues( menu,
			XmNleftOffset,		0,
			XmNrightOffset,		3,
			NULL );
    /**************/
    appMakeRowText( &text, row, 10, 0 );

    if  ( callback )
	{ XtAddCallback( text, XmNactivateCallback, callback, through ); }

    /* NO: XtManageChild( pulldown ); */

    XtManageChild( menu );
    XtManageChild( menuForm );
    XtManageChild( row );

    *pRow= row;
    hc->hcMenu= menu;
    hc->hcPulldown= pulldown;
    hc->hcText= text;

    return;
    }

static void tedTableFillPages(	TedFormatToolResources *	tftr,
				TedFormatTool *			tft,
				AppInspector *			ai )
    {
    tedFormatFillTablePage( &(tft->tftTableTool),
			    &(tftr->tttrTablePageResources),
			    &(ai->aiSubjects[TEDtsiTABLE]),
			    ai->aiSubjects[TEDtsiTABLE].isPagePaned,
			    &(tftr->tttrSubjectResources[TEDtsiTABLE]) );

    tedFormatFillRowPage( &(tft->tftRowTool),
			    &(tftr->tttrRowPageResources),
			    &(ai->aiSubjects[TEDtsiROW]),
			    ai->aiSubjects[TEDtsiROW].isPagePaned,
			    &(tftr->tttrSubjectResources[TEDtsiROW]) );

    tedFormatFillColumnPage( &(tft->tftColumnTool),
			    &(tftr->tttrColumnPageResources),
			    &(ai->aiSubjects[TEDtsiCOLUMN]),
			    ai->aiSubjects[TEDtsiCOLUMN].isPagePaned,
			    &(tftr->tttrSubjectResources[TEDtsiCOLUMN]) );

    tedFormatFillParagraphPage( &(tft->tftParagraphTool),
			    &(tftr->tttrParaPageResources),
			    &(ai->aiSubjects[TEDtsiPARA]),
			    ai->aiSubjects[TEDtsiPARA].isPagePaned,
			    &(tftr->tttrSubjectResources[TEDtsiPARA]) );

    tedFormatFillSectionPage( &(tft->tftSectionTool),
			    &(tftr->tttrSectPageResources),
			    &(ai->aiSubjects[TEDtsiSECT]),
			    ai->aiSubjects[TEDtsiSECT].isPagePaned,
			    &(tftr->tttrSubjectResources[TEDtsiSECT]) );

    return;
    }

/************************************************************************/
/*  Fill the menu in a height chooser.					*/
/************************************************************************/

void tedFormatFillHeightChooser(	HeightChooser *		hc,
					XtCallbackProc		callback,
					void *			voidtft,
					const char *		freeText,
					const char *		atLeastText,
					const char *		exactlyText )
    {
    Dimension		width;

    XtVaGetValues( hc->hcMenu,
			XmNwidth,		&width,
			NULL );

    appEmptyPulldownList( hc->hcPulldown );

    hc->hcFreeItem= appPulldownMakeOption( hc->hcPulldown,
				    freeText, width, callback, voidtft );

    hc->hcAtLeastItem= appPulldownMakeOption( hc->hcPulldown,
				    atLeastText, width, callback, voidtft );

    hc->hcExactlyItem= appPulldownMakeOption( hc->hcPulldown,
				    exactlyText, width, callback, voidtft );

    XtVaSetValues( hc->hcMenu,
			    XmNmenuHistory,	hc->hcFreeItem,
			    NULL );

    appPulldownSetWidth( hc->hcMenu, width );
    }

/************************************************************************/
/*									*/
/*  The user destroys the table tool.					*/
/*									*/
/************************************************************************/

static void tedDestroyFormatTool(	void *	voidtft )
    {
    TedFormatTool *	tft= (TedFormatTool *)voidtft;
    ParagraphTool *	pt= &(tft->tftParagraphTool);
    RowTool *		rt= &(tft->tftRowTool);
    ColumnTool *	ct= &(tft->tftColumnTool);
    TableTool *		tt= &(tft->tftTableTool);

    docCleanParagraphProperties( &(pt->ptPropertiesSet) );
    docCleanParagraphProperties( &(pt->ptPropertiesChosen) );

    docCleanRowProperties( &(rt->rtPropertiesSet) );
    docCleanRowProperties( &(rt->rtPropertiesChosen) );

    docCleanRowProperties( &(ct->ctPropertiesSet) );
    docCleanRowProperties( &(ct->ctPropertiesChosen) );

    docCleanRowProperties( &(tt->ttPropertiesSet) );
    docCleanRowProperties( &(tt->ttPropertiesChosen) );

    free( tft );

    TED_Inspector= (void *)0;

    return;
    }

/************************************************************************/
/*									*/
/*  make/show the format tool.						*/
/*									*/
/************************************************************************/

void tedShowFormatTool(	Widget			option,
			EditApplication *	ea,
			const char *		widgetName,
			const char *		pixmapName )
    {
    AppInspector *			ai;
    TedFormatTool *			tft;
   	 
    static TedFormatToolResources	tftr;
    static int				gotResources;

    if  ( TED_Inspector )
	{
	ai= (AppInspector *)TED_Inspector;

	XtVaSetValues( ai->aiTopWidget, XmNinitialState, NormalState, NULL );
	XtMapWidget( ai->aiTopWidget );
	XRaiseWindow( XtDisplay( ai->aiTopWidget ),
					    XtWindow( ai->aiTopWidget ) );
	return;
	}

    if  ( ! gotResources )
	{
	XtGetApplicationResources( ea->eaTopWidget, (void *)&tftr,
	    TED_TedFormatToolResourceTable,
	    XtNumber(TED_TedFormatToolResourceTable),
	    NULL, 0 );

	gotResources= 1;
	}

    tft= (TedFormatTool *)malloc( sizeof(TedFormatTool) );
    if  ( ! tft )
	{ XDEB(tft); return;	}

    tft->tftSectionTool.stApplication= ea;
    tft->tftParagraphTool.ptApplication= ea;
    tft->tftColumnTool.ctApplication= ea;
    tft->tftRowTool.rtApplication= ea;
    tft->tftTableTool.ttApplication= ea;

    ai= appMakeInspector( ea, option, pixmapName, widgetName,
				    tftr.tttrSubjectResources, TEDtsiCOUNT,
				    tedDestroyFormatTool, (void *)tft );
    tedTableFillPages( &tftr, tft, ai );

    tedFormatFillRowChoosers( &(tft->tftRowTool),
					    &(tftr.tttrRowPageResources) );
    tedFormatFillParagraphChoosers( &(tft->tftParagraphTool),
					    &(tftr.tttrParaPageResources) );
    tedFormatFillSectionChoosers( &(tft->tftSectionTool),
					    &(tftr.tttrSectPageResources) );

    appFinishInspector( ai );

    tedFormatFinishRowPage( &(tft->tftRowTool),
					    &(tftr.tttrRowPageResources) );
    tedFormatFinishParagraphPage( &(tft->tftParagraphTool),
					    &(tftr.tttrParaPageResources) );
    tedFormatFinishSectionPage( &(tft->tftSectionTool),
					    &(tftr.tttrSectPageResources) );

    TED_Inspector= (void *)ai;

    return;
    }

/************************************************************************/
/*									*/
/*  Ask the table tool to adapt itself to the selection.		*/
/*									*/
/************************************************************************/

void tedFormatToolAdaptToSelection( void *			voidai,
				const BufferSelection *		bs,
				int				fileReadonly,
				const DocumentGeometry *	dg )
    {
    AppInspector *	ai= (AppInspector *)voidai;
    TedFormatTool *	tft= (TedFormatTool *)ai->aiTarget;
    const int		andMenu= 1;

    int			col0;
    int			col1;
    int			col11;
    int			row0;
    int			row1;
    int			row00;
    int			row11;
    BufferItem *	sectBi;

    Boolean		tableRectangle= False;

    InspectorSubject *	is;
    int			subject;

    int			pageRight;
    int			pageLeftMargin;

    pageRight= dg->dgPaperWideTwips-
			    dg->dgLeftMarginTwips- dg->dgRightMarginTwips;
    pageLeftMargin= dg->dgLeftMarginTwips;

    tft->tftRowTool.rtPageRight= pageRight;
    tft->tftRowTool.rtPageLeftMargin= pageLeftMargin;

    tft->tftColumnTool.ctPageRight= pageRight;
    tft->tftColumnTool.ctPageLeftMargin= pageLeftMargin;

    tft->tftTableTool.ttPageRight= pageRight;
    tft->tftTableTool.ttPageLeftMargin= pageLeftMargin;

    if  ( ! bs->bsBegin.bpBi )
	{ appEnableInspector( voidai, False ); return; }
    else{
	ParagraphTool *			pt= &(tft->tftParagraphTool);
	SectionTool *			st= &(tft->tftSectionTool);

	tedFormatToolRefreshParagraphTool( pt,
					    ai->aiSubjects+ TEDtsiPARA, bs );

	tedFormatToolRefreshSectionTool( st,
					    ai->aiSubjects+ TEDtsiSECT, bs );

	if  ( fileReadonly )
	    { appEnableInspector( voidai, False );	}
	else{ appEnableInspector( voidai, True );	}
	}

    is= ai->aiSubjects;
    for ( subject= 0; subject < TEDtsiCOUNT; is++, subject++ )
	{ XtSetSensitive( is->isPageFrame, fileReadonly == 0 );	}

    if  ( ! tedGetTableRectangle( bs, &sectBi, &col0, &col1, &col11,
					    &row00, &row0, &row1, &row11 ) )
	{ tableRectangle= True;	}

    if  ( tableRectangle )
	{
	TableTool *	tt= &(tft->tftTableTool);
	ColumnTool *	ct= &(tft->tftColumnTool);
	RowTool *	rt= &(tft->tftRowTool);

	ct->ctCol0=	rt->rtCol0=	tt->ttCol0=	col0;
	ct->ctCol1=	rt->rtCol1=	tt->ttCol1=	col1;
	ct->ctRow00=	rt->rtRow00=	tt->ttRow00=	row00;
	ct->ctRow0=	rt->rtRow0=	tt->ttRow0=	row0;
	ct->ctRow1=	rt->rtRow1=	tt->ttRow1=	row1;
	ct->ctRow11=	rt->rtRow11=	tt->ttRow11=	row11;

	if  ( tedFormatSaveRowProperties( tft, sectBi, row0 ) )
	    { LDEB(1); return;	}

	tedFormatToolRefreshTablePage( tt );

	appEnableInspectorSubject( ai, TEDtsiTABLE, True );

	if  ( col1 == col0 )
	    {
	    tedFormatToolRefreshColumnPage( ct );

	    appEnableInspectorSubject( ai, TEDtsiCOLUMN, True );

	    is= ai->aiSubjects+ TEDtsiCOLUMN;
	    XtSetSensitive( is->isPrevButton, col0 > 0 );
	    XtSetSensitive( is->isNextButton, col1 < col11 );
	    }
	else{
	    appEnableInspectorSubject( ai, TEDtsiCOLUMN, False );
	    }

	if  ( row1 == row0 )
	    {
	    appEnableInspectorSubject( ai, TEDtsiROW, True );

	    tedFormatToolRefreshRowPage( rt );

	    is= ai->aiSubjects+ TEDtsiROW;
	    XtSetSensitive( is->isPrevButton, row0 > row00 );
	    XtSetSensitive( is->isNextButton, row1 < row11 );
	    }
	else{
	    appEnableInspectorSubject( ai, TEDtsiROW, False );
	    }

	if  ( ai->aiCurrentSubject < 0 )
	    { appInspectorSelectSubject( ai, andMenu, TEDtsiTABLE );	}
	else{
	    is= ai->aiSubjects+ ai->aiCurrentSubject;

	    if  ( ! XtIsSensitive( is->isItem ) )
		{ appInspectorSelectSubject( ai, andMenu, TEDtsiTABLE ); }
	    }
	}
    else{
	appEnableInspectorSubject( ai, TEDtsiTABLE, False );
	appEnableInspectorSubject( ai, TEDtsiROW, False );
	appEnableInspectorSubject( ai, TEDtsiCOLUMN, False );

	if  ( ai->aiCurrentSubject != TEDtsiSECT )
	    { appInspectorSelectSubject( ai, andMenu, TEDtsiPARA );	}
	}
    }
