/************************************************************************/
/*									*/
/*  Format tool:							*/
/*									*/
/************************************************************************/

/************************************************************************/
/*									*/
/*  Resources for the 'Row' page.					*/
/*									*/
/************************************************************************/

typedef struct RowPageResources
    {
    char *		rprHeightFree;
    char *		rprHeightAtLeast;
    char *		rprHeightExactly;

    char *		rprTopBorder;
    char *		rprBottomBorder;

    char *		rprCellMargin;
    char *		rprLeftIndent;

    char *		rprKeepOnPage;
    char *		rprIsTableHeader;
    } RowPageResources;

/************************************************************************/
/*									*/
/*  Resources for the 'Column' page.					*/
/*									*/
/************************************************************************/

typedef struct ColumnPageResources
    {
    char *		cprWidth;
    char *		cprLeftBorder;
    char *		cprRightBorder;
    } ColumnPageResources;

/************************************************************************/
/*									*/
/*  Resources for the 'Table' page.					*/
/*									*/
/************************************************************************/

typedef struct TablePageResources
    {
    char *		tprCellMargin;
    char *		tprLeftIndent;
    } TablePageResources;

/************************************************************************/
/*									*/
/*  Resources for the 'Section; page.					*/
/*									*/
/************************************************************************/

typedef struct SectionPageResources
    {
    char *		sprPageWide;
    char *		sprPageHigh;

    char *		sprLeftMargin;
    char *		sprRightMargin;
    char *		sprTopMargin;
    char *		sprBottomMargin;
    char *		sprHeaderY;
    char *		sprFooterY;

    char *		sprPageNumberStyle;
    char *		sprNumberStyleMenuTexts[DOCpgn_COUNT];

    char *		sprBreakKind;
    char *		sprBreakKindMenuTexts[DOCsbk_COUNT];
    } SectionPageResources;

/************************************************************************/
/*									*/
/*  Resources for the 'Paragraph; page.					*/
/*									*/
/************************************************************************/

typedef struct ParagraphPageResources
    {
    char *		pprParaFirstIndent;
    char *		pprParaLeftIndent;
    char *		pprParaRightIndent;

    char *		pprParaSpaceAbove;
    char *		pprParaSpaceBelow;

    char *		pprOnNewPage;
    char *		pprOnOnePage;

    char *		pprParaLineDistFree;
    char *		pprParaLineDistAtLeast;
    char *		pprParaLineDistExactly;

    char *		pprTopBorder;
    char *		pprBottomBorder;

    char *		pprParaAlignment;
    char *		pprParaAlignMenuTexts[DOCia_COUNT];
    } ParagraphPageResources;

/************************************************************************/
/*									*/
/*  A compination of a pulldown menu and a text widget to manage a	*/
/*  height property.							*/
/*									*/
/************************************************************************/

typedef struct HeightChooser
    {
    Widget	hcMenu;
    Widget	hcPulldown;
    Widget	hcText;

    Widget	hcFreeItem;
    Widget	hcExactlyItem;
    Widget	hcAtLeastItem;
    } HeightChooser;

/************************************************************************/
/*									*/
/*  A section tool, i.e. the 'Section' page of the format tool.		*/
/*									*/
/************************************************************************/

typedef struct SectionTool
    {
    EditApplication *		stApplication;

    SectionProperties		stPropertiesSet;
    SectionProperties		stPropertiesChosen;

    int				stSectionNumber;

    Widget			stNumberText;

    Widget			stPageWideText;
    Widget			stPageHighText;

    Widget			stLeftMarginText;
    Widget			stRightMarginText;
    Widget			stTopMarginText;
    Widget			stBottomMarginText;
    Widget			stHeaderYText;
    Widget			stFooterYText;

    Widget			stNumberStyleMenu;
    Widget			stNumberStylePulldown;
    Widget			stNumberStyleItems[DOCpgn_COUNT];

    Widget			stBreakKindMenu;
    Widget			stBreakKindPulldown;
    Widget			stBreakKindItems[DOCsbk_COUNT];
    } SectionTool;

/************************************************************************/
/*									*/
/*  A paragraph tool, i.e. the 'Paragraph' page of the format tool.	*/
/*									*/
/************************************************************************/

typedef struct ParagraphTool
    {
    EditApplication *		ptApplication;

    ParagraphProperties		ptPropertiesSet;
    ParagraphProperties		ptPropertiesChosen;

    Widget			ptFirstIndentText;
    Widget			ptRightIndentText;
    Widget			ptLeftIndentText;

    Widget			ptOnNewPageToggle;
    Widget			ptOnOnePageToggle;

    Widget			ptSpaceAboveToggle;
    Widget			ptSpaceAboveText;

    Widget			ptSpaceBelowToggle;
    Widget			ptSpaceBelowText;

    Widget			ptTopBorderToggle;
    Widget			ptBottomBorderToggle;

    HeightChooser		ptLineDistChooser;

    Widget			ptAlignMenu;
    Widget			ptAlignPulldown;
    Widget			ptAlignItems[DOCia_COUNT];
    } ParagraphTool;

/************************************************************************/
/*									*/
/*  A row tool, i.e. the 'Row' page of the format tool.			*/
/*									*/
/************************************************************************/

typedef struct RowTool
    {
    EditApplication *		rtApplication;

    RowProperties		rtPropertiesSet;
    RowProperties		rtPropertiesChosen;

    int				rtCol0;
    int				rtCol1;
    int				rtRow00;
    int				rtRow0;
    int				rtRow1;
    int				rtRow11;

    int				rtPageRight;
    int				rtPageLeftMargin;

    Widget			rtNumberText;
    HeightChooser		rtHeightChooser;
    Widget			rtCellMarginText;
    Widget			rtLeftIndentText;

    Widget			rtHeightFreeItem;
    Widget			rtHeightAtLeastItem;
    Widget			rtHeightExactlyItem;

    Widget			rtTopBorderToggle;
    Widget			rtBottomBorderToggle;

    Widget			rtKeepOnPageToggle;
    Widget			rtIsTableHeaderToggle;
    } RowTool;

/************************************************************************/
/*									*/
/*  A column tool, i.e. the 'Column' page of the format tool.		*/
/*									*/
/************************************************************************/

typedef struct ColumnTool
    {
    EditApplication *		ctApplication;

    int				ctCol0;
    int				ctCol1;
    int				ctRow00;
    int				ctRow0;
    int				ctRow1;
    int				ctRow11;

    int				ctPageRight;
    int				ctPageLeftMargin;

    RowProperties		ctPropertiesSet;
    RowProperties		ctPropertiesChosen;

    Widget			ctNumberText;
    Widget			ctWidthText;

    Widget			ctLeftBorderToggle;
    Widget			ctRightBorderToggle;
    } ColumnTool;

/************************************************************************/
/*									*/
/*  A table tool, i.e. the 'Table' page of the format tool.		*/
/*									*/
/************************************************************************/

typedef struct TableTool
    {
    EditApplication *		ttApplication;

    RowProperties		ttPropertiesSet;
    RowProperties		ttPropertiesChosen;

    int				ttCol0;
    int				ttCol1;
    int				ttRow00;
    int				ttRow0;
    int				ttRow1;
    int				ttRow11;

    int				ttPageRight;
    int				ttPageLeftMargin;

    Widget			ttCellMarginText;
    Widget			ttLeftIndentText;
    } TableTool;

/************************************************************************/
/*									*/
/*  The different pages in a Format tool:				*/
/*									*/
/************************************************************************/

typedef enum ToolSubjectIndex
    {
    TEDtsiTABLE= 0,
    TEDtsiROW,
    TEDtsiCOLUMN,
#   if 0
    TEDtsiCELL,
#   endif
    TEDtsiPARA,
    TEDtsiSECT,

    TEDtsiCOUNT

    } ToolSubjectIndex;

/************************************************************************/
/*									*/
/*  Represents a format tool.						*/
/*									*/
/************************************************************************/

typedef struct TedFormatTool
    {
    TableTool			tftTableTool;
    ColumnTool			tftColumnTool;
    RowTool			tftRowTool;
    ParagraphTool		tftParagraphTool;
    SectionTool			tftSectionTool;
    } TedFormatTool;

/************************************************************************/
/*									*/
/*  Routine declarations.						*/
/*									*/
/************************************************************************/

extern void tedFormatRefreshHeightChooser(	HeightChooser *	hc,
						int		height );

extern void tedFormatHeightChosen(	Widget			w,
					HeightChooser *		hc,
					int			defaultValue );

extern void tedFormatMakeHeightRow(	Widget *		pRow,
					void *			through,
					Widget			parent,
					HeightChooser *		hc,
					XtCallbackProc		callback );

extern void tedFormatFillParagraphPage(	ParagraphTool *			pt,
					const ParagraphPageResources * ppr,
					InspectorSubject *			is,
					Widget				pageWidget,
					const InspectorSubjectResources *	fpr );

extern void tedFormatFinishParagraphPage( ParagraphTool *		pt,
					const ParagraphPageResources *	ppr );

extern void tedFormatFinishRowPage( 	RowTool *			rt,
					const RowPageResources *	rpr );

extern void tedFormatFillParagraphChoosers( ParagraphTool *		pt,
					const ParagraphPageResources *	ppr );

extern void tedFormatFillHeightChooser(	HeightChooser *		hc,
					XtCallbackProc		callback,
					void *			voidtft,
					const char *		freeText,
					const char *		atLeastText,
					const char *		exactlyText );

extern int tedFormatToolGetHeight(	int *		pHeight,
					HeightChooser *	hc );

extern void tedFormatReflectBorder(	Widget				toggle,
					const BorderProperties *	bp );

extern void tedFormatFillRowPage( RowTool *			rt,
				const RowPageResources *	rpr,
				InspectorSubject *		is,
				Widget				pageWidget,
				const InspectorSubjectResources * isr );

extern void tedFormatFillRowChoosers(	RowTool *			rt,
					const RowPageResources *	rpr );

extern void tedFormatToolRefreshParagraphTool(	ParagraphTool *		pt,
						InspectorSubject *	is,
						const BufferSelection *	bs );

extern void tedFormatToolRefreshSectionTool(	SectionTool *		st,
						InspectorSubject *	is,
						const BufferSelection *	bs );

extern void tedFormatFillTablePage( TableTool *			tt,
				TablePageResources *		tpr,
				InspectorSubject *		is,
				Widget				pageWidget,
				const InspectorSubjectResources * isr );

extern int tedFormatToolGetRowLeftIndent( const RowProperties *	rp,
					const int		pageLeftMargin,
					int *			pValue,
					Widget			w );

extern int tedFormatToolGetGapWidth(	const RowProperties *	rp,
					int *			pValue,
					Widget			w );

extern void tedFormatFillColumnPage(	ColumnTool *		ct,
				ColumnPageResources *		cpr,
				InspectorSubject *		is,
				Widget				pageWidget,
				const InspectorSubjectResources * isr );

extern void tedFormatToolRefreshTablePage(	TableTool *	tt );
extern void tedFormatToolRefreshColumnPage(	ColumnTool *	ct );
extern void tedFormatToolRefreshRowPage(	RowTool *	rt );

extern void tedFormatToggleBorder(	BorderProperties *	bp,
					int			set );

extern int tedFormatSaveRowProperties(	TedFormatTool *		tft,
					const BufferItem *	sectBi,
					int			row0 );

extern void tedFormatFillSectionPage(	SectionTool *		st,
				const SectionPageResources *	spr,
				InspectorSubject *		is,
				Widget				pageWidget,
				const InspectorSubjectResources * isr );

extern void tedFormatFillSectionChoosers( SectionTool *			st,
					const SectionPageResources *	spr );

extern void tedFormatFinishSectionPage(	SectionTool *			st,
					const SectionPageResources *	spr );

