/************************************************************************/
/*									*/
/*  Geometry calculations.						*/
/*									*/
/************************************************************************/

#   include	"config.h"

#   include	<stddef.h>
#   include	<stdio.h>
#   include	<stdlib.h>
#   include	<limits.h>
#   define	y0	math_y0
#   define	y1	math_y1
#   include	<math.h>
#   undef	y0
#   undef	y1

#   include	"tedApp.h"
#   include	"docPs.h"

#   include	<debugon.h>

/************************************************************************/
/*									*/
/*  Calculate border width.						*/
/*									*/
/************************************************************************/

int tedBorderThick(		int *				pWide,
				const BorderProperties *	bp,
				const AppDrawingData *		add )
    {
    int			wide= bp->bpWidthTwips+ bp->bpSpacingTwips;
    int			thick= bp->bpWidthTwips;

    if  ( ! bp->bpIsSet )
	{ thick= 0;	}

    if  ( thick > 0 )
	{
	thick *= add->addMagnifiedPixelsPerTwip;
	wide  *= add->addMagnifiedPixelsPerTwip;

	if  ( thick == 0 )
	    { thick= 1;	}

	*pWide= wide;
	}
    else{ *pWide= 0;	}

    return thick;
    }

/************************************************************************/
/*									*/
/*  Calculate the width of a string.					*/
/*									*/
/************************************************************************/
static int tedTextWidth(	AppPhysicalFont *		apf,
				const unsigned char *		s,
				int				len )
    {
    if  ( apf->apfFontSet )
	{ return XmbTextEscapement( apf->apfFontSet, (const char *)s, len ); }
    else{ return XTextWidth( apf->apfFontStruct, (const char *)s, len ); }
    }

static int tedVisibleTextWidth(	AppPhysicalFont *		apf,
				const unsigned char *		s,
				int				len )
    {
    while( len > 0 && s[len- 1] == ' ' )
	{ len--;	}

    if  ( len == 0 )
	{ return 0;	}

    if  ( apf->apfFontSet )
	{ return XmbTextEscapement( apf->apfFontSet, (const char *)s, len ); }
    else{ return XTextWidth( apf->apfFontStruct, (const char *)s, len ); }
    }

/************************************************************************/
/*									*/
/*  Place as many particules on a line as possible.			*/
/*									*/
/*  1)  Accept at least one particule.					*/
/*  9)  Predictable undefied behaviour.					*/
/*									*/
/************************************************************************/
static int tedLayoutParticules(	const BufferItem *		bi,
				AppDrawingData *		add,
				const FormattingFrame *		ffPixels,
				TextParticule *			tp,
				ParticuleData *			pd,
				int				particuleCount,
				int				x0 )
    {
    int				accepted;
    int				width;

    int				len= 0;
    int				xFrom= x0;
    const unsigned char *	from= bi->biParaString+ tp->tpStroff;

    int				physicalFont= tp->tpPhysicalFont;
    AppPhysicalFontList *	apfl= &(add->addPhysicalFontList);
    AppPhysicalFont *		apf= apfl->apflFonts+ physicalFont;

    accepted= 0;
    while( accepted < particuleCount )
	{
	switch( tp->tpKind )
	    {
	    int		x1;

	    case DOCkindTEXT:
		while( accepted < particuleCount		&&
		       tp->tpPhysicalFont == physicalFont	&&
		       tp->tpKind == DOCkindTEXT		)
		    {
		    len += tp->tpStrlen;
		    width= tedTextWidth( apf, from, len );

		    pd->pdVisiblePixels= tedVisibleTextWidth(
						apf,
						bi->biParaString+ tp->tpStroff,
						tp->tpStrlen );
		    pd->pdWhiteUnits= xFrom+ width- x0- pd->pdVisiblePixels;

		    tp->tpX0= x0;
		    tp->tpPixelsWide= xFrom+ width- x0;
		    x0= xFrom+ width;

		    accepted++; tp++;
		    pd++; break;
		    }

		break;
	    case DOCkindTAB:
		x1= TWIPStoPIXELS( add->addMagnifiedPixelsPerTwip,
							pd->pdTabPosition );

		pd->pdVisiblePixels= 0;
		pd->pdWhiteUnits= x1- x0;
		tp->tpObjectNumber= pd->pdTabNumber;

		tp->tpX0= x0;
		tp->tpPixelsWide= x1- x0;
		x0= x1;

		accepted++; tp++;
		pd++;
		break;

	    case DOCkindOBJECT:
		width= bi->biParaObjects[tp->tpObjectNumber].ioPixelsWide;

		pd->pdVisiblePixels= width;
		pd->pdWhiteUnits= 0;

		tp->tpX0= x0;
		tp->tpPixelsWide= width;
		x0 += width;

		accepted++; tp++;
		pd++;
		break;

	    case DOCkindFIELDSTART:
	    case DOCkindFIELDEND:
	    case DOCkindBKMKSTART:
	    case DOCkindBKMKEND:
	    case DOCkindXE:
	    case DOCkindTC:
	    case DOCkindLINEBREAK:
	    case DOCkindPAGEBREAK:
		if  ( pd )
		    { pd->pdVisiblePixels= 0; pd->pdWhiteUnits= 0;	}

		tp->tpX0= x0;
		tp->tpPixelsWide= 0;
		accepted++; tp++;
		if  ( pd )
		    { pd++;	}
		break;

	    default:
		LDEB(tp->tpKind); return -1;
	    }

	if  ( accepted >= particuleCount )
	    { return accepted;	}

	physicalFont= tp->tpPhysicalFont;
	apf= apfl->apflFonts+ physicalFont;
	from= bi->biParaString+ tp->tpStroff;
	xFrom= x0;
	len= 0;
	}

    return accepted;
    }

/************************************************************************/
/*									*/
/*  Determine the height of a line.					*/
/*									*/
/************************************************************************/
static void tedLineHeight(	const BufferItem *	bi,
				const TextParticule *	tp,
				const AppDrawingData *	add,
				int			particuleCount,
				int *			pAscent,
				int *			pDescent )
    {
    const AppPhysicalFontList *	apfl= &(add->addPhysicalFontList);
    int				physicalFont= -1;
    AppPhysicalFont *		apf;
    int				ascent= 0;
    int				descent= 0;
    int				objectHeight= 0;

    InsertedObject *		io;

    if  ( particuleCount <= 0 )
	{ LDEB(particuleCount); return;	}

    while( particuleCount > 0 )
	{
	switch( tp->tpKind )
	    {
	    case DOCkindTEXT:
	    case DOCkindTAB:
		physicalFont= tp->tpPhysicalFont;
		apf= apfl->apflFonts+ physicalFont;

		if  ( ascent < apf->apfAscentPixels )
		    { ascent=  apf->apfAscentPixels;	}
		if  ( descent < apf->apfDescentPixels )
		    { descent=  apf->apfDescentPixels; }
		break;

	    case DOCkindOBJECT:
		physicalFont= tp->tpPhysicalFont;
		apf= apfl->apflFonts+ physicalFont;

		if  ( ascent < apf->apfAscentPixels )
		    { ascent=  apf->apfAscentPixels;	}
		if  ( descent < apf->apfDescentPixels )
		    { descent=  apf->apfDescentPixels; }

		io= bi->biParaObjects+ tp->tpObjectNumber;

		if  ( objectHeight < io->ioPixelsHigh )
		    { objectHeight=  io->ioPixelsHigh;	}
		break;

	    case DOCkindFIELDSTART:
	    case DOCkindFIELDEND:
	    case DOCkindBKMKSTART:
	    case DOCkindBKMKEND:
	    case DOCkindXE:
	    case DOCkindTC:
	    case DOCkindLINEBREAK:
	    case DOCkindPAGEBREAK:
		break;

	    default:
		LDEB(tp->tpKind); break;
	    }

	tp++; particuleCount--;
	}

    if  ( ascent < objectHeight )
	{ ascent=  objectHeight; }

    *pAscent= ascent;
    *pDescent= descent;
    }

/************************************************************************/
/*									*/
/*  Justify the particules in a line of text.				*/
/*									*/
/*  1)  Start justification after the last tab of the line. Justifying	*/
/*	after anything else than a left tab is ridiculous. Simply	*/
/*	refuse to.							*/
/*  2)  Ignore organisational particules such as the delimitation of	*/
/*	links and bookmarks at the end of the line.			*/
/*									*/
/************************************************************************/

static void tedJustifyLine(	const BufferItem *	bi,
				TextParticule *		tp,
				ParticuleData *		pd,
				int			accepted,
				int			x1TextLines )
    {
    TextParticule *	tpp;
    ParticuleData *	pdd;

    int			i;
    int			extra;
    int			totalWeight;
    int			step;
    int			left;

    /*  1  */
    left= 0;
    tpp= tp; pdd= pd;
    for ( i= 0; i < accepted- 1; tpp++, pdd++, i++ )
	{
	if  ( tpp->tpKind == DOCkindTAB )
	    { left= i+ 1;	}
	}

    if  ( left > 0 )
	{
	if  ( pd[left-1].pdTabKind != DOCtkLEFT )
	    { LDEB(pd[left-1].pdTabKind); return;	}

	tp += left; pd += left; accepted -= left;
	}

    /*  2  */
    while( accepted > 0 && tp[accepted- 1].tpStrlen == 0 )
	{ accepted--;	}

    if  ( accepted < 2 )
	{ /*LDEB(accepted);*/ return;	}

    extra= x1TextLines- tp[accepted- 1].tpX0- pd[accepted- 1].pdVisiblePixels;

    totalWeight= 0;
    tpp= tp; pdd= pd;
    for ( i= 0; i < accepted- 1; tpp++, pdd++, i++ )
	{
	pdd->pdWhiteUnits= 0;
	pdd->pdCorrectBy= 0;

	if  ( bi->biParaString[tpp->tpStroff+tpp->tpStrlen-1] == ' ' )
	    {
	    pdd->pdWhiteUnits=
		sqrt( (double)pdd[0].pdWidth+ (double)pdd[1].pdVisibleWidth );

	    totalWeight += pdd->pdWhiteUnits;
	    }
	}

    left= extra;
    tpp= tp; pdd= pd;
    for ( i= 0; i < accepted- 1; tpp++, pdd++, i++ )
	{
	if  ( pdd->pdWhiteUnits > 0 )
	    {
	    step= ( extra* pdd->pdWhiteUnits )/ totalWeight;
	    if  ( step > left )
		{ step= left;	}

	    pdd->pdCorrectBy += step;

	    left -= step;
	    }
	}

    tpp= tp; pdd= pd;
    for ( i= 0; i < accepted- 1; tpp++, pdd++, i++ )
	{
	if  ( pdd->pdWhiteUnits > 0 )
	    {
	    step= 1;
	    if  ( step > left )
		{ step= left;	}

	    pdd->pdCorrectBy += step;

	    left -= step;
	    }
	}

    step= 0;
    tpp= tp; pdd= pd;
    for ( i= 0; i < accepted- 1; tpp++, pdd++, i++ )
	{
	tpp->tpX0 += step;
	if  ( pdd->pdWhiteUnits > 0 )
	    {
	    tpp->tpPixelsWide += pdd->pdCorrectBy;
	    step += pdd->pdCorrectBy;
	    }
	}

    tpp->tpX0 += step;

    return;
    }

/************************************************************************/
/*									*/
/*  Recalculate the geometry from the start of the selection to the	*/
/*  end of the line.							*/
/*									*/
/*  'part'	is known to be the first particule of the line.		*/
/*									*/
/************************************************************************/

static void tedSqueezeParticules(	TextParticule *		tp,
					ParticuleData *		pd,
					int			accepted,
					int			x0,
					int			x1 )
    {
    TextParticule *	tpp= tp+ accepted- 1;
    ParticuleData *	pdd= pd+ accepted- 1;
    int			i;
    int			whitePixels= 0;
    int			shortage= tpp->tpX0+ pdd->pdVisiblePixels- x1;
    int			xx0;
    int			done;

    for ( i= 0; i < accepted- 1; i++ )
	{
	if  ( pd[i].pdWhiteUnits > 0 )
	    { whitePixels += pd[i].pdWhiteUnits; }

	pd[i].pdCorrectBy= 0;
	}

    if  ( whitePixels < shortage )
	{
	/*
	LLDEB(whitePixels,shortage);
	LLLDEB(x0,x1,tpp->tpX0+ pdd->pdVisiblePixels);
	LLDEB(whitePixels,tpp->tpX0+ pdd->pdVisiblePixels-x1);

	for ( i= 0; i < accepted; i++ )
	    { docListParticule( 0, "SH", part+ i, bi, tp+ i ); }
	*/
	}
    else{
	done= 0;

	for ( i= 0; i < accepted- 1; i++ )
	    {
	    int		step;

	    if  ( pd[i].pdWhiteUnits > 0 )
		{
		step= ( pd[i].pdWhiteUnits* shortage )/ whitePixels;

		if  ( step > 0 && done < shortage )
		    {
		    if  ( step > shortage- done )
			{ step= shortage- done;	}

		    pd[i].pdCorrectBy += step;
		    done += step;
		    }
		}
	    }

	for ( i= 0; i < accepted- 1 && done < shortage; i++ )
	    {
	    if  ( pd[i].pdWhiteUnits > pd[i].pdCorrectBy )
		{ pd[i].pdCorrectBy++; done++; }
	    }

	if  ( done != shortage )
	    { LLDEB(done,shortage);	}

	xx0= x0; whitePixels= 0;
	for ( i= 0; i < accepted- 1; i++ )
	    {
	    tp[i].tpX0 -= whitePixels;

	    if  ( pd[i].pdCorrectBy > 0 )
		{
		tp[i].tpPixelsWide -= pd[i].pdCorrectBy;
		whitePixels += pd[i].pdCorrectBy;
		}
	    }

	tp[i].tpX0 -= whitePixels;
	tp[i].tpPixelsWide= x1- tp[i].tpX0;
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Layout one line on screen, reconciling PostScript and X11 geometry.	*/
/*									*/
/************************************************************************/

static int tedLayoutLine(	const BufferItem *	bi,
				AppDrawingData *	add,
				const FormattingFrame *	ffPixels,
				TextParticule *		tp,
				ParticuleData *		pd,
				int			particuleCount,
				int			x0 )
    {
    int			past= 0;
    int			part= 0;
    int			x1;

    int			accepted;

    while( past < particuleCount )
	{
	int		squeeze= 0;
	int		includeTab= 0;
	int		j;

	TextParticule *	tpp;
	ParticuleData *	pdd;

	j= 0;
	while( past < particuleCount && tp[j].tpKind != DOCkindTAB )
	    { j++; past++;	}

	if  ( past < particuleCount )
	    { includeTab= 1; past++;	}

	accepted= tedLayoutParticules( bi, add, ffPixels, tp, pd, past- part, x0 );

	if  ( accepted != past- part )
	    { LLDEB(accepted,past- part); return -1;	}

	tpp= tp+ accepted- 1;
	pdd= pd+ accepted- 1;

	if  ( includeTab )
	    {
	    x1= TWIPStoPIXELS( add->addMagnifiedPixelsPerTwip,
						    pdd->pdX0+ pdd->pdWidth );

	    if  ( tpp->tpX0+ pdd->pdVisiblePixels > x1	)
		{
		/*
		LLDEB(tpp->tpX0+ tpp->tpPixelsWide,x1);
		LLDEB(tpp->tpX0+ pdd->pdVisiblePixels,x1);
		docListParticule( 0, "WIDE", part, bi, tpp );
		*/

		squeeze= 1;
		}
	    else{
		if  ( tpp->tpX0+ tpp->tpPixelsWide != x1 )
		    {
		    if  ( tpp->tpX0 > x1 )
			{
			LLDEB(tpp->tpX0+ tpp->tpPixelsWide,x1);
			tpp->tpX0= x1;
			tpp->tpPixelsWide= 0;
			}
		    else{
			tpp->tpPixelsWide= x1- tpp->tpX0;
			}
		    }
		}
	    }
	else{
	    x1= ffPixels->ffX1TextLines;

	    if  ( tpp->tpX0+ pdd->pdVisiblePixels > x1	)
		{
		/*
		LLDEB(tpp->tpX0+ tpp->tpPixelsWide,x1);
		LLDEB(tpp->tpX0+ pdd->pdVisiblePixels,x1);
		docListParticule( 0, "WIDE", part, bi, tpp );
		*/

		squeeze= 1;
		}
	    }

	if  ( squeeze )
	    { tedSqueezeParticules( tp, pd, accepted, x0, x1 ); }

	part += accepted; tp += accepted; pd += accepted;
	x0= x1; past= part;
	}

    return 0;
    }

static int tedPlaceScreenLine(	TextLine *			tl,
				const BufferItem *		bi,
				int				part,
				int				accepted,
				AppDrawingData *		add,
				const FormattingFrame *		ffPixels,
				TextParticule *			tp,
				ParticuleData *			pd,
				int				y )
    {
    tl->tlTopPosition.lpYPixels= y;

    return tl->tlParticuleCount;
    }

static int tedLayoutScreenLine(	TextLine *			tl,
				const BufferItem *		bi,
				int				part,
				int				accepted,
				AppDrawingData *		add,
				const FormattingFrame *		ffPixels,
				TextParticule *			tp,
				ParticuleData *			pd,
				int				y )
    {
    int		xShift;
    int		x1;

    int		x0Pixels;
    int		ascentPixels;
    int		descentPixels;
    int		lineDistPixels;

    if  ( part == 0 )
	{ x0Pixels= ffPixels->ffX0FirstLine;	}
    else{ x0Pixels= ffPixels->ffX0TextLines;	}

    if  ( tedLayoutLine( bi, add, ffPixels, tp, pd, accepted, x0Pixels ) )
	{ LDEB(1); return -1;	}

    switch( bi->biParaAlignment )
	{
	case DOCiaLEFT:
	    xShift= 0; break;
	case DOCiaRIGHT:
	    x1= tp[accepted-1].tpX0+ pd[accepted-1].pdVisiblePixels;
	    xShift= ffPixels->ffX1TextLines- x1;
	    break;
	case DOCiaCENTERED:
	    x1= tp[accepted-1].tpX0+ pd[accepted-1].pdVisiblePixels;
	    xShift= ( ffPixels->ffX1TextLines- x1 )/ 2;
	    break;
	case DOCiaJUSTIFIED:
	    xShift= 0;
	    if  ( part+ accepted < bi->biParaParticuleCount )
		{
		tedJustifyLine( bi, tp, pd, accepted,
						ffPixels->ffX1TextLines );
		}
	    break;
	default:
	    LDEB(bi->biParaAlignment); xShift= 0; break;
	}

    if  ( xShift > 0 )
	{
	int	i;

	for ( i= 0; i < accepted; i++ )
	    { tp[i].tpX0 += xShift; }
	}

    tedLineHeight( bi, bi->biParaParticules+ part, add, accepted,
					    &ascentPixels, &descentPixels );

    lineDistPixels= TWIPStoPIXELS( add->addMagnifiedPixelsPerTwip,
						tl->tlLineSpacingTwips );

    tl->tlLineAscentPixels= ascentPixels;
    tl->tlTopPosition.lpYPixels= y;
    /*
    tl->tlLineHeightPixels= ascentPixels+ descentPixels;
    */

    return tl->tlParticuleCount;
    }

/************************************************************************/
/*									*/
/*  Derive the frame for a paragraph from the page rectangle and the	*/
/*  paragraph properties.						*/
/*									*/
/*  For paragraphs inside a table cell, geometry is derived from the	*/
/*  table column.							*/
/*									*/
/************************************************************************/

void tedParagraphFrame(		FormattingFrame *	ffPixels,
				const AppDrawingData *	add,
				int			bottom,
				const BufferItem *	bi )
    {
    double		xfac= add->addMagnifiedPixelsPerTwip;
    int			leftIndentPixels;
    int			firstIndentPixels;
    int			rightIndentPixels;

    ffPixels->ffX0Geometry= add->addDocRect.drX0;
    ffPixels->ffX1Geometry= add->addDocRect.drX1;

    ffPixels->ffY1= -1;

    if  ( bottom > 0 )
	{ ffPixels->ffY1= bottom;	}

    if  ( bi->biParaInTable )
	{
	int			col;
	BufferItem *		cellBi= bi->biParent;
	BufferItem *		rowBi= cellBi->biParent;
	RowProperties *		rp= &(rowBi->biRowProperties);
	CellProperties *	cp;

	col= cellBi->biNumberInParent;

	cp= rp->rpCells+ col;
	ffPixels->ffX1Geometry= add->addDocRect.drX0+ cp->cpRightBoundaryPixels;

	ffPixels->ffX1Geometry -= rowBi->biRowHalfGapWidthPixels;

	if  ( col == 0 )
	    { ffPixels->ffX0Geometry += rowBi->biRowLeftIndentPixels; }
	else{
	    BufferItem *		prevCell;

	    prevCell= rowBi->biGroupChildren[col- 1];

	    cp= rp->rpCells+ col- 1;
	    ffPixels->ffX0Geometry += cp->cpRightBoundaryPixels;
	    }

	ffPixels->ffX0Geometry += rowBi->biRowHalfGapWidthPixels;
	}

    leftIndentPixels= TWIPStoPIXELS( xfac, bi->biParaLeftIndentTwips );
    firstIndentPixels= TWIPStoPIXELS( xfac, bi->biParaFirstIndentTwips );
    rightIndentPixels= TWIPStoPIXELS( xfac, bi->biParaRightIndentTwips );

    ffPixels->ffX0TextLines= ffPixels->ffX0Geometry+ leftIndentPixels;
    ffPixels->ffX1TextLines= ffPixels->ffX1Geometry- rightIndentPixels;

    ffPixels->ffX0FirstLine= ffPixels->ffX0TextLines+ firstIndentPixels;

    return;
    }

/************************************************************************/
/*									*/
/*  Prepare for the layout of a paragraph for displaying.		*/
/*									*/
/************************************************************************/

void tedParaScreenGeometry(	BufferItem *		bi,
				const AppDrawingData *	add )
    {
    double		xfac= add->addMagnifiedPixelsPerTwip;

    int			tab;
    TabStop *		ts;
    int			bottom= -1;

    FormattingFrame	ffPixels;

    tedParagraphFrame( &ffPixels, add, bottom, bi );

    ts= bi->biParaTabStops;
    for ( tab= 0; tab < bi->biParaTabCount; ts++, tab++ )
	{
	ts->tsPixels= ffPixels.ffX0Geometry+
					TWIPStoPIXELS( xfac, ts->tsTwips );
	}

    return;
    }

static int tedLayoutStartScreenPara( FormattingFrame *		ffPixels,
				int				bottom,
				BufferItem *			bi,
				AppDrawingData *		add,
				const DocumentProperties *	dp )
    {
    int			part;
    TextParticule *	tp;

    double		xfac= add->addMagnifiedPixelsPerTwip;

    tp= bi->biParaParticules;
    for ( part= 0; part < bi->biParaParticuleCount; tp++, part++ )
	{
	if  ( tp->tpKind != DOCkindOBJECT )
	    { continue;	}

	if  ( 1 || tp->tpPhysicalFont < 0 )
	    {
	    InsertedObject *	io= bi->biParaObjects+ tp->tpObjectNumber;

	    if  ( io->ioKind == DOCokMACPICT )
		{
		if  ( io->ioTwipsWide == 0 )
		    { io->ioTwipsWide= 20* io->ioXExtent;	}
		if  ( io->ioTwipsHigh == 0 )
		    { io->ioTwipsHigh= 20* io->ioYExtent;	}
		}

	    io->ioPixelsWide= TWIPStoPIXELS( xfac,
				( io->ioScaleX* io->ioTwipsWide )/ 100 );
	    io->ioPixelsHigh= TWIPStoPIXELS( xfac,
				(  io->ioScaleY* io->ioTwipsHigh )/ 100 );
	    }
	}

    tedParagraphFrame( ffPixels, add, bottom, bi );

    tedParaScreenGeometry( bi, add );

    return 0;
    }

static int tedPlaceStartScreenPara( FormattingFrame *		ffPixels,
				int				bottom,
				BufferItem *			bi,
				AppDrawingData *		add,
				const DocumentProperties *	dp )
    {
    tedParagraphFrame( ffPixels, add, bottom, bi );

    return 0;
    }

static int tedLayoutStartScreenRow(	BufferItem *			rowBi,
					AppDrawingData *		add,
					const DocumentProperties *	dp )
    {
    int			i;
    CellProperties *	cp;

    double		xfac= add->addMagnifiedPixelsPerTwip;

    rowBi->biRowHalfGapWidthPixels= TWIPStoPIXELS( xfac,
						rowBi->biRowHalfGapWidthTwips );
    rowBi->biRowLeftIndentPixels= TWIPStoPIXELS( xfac,
						rowBi->biRowLeftIndentTwips );

    cp= rowBi->biRowCells;
    for ( i= 0; i < rowBi->biRowCellCount; cp++, i++ )
	{
	cp->cpRightBoundaryPixels=
			    TWIPStoPIXELS( xfac, cp->cpRightBoundaryTwips );
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Various layout routines.						*/
/*									*/
/************************************************************************/

static void tedFillLayoutJob(	LayoutJob *	lj )
    {
    lj->ljInHeaderFooter= 0;

    lj->ljLayoutScreen.slStartRow= tedLayoutStartScreenRow;
    lj->ljLayoutScreen.slStartParagraph= tedLayoutStartScreenPara;
    lj->ljLayoutScreen.slLayoutLine= tedLayoutScreenLine;

    lj->ljPlaceScreen.slStartRow= (START_ROW)0;
    lj->ljPlaceScreen.slStartParagraph= tedPlaceStartScreenPara;
    lj->ljPlaceScreen.slLayoutLine= tedPlaceScreenLine;

    return;
    }

int tedLayoutItem(	BufferItem *		bi,
			DocumentProperties *	dp,
			AppDrawingData *	add,
			DocumentRectangle *	drChanged )
    {
    const int		andParents= 1;
    LayoutJob		lj;

    tedFillLayoutJob( &lj );

    if  ( bi->biNumberInParent == 0 )
	{
	if  ( bi->biParent )
	    { lj.ljPosition= bi->biParent->biTopPosition;		}
	else{ docInitLayoutPosition( &(lj.ljPosition) );		}
	}
    else{
	BufferItem *	prevBi;

	prevBi= bi->biParent->biGroupChildren[bi->biNumberInParent- 1];

	lj.ljPosition= prevBi->biBelowPosition;
	}

    lj.ljChangedRectanglePixels= drChanged;

    if  ( docPsLayoutItemAndParents( bi, add, dp, &lj, andParents ) )
	{ LDEB(1); return -1;	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Calculate the layout of a page header.				*/
/*									*/
/************************************************************************/

int tedLayoutHeaderItem(	BufferItem *			bi,
				int				page,
				const SectionProperties *	sp,
				const DocumentProperties *	dp,
				AppDrawingData *		add,
				DocumentRectangle *		drChanged )
    {
    const int			andParents= 0;
    LayoutJob			lj;

    double			xfac= add->addMagnifiedPixelsPerTwip;

    tedFillLayoutJob( &lj );

    lj.ljInHeaderFooter= 1;

    lj.ljPosition.lpPage= page;
    lj.ljPosition.lpPageYTwips= sp->spHeaderYTwips;
    lj.ljPosition.lpYPixels= page* add->addPageStepPixels+ 
				    TWIPStoPIXELS( xfac, sp->spHeaderYTwips );

    lj.ljChangedRectanglePixels= drChanged;

    if  ( docPsLayoutItemAndParents( bi, add, dp, &lj, andParents ) )
	{ LDEB(1); return -1;	}

    return 0;
    }

int tedLayoutFooterItem(	BufferItem *			bi,
				int				page,
				const SectionProperties *	sp,
				const DocumentProperties *	dp,
				AppDrawingData *		add,
				DocumentRectangle *		drChanged )
    {
    const int			andParents= 0;
    LayoutJob			lj;

    double			xfac= add->addMagnifiedPixelsPerTwip;

    tedFillLayoutJob( &lj );

    lj.ljInHeaderFooter= -1;

    lj.ljPosition.lpPage= page;
    lj.ljPosition.lpPageYTwips= sp->spPageHighTwips-
						sp->spFooterYTwips;
    lj.ljPosition.lpYPixels= page* add->addPageStepPixels+ 
			TWIPStoPIXELS( xfac, lj.ljPosition.lpPageYTwips );

    lj.ljChangedRectanglePixels= drChanged;

    if  ( docPsLayoutItemAndParents( bi, add, dp, &lj, andParents ) )
	{ LDEB(1); return -1;	}

    return 0;
    }

int tedAdjustParagraphLayout(	BufferItem *			bi,
				int				line,
				int				stroffShift,
				int				upto,
				AppDrawingData *		add,
				const DocumentProperties *	dp,
				DocumentRectangle *		drChanged )
    {
    LayoutJob		lj;

    tedFillLayoutJob( &lj );

    /*  bogus  */
    lj.ljPosition.lpPage= 0;
    lj.ljPosition.lpPageYTwips= 0;
    lj.ljPosition.lpYPixels= 0;

    lj.ljChangedRectanglePixels= drChanged;

    if  ( docPsAdjustParaLayout( bi, line, stroffShift, upto, add, dp, &lj ) )
	{ LDEB(1); return -1;	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Find the object in the document for an x,y position.		*/
/*									*/
/************************************************************************/

static BufferItem * tedFindItem(	BufferItem *		bi,
					const AppDrawingData *	add,
					int			x,
					int			y	)
    {
    int			i;
    BufferItem *	child;
    BufferItem *	found;
    CellProperties *	cp;
    int			page= y/ add->addPageStepPixels;

    switch( bi->biLevel )
	{
	case DOClevDOC:
	case DOClevSECT:
	case DOClevCELL:
	rowAsGroup:
	    for ( i= 0; i < bi->biGroupChildCount; i++ )
		{
		BufferItem *	next= (BufferItem *)0;

		child= bi->biGroupChildren[i];

		if  ( i+ 1 < bi->biGroupChildCount )
		    { next= bi->biGroupChildren[i+ 1];	}

		if  ( y < child->biBelowPosition.lpYPixels		||
		      ( next					&&
		        next->biTopPosition.lpPage > page	)	)
		    {
		    found= tedFindItem( child, add, x, y );
		    return found;
		    }
		}
	    if  ( bi->biGroupChildCount > 0 )
		{
		child= bi->biGroupChildren[bi->biGroupChildCount- 1];

		found= tedFindItem( child, add, x, y );
		return found;
		}
	    LDEB(bi->biGroupChildCount);
	    break;

	case DOClevROW:
	    if  ( ! bi->biRowHasTableParagraphs )
		{ goto rowAsGroup;	}

	    cp= bi->biRowProperties.rpCells;
	    for ( i= 0; i < bi->biGroupChildCount; cp++, i++ )
		{
		BufferItem *	paraBi;

		if  ( x <= add->addDocRect.drX0+
					cp->cpRightBoundaryPixels	||
		      i == bi->biGroupChildCount- 1			)
		    {
		    child= bi->biGroupChildren[i];
		    paraBi= tedFindItem( child, add, x, y );

		    if  ( paraBi )
			{ return paraBi;	}

		    if  ( child->biGroupChildCount > 0 )
			{
			found= child->biGroupChildren[
						child->biGroupChildCount-1];
			return found;
			}
		    }
		}
	    LLDEB(bi->biTopPosition.lpYPixels,bi->biBelowPosition.lpYPixels);
	    LLDEB(x,y);
	    return (BufferItem *)0;

	case DOClevPARA:
	    return bi;

	default:
	    LDEB(bi->biLevel); return (BufferItem *)0;
	}

    LSDEB(y,docLevelStr(bi->biLevel)); return (BufferItem *)0;
    }

static int tedFindLine(	const BufferItem *	bi,
			const AppDrawingData *	add,
			int			x,
			int			y	)
    {
    double		xfac= add->addMagnifiedPixelsPerTwip;
    int			page= y/ add->addPageStepPixels;

    int			i;
    TextLine *		tl;
    int			y1;

    if  ( y < bi->biTopPosition.lpYPixels )
	{ return 0;	}

    if  ( bi->biLevel != DOClevPARA )
	{ LLDEB(bi->biLevel,DOClevPARA); return -1;	}

    tl= bi->biParaLines;
    for ( i= 0; i < bi->biParaLineCount; tl++, i++ )
	{
	y1= tl->tlTopPosition.lpYPixels+
			    TWIPStoPIXELS( xfac, tl->tlLineSpacingTwips );

	if  ( y < y1						||
	      ( i+ 1 < bi->biParaLineCount		&&
	        tl[1].tlTopPosition.lpPage > page	)	)
	    { return i;	}
	}

    if  ( bi->biParaLineCount > 0 )
	{ return bi->biParaLineCount- 1;	}

    LLDEB(y,bi->biParaLineCount); return -1;
    }

int tedFindParticule(	TextLine *	tl,
			TextParticule *	tp,
			int		x,
			int		y	)
    {
    int				i;

    tp += tl->tlFirstParticule;

    if  ( x < tp->tpX0 )
	{ return tl->tlFirstParticule;	}

    for ( i= 0; i < tl->tlParticuleCount; tp++, i++ )
	{
	if  ( tp->tpX0 <= x &&  x < tp->tpX0+ tp->tpPixelsWide )
	    { return tl->tlFirstParticule+ i;	}
	}

    return tl->tlFirstParticule+ tl->tlParticuleCount- 1;
    }

/************************************************************************/
/*  Return the X position for an I bar.					*/
/************************************************************************/
int tedCalculateX(	const BufferItem *	bi,
			const TextParticule *	tp,
			const AppDrawingData *	add,
			int			stroff	)
    {
    const AppPhysicalFontList *	apfl= &(add->addPhysicalFontList);
    AppPhysicalFont *		apf;
    const unsigned char *	s;
    int				len;
    int				x;

    switch( tp->tpKind )
	{
	case DOCkindOBJECT:
	case DOCkindTAB:
	    len= stroff- tp->tpStroff;

	    if  ( len == 0 )
		{ x= tp->tpX0;	}
	    else{
		if  ( len == 1 )
		    { x= tp->tpX0+ tp->tpPixelsWide;	}
		else{
		    LLDEB(stroff,tp->tpStroff);
		    docListParticule( 0, "???", tp-bi->biParaParticules,
								    bi, tp );
		    docListItem( 0, bi );
		    x= tp->tpX0;
		    }
		}
	    return x;
	case DOCkindTEXT:
	    if  ( tp->tpPhysicalFont < 0 )
		{ LDEB(tp->tpPhysicalFont); return tp->tpX0;	}

	    s= bi->biParaString+ tp->tpStroff;
	    apf= apfl->apflFonts+ tp->tpPhysicalFont;
	    len= stroff- tp->tpStroff;

	    if  ( len == tp->tpStrlen )
		{ x= tp->tpX0+ tp->tpPixelsWide;		}
	    else{ x= tp->tpX0+ tedTextWidth( apf, s, len );	}
	    return x;
	case DOCkindFIELDSTART:
	case DOCkindFIELDEND:
	case DOCkindBKMKSTART:
	case DOCkindBKMKEND:
	case DOCkindXE:
	case DOCkindTC:
	case DOCkindLINEBREAK:
	case DOCkindPAGEBREAK:
	    x= tp->tpX0;
	    return x;
	default:
	    LDEB(tp->tpKind);
	    x= tp->tpX0;
	    return x;
	}
    }

/************************************************************************/
/*									*/
/*  Return the offset in the paragraph string of a mouse event		*/
/*									*/
/*  1)  For Empty particules report the position of the beginning of	*/
/*	the particule.							*/
/*	For clicks in the left margin, report the position of the	*/
/*	particule.							*/
/*  2)  Note the <=, also consider the position after the last		*/
/*	character.							*/
/*									*/
/************************************************************************/
int tedFindStringOffset(	const TextParticule *	tp,
				const unsigned char *	paraString,
				const AppDrawingData *	add,
				int *			pBarX,
				int			x,
				int			y	)
    {
    const AppPhysicalFontList *	apfl= &(add->addPhysicalFontList);
    int				len;
    int				stroff= tp->tpStroff+ 1;
    const unsigned char *	s= paraString+ tp->tpStroff;
    AppPhysicalFont *		apf;
    int				before= tp->tpX0;
    int				past= tp->tpX0;

    /*  1  */
    if  ( x < before || tp->tpStrlen == 0 )
	{ *pBarX= tp->tpX0; return tp->tpStroff; }

    if  ( tp->tpPhysicalFont < 0 )
	{ LDEB(tp->tpPhysicalFont); return tp->tpStroff;	}

    apf= apfl->apflFonts+ tp->tpPhysicalFont;

    /*  2  */
    for ( len= 1; len <= tp->tpStrlen; stroff++, len++ )
	{
	int	between;

	if  ( len == tp->tpStrlen )
	    { past= tp->tpX0+ tp->tpPixelsWide;			}
	else{ past= tp->tpX0+ tedTextWidth( apf, s, len );	}

	between= ( before+ past )/ 2;

	if  ( before <= x && x < between )
	    { *pBarX= before; return stroff- 1;	}
	if  ( between <= x && x < past )
	    { *pBarX= past; return stroff;		}

	before= past;
	}

    *pBarX= tp->tpX0+ tp->tpPixelsWide; return tp->tpStroff+ tp->tpStrlen;
    }

/************************************************************************/
/*									*/
/*  Translate a Window coordinate to a position in a text buffer.	*/
/*									*/
/************************************************************************/

int tedFindPosition(	BufferDocument *	bd,
			const AppDrawingData *	add,
			int			x,
			int			y,
			TextParticule **	pTp,
			TextLine **		pTl,
			BufferPosition *	bp	)
    {
    const BufferItem *	bi;
    double		xfac= add->addMagnifiedPixelsPerTwip;

    int			part;
    int			stroff;
    int			barX;
    int			line;

    TextLine *		tl;

    bi= tedFindItem( &(bd->bdItem), add, x, y );
    if  ( ! bi )
	{
	if  ( y < bd->bdItem.biTopPosition.lpYPixels )
	    {
	    if  ( docFirstPosition( &(bd->bdItem), bp ) )
		{ LLDEB(bd->bdItem.biTopPosition.lpYPixels,y); return -1; }

	    bi= bp->bpBi;
	    *pTl= bi->biParaLines+ bp->bpLine;
	    *pTp= bi->biParaParticules+ bp->bpParticule;

	    return 0;
	    }

	if  ( y >= bd->bdItem.biBelowPosition.lpYPixels )
	    {
	    if  ( docLastPosition( &(bd->bdItem), bp ) )
		{ LLDEB(bd->bdItem.biTopPosition.lpYPixels,y); return -1; }

	    bi= bp->bpBi;
	    *pTl= bi->biParaLines+ bp->bpLine;
	    *pTp= bi->biParaParticules+ bp->bpParticule;

	    return 0;
	    }

	LDEB(y);
	LDEB(bd->bdItem.biTopPosition.lpYPixels);
	LDEB(bd->bdItem.biBelowPosition.lpYPixels);
	return -1;
	}

    if  ( bi->biLevel != DOClevPARA )
	{ LLDEB(bi->biLevel,DOClevPARA); return -1;	}

    if  ( bi->biParaLineCount > 0				&&
	  y < bi->biParaLines[0].tlTopPosition.lpYPixels	)
	{
	line= 0;
	part= 0;
	stroff= 0;
	barX= bi->biParaParticules[0].tpX0;
	}
    else{
	line= tedFindLine( bi, add, x, y );
	if  ( line < 0 )
	    {
	    if  ( y > bi->biBelowPosition.lpYPixels	&&
		  bi->biParaLineCount > 0		&&
		  bi->biParaParticuleCount > 0		)
		{
		line= bi->biParaLineCount- 1;
		part= bi->biParaParticuleCount- 1;
		stroff= bi->biParaStrlen;
		barX= bi->biParaParticules[part].tpX0+
				    bi->biParaParticules[part].tpPixelsWide;
		}
	    else{
		line= 0; part= 0; stroff= 0;
		if  ( bi->biParaParticuleCount == 0 )
		    { LLDEB(line,bi->biParaParticuleCount); return -1;	}
		barX= bi->biParaParticules[part].tpX0;
		}
	    }
	else{
	    part= tedFindParticule( bi->biParaLines+ line,
						bi->biParaParticules, x, y );
	    if  ( part < 0 )
		{ LLDEB(x,part); return -1;			}

	    stroff= tedFindStringOffset( bi->biParaParticules+ part,
					bi->biParaString, add, &barX, x, y );
	    if  ( stroff < 0 )
		{ LDEB(stroff); return -1;	}
	    }
	}

#   if 0
    {
    TextParticule *	tp= bi->biParaParticules+ part;

    appDebug( "(%3d,%3d): %.*s|%.*s\n", x, y,
			stroff- tp->tpStroff, bi->biParaString+ tp->tpStroff,
			tp->tpStroff+ tp->tpStrlen- stroff,
			bi->biParaString+ stroff );
    }
#   endif

    tl= bi->biParaLines+ line;

    *pTl= tl;
    *pTp= bi->biParaParticules+ part;

    bp->bpBi= (BufferItem *)bi;	/*  Discards const */
    bp->bpLine= line;
    bp->bpParticule= part;
    bp->bpStroff= stroff;

    bp->bpX= barX;
    bp->bpTopPosition= tl->tlTopPosition;
    bp->bpY1Pixels= tl->tlTopPosition.lpYPixels+
				TWIPStoPIXELS( xfac, tl->tlLineSpacingTwips );

    return 0;
    }

/************************************************************************/
/*									*/
/*  Calculate the smallest rectangle that contains the selection.	*/
/*									*/
/*  1)  Same paragraph?							*/
/*  1b) For a single position, widen to contain the whole I-Bar.	*/
/*  2)  Same table row?							*/
/*  3)  Expose the whole cell.						*/
/*									*/
/************************************************************************/

#   define	IW	5

void tedSelectionRectangle(	DocumentRectangle *	drClip,
				const AppDrawingData *	add,
				const BufferSelection *	bs )
    {
    FormattingFrame	ffPixels;
    DocumentRectangle	dr;

    drClip->drY0= bs->bsBegin.bpTopPosition.lpYPixels;
    drClip->drY1= bs->bsEnd.bpY1Pixels;

    /*  1  */
    if  ( bs->bsEnd.bpBi == bs->bsBegin.bpBi )
	{
	if  ( bs->bsEnd.bpLine == bs->bsBegin.bpLine )
	    {
	    /*  1b  */
	    if  ( bs->bsEnd.bpStroff == bs->bsBegin.bpStroff )
		{
		drClip->drX0= bs->bsBegin.bpX- IW;
		drClip->drX1= bs->bsEnd.bpX+ IW;
		}
	    else{
		drClip->drX0= bs->bsBegin.bpX;
		drClip->drX1= bs->bsEnd.bpX;
		}

	    return;
	    }

	if  ( bs->bsBegin.bpBi->biParaInTable )
	    {
	    tedParagraphFrame( &ffPixels, add, -1, bs->bsBegin.bpBi );

	    drClip->drX0= ffPixels.ffX0TextLines;
	    drClip->drX1= ffPixels.ffX1TextLines;

	    return;
	    }

	drClip->drX0= add->addBackRect.drX0;
	drClip->drX1= add->addBackRect.drX1;

	return;
	}

    /*  2  */
    if  ( ! bs->bsBegin.bpBi->biParaInTable		||
	  ! bs->bsEnd.bpBi->biParaInTable		||
	  ! docSelectionInsideRow( bs )			)
	{
	drClip->drX0= add->addBackRect.drX0;
	drClip->drX1= add->addBackRect.drX1;
	}

    /*  3  */
    if  ( bs->bsBegin.bpBi->biParaInTable )
	{
	tedParagraphFrame( &ffPixels, add, -1, bs->bsBegin.bpBi );

	dr.drX0= ffPixels.ffX0TextLines;
	dr.drX1= ffPixels.ffX1TextLines;

	if  ( docSelectionInsideCell( bs ) )
	    {
	    dr.drY0= bs->bsBegin.bpBi->biTopPosition.lpYPixels;
	    dr.drY1= bs->bsBegin.bpBi->biBelowPosition.lpYPixels;
	    }
	else{
	    BufferItem *	rowBi= bs->bsBegin.bpBi->biParent->biParent;

	    dr.drY0= rowBi->biTopPosition.lpYPixels;
	    dr.drY1= rowBi->biBelowPosition.lpYPixels;
	    }

	docUnionRectangle( drClip, drClip, &dr );
	}

    /*  3  */
    if  ( bs->bsEnd.bpBi->biParaInTable )
	{
	tedParagraphFrame( &ffPixels, add, -1, bs->bsEnd.bpBi );

	dr.drX0= ffPixels.ffX0TextLines;
	dr.drX1= ffPixels.ffX1TextLines;

	if  ( docSelectionInsideCell( bs ) )
	    {
	    dr.drY0= bs->bsEnd.bpBi->biTopPosition.lpYPixels;
	    dr.drY1= bs->bsEnd.bpBi->biBelowPosition.lpYPixels;
	    }
	else{
	    BufferItem *	rowBi= bs->bsEnd.bpBi->biParent->biParent;

	    dr.drY0= rowBi->biTopPosition.lpYPixels;
	    dr.drY1= rowBi->biBelowPosition.lpYPixels;
	    }

	docUnionRectangle( drClip, drClip, &dr );
	}

    return;
    }

/************************************************************************/
/*									*/
/*  A selection was found, do bookkeeping.				*/
/*									*/
/************************************************************************/

void tedCalculateSelectedLines(	TedDocument *	td	)
    {
    BufferItem *	bi;
    TextLine *		tl;
    int			line;

    BufferPosition *	bp;

    bp= &(td->tdSelection.bsBegin);
    bi= bp->bpBi;
    tl= bi->biParaLines; line= 0;
    while( tl->tlFirstParticule+ tl->tlParticuleCount < bp->bpParticule )
	{ tl++; line++;	}

    if  ( line < bi->biParaLineCount- 1					&&
	  tl->tlFirstParticule+ tl->tlParticuleCount == bp->bpParticule	)
	{ tl++; line++;	}

    bp->bpLine= line;

    bp= &(td->tdSelection.bsEnd);
    bi= bp->bpBi;
    tl= bi->biParaLines; line= 0;
    while( tl->tlFirstParticule+ tl->tlParticuleCount < bp->bpParticule )
	{ tl++; line++;	}

    if  ( line < bi->biParaLineCount- 1					&&
	  tl->tlFirstParticule+ tl->tlParticuleCount == bp->bpParticule	)
	{ tl++; line++;	}

    bp->bpLine= line;
    }

void tedPositionCoordinates(	BufferPosition *	bp,
				const AppDrawingData *	add )
    {
    BufferItem *	bi;
    TextLine *		tl;
    TextParticule *	tp;

    double		xfac= add->addMagnifiedPixelsPerTwip;

    bi= bp->bpBi;
    tl= bi->biParaLines+ bp->bpLine;
    tp= bi->biParaParticules+ bp->bpParticule;

    bp->bpX= tedCalculateX( bi, tp, add, bp->bpStroff );
    bp->bpTopPosition= tl->tlTopPosition;
    bp->bpY1Pixels= tl->tlTopPosition.lpYPixels+
				TWIPStoPIXELS( xfac, tl->tlLineSpacingTwips );

#   if 0
    Gives silly high I Bar.
    if  ( bp->bpLine == bi->biParaLineCount- 1 )
	{ bp->bpY1 += bi->biParaSpaceAfterPixels; }
#   endif

    return;
    }

void tedSelectionCoordinates(	BufferSelection *	bs,
				const AppDrawingData *	add )
    {
    tedPositionCoordinates( &(bs->bsBegin), add );
    tedPositionCoordinates( &(bs->bsEnd), add );
    }

/************************************************************************/
/*									*/
/*  Use the new ruler settings, or other geometry properties of the	*/
/*  paragraph.								*/
/*									*/
/************************************************************************/

void tedApplyItemFormat(	BufferItem *			bi,
				DocumentProperties *		dp,
				AppDrawingData *		add,
				DocumentRectangle *		drChanged )
    {
    DocumentRectangle	drLocal;

    drLocal.drX0= add->addBackRect.drX0;
    drLocal.drX1= add->addBackRect.drX1;
    drLocal.drY0= bi->biTopPosition.lpYPixels;
    drLocal.drY1= bi->biBelowPosition.lpYPixels;

    if  ( tedLayoutItem( bi, dp, add, &drLocal ) )
	{ LDEB(1); return;	}

    docUnionRectangle( drChanged, drChanged, &drLocal );

    drLocal.drX0= add->addBackRect.drX0;
    drLocal.drX1= add->addBackRect.drX1;
    drLocal.drY0= bi->biTopPosition.lpYPixels;
    drLocal.drY1= bi->biBelowPosition.lpYPixels;

    docUnionRectangle( drChanged, drChanged, &drLocal );

    return;
    }

