/************************************************************************/
/*  Ted: interaction with the page layout tool.				*/
/************************************************************************/

#   include	"config.h"

#   include	"tedApp.h"
#   include	"tedRuler.h"
#   include	"docPs.h"

#   include	<debugon.h>

/************************************************************************/
/*									*/
/*  Show or create a print tool.					*/
/*									*/
/************************************************************************/

static void tedSetPageLayout(	void *				voidea,
				const DocumentGeometry *	dg )
    {
    EditApplication *		ea= (EditApplication *)voidea;
    Display *			display= ea->eaDisplay;
    EditDocument *		ed= ea->eaCurrentDocument;
    AppDrawingData *		add;
    TedDocument *		td;
    BufferDocument *		bd;
    DocumentProperties *	dp;

    double			horPixPerMM;
    double			verPixPerMM;
    double			xfac;
    double			yfac;

    int				changed= 0;

    int				scrolledX;
    int				scrolledY;

    if  ( ! ed )
	{ XDEB(ed); return;	}

    add= &(ed->edDrawingData);
    td= (TedDocument *)ed->edPrivateData;
    bd= td->tdDocument;
    dp= &(bd->bdProperties);

    appSetDocumentGeometry( &(dp->dpGeometry), dg, &changed );

    if  ( changed )
	{
	int			hasSelection= tedHasSelection( td );

	appDocumentChanged( ed, 1 );

	appGetFactors( ea, &horPixPerMM, &verPixPerMM, &xfac, &yfac );
	tedSetRectangles( ed, &(dp->dpGeometry), xfac );

	tedLayoutDocument( ed );

	if  ( hasSelection )
	    {
	    tedCalculateSelectedLines( td );
	    tedSelectionCoordinates( &(td->tdSelection), add );

	    tedSelectionRectangle( &(td->tdSelectedRectangle),
						 add, &(td->tdSelection) );
	    }

	appDocSetScrollbarValues( ed );
	appSetShellConstraints( ed );

	if  ( TED_Inspector )
	    { tedAdaptFormatToolToDocument( TED_Inspector, ed );	}

	if  ( hasSelection )
	    {
	    tedDocAdaptHorizontalRuler( ed, td->tdSelection.bsBegin.bpBi, 1 );

	    tedScrollToSelection( ed, &scrolledX, &scrolledY );

	    if  ( td->tdObjectSelected )
		{ tedMoveObjectWindows( ed->edDocumentWidget, ed );	}
	    }

	XClearArea( display, XtWindow( ed->edDocumentWidget ),
							0, 0, 0, 0, True );
	tedAdaptLeftRuler( ed->edLeftRuler, ed->edLeftRulerWidget,
			    ed->edVisibleRect.drY0, ed->edVisibleRect.drY1,
			    add->addPageStepPixels );
	}

    return;
    }

void tedAppShowPageTool(		Widget		pageOption,
					XtPointer	voidea,
					XtPointer	voidcbs	 )
    {
    EditApplication *	ea=	(EditApplication *)voidea;
    EditDocument *	ed= ea->eaCurrentDocument;

    if  ( ! ed )
	{ XDEB(ed); return;	}

    if  ( ! ea->eaPageTool )
	{
	char *			pixmapName= "page";

	ea->eaPageTool= appMakePageTool( ea, pageOption,
					    pixmapName, tedSetPageLayout );
	}
    else{ appShowPageTool( ea->eaPageTool ); }

    tedAdaptPageToolToDocument( ea, ed );

    return;
    }

/************************************************************************/
/*									*/
/*  Adapt the page tool to a document.					*/
/*									*/
/************************************************************************/
void tedAdaptPageToolToDocument(	EditApplication *	ea,
					EditDocument *		ed )
    {
    TedDocument *		td;
    BufferDocument *		bd;
    DocumentProperties *	dp;

    if  ( ! ea->eaPageTool )
	{ return;	}

    td= (TedDocument *)ed->edPrivateData;
    bd= td->tdDocument;
    dp= &(bd->bdProperties);

    appPageToolSetProperties( ea->eaPageTool, &(dp->dpGeometry) );

    appEnablePageTool( ea->eaPageTool, ! ed->edIsReadonly );

    return;
    }

/************************************************************************/
/*									*/
/*  Derive private ractangles from document properties.			*/
/*									*/
/*  1)  HACK: IMAGEABLE AREA OF PRINTER					*/
/*									*/
/************************************************************************/
# define MINMARG	300

void tedSetRectangles(	EditDocument *		ed,
			DocumentGeometry *	dg,
			double			xfac )
    {
    AppDrawingData *		add= &(ed->edDrawingData);
    double			yfac= xfac;

    int				leftMarginPixels;
    int				rightMarginPixels;

    /*  1  */
    if  ( dg->dgLeftMarginTwips < MINMARG )
	{ dg->dgLeftMarginTwips= MINMARG;	}
    if  ( dg->dgTopMarginTwips < MINMARG )
	{ dg->dgTopMarginTwips= MINMARG;	}
    if  ( dg->dgRightMarginTwips < MINMARG )
	{ dg->dgRightMarginTwips= MINMARG;	}
    if  ( dg->dgBottomMarginTwips < MINMARG )
	{ dg->dgBottomMarginTwips= MINMARG;	}

    leftMarginPixels= TWIPStoPIXELS( xfac, dg->dgLeftMarginTwips );
    rightMarginPixels= TWIPStoPIXELS( xfac, dg->dgRightMarginTwips );
    add->addBottomMarginPixels= TWIPStoPIXELS( xfac, dg->dgBottomMarginTwips );

    add->addBackRect.drX0= 0;
    add->addBackRect.drY0= 0;
    add->addBackRect.drX1= TWIPStoPIXELS( xfac, dg->dgPaperWideTwips );
    add->addBackRect.drY1= TWIPStoPIXELS( yfac, dg->dgPaperHighTwips );

    add->addPaperRect.drX0= 0;
    add->addPaperRect.drY0= 0;
    add->addPaperRect.drX1= TWIPStoPIXELS( xfac, dg->dgPaperWideTwips );
    add->addPaperRect.drY1= TWIPStoPIXELS( yfac, dg->dgPaperHighTwips );

    add->addDocRect.drX0= leftMarginPixels;
    add->addDocRect.drY0= TWIPStoPIXELS( xfac, dg->dgTopMarginTwips );
    add->addDocRect.drX1= add->addBackRect.drX1- rightMarginPixels;
    add->addDocRect.drY1= add->addBackRect.drY1- add->addBottomMarginPixels;

    add->addPageStepPixels= add->addPaperRect.drY1- add->addPaperRect.drY0+
							add->addPageGapPixels;

    appSetShellConstraints( ed );

    return;
    }

/************************************************************************/
/*									*/
/*  (re)Calculate the layout of a whole document.			*/
/*									*/
/************************************************************************/

int tedLayoutDocument(	EditDocument *		ed )
    {
    TedDocument *		td= (TedDocument *)ed->edPrivateData;
    BufferDocument *		bd= td->tdDocument;

    DocumentProperties *	dp= &(bd->bdProperties);
    AppDrawingData *		add= &(ed->edDrawingData);

    DocumentRectangle		drChanged;

    drChanged= add->addBackRect;

    if  ( tedLayoutItem( &(bd->bdItem), dp, add, &drChanged ) )
	{ LDEB(1); return -1;	}

    /* NO.. Whole pages, done by tedLayoutItem()

    add->addDocRect.drY1= bd->bdItem.biBelowPosition.lpYPixels;
    add->addBackRect.drY1= bd->bdItem.biBelowPosition.lpYPixels+
						add->addBottomMarginPixels;
    */

    return 0;
    }

/************************************************************************/
/*									*/
/*  Show the page setup dialog for a certain document.			*/
/*									*/
/************************************************************************/
void tedDocShowPageTool(		Widget		pageOption,
					XtPointer	voided,
					XtPointer	voidcbs	 )
    {
    EditDocument *	ed= (EditDocument *)voided;

    tedAppShowPageTool( pageOption, (void *)ed->edApplication, voidcbs );
    }
