/************************************************************************/
/*									*/
/*  The Row page of the Format Tool.					*/
/*									*/
/************************************************************************/

#   include	"config.h"

#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<debugon.h>

#   include	<Xm/Text.h>
#   include	<Xm/ToggleB.h>

#   include	<appUtil.h>
#   include	<appGeoString.h>
#   include	<appUnit.h>

#   include	"tedApp.h"
#   include	"tedFormatTool.h"

/************************************************************************/
/*									*/
/*  Refresh the row page.						*/
/*									*/
/************************************************************************/

void tedFormatToolRefreshRowPage(	RowTool *	rt )
    {
    RowProperties *	rp= &(rt->rtPropertiesChosen);
    CellProperties *	cp;

    char		scratch[50];

    appIntegerToTextWidget( rt->rtNumberText, rt->rtRow0- rt->rtRow00+ 1 );

    tedFormatRefreshHeightChooser( &(rt->rtHeightChooser),
							rp->rpHeightTwips );

    appGeoLengthToString( scratch, rp->rpHalfGapWidthTwips, UNITtyPOINTS );
    XmTextSetString( rt->rtCellMarginText, scratch );

    appGeoLengthToString( scratch, rp->rpLeftIndentTwips, UNITtyPOINTS );
    XmTextSetString( rt->rtLeftIndentText, scratch );

    cp= rp->rpCells;
    tedFormatReflectBorder( rt->rtTopBorderToggle,
					&(cp[rt->rtCol0].cpTopBorder) );
    tedFormatReflectBorder( rt->rtBottomBorderToggle,
					&(cp[rt->rtCol0].cpBottomBorder) );

    XmToggleButtonSetState( rt->rtKeepOnPageToggle,
						rp->rpKeepOnPage, False );
    XmToggleButtonSetState( rt->rtIsTableHeaderToggle,
						rp->rpIsTableHeader, False );
    }

/************************************************************************/
/*									*/
/*  The 'Change' button of the row tool has been pushed.		*/
/*									*/
/************************************************************************/

static void tedFormatChangeRowPushed(	Widget		w,
					XtPointer	voidrt,
					XtPointer	voidpbcs )
    {
    RowTool *		rt= (RowTool *)voidrt;
    RowProperties *	rp= &(rt->rtPropertiesChosen);

    int			width;
    int			changed;

    int			res;

    if  ( tedFormatToolGetHeight( &(rp->rpHeightTwips),
						&(rt->rtHeightChooser) ) )
	{ return;	}

    width= rp->rpHalfGapWidthTwips;
    if  ( appGetLengthFromTextWidget( rt->rtCellMarginText, &width, &changed,
							UNITtyPOINTS, 1 ) )
	{ return;	}
    rp->rpHalfGapWidthTwips= width;

    res= tedFormatToolGetRowLeftIndent( rp, rt->rtPageLeftMargin, &width,
							rt->rtLeftIndentText );
    if  ( res != 0 )
	{ return;	}
    rp->rpLeftIndentTwips= width;

    tedSetTableProperties( rt->rtApplication,
				0, rt->rtPropertiesSet.rpCellCount- 1,
				rt->rtRow0, rt->rtRow1,
				rp );

    return;
    }

/************************************************************************/
/*									*/
/*  The row height changed, check it.					*/
/*									*/
/************************************************************************/

static void tedRowHeightChanged(	Widget		w,
					XtPointer	voidrt,
					XtPointer	voidpbcs )
    {
    RowTool *		rt= (RowTool *)voidrt;
    int			value;

    if  ( ! tedFormatToolGetHeight( &value, &(rt->rtHeightChooser) ) )
	{
	tedFormatRefreshHeightChooser( &(rt->rtHeightChooser), value );
	}
    }

static void tedRowCellMarginChanged(	Widget		w,
					XtPointer	voidrt,
					XtPointer	voidpbcs )
    {
    RowTool *		rt= (RowTool *)voidrt;
    RowProperties *	rp= &(rt->rtPropertiesChosen);

    int			width;

    if  ( ! tedFormatToolGetGapWidth( rp, &width, rt->rtCellMarginText ) )
	{
	char	scratch[50];

	appGeoLengthToString( scratch, width, UNITtyPOINTS );
	XmTextSetString( rt->rtCellMarginText, scratch );
	}
    }

static void tedRowLeftIndentChanged(	Widget		w,
					XtPointer	voidrt,
					XtPointer	voidpbcs )
    {
    RowTool *		rt= (RowTool *)voidrt;
    RowProperties *	rp= &(rt->rtPropertiesChosen);

    int			value;

    if  ( ! tedFormatToolGetRowLeftIndent( rp, rt->rtPageLeftMargin, &value,
						    rt->rtLeftIndentText ) )
	{
	char	scratch[50];

	appGeoLengthToString( scratch, value, UNITtyPOINTS );
	XmTextSetString( rt->rtLeftIndentText, scratch );
	}
    }

/************************************************************************/
/*									*/
/*  The 'Revert' button of the row tool has been pushed.		*/
/*									*/
/************************************************************************/

static void tedFormatRevertRowPushed(	Widget		w,
					XtPointer	voidrt,
					XtPointer	voidpbcs	 )
    {
    RowTool *	rt= (RowTool *)voidrt;

    docCopyRowProperties( &(rt->rtPropertiesChosen), &(rt->rtPropertiesSet) );

    tedFormatToolRefreshRowPage( rt );

    return;
    }

static void tedFormatPreviousRow(	Widget		w,
					XtPointer	voidrt,
					XtPointer	voidpbcs	 )
    {
    RowTool *		rt= (RowTool *)voidrt;
    EditApplication *	ea= rt->rtApplication;
    EditDocument *	ed= ea->eaCurrentDocument;

    if  ( ! ed )
	{ XDEB(ed); return;	}

    tedAppSetTableSelection( ed, rt->rtCol0, rt->rtCol1,
					    rt->rtRow0- 1, rt->rtRow1- 1 );
    }

static void tedFormatSelectRow(		Widget		w,
					XtPointer	voidrt,
					XtPointer	voidpbcs	 )
    {
    RowTool *		rt= (RowTool *)voidrt;
    EditApplication *	ea= rt->rtApplication;
    EditDocument *	ed= ea->eaCurrentDocument;

    if  ( ! ed )
	{ XDEB(ed); return;	}

    tedAppSetTableSelection( ed, 0,
			    rt->rtPropertiesChosen.rpCellCount -1,
			    rt->rtRow0, rt->rtRow1 );
    }

static void tedFormatNextRow(		Widget		w,
					XtPointer	voidrt,
					XtPointer	voidpbcs	 )
    {
    RowTool *		rt= (RowTool *)voidrt;
    EditApplication *	ea= rt->rtApplication;
    EditDocument *	ed= ea->eaCurrentDocument;

    if  ( ! ed )
	{ XDEB(ed); return;	}

    tedAppSetTableSelection( ed, rt->rtCol0, rt->rtCol1,
					    rt->rtRow0+ 1, rt->rtRow1+ 1 );
    }

/************************************************************************/
/*									*/
/*  The 'Delete' button of the row tool has been pushed.		*/
/*									*/
/************************************************************************/

static void tedFormatDeleteRow(		Widget		w,
					XtPointer	voidrt,
					XtPointer	voidpbcs	 )
    {
    RowTool *		rt= (RowTool *)voidrt;
    EditApplication *	ea= rt->rtApplication;
    EditDocument *	ed= ea->eaCurrentDocument;

    if  ( tedDeleteRowsFromTable( ed, rt->rtRow0, rt->rtRow1 ) )
	{ LLDEB(rt->rtRow0,rt->rtRow1); return;	}

    appDocumentChanged( ed, 1 );

    return;
    }

/************************************************************************/
/*									*/
/*  Insert/Append Row buttons have been pushed.				*/
/*									*/
/************************************************************************/

static void tedFormatInsertRow(	Widget		w,
				XtPointer	voidrt,
				XtPointer	voidpbcs	 )
    {
    RowTool *		rt= (RowTool *)voidrt;
    EditApplication *	ea= rt->rtApplication;

    EditDocument *	ed= ea->eaCurrentDocument;

    if  ( ! ed )
	{ XDEB(ed); return;	}

    tedInsertRowInTable( ed->edDocumentWidget, ed );

    appDocumentChanged( ed, 1 );
    }

static void tedFormatAppendRow(	Widget		w,
				XtPointer	voidrt,
				XtPointer	voidpbcs	 )
    {
    RowTool *		rt= (RowTool *)voidrt;
    EditApplication *	ea= rt->rtApplication;
    EditDocument *	ed= ea->eaCurrentDocument;

    if  ( ! ed )
	{ XDEB(ed); return;	}

    tedAppendRowToTable( ed->edDocumentWidget, ed );

    appDocumentChanged( ed, 1 );
    }

static void tedRowHeightChosen(	Widget		w,
				XtPointer	voidrt,
				XtPointer	voidtbcs )
    {
    RowTool *		rt= (RowTool *)voidrt;

    int			defaultValue;

    defaultValue= rt->rtPropertiesChosen.rpHeightTwips;
    if  ( defaultValue == 0 )
	{ defaultValue= rt->rtPropertiesSet.rpHeightTwips;	}
    if  ( defaultValue == 0 )
	{ defaultValue= 240;	}

    tedFormatHeightChosen( w, &(rt->rtHeightChooser), defaultValue );

    return;
    }

/************************************************************************/
/*									*/
/*  A toggle was activated.						*/
/*									*/
/************************************************************************/

static void tedRowToggleKeepOnPage(	Widget		w,
					XtPointer	voidrt,
					XtPointer	voidtbcs )
    {
    RowTool *				rt= (RowTool *)voidrt;
    RowProperties *			rp= &(rt->rtPropertiesChosen);
    XmToggleButtonCallbackStruct *	tbcs;

    tbcs= (XmToggleButtonCallbackStruct *)voidtbcs;

    rp->rpKeepOnPage= tbcs->set;

    return;
    }

static void tedRowToggleIsTableHeader(	Widget		w,
					XtPointer	voidrt,
					XtPointer	voidtbcs )
    {
    RowTool *				rt= (RowTool *)voidrt;
    RowProperties *			rp= &(rt->rtPropertiesChosen);
    XmToggleButtonCallbackStruct *	tbcs;

    tbcs= (XmToggleButtonCallbackStruct *)voidtbcs;

    rp->rpIsTableHeader= tbcs->set;

    return;
    }

/************************************************************************/
/*									*/
/*  One of the 'Row Border' toggles has been activated.			*/
/*									*/
/************************************************************************/

static void tedRowToggleBorder(	Widget		w,
				XtPointer	voidrt,
				XtPointer	voidtbcs )
    {
    RowTool *				rt= (RowTool *)voidrt;
    RowProperties *			rp= &(rt->rtPropertiesChosen);
    CellProperties *			cp;
    XmToggleButtonCallbackStruct *	tbcs;
    int					col;

    tbcs= (XmToggleButtonCallbackStruct *)voidtbcs;

    if  ( w == rt->rtTopBorderToggle )
	{
	cp= rp->rpCells;
	for ( col= 0; col < rp->rpCellCount; cp++, col++ )
	    { tedFormatToggleBorder( &(cp->cpTopBorder), tbcs->set ); }

	return;
	}

    if  ( w == rt->rtBottomBorderToggle )
	{
	cp= rp->rpCells;
	for ( col= 0; col < rp->rpCellCount; cp++, col++ )
	    { tedFormatToggleBorder( &(cp->cpBottomBorder), tbcs->set ); }

	return;
	}

    XDEB(w); return;
    }

/************************************************************************/
/*									*/
/*  Make a row tool. I.E. the 'Row' page of the format tool.		*/
/*									*/
/************************************************************************/

void tedFormatFillRowPage(	RowTool *			rt,
				const RowPageResources *	rpr,
				InspectorSubject *			is,
				Widget				pageWidget,
				const InspectorSubjectResources *	isr )
    {
    Widget	rowLabel;
    Widget	widthLabel;
    Widget	leftLabel;

    Widget	row;

    const int	textColumns= 10;

    /**************/
    rt->rtCol0= -1;
    rt->rtCol1= -1;
    rt->rtRow00= -1;
    rt->rtRow0= -1;
    rt->rtRow1= -1;
    rt->rtRow11= -1;

    docInitRowProperties( &(rt->rtPropertiesSet) );
    docInitRowProperties( &(rt->rtPropertiesChosen) );

    /**************/
    appMakeLabelAndTextRow( &row, &rowLabel, &(rt->rtNumberText),
			    pageWidget, isr->isrSubjectName, textColumns, 0 );

    /**************/
    tedFormatMakeHeightRow( &row, rt, pageWidget,
			&(rt->rtHeightChooser), tedRowHeightChanged );

    /**************/
    appMakeLabelAndTextRow( &row, &leftLabel, &(rt->rtLeftIndentText),
			    pageWidget, rpr->rprLeftIndent, textColumns, 1 );

    XtAddCallback( rt->rtLeftIndentText, XmNactivateCallback,
				tedRowLeftIndentChanged, (void *)rt );

    /**************/
    appMakeLabelAndTextRow( &row, &widthLabel, &(rt->rtCellMarginText),
			    pageWidget, rpr->rprCellMargin, textColumns, 1 );

    XtAddCallback( rt->rtCellMarginText, XmNactivateCallback,
				tedRowCellMarginChanged, (void *)rt );

    /**************/
    appInspectorMakeToggleRow( &row, pageWidget,
	    &(rt->rtTopBorderToggle), &(rt->rtBottomBorderToggle),
	    rpr->rprTopBorder, rpr->rprBottomBorder,
	    tedRowToggleBorder, tedRowToggleBorder, (void *)rt );

    /**************/
    appInspectorMakeToggleRow( &row, pageWidget,
	    &(rt->rtKeepOnPageToggle), &(rt->rtIsTableHeaderToggle),
	    rpr->rprKeepOnPage, rpr->rprIsTableHeader,
	    tedRowToggleKeepOnPage, tedRowToggleIsTableHeader, (void *)rt );

    appInspectorMakeButtonRow( &row, pageWidget,
	    &(is->isPrevButton), &(is->isNextButton),
	    isr->isrPrevButtonText, isr->isrNextButtonText,
	    tedFormatPreviousRow, tedFormatNextRow, (void *)rt );

    appInspectorMakeButtonRow( &row, pageWidget,
	    &(is->isSelectButton), &(is->isDeleteButton),
	    isr->isrSelectButtonText, isr->isrDeleteButtonText,
	    tedFormatSelectRow, tedFormatDeleteRow, (void *)rt );

    appInspectorMakeButtonRow( &row, pageWidget,
	    &(is->isInsertButton), &(is->isAppendButton),
	    isr->isrInsertButtonText, isr->isrAppendButtonText,
	    tedFormatInsertRow, tedFormatAppendRow, (void *)rt );

    appInspectorMakeButtonRow( &row, pageWidget,
	    &(is->isRevertButton), &(is->isApplyButton),
	    isr->isrRevert, isr->isrApplyToSubject,
	    tedFormatRevertRowPushed, tedFormatChangeRowPushed, (void *)rt );

    XtSetSensitive( rt->rtIsTableHeaderToggle, False );

    return;
    }

void tedFormatFillRowChoosers(		RowTool *			rt,
					const RowPageResources *	rpr )
    {
    tedFormatFillHeightChooser( &(rt->rtHeightChooser),
					    tedRowHeightChosen, (void *)rt,
					    rpr->rprHeightFree,
					    rpr->rprHeightAtLeast,
					    rpr->rprHeightExactly );
    }

void tedFormatFinishRowPage( 	RowTool *			rt,
				const RowPageResources *	rpr )
    {
    appPulldownRefreshWidth( rt->rtHeightChooser.hcMenu );
    }

int tedFormatSaveRowProperties(	TedFormatTool *		tft,
				const BufferItem *	sectBi,
				int			row0 )
    {
    const RowProperties *	rp;

    rp= &(sectBi->biGroupChildren[row0]->biRowProperties);

    if  ( docCopyRowProperties( &(tft->tftRowTool.rtPropertiesChosen), rp ) )
	{ LDEB(1); return -1;	}
    if  ( docCopyRowProperties( &(tft->tftRowTool.rtPropertiesSet), rp ) )
	{ LDEB(1); return -1;	}

    if  ( docCopyRowProperties( &(tft->tftColumnTool.ctPropertiesChosen), rp ) )
	{ LDEB(1); return -1;	}
    if  ( docCopyRowProperties( &(tft->tftColumnTool.ctPropertiesSet), rp ) )
	{ LDEB(1); return -1;	}

    if  ( docCopyRowProperties( &(tft->tftTableTool.ttPropertiesChosen), rp ) )
	{ LDEB(1); return -1;	}
    if  ( docCopyRowProperties( &(tft->tftTableTool.ttPropertiesSet), rp ) )
	{ LDEB(1); return -1;	}

    return 0;
    }
