#   include	"config.h"

#   include	<stdlib.h>
#   include	<stdio.h>

#   include	<X11/IntrinsicP.h>
#   include	<X11/StringDefs.h>

#   include	"tedRuler.h"
#   include	<appDraw.h>

#   include	<debugon.h>

#   define	BACK_MARG(s)	((s)/3)
#   define	WHITE_TOP(s)	BACK_MARG(s)
#   define	WHITE_BOTTOM(s)	((s)-BACK_MARG(s))
#   define	WHITE_HEIGHT(s)	(WHITE_BOTTOM(s)-WHITE_TOP(s))

#   define	TAG_DIST(s)	1
#   define	TAG_TOP(s)	(WHITE_TOP(s)+TAG_DIST(s))
#   define	TAG_BOTTOM(s)	(WHITE_BOTTOM(s)-TAG_DIST(s))
#   define	TAG_HEIGHT(s)	(TAG_BOTTOM(s)-TAG_TOP(s))

#   define	FONT_HEIGHT(s)	(WHITE_HEIGHT(s)-2)
#   define	FONT_BOTTOM(s)	(TAG_BOTTOM(s)-1)

#   define	TICK_DIST(s)	((s)/12)
#   define	TICK_TOP(s)	(WHITE_TOP(s)+TICK_DIST(s))
#   define	TICK_BOTTOM(s)	(WHITE_BOTTOM(s)-TICK_DIST(s))
#   define	TICK_HEIGHT(s)	(TICK_BOTTOM(s)-TICK_TOP(s))

#   define	BUTTON_MARG		1

#   define	BUTTON_DIST(s)		(((s)/6)-1)
#   define	BUTTON_RECTH(s)		((s)/6)
#   define	BUTTON_TRIH(s)		((s)/6)
#   define	BUTTON_HWIDTH(s)	((s)/6)


#   define	HBUTTON_BASE(s)		BUTTON_DIST(s)
#   define	HBUTTON_APEX(s)		(HBUTTON_BASE(s)+ \
					    BUTTON_RECTH(s)+BUTTON_TRIH(s))

#   define	SBUTTON_BASE(s)		((s)-BUTTON_DIST(s))
#   define	SBUTTON_APEX(s)		(SBUTTON_BASE(s)- \
					    BUTTON_RECTH(s)-BUTTON_TRIH(s))

#   define	BUTTON_LEFT(v,s)	((v)-BUTTON_HWIDTH((s))-BUTTON_MARG)
#   define	BUTTON_RIGHT(v,s)	((v)+BUTTON_HWIDTH((s))+BUTTON_MARG)

#   define	TAB_HEIGHT(s)		((s)/6)
#   define	TAB_TOP(s)		(WHITE_BOTTOM(s)-TAB_HEIGHT(s))
#   define	TAB_WIDTH(s)		((s)/6)
#   define	TAB_THICK(s)		3

typedef struct TedTopRuler
    {
    RulerData		ttrRulerData;
					/********************************/
					/*  Describe an area that is	*/
					/*  not to be used.		*/
					/********************************/
    int			ttrColX0;
    int			ttrColX1;
					/********************************/
					/*  Describe the area that is	*/
					/*  accessible as values.	*/
					/********************************/

    int			ttrFirstIndent;		/*  pixels		*/
    int			ttrLeftIndent;		/*  pixels		*/
    int			ttrRightIndent;		/*  pixels		*/

    int			ttrTabCount;
    TabStop *		ttrTabStops;
    int			ttrTabKind;		/************************/
						/*  tsKind for fresh	*/
						/*  tabs.		*/
						/************************/
    int			ttrColumnCount;
    ColumnSeparator *	ttrColumns;
    } TedTopRuler;

void * tedMakeTopRuler(	int			height,
			double			pixPerMM,
			double			magnification,

			Pixel			forePixel,
			Pixel			backPixel,
			Pixel			topPixel,
			Pixel			bottomPixel,

			int			leftRulerWidth,
			int			rightRulerWidth,

			int			documentC0,
			int			documentC1,
			int			visibleC0,
			int			visibleC1,
			int			rulerC1,

			int			whatUnit )
    {
    TedTopRuler *		ttr;
    RulerData *			rd;

    ttr= (TedTopRuler *)malloc( sizeof( TedTopRuler ) );
    if  ( ! ttr )
	{ XDEB(ttr); return (void *)0;	}
    rd= &(ttr->ttrRulerData);

    appInitRulerData( &(ttr->ttrRulerData) );

    rd->rdUnit= whatUnit;
    appRulerCalculateIncrements( rd, pixPerMM, magnification );

    rd->rdMinUsed= leftRulerWidth;
    rd->rdMaxUsed= 0;
    rd->rdExtraAfterMaxUsed= rightRulerWidth;

    rd->rdDocumentC0= documentC0;
    rd->rdDocumentC1= documentC1;
    rd->rdVisibleC0= visibleC0;
    rd->rdVisibleC1= visibleC1;
    rd->rdRulerC1= rulerC1;

    ttr->ttrColX0= documentC0;
    ttr->ttrColX1= documentC1;

    rd->rdSizeAcross= height;
    rd->rdSizeAlong= 0;

    rd->rdForePixel= forePixel;
    rd->rdBackPixel= backPixel;
    rd->rdTopPixel= topPixel;
    rd->rdBottomPixel= bottomPixel;

    ttr->ttrTabCount= 0;
    ttr->ttrTabStops= (TabStop *)0;
    ttr->ttrTabKind= DOCtkLEFT;

    ttr->ttrColumnCount= 0;
    ttr->ttrColumns= (ColumnSeparator *)0;

    ttr->ttrFirstIndent= -1;
    ttr->ttrLeftIndent= -1;
    ttr->ttrRightIndent= -1;

    return (void *)ttr;
    }

void tedFreeHorizontalRuler(	Display *	display,
				void *		voidttr )
    {
    TedTopRuler *		ttr= (TedTopRuler *)voidttr;

    if  ( ttr->ttrTabStops )
	{ free( ttr->ttrTabStops );			}

    appCleanRulerData( display, &(ttr->ttrRulerData) );

    free( ttr );

    return;
    }

/************************************************************************/
/*									*/
/*  Work around a bug in XDrawLines().					*/
/*									*/
/************************************************************************/
static void editDrawLines(	Display *	display,
				Window		win,
				GC		gc,
				XPoint *	points,
				int		npoints )
    {
    int		i;

    /*
    XDrawLines( display, win, gc, points, npoints, CoordModeOrigin );
    */

    npoints--;

    for ( i= 0; i < npoints; points++, i++ )
	{
	XDrawLine( display, win, gc,
				points[0].x, points[0].y,
				points[1].x, points[1].y );
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Draw the little button for the 'First Indent' of a paragraph	*/
/*									*/
/*	       v							*/
/*	4 ---- . ---- 0 = 5						*/
/*	|      .      |							*/
/*	|      .      |							*/
/*	3      .      1							*/
/*	 \     .     / 							*/
/*	  \    .    /  							*/
/*	   \   .   /   							*/
/*	    \  .  /    							*/
/*	     \ . /     							*/
/*	       2       							*/
/*									*/
/************************************************************************/

static void tedHangingButtonRect(	DocumentRectangle *	drButton,
					int			value,
					int			sizeAcross )
    {
    drButton->drX0= value- BUTTON_HWIDTH(sizeAcross)- 1;
    drButton->drX1= value+ BUTTON_HWIDTH(sizeAcross)+ 1;

    drButton->drY0= BUTTON_HWIDTH(sizeAcross)- 1;
    drButton->drY1= HBUTTON_APEX(sizeAcross)+ 1;

    return;
    }

static void tedRulerDrawHangingButton(	int		value,
					int		ox,
					int		sizeAcross,
					Pixel		forePixel,
					Pixel		backPixel,
					Pixel		topPixel,
					Pixel		bottomPixel,
					Pixmap		stipple,
					Display *	display,
					Window		win,
					GC		gc )
    {
    XPoint	points[6];
    int		w= BUTTON_MARG;

    value -= ox;

    points[0].x= value+ BUTTON_HWIDTH(sizeAcross);
    points[0].y= HBUTTON_BASE(sizeAcross);

    points[1].x= points[0].x;
    points[1].y= HBUTTON_BASE(sizeAcross)+ BUTTON_RECTH(sizeAcross);

    points[2].x= value;
    points[2].y= HBUTTON_APEX(sizeAcross);

    points[3].x= value- BUTTON_HWIDTH(sizeAcross);
    points[3].y= points[1].y;

    points[4].x= points[3].x;
    points[4].y= points[0].y;

    points[5]= points[0];

    XSetForeground( display, gc, backPixel );
    if  ( stipple )
	{ XSetFillStyle( display, gc, FillStippled ); }

    XFillPolygon( display, win, gc, points, 6, Convex, CoordModeOrigin );

    if  ( stipple )
	{ XSetFillStyle( display, gc, FillSolid ); }

    XSetForeground( display, gc, forePixel );
    editDrawLines( display, win, gc, points, 6 );

    if  ( ! stipple )
	{
	points[0].x -= w; points[0].y += (w+1)/2;
	points[1].x -= w;
	points[2].y -= w;
	points[3].x += w;
	points[4].x += w; points[4].y += w;
	points[5].x -= w; points[5].y += w;

	XSetForeground( display, gc, bottomPixel );
	editDrawLines( display, win, gc, points, 4 );
	XSetForeground( display, gc, topPixel );
	editDrawLines( display, win, gc, points+ 3, 3 );
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Draw the little buttons for the 'Left, Right Indent' of a paragraph	*/
/*									*/
/*	       3   							*/
/*	     / . \     							*/
/*	    /  .  \    							*/
/*	   /   .   \   							*/
/*	  /    .    \  							*/
/*	 /     .     \ 							*/
/*	4      .      2							*/
/*	|      .      |							*/
/*	|      .      |							*/
/*  5 = 0 ---- . ---- 1							*/
/*	       v							*/
/*									*/
/************************************************************************/

static void tedStandingButtonRect(	DocumentRectangle *	drButton,
					int			value,
					int			sizeAcross )
    {
    drButton->drX0= value- BUTTON_HWIDTH(sizeAcross)- 1;
    drButton->drX1= value+ BUTTON_HWIDTH(sizeAcross)+ 1;

    drButton->drY0= SBUTTON_APEX(sizeAcross)- 1;
    drButton->drY1= SBUTTON_BASE(sizeAcross)+ 1;

    return;
    }

static void tedRulerDrawStandingButton(		int		value,
						int		ox,
						int		sizeAcross,
						Pixel		forePixel,
						Pixel		backPixel,
						Pixel		topPixel,
						Pixel		bottomPixel,
						Pixmap		stipple,
						Display *	display,
						Window		win,
						GC		gc )
    {
    XPoint	points[6];
    int		w= BUTTON_MARG;

    value -= ox;

    points[0].x= value- BUTTON_HWIDTH(sizeAcross);
    points[0].y= SBUTTON_BASE(sizeAcross);

    points[1].x= value+ BUTTON_HWIDTH(sizeAcross);
    points[1].y= points[0].y;

    points[2].x= points[1].x;
    points[2].y= SBUTTON_BASE(sizeAcross)- BUTTON_RECTH(sizeAcross);

    points[3].x= value;
    points[3].y= SBUTTON_APEX(sizeAcross);

    points[4].x= points[0].x;
    points[4].y= points[2].y;

    points[5]= points[0];

    XSetForeground( display, gc, backPixel );
    if  ( stipple )
	{ XSetFillStyle( display, gc, FillStippled ); }

    XFillPolygon( display, win, gc, points, 6, Convex, CoordModeOrigin );

    if  ( stipple )
	{ XSetFillStyle( display, gc, FillSolid ); }

    XSetForeground( display, gc, forePixel );
    editDrawLines( display, win, gc, points, 6 );

    if  ( ! stipple )
	{
	points[0].x += (w+1)/2; points[0].y -= w;
	points[1].x -= w; points[1].y -= w;
	points[2].x -= w;
	points[3].y += (w+1)/2;
	points[4].x += w;
	points[5].x += w; points[5].y -= w;

	XSetForeground( display, gc, bottomPixel );
	editDrawLines( display, win, gc, points, 4 );
	XSetForeground( display, gc, topPixel );
	editDrawLines( display, win, gc, points+ 3, 3 );
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Draw the little holes for the table columns on the ruler.		*/
/*									*/
/*	3-------------2							*/
/*	|      .      |							*/
/*	|      .      |							*/
/*	|      .      |							*/
/*	|      .      |							*/
/*	|      .      |							*/
/*  4 = 0 ---- . ---- 1							*/
/*	       v							*/
/*									*/
/************************************************************************/

static void tedRulerColumnControlRect(	DocumentRectangle *	drButton,
					int			x0,
					int			x1,
					int			sizeAcross )
    {
    if  ( x1- x0 < BUTTON_HWIDTH( sizeAcross ) )
	{
	int	m= ( x1+ x0 )/ 2;

	x0= m- BUTTON_HWIDTH( sizeAcross )/2;
	x1= m+ BUTTON_HWIDTH( sizeAcross )/2;
	}

    drButton->drX0= x0- 1;
    drButton->drX1= x1+ 1;

    drButton->drY0= WHITE_TOP(sizeAcross)- 1;
    drButton->drY1= WHITE_BOTTOM(sizeAcross)+ 1;

    return;
    }

static void tedRulerDrawColumnControl(	int		x0,
					int		x1,
					int		ox,
					int		sizeAcross,
					Pixel		forePixel,
					Pixel		backPixel,
					Pixel		topPixel,
					Pixel		bottomPixel,
					Pixmap		stipple,
					Display *	display,
					Window		win,
					GC		gc )
    {
    XPoint	points[5];
    int		w= BUTTON_MARG;

    x0 -= ox;
    x1 -= ox;

    if  ( x1- x0 < BUTTON_HWIDTH( sizeAcross ) )
	{
	int	m= ( x1+ x0 )/ 2;

	x0= m- BUTTON_HWIDTH( sizeAcross )/2;
	x1= m+ BUTTON_HWIDTH( sizeAcross )/2;
	}

    points[0].x= x0;
    points[0].y= WHITE_BOTTOM(sizeAcross);

    points[1].x= x1;
    points[1].y= points[0].y;

    points[2].x= points[1].x;
    points[2].y= WHITE_TOP(sizeAcross);

    points[3].x= points[0].x;
    points[3].y= points[2].y;

    points[4]= points[0];

    XSetForeground( display, gc, backPixel );
    if  ( stipple )
	{ XSetFillStyle( display, gc, FillStippled ); }

    XFillPolygon( display, win, gc, points, 5, Convex, CoordModeOrigin );

    if  ( stipple )
	{ XSetFillStyle( display, gc, FillSolid ); }

    XSetForeground( display, gc, forePixel );
    editDrawLines( display, win, gc, points, 5 );

    if  ( ! stipple )
	{
	points[0].x += w; points[0].y -= w;
	points[1].x -= w; points[1].y -= w;
	points[2].x -= w; points[2].y += w;
	points[3].x += w; points[3].y += w;
	points[4].x += w; points[4].y -= w;

	XSetForeground( display, gc, topPixel );
	editDrawLines( display, win, gc, points, 3 );
	XSetForeground( display, gc, bottomPixel );
	editDrawLines( display, win, gc, points+ 2, 2 );
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Draw a 'Tab' symbol.						*/
/*									*/
/************************************************************************/

static void tedTabSymbolRect(	DocumentRectangle *	drButton,
				int			x,
				int			y,
				int			sizeAcross )
    {
    int		thick= TAB_THICK(sizeAcross);
    int		tabWidth= TAB_WIDTH(sizeAcross);
    int		tabHeight= TAB_HEIGHT(sizeAcross);

    drButton->drX0= x+ (thick+1)/2- tabWidth;
    drButton->drX1= x- thick/2+ tabWidth;

    drButton->drY0= y- tabHeight- 1;
    drButton->drY1= y+ 1;

    return;
    }

static void tedDrawTabSymbol(	Display *	display,
				Window		win,
				GC		gc,
				int		x,
				int		ox,
				int		y,
				int		sizeAcross,
				int		kind )
    {
    int		thick= TAB_THICK(sizeAcross);
    int		tabWidth= TAB_WIDTH(sizeAcross);
    int		tabHeight= TAB_HEIGHT(sizeAcross);

    int		xx;

    x -= ox;

    xx= x- thick/2;
    XFillRectangle( display, win, gc, xx, y- tabHeight, thick, tabHeight );

    if  ( kind == DOCtkLEFT		||
	  kind == DOCtkCENTRE		||
	  kind == DOCtkDECIMAL		)
	{
	xx= x- thick/2;

	/*  laying right	*/
	XFillRectangle( display, win, gc, xx, y- thick, tabWidth, thick );
	}

    if  ( kind == DOCtkRIGHT		||
	  kind == DOCtkCENTRE		||
	  kind == DOCtkDECIMAL		)
	{
	xx= x+ (thick+1)/2- tabWidth;

	/*  laying left	*/
	XFillRectangle( display, win, gc, xx, y- thick, tabWidth, thick );
	}

    if  ( kind == DOCtkDECIMAL	)
	{
	int	dec= ( thick+ 1 )/ 2;

	xx= x- thick/2+ tabWidth- dec;

	/*  laying right	*/
	XFillRectangle( display, win, gc, xx, y- tabHeight, dec, dec );
	}

    return;
    }

static void tedTabButtonRectangle(	DocumentRectangle *	drButton,
					int			x,
					int			y,
					int			sizeAcross )
    {
    tedTabSymbolRect( drButton, x, y, sizeAcross );

    drButton->drX0 -= 3;
    drButton->drX1 += 3;

    drButton->drY0 -= 5;
    drButton->drY1 += 0;

    return;
    }

static void tedDrawTabButton(	Display *	display,
				Window		win,
				RulerData *	rd,
				int		x,
				int		y,
				int		sizeAcross,
				int		kind )
    {
    int			tw= TAB_WIDTH(sizeAcross);
    DocumentRectangle	dr;
    int			ox= 0;

    tedTabButtonRectangle( &dr, x, y, sizeAcross );


#   define xfr( x, y, w, h ) \
    XFillRectangle( display, win, rd->rdGc, (x), (y), (w), (h) );

    /*  top, bottom */
    XSetForeground( display, rd->rdGc, rd->rdForePixel );

    xfr(	dr.drX0,	dr.drY0,	dr.drX1- dr.drX0, 1 );
    xfr(	dr.drX0,	dr.drY1,	dr.drX1- dr.drX0, 1 );

    /*  left, right */
    xfr(	dr.drX0,	dr.drY0,	1, dr.drY1- dr.drY0+ 1 );
    xfr(	dr.drX1- 1,	dr.drY0,	1, dr.drY1- dr.drY0 );


    /*  top, left */
    XSetForeground( display, rd->rdGc, rd->rdTopPixel );

    xfr(	dr.drX0+ 1,	dr.drY0+ 1,	dr.drX1- dr.drX0- 2, 1 );
    xfr(	dr.drX0+ 1,	dr.drY0+ 1,	1, dr.drY1- dr.drY0- 1 );

    /*  bottom, right */
    XSetForeground( display, rd->rdGc, rd->rdBottomPixel );

    xfr(	dr.drX0+ 2,	dr.drY1- 1,	dr.drX1- dr.drX0- 3, 1 );
    xfr(	dr.drX1- 2,	dr.drY0+ 2,	1, dr.drY1- dr.drY0- 3 );

    XSetForeground( display, rd->rdGc, rd->rdForePixel );

    tedDrawTabSymbol( display, win, rd->rdGc,
				    x, ox, y- tw/2+ 1, sizeAcross, kind );
    }

/************************************************************************/
/*									*/
/*  Draw a horizontal ruler.						*/
/*									*/
/************************************************************************/

static void tedDrawHorizontalRuler(	Widget			w,
					TedTopRuler *		ttr,
					DocumentRectangle *	drClip,
					int			ox,
					Window			win,
					Display *		display )
    {
    RulerData *		rd= &(ttr->ttrRulerData);
    int			sizeAcross= rd->rdSizeAcross;

    double		item;
    int			tick;
    double		units;

    int			tab;
    TabStop *		ts;

    int			screen= DefaultScreen( display );

    int			whiteY= WHITE_TOP(sizeAcross);
    int			whiteH= WHITE_HEIGHT(sizeAcross);

    int			tagY= TAG_TOP(sizeAcross);
    int			tagH= TAG_HEIGHT(sizeAcross);

    int			tickY= TICK_TOP(sizeAcross);
    int			tickH= TICK_HEIGHT(sizeAcross);

    int			fontBottom= FONT_BOTTOM(sizeAcross);

    ColumnSeparator *	cs;

    DocumentRectangle	drWhite;
    DocumentRectangle	drButton;

    if  ( rd->rdSizeAlong == 0 )
	{
	XtVaGetValues( w, XtNwidth, &rd->rdSizeAlong, NULL );
	rd->rdMaxUsed= rd->rdSizeAlong- rd->rdExtraAfterMaxUsed;
	}

    appRulerDrawBackground( display, win, &(ttr->ttrRulerData),
		    drClip->drX0- ox, drClip->drY0,
		    drClip->drX1- drClip->drX0, drClip->drY1- drClip->drY0 );

    drWhite.drX0= 0;
    drWhite.drX1= rd->rdRulerC1;
    drWhite.drY0= whiteY;
    drWhite.drY1= whiteY+ whiteH;

    if  ( docIntersectRectangle( &drWhite, &drWhite, drClip ) )
	{
	int	x0= drWhite.drX0- ox;
	int	x1= drWhite.drX1- ox;

	XSetForeground( display, rd->rdGc, WhitePixel( display, screen ) );

	if  ( x0 < rd->rdMinUsed )
	    { x0=  rd->rdMinUsed;	}
	if  ( x1 > rd->rdMaxUsed )
	    { x1=  rd->rdMaxUsed;	}

	XFillRectangle( display, win, rd->rdGc,
		    x0, drWhite.drY0,
		    x1- x0+ 1, drWhite.drY1- drWhite.drY0 );
	}

    tedDrawTabButton( display, win, rd, rd->rdMinUsed, whiteY,
						sizeAcross, ttr->ttrTabKind );

    XSetForeground( display, rd->rdGc, rd->rdForePixel );

    item= 0; units= 0; tick= 0;
    while( item <= rd->rdVisibleC1 )
	{
	if  ( item < rd->rdVisibleC0 )
	    {
	    tick++;
	    units += rd->rdUnitsPerTick;
	    item += rd->rdTickDistancePixels;
	    continue;
	    }

	if  ( item != 0 && tick % rd->rdTicksPerText == 0 )
	    {
	    char	scratch[30];
	    int		len;
	    int		wide;

	    appRulerTagText( scratch, &wide, &len, rd, units );

	    if  ( item- ox- wide/2 >= rd->rdMinUsed	&&
		  item- ox+ wide/2 <  rd->rdMaxUsed	)
		{
		XDrawString( display, win, rd->rdGc, 
			    (int)item- ox- wide/2, fontBottom, scratch, len );
		}
	    else{
		XFillRectangle( display, win, rd->rdGc,
					    (int)item- ox, tagY, 1, tagH );
		}
	    }
	else{
	    if  ( tick % rd->rdTicksPerTag == 0 )
		{
		XFillRectangle( display, win, rd->rdGc,
						(int)item- ox, tagY, 1, tagH );
		}
	    else{
		XFillRectangle( display, win, rd->rdGc,
					    (int)item- ox, tickY, 1, tickH );
		}
	    }

	tick++;
	units += rd->rdUnitsPerTick;
	item += rd->rdTickDistancePixels;
	continue;
	}

    cs= ttr->ttrColumns;
    for ( tab= 0; tab < ttr->ttrColumnCount; cs++, tab++ )
	{
	int	m= ( cs->csX0+ cs->csX1 )/ 2;

	if  ( m > rd->rdVisibleC1	||
	      m < rd->rdVisibleC0	)
	    { continue;	}

	tedRulerColumnControlRect( &drButton, cs->csX0, cs->csX1, sizeAcross );

	if  ( ! docIntersectRectangle( &drButton, &drButton, drClip ) )
	    { continue;	}

	tedRulerDrawColumnControl( cs->csX0, cs->csX1, ox, sizeAcross,
			    rd->rdForePixel, rd->rdBackPixel,
			    rd->rdTopPixel, rd->rdBottomPixel,
			    rd->rdBackStipple,
			    display, win, rd->rdGc );
	}

    if  ( ttr->ttrFirstIndent >= 0					&&
	  ttr->ttrFirstIndent >= rd->rdVisibleC0			&&
	  ttr->ttrFirstIndent <= rd->rdVisibleC1			)
	{
	tedHangingButtonRect( &drButton, ttr->ttrFirstIndent, sizeAcross );

	if  ( docIntersectRectangle( &drButton, &drButton, drClip ) )
	    {
	    tedRulerDrawHangingButton( ttr->ttrFirstIndent, ox, sizeAcross,
				rd->rdForePixel, rd->rdBackPixel,
				rd->rdTopPixel, rd->rdBottomPixel,
				rd->rdBackStipple,
				display, win, rd->rdGc );
	    }
	}

    if  ( ttr->ttrLeftIndent >= 0					&&
	  ttr->ttrLeftIndent >= rd->rdVisibleC0			&&
	  ttr->ttrLeftIndent <= rd->rdVisibleC1			)
	{
	tedStandingButtonRect( &drButton, ttr->ttrLeftIndent, sizeAcross );

	if  ( docIntersectRectangle( &drButton, &drButton, drClip ) )
	    {
	    tedRulerDrawStandingButton( ttr->ttrLeftIndent, ox, sizeAcross,
				rd->rdForePixel, rd->rdBackPixel,
				rd->rdTopPixel, rd->rdBottomPixel,
				rd->rdBackStipple,
				display, win, rd->rdGc );
	    }
	}

    if  ( ttr->ttrRightIndent >= 0					&&
	  ttr->ttrRightIndent >= rd->rdVisibleC0			&&
	  ttr->ttrRightIndent <= rd->rdVisibleC1			)
	{
	tedStandingButtonRect( &drButton, ttr->ttrRightIndent, sizeAcross );

	if  ( docIntersectRectangle( &drButton, &drButton, drClip ) )
	    {
	    tedRulerDrawStandingButton( ttr->ttrRightIndent, ox, sizeAcross,
				rd->rdForePixel, rd->rdBackPixel,
				rd->rdTopPixel, rd->rdBottomPixel,
				rd->rdBackStipple,
				display, win, rd->rdGc );
	    }
	}

    XSetForeground( display, rd->rdGc, rd->rdForePixel );

    ts= ttr->ttrTabStops;
    for ( tab= 0; tab < ttr->ttrTabCount; ts++, tab++ )
	{
	if  ( ts->tsPixels < rd->rdVisibleC0	||
	      ts->tsPixels > rd->rdVisibleC1	)
	    { continue;	}

	tedTabSymbolRect( &drButton, ts->tsPixels, whiteY+ whiteH, sizeAcross );

	if  ( ! docIntersectRectangle( &drButton, &drButton, drClip ) )
	    { continue;	}

	tedDrawTabSymbol( display, win, rd->rdGc,
				ts->tsPixels, ox, whiteY+ whiteH,
				sizeAcross, ts->tsKind );
	}

    XSetClipMask( display, rd->rdGc, None );

    return;
    }

void tedExposeHorizontalRuler(	Widget			w,
				void *			voidttr,
				XEvent *		event )
    {
    TedTopRuler *	ttr= (TedTopRuler *)voidttr;
    RulerData *		rd= &(ttr->ttrRulerData);

    Display *		display= XtDisplay( w );
    Window		win= XtWindow( w );

    int			ox= rd->rdVisibleC0- rd->rdMinUsed;
    int			oy= 0;

    DocumentRectangle	drClip;

    if  ( ! rd->rdGc )
	{
	const int	vertical= 0;

	appOpenRulerGc( &(ttr->ttrRulerData),
				vertical, FONT_HEIGHT(rd->rdSizeAcross), w );
	}

    appCollectExposures( &drClip, ox, oy, display, win,
					    rd->rdGc, event );

    tedDrawHorizontalRuler( w, ttr, &drClip, ox, win, display );
    }

void tedHorizontalRulerConfigure(	Widget			w,
					void *			voidttr,
					XEvent *		event,
					Boolean *		pRefused )
    {
    XConfigureEvent *	cevent= &(event->xconfigure);
    TedTopRuler *	ttr= (TedTopRuler *)voidttr;
    RulerData *		rd= &(ttr->ttrRulerData);

    if  ( cevent->type != ConfigureNotify )
	{ return;	}

    rd->rdSizeAlong= cevent->width;
    rd->rdMaxUsed= cevent->width- rd->rdExtraAfterMaxUsed;

    return;
    }

/************************************************************************/
/*									*/
/*  Scroll the horizontal ruler.					*/
/*									*/
/*  1)  Stretch the rectangle to contain whole controls.		*/
/*									*/
/************************************************************************/

static void tedRulerChangeRect(	int *				pChanged,
				DocumentRectangle *		drRedraw,
				const DocumentRectangle *	drClip,
				int				d,
				const TedTopRuler *		ttr )
    {
    const RulerData *		rd= &(ttr->ttrRulerData);
    int				sizeAcross= rd->rdSizeAcross;
    int				whY= WHITE_TOP(sizeAcross);

    TabStop *			ts;
    ColumnSeparator *		cs;
    int				tab;

    int				changed= 0;

    DocumentRectangle		drButMore;
    DocumentRectangle		drButLess;
    DocumentRectangle		drScratch;

    *drRedraw= *drClip;

    tedHangingButtonRect( &drButMore, ttr->ttrFirstIndent+ d, sizeAcross );
    tedHangingButtonRect( &drButLess, ttr->ttrFirstIndent- d, sizeAcross );
    docUnionRectangle( &drButMore, &drButMore, &drButLess );
    if  ( docIntersectRectangle( &drScratch, drClip, &drButMore ) )
	{ docUnionRectangle( drRedraw, drRedraw, &drButMore ); changed= 1; }

    tedStandingButtonRect( &drButMore, ttr->ttrLeftIndent+ d, sizeAcross );
    tedStandingButtonRect( &drButLess, ttr->ttrLeftIndent- d, sizeAcross );
    docUnionRectangle( &drButMore, &drButMore, &drButLess );
    if  ( docIntersectRectangle( &drScratch, drClip, &drButMore ) )
	{ docUnionRectangle( drRedraw, drRedraw, &drButMore ); changed= 1; }

    tedStandingButtonRect( &drButMore, ttr->ttrRightIndent+ d, sizeAcross );
    tedStandingButtonRect( &drButLess, ttr->ttrRightIndent- d, sizeAcross );
    docUnionRectangle( &drButMore, &drButMore, &drButLess );
    if  ( docIntersectRectangle( &drScratch, drClip, &drButMore ) )
	{ docUnionRectangle( drRedraw, drRedraw, &drButMore ); changed= 1; }

    tedTabButtonRectangle( &drButMore, rd->rdVisibleC0+ d, whY, sizeAcross );
    tedTabButtonRectangle( &drButLess, rd->rdVisibleC0- d, whY, sizeAcross );
    docUnionRectangle( &drButMore, &drButMore, &drButLess );
    if  ( docIntersectRectangle( &drScratch, drClip, &drButMore ) )
	{ docUnionRectangle( drRedraw, drRedraw, &drButMore ); changed= 1; }

    cs= ttr->ttrColumns;
    for ( tab= 0; tab < ttr->ttrColumnCount; cs++, tab++ )
	{
	tedRulerColumnControlRect( &drButMore,
					cs->csX0+ d, cs->csX1+ d, sizeAcross );
	tedRulerColumnControlRect( &drButLess,
					cs->csX0- d, cs->csX1- d, sizeAcross );

	docUnionRectangle( &drButMore, &drButMore, &drButLess );
	if  ( docIntersectRectangle( &drScratch, drClip, &drButMore ) )
	    { docUnionRectangle( drRedraw, drRedraw, &drButMore ); changed= 1; }
	}

    ts= ttr->ttrTabStops;
    for ( tab= 0; tab < ttr->ttrTabCount; ts++, tab++ )
	{
	int	y= whY+ WHITE_HEIGHT(sizeAcross);

	tedTabSymbolRect( &drButMore, ts->tsPixels+ d, y, sizeAcross );
	tedTabSymbolRect( &drButLess, ts->tsPixels- d, y, sizeAcross );

	docUnionRectangle( &drButMore, &drButMore, &drButLess );
	if  ( docIntersectRectangle( &drScratch, drClip, &drButMore ) )
	    { docUnionRectangle( drRedraw, drRedraw, &drButMore ); changed= 1; }
	}

    *pChanged= changed;
    return;
    }

void tedScrollHorizontalRuler(	void *			voidttr,
				Widget			w,
				int			d	)
    {
    TedTopRuler *		ttr= (TedTopRuler *)voidttr;
    RulerData *			rd= &(ttr->ttrRulerData);

    Display *			display= XtDisplay( w );
    Window			win= XtWindow( w );
    GC				gc= rd->rdGc;

    int				changed;

    DocumentRectangle		drClip;
    DocumentRectangle		drRedraw;

    int				ox;

    appScrollHorizontalRuler( rd, &drClip, d, display, win, gc );

    ox= rd->rdVisibleC0- rd->rdMinUsed;

    changed= 0;
    tedRulerChangeRect( &changed, &drRedraw, &drClip, d, ttr );

    tedDrawHorizontalRuler( w, ttr, &drRedraw, ox, win, display );

    if  ( d > 0 )
	{ drClip.drX0= rd->rdVisibleC0- 1; drClip.drX1= rd->rdVisibleC0+ 1; }
    else{ drClip.drX0= rd->rdVisibleC1- 1; drClip.drX1= rd->rdVisibleC1+ 1; }

    changed= 0;
    tedRulerChangeRect( &changed, &drRedraw, &drClip, d, ttr );

    if  ( changed )
	{ tedDrawHorizontalRuler( w, ttr, &drRedraw, ox, win, display ); }
    }

void tedSetHorizontalRulerRange(	void *			voidttr,
					Widget			w,
					int			docVisX0,
					int			docVisX1,
					int			docBackX1 )
    {
    TedTopRuler *		ttr= (TedTopRuler *)voidttr;
    RulerData *			rd= &(ttr->ttrRulerData);

    Display *			display= XtDisplay( w );
    Window			win= XtWindow( w );

    rd->rdVisibleC0= docVisX0;
    rd->rdVisibleC1= docVisX1;
    rd->rdMaxUsed= docBackX1;

    XClearArea( display, win, 0, 0, 0, 0, True );

    return;
    }

/************************************************************************/
/*									*/
/*  Let the user drag one of the controls on the top ruler.		*/
/*									*/
/*  1)  Draw a 'hair' over the document that corresponds to the		*/
/*	position of the control.					*/
/*  2)  Until the mouse button is released, observe events.		*/
/*  3)  Merge as many mouse motion events as possible.			*/
/*  4)  If the event is a mouse motion or a button release, redraw the	*/
/*	old an new position of the control on the ruler. (Assume that	*/
/*	the buttons are wider than the tab marks.			*/
/*  5)  Let the document move the hair.					*/
/*  6)  When the mouse button is released, the loop ends.		*/
/*  7)  Let the toolkit dispatch all events that are not relevant for	*/
/*	this routine.							*/
/*									*/
/************************************************************************/

static void tedHorizontalRulerDragItem(		int *		pV0,
						int *		pV1,
						int *		pChanged,
						TedTopRuler *	ttr,
						int		x0,
						int		x1,
						int		ox,
						void *		voided,
						DragHair	dragHair,
						XEvent *	event,
						Widget		w )
    {
    RulerData *		rd= &(ttr->ttrRulerData);
    Display *		display= XtDisplay( w );
    Window		win= XtWindow( w );

    int			v0= *pV0;
    int			v1= *pV0;
    int			v= event->xbutton.x+ ox;

    DocumentRectangle	drClip;

    const int		asUpper= 0;

    /*  1  */
    (*dragHair)( voided, -1, v0, asUpper );
    if  ( pV1 )
	{
	v1= *pV1;
	(*dragHair)( voided, -1, v1, asUpper );
	}

    /*  2  */
    for (;;)
	{
	/*  3  */
	if  ( QLength( display ) > 0 )
	    {
	    while( QLength( display ) > 0 )
		{
		XNextEvent( display, event );

		if  ( event->type != MotionNotify )
		    { break;	}
		}
	    }
	else{ XNextEvent( display, event );	}

	if  ( event->type == MotionNotify )
	    { *pChanged= 1;	}

	/*  4  */
	if  ( event->type == ButtonRelease	||
	      event->type == MotionNotify	)
	    {
	    int		nv= event->xbutton.x+ ox;
	    int		nv0= v0+ nv- v;
	    int		nv1= v1+ nv- v;

	    if  ( nv0 < x0 )
		{ nv1 += x0- nv0; nv0 += x0- nv0; }
	    if  ( nv1 > x1 )
		{ nv0 -= nv1- x1; nv1 -= nv1- x1; }

	    if  ( nv0 == v0 && nv1 == v1 && event->type == MotionNotify )
		{ continue;	}

	    *pV0= nv0;
	    if  ( pV1 )
		{ *pV1= nv1;	}

	    if  ( v0 < nv0 )
		{
		drClip.drX0= v0-  BUTTON_HWIDTH(rd->rdSizeAcross)- BUTTON_MARG;
		drClip.drX1= nv1+ BUTTON_HWIDTH(rd->rdSizeAcross)+ BUTTON_MARG;
		drClip.drY0= 0;
		drClip.drY1= rd->rdSizeAcross;
		}
	    else{
		drClip.drX0= nv0- BUTTON_HWIDTH(rd->rdSizeAcross)- BUTTON_MARG;
		drClip.drX1= v1+  BUTTON_HWIDTH(rd->rdSizeAcross)+ BUTTON_MARG;
		drClip.drY0= 0;
		drClip.drY1= rd->rdSizeAcross;
		}

	    tedDrawHorizontalRuler( w, ttr, &drClip, ox, win, display );

	    /*  5  */
	    (*dragHair)( voided, v0, nv0, asUpper );
	    if  ( pV1 )
		{ (*dragHair)( voided, v1, nv1, asUpper ); }

	    v= nv; v0= nv0; v1= nv1;
	    }

	if  ( event->type == MotionNotify )
	    { continue;	}

	/*  6  */
	if  ( event->type == ButtonRelease )
	    {
	    (*dragHair)( voided, *pV0, -1, asUpper );
	    if  ( pV1 )
		{ (*dragHair)( voided, *pV1, -1, asUpper );	}
	    return;
	    }

	/*  7  */
	XtDispatchEvent( event );
	}
    }

/************************************************************************/
/*									*/
/*  The user clicked on the top ruler, make her happy.			*/
/*									*/
/*  1)  Only accept 'left mouse down' events.				*/
/*  2)  If the click is on one of the tab marks, apparently the user	*/
/*	wants to move the tab.						*/
/*  3)  If the tab was dragged from the ruler, assume that the user	*/
/*	wants to delete it.						*/
/*  4)  Move first indent?						*/
/*  5)  Move left indent?						*/
/*  6)  Move right indent?						*/
/*  7)  Not on one of the controls.. If the click is on the white band,	*/
/*	assume she wants to set a new tab.				*/
/*  8)  Is it on one of the table column separators? Allow the user to	*/
/*	drag it.							*/
/*  9)  Is it on the button with the kind of thabs that is to be added?	*/
/*	Rotate the type.						*/
/*									*/
/************************************************************************/

static int tedCompareTabStopsPixels(	const void *	voidts1,
					const void *	voidts2	)
    {
    const TabStop *	ts1= (const TabStop *)voidts1;
    const TabStop *	ts2= (const TabStop *)voidts2;

    return ts1->tsPixels- ts2->tsPixels;
    }

/*  2  */
static void tedHorizontalRulerDragTab(		int		item,
						int *		pChanged,
						TedTopRuler *	ttr,
						int		ox,
						void *		voided,
						DragHair	dragHair,
						XEvent *	event,
						Widget		w )
    {
    RulerData *		rd= &(ttr->ttrRulerData);
    XButtonEvent *	bevent= &(event->xbutton);
    int			sizeAcross= rd->rdSizeAcross;

    TabStop *		ts= ttr->ttrTabStops+ item;

    tedHorizontalRulerDragItem( &(ts->tsPixels), (int *)0, pChanged,
			ttr, ttr->ttrColX0, ttr->ttrColX1, ox,
			voided, dragHair, event, w );

    /*  3  */
    if  ( bevent->y <  WHITE_TOP(sizeAcross)				||
	  bevent->y >  WHITE_TOP(sizeAcross)+ WHITE_HEIGHT(sizeAcross)	)
	{
	int		i;

	Display *	display= XtDisplay( w );
	Window		win= XtWindow( w );

	ttr->ttrTabCount--;
	for ( i= item; i < ttr->ttrTabCount; i++ )
	    { ttr->ttrTabStops[i]= ttr->ttrTabStops[i+ 1]; }

	XClearArea( display, win, 0, 0, 0, 0, True );
	}

    qsort( ttr->ttrTabStops, ttr->ttrTabCount, sizeof(TabStop),
						tedCompareTabStopsPixels );

    return;
    }

void tedHorizontalRulerTrackMouse(	Widget		w,
					void *		voidttr,
					XEvent *	pEvent,
					unsigned int *	pChange,
					int *		pFirstIndent,
					int *		pLeftIndent,
					int *		pRightIndent,
					int *		pTabCount,
					TabStop **	pTabStops,
					int *		pCsCount,
					ColumnSeparator ** pCs,
					void *		voided,
					DragHair	dragHair )
    {
    Display *		display= XtDisplay( w );
    Window		win= XtWindow( w );

    TedTopRuler  *	ttr= (TedTopRuler *)voidttr;
    RulerData *		rd= &(ttr->ttrRulerData);
    int			sizeAcross= rd->rdSizeAcross;

    int			ox= rd->rdVisibleC0- rd->rdMinUsed;

    int			item;
    int			tabWidth= TAB_WIDTH(sizeAcross);

    int			whiteY= WHITE_TOP(sizeAcross);
    int			whiteH= WHITE_HEIGHT(sizeAcross);

    XEvent		event= *pEvent;
    XButtonEvent *	bevent= &(event.xbutton);

    int			changed= 0;

    DocumentRectangle	dr;

    /*  1  */
    if  ( event.type != ButtonPress	||
	  bevent->button != Button1	)
	{ *pChange= PPupdNONE; return; }

    /*  2  */
    if  ( bevent->y >= whiteY		&&
	  bevent->y <  whiteY+ whiteH	)
	{
	TabStop *		ts;

	ts= ttr->ttrTabStops;
	for ( item= 0; item < ttr->ttrTabCount; ts++, item++ )
	    {
	    if  ( bevent->x+ ox >= ts->tsPixels			&&
		  bevent->x+ ox <  ts->tsPixels+ tabWidth	)
		{
		tedHorizontalRulerDragTab( item, &changed, ttr, ox,
						voided, dragHair, &event, w );

		if  ( changed )
		    {
		    *pTabCount= ttr->ttrTabCount;
		    *pTabStops= ttr->ttrTabStops;
		    *pChange= PPupdTAB_STOPS;
		    }
		else{ *pChange= PPupdNONE; }

		return;
		}
	    }
	}

    /*  4  */
    if  ( ttr->ttrFirstIndent >= 0					&&
	  bevent->y < HBUTTON_APEX(sizeAcross)				&&
	  bevent->x+ ox >= ttr->ttrFirstIndent- BUTTON_HWIDTH(sizeAcross)	&&
	  bevent->x+ ox <= ttr->ttrFirstIndent+ BUTTON_HWIDTH(sizeAcross)	)
	{
	tedHorizontalRulerDragItem( &ttr->ttrFirstIndent, (int *)0, &changed,
				ttr, ttr->ttrColX0, ttr->ttrColX1, ox,
				voided, dragHair, &event, w );

	if  ( changed )
	    {
	    *pFirstIndent= ttr->ttrFirstIndent;
	    *pTabCount= ttr->ttrTabCount;
	    *pTabStops= ttr->ttrTabStops;
	    *pChange= PPupdFIRST_INDENT;
	    }
	else{ *pChange= PPupdNONE; }
	return;
	}

    /*  5  */
    if  ( ttr->ttrLeftIndent >= 0						&&
	  bevent->y >= SBUTTON_APEX(sizeAcross)				&&
	  bevent->y <= SBUTTON_BASE(sizeAcross)				&&
	  bevent->x+ ox >= ttr->ttrLeftIndent- BUTTON_HWIDTH(sizeAcross)	&&
	  bevent->x+ ox <= ttr->ttrLeftIndent+ BUTTON_HWIDTH(sizeAcross)	)
	{
	tedHorizontalRulerDragItem( &ttr->ttrLeftIndent, (int *)0, &changed,
				ttr, ttr->ttrColX0, ttr->ttrColX1, ox,
				voided, dragHair, &event, w );

	if  ( changed )
	    {
	    *pLeftIndent= ttr->ttrLeftIndent;
	    *pTabCount= ttr->ttrTabCount;
	    *pTabStops= ttr->ttrTabStops;
	    *pChange= PPupdLEFT_INDENT;
	    }
	else{ *pChange= PPupdNONE; }
	return;
	}

    /*  6  */
    if  ( ttr->ttrRightIndent >= 0					&&
	  bevent->y >= SBUTTON_APEX(sizeAcross)				&&
	  bevent->y <= SBUTTON_BASE(sizeAcross)				&&
	  bevent->x+ ox >= ttr->ttrRightIndent- BUTTON_HWIDTH(sizeAcross)	&&
	  bevent->x+ ox <= ttr->ttrRightIndent+ BUTTON_HWIDTH(sizeAcross)	)
	{
	tedHorizontalRulerDragItem( &ttr->ttrRightIndent, (int *)0, &changed,
				ttr, ttr->ttrColX0, ttr->ttrColX1, ox,
				voided, dragHair, &event, w );

	if  ( changed )
	    {
	    *pRightIndent= ttr->ttrRightIndent;
	    *pTabCount= ttr->ttrTabCount;
	    *pTabStops= ttr->ttrTabStops;
	    *pChange= PPupdRIGHT_INDENT;
	    }
	else{ *pChange= PPupdNONE; }
	return;
	}

    /*  7  */
    if  ( bevent->y >= whiteY			&&
	  bevent->y <  whiteY+ whiteH		&&
	  bevent->x+ ox >= ttr->ttrColX0		&&
	  bevent->x+ ox <= ttr->ttrColX1		)
	{
	int		i;
	TabStop *	fresh;

	fresh= (TabStop *)realloc( ttr->ttrTabStops,
				    ( ttr->ttrTabCount+ 1)* sizeof(TabStop) );
	if  ( ! fresh )
	    { LXDEB(ttr->ttrTabCount,fresh); *pChange= PPupdNONE; return; }

	ttr->ttrTabStops= fresh;

	for ( item= 0; item < ttr->ttrTabCount; item++ )
	    {
	    if  ( ttr->ttrTabStops[item].tsPixels >= bevent->x+ ox )
		{ break;	}
	    }

	for ( i= ttr->ttrTabCount; i > item; i-- )
	    { ttr->ttrTabStops[i]= ttr->ttrTabStops[i-1];	}

	docInitTabStop( ttr->ttrTabStops+ item );

	ttr->ttrTabStops[item].tsPixels= bevent->x+ ox;
	ttr->ttrTabStops[item].tsKind= ttr->ttrTabKind;
	ttr->ttrTabCount++;

	XClearArea( display, win, 0, 0, 0, 0, True );

	tedHorizontalRulerDragTab( item, &changed, ttr, ox,
				voided, dragHair, &event, w );

	*pTabCount= ttr->ttrTabCount;
	*pTabStops= ttr->ttrTabStops;
	*pChange= PPupdTAB_STOPS; return;
	}

    /*  8  */
    if  ( bevent->y >= whiteY			&&
	  bevent->y <  whiteY+ whiteH		)
	{
	for ( item= 0; item < ttr->ttrColumnCount; item++ )
	    {
	    if  ( bevent->x+ ox >= ttr->ttrColumns[item].csX0		&&
		  bevent->x+ ox <  ttr->ttrColumns[item].csX1		)
		{
		int		wide;
		int		x0;
		int		x1;

		wide= ttr->ttrColumns[item].csX1- ttr->ttrColumns[item].csX0;
		x0= rd->rdDocumentC0- wide;
		x1= rd->rdDocumentC1+ wide;

		if  ( item > 0 )
		    { x0= ttr->ttrColumns[item- 1].csX1; }
		if  ( item < ttr->ttrColumnCount- 1 )
		    { x1= ttr->ttrColumns[item+ 1].csX0; }

		tedHorizontalRulerDragItem( &ttr->ttrColumns[item].csX0,
				    &ttr->ttrColumns[item].csX1,
				    &changed, ttr, x0, x1, ox,
				    voided, dragHair, &event, w );

		if  ( changed )
		    {
		    *pCsCount= ttr->ttrColumnCount; *pCs= ttr->ttrColumns;
		    *pChange= PPupdCOLUMNS;
		    }
		else{ *pChange= PPupdNONE;	}
		return;
		}
	    }
	}

    /*  9  */
    tedTabButtonRectangle( &dr, rd->rdMinUsed, whiteY, sizeAcross );

    if  ( bevent->x >= dr.drX0 && bevent->x <= dr.drX1	&&
	  bevent->y >= dr.drY0 && bevent->y <= dr.drY1	)
	{
	switch( ttr->ttrTabKind )
	    {
	    case DOCtkLEFT:	ttr->ttrTabKind= DOCtkRIGHT;	break;
	    case DOCtkRIGHT:	ttr->ttrTabKind= DOCtkCENTRE;	break;
	    case DOCtkCENTRE:	ttr->ttrTabKind= DOCtkLEFT;	break;
	    default:
		LDEB(ttr->ttrTabKind);
	    }

	XClearArea( display, win,
			dr.drX0, dr.drY0,
			dr.drX1- dr.drX0+ 1, dr.drY1- dr.drY0+ 1, True );
	}

    *pChange= PPupdNONE; return;
    }

/************************************************************************/
/*									*/
/*  Adapt an horizontal ruler to the paragraph.				*/
/*									*/
/************************************************************************/

void tedAdaptHorizontalRuler(	void *		voidttr,
				Widget		w,
				int		documentC0,
				int		documentC1,
				int		firstIndent,
				int		leftIndent,
				int		rightIndent,
				int		rulerC1,
				int		tabCount,
				TabStop *	tabStops,
				int		andExpose )
    {
    TedTopRuler *	ttr= (TedTopRuler *)voidttr;
    RulerData *		rd= &(ttr->ttrRulerData);
    TabStop *		ts;

    int			changed= 0;

    int			i;

    rd->rdDocumentC0= documentC0;
    rd->rdDocumentC1= documentC1;

    if  ( ttr->ttrFirstIndent != firstIndent )
	{ ttr->ttrFirstIndent=   firstIndent; changed= 1;	}

    if  ( ttr->ttrLeftIndent != leftIndent )
	{ ttr->ttrLeftIndent=   leftIndent; changed= 1;	}

    if  ( ttr->ttrRightIndent != rightIndent )
	{ ttr->ttrRightIndent=   rightIndent; changed= 1;	}

    if  ( rd->rdRulerC1 != rulerC1 )
	{ rd->rdRulerC1=   rulerC1; changed= 1;	}

    if  ( tabCount != ttr->ttrTabCount )
	{
	if  ( tabCount > ttr->ttrTabCount )
	    {
	    TabStop *	fresh;

	    fresh= (TabStop *)realloc( ttr->ttrTabStops,
						tabCount* sizeof(TabStop) );
	    if  ( ! fresh )
		{ LXDEB(tabCount,fresh); return;	}

	    ttr->ttrTabStops= fresh;
	    }

	ttr->ttrTabCount= tabCount;
	changed= 1;
	}

    ts= ttr->ttrTabStops;
    for ( i= 0; i < tabCount; ts++, tabStops++, i++ )
	{
	int		different= 0;

	if  ( ts->tsPixels != tabStops->tsPixels )
	    { different= 1; changed= 1;	}
	if  ( ts->tsKind != tabStops->tsKind )
	    { different= 1; changed= 1;	}
	if  ( ts->tsLeader != tabStops->tsLeader )
	    { different= 1;	}

	if  ( different )
	    { *ts= *tabStops;	}
	}

    if  ( changed && andExpose )
	{
	Display *	display= XtDisplay( w );
	Window		win= XtWindow( w );

	XClearArea( display, win, 0, 0, 0, 0, True );
	}

    return;
    }

int tedSetRulerColumns(		Widget			w,
				void *			voidttr,
				int			colX0,
				int			colX1,
				ColumnSeparator *	cs,
				int			csCount )
    {
    TedTopRuler *	ttr= (TedTopRuler *)voidttr;
    int			changed= 0;
    int			i;

    ttr->ttrColX0= colX0;
    ttr->ttrColX1= colX1;

    if  ( csCount != ttr->ttrColumnCount )
	{
	if  ( csCount > ttr->ttrColumnCount )
	    {
	    ColumnSeparator *	fresh;

	    fresh= (ColumnSeparator *)
		realloc( ttr->ttrColumns, csCount* sizeof(ColumnSeparator) );
	    if  ( ! fresh )
		{ LXDEB(csCount,fresh); return -1;	}

	    ttr->ttrColumns= fresh;
	    }

	ttr->ttrColumnCount= csCount;
	changed= 1;
	}

    for ( i= 0; i < csCount; i++ )
	{
	if  ( ttr->ttrColumns[i].csX0 != cs[i].csX0 )
	    { ttr->ttrColumns[i].csX0=   cs[i].csX0; changed= 1; }
	if  ( ttr->ttrColumns[i].csX1 != cs[i].csX1 )
	    { ttr->ttrColumns[i].csX1=   cs[i].csX1; changed= 1; }
	}

    if  ( changed && XtIsRealized( w ) )
	{
	Display *	display= XtDisplay( w );
	Window		win= XtWindow( w );

	XClearArea( display, win, 0, 0, 0, 0, True );
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Left Ruler..							*/
/*									*/
/************************************************************************/

typedef struct TedLeftRuler
    {
    RulerData		tlrRulerData;

    int			tlrPageStep;
    int			tlrTopPage;
    int			tlrBottomPage;

    const char *	tlrPageFormat;
    char *		tlrPageScratch;
    } TedLeftRuler;

void * tedMakeLeftRuler(	int			width,
				double			pixPerMM,
				double			magnification,

				Pixel			forePixel,
				Pixel			backPixel,
				Pixel			topPixel,
				Pixel			bottomPixel,

				int			minUnused,
				int			maxUnused,

				int			documentC0,
				int			documentC1,
				int			visibleC0,
				int			visibleC1,
				int			rulerC1,

				int			whatUnit,
				int			pageStep,
				const char *		pageFormat )
    {
    TedLeftRuler *		tlr;
    RulerData *			rd;

    int				l;
    const char *		s;

    tlr= (TedLeftRuler *)malloc( sizeof( TedLeftRuler ) );
    if  ( ! tlr )
	{ XDEB(tlr); return (void *)0;	}
    rd= &(tlr->tlrRulerData);

    tlr->tlrPageFormat= pageFormat;
    tlr->tlrPageStep= pageStep;
    tlr->tlrTopPage= -1;
    tlr->tlrBottomPage= -1;

    l= 0; s= tlr->tlrPageFormat;
    while( *s )
	{
	if  ( s[0] == '%' && s[1] == 'd' )
	    { l += 12;	}
	else{ l++;	}

	s++;
	}

    tlr->tlrPageScratch= malloc( l+ 1 );
    if  ( ! tlr->tlrPageScratch )
	{ LXDEB(l,tlr->tlrPageScratch);	}
    tlr->tlrPageScratch[0]= '\0';

    appInitRulerData( &(tlr->tlrRulerData) );

    rd->rdUnit= whatUnit;
    appRulerCalculateIncrements( rd, pixPerMM, magnification );

    rd->rdMinUsed= minUnused;
    rd->rdMaxUsed= 0;
    rd->rdExtraAfterMaxUsed= maxUnused;

    rd->rdDocumentC0= documentC0;
    rd->rdDocumentC1= documentC1;
    rd->rdVisibleC0= visibleC0;
    rd->rdVisibleC1= visibleC1;
    rd->rdRulerC1= rulerC1;

    rd->rdSizeAcross= width;
    rd->rdSizeAlong= 0;

    rd->rdForePixel= forePixel;
    rd->rdBackPixel= backPixel;
    rd->rdTopPixel= topPixel;
    rd->rdBottomPixel= bottomPixel;

    return (void *)tlr;
    }

void tedVerticalRulerConfigure(		Widget			w,
					void *			voidtlr,
					XEvent *		event,
					Boolean *		pRefused )
    {
    XConfigureEvent *	cevent= &(event->xconfigure);
    TedLeftRuler *	tlr= (TedLeftRuler *)voidtlr;
    RulerData *		rd= &(tlr->tlrRulerData);

    if  ( cevent->type != ConfigureNotify )
	{ return;	}

    rd->rdSizeAlong= cevent->height;
    rd->rdMaxUsed= cevent->height- rd->rdExtraAfterMaxUsed;

    return;
    }

void tedSetLeftRulerRange(	void *			voidtlr,
				Widget			w,
				int			docVisV0,
				int			docVisV1,
				int			docBackV1 )
    {
    TedLeftRuler *		tlr= (TedLeftRuler *)voidtlr;
    RulerData *			rd= &(tlr->tlrRulerData);

    Display *			display= XtDisplay( w );
    Window			win= XtWindow( w );

    rd->rdVisibleC0= docVisV0;
    rd->rdVisibleC1= docVisV1;
    rd->rdMaxUsed= docBackV1;

    tlr->tlrTopPage=	  rd->rdVisibleC0    / tlr->tlrPageStep;
    tlr->tlrBottomPage=	( rd->rdVisibleC1- 1)/ tlr->tlrPageStep;

    XClearArea( display, win, 0, 0, 0, 0, True );

    return;
    }

void tedScrollLeftRuler(	void *			voidtlr,
				Widget			w,
				int			d	)
    {
    TedLeftRuler *		tlr= (TedLeftRuler *)voidtlr;
    RulerData *			rd= &(tlr->tlrRulerData);
    Display *			display= XtDisplay( w );
    Window			win= XtWindow( w );

    int				oldTopPage= tlr->tlrTopPage;
    int				oldBottomPage= tlr->tlrBottomPage;

    rd->rdVisibleC0 += d;
    rd->rdVisibleC1 += d;

    tlr->tlrTopPage=	  rd->rdVisibleC0    / tlr->tlrPageStep;
    tlr->tlrBottomPage=	( rd->rdVisibleC1- 1)/ tlr->tlrPageStep;

    if  ( tlr->tlrTopPage != oldTopPage		||
	  tlr->tlrBottomPage != oldBottomPage	)
	{ XClearArea( display, win, 0, 0, 0, 0, True );	}

    return;
    }

void tedAdaptLeftRuler(		void *		voidtlr,
				Widget		w,
				int		visisbleC0,
				int		visisbleC1,
				int		pageStep )
    {
    TedLeftRuler *		tlr= (TedLeftRuler *)voidtlr;
    RulerData *			rd= &(tlr->tlrRulerData);
    Display *			display= XtDisplay( w );
    Window			win= XtWindow( w );

    rd->rdVisibleC0= visisbleC0;
    rd->rdVisibleC1= visisbleC1;

    tlr->tlrPageStep= pageStep;

    tlr->tlrTopPage=	  rd->rdVisibleC0    / tlr->tlrPageStep;
    tlr->tlrBottomPage=	( rd->rdVisibleC1- 1)/ tlr->tlrPageStep;

    XClearArea( display, win, 0, 0, 0, 0, True );
    }

void tedFreeVerticalRuler(	Display *	display,
				void *		voidtlr )
    {
    TedLeftRuler *		tlr= (TedLeftRuler *)voidtlr;

    if  ( tlr->tlrPageScratch )
	{ free( tlr->tlrPageScratch );	}

    appCleanRulerData( display, &(tlr->tlrRulerData) );

    free( tlr );

    return;
    }

/************************************************************************/
/*									*/
/*  Draw a vertical ruler.						*/
/*									*/
/************************************************************************/

static void tedFormatPageNumber(	TedLeftRuler *		tlr,
					int *			pLen,
					int *			pWide,
					int			page )
    {
    RulerData *		rd= &(tlr->tlrRulerData);

    char		scratch[11];
    const char *	from;
    char *		to;

    int			len;
    int			wide;

    sprintf( scratch, "%d", page+ 1 );

    from= tlr->tlrPageFormat;
    to= tlr->tlrPageScratch;
    while( *from )
	{
	if  ( from[0] == '%' && from[1] == 'd' )
	    { strcpy( to, scratch ); to += strlen( to ); from += 2;	}
	else{ *(to++)= *(from++);					}
	}

    *to= '\0';

    len= strlen( tlr->tlrPageScratch );
    wide= appRulerTextWidth( rd, tlr->tlrPageScratch, len );

    *pLen= len; *pWide= wide; return;
    }

static void tedDrawVerticalRuler(	Widget			w,
					TedLeftRuler *		tlr,
					DocumentRectangle *	drClip,
					int			oy,
					Window			win,
					Display *		display )
    {
    RulerData *		rd= &(tlr->tlrRulerData);

    int			len;
    int			wide;

    int			x;

    if  ( rd->rdSizeAlong == 0 )
	{
	XtVaGetValues( w, XtNheight, &rd->rdSizeAlong, NULL );
	rd->rdMaxUsed= rd->rdSizeAlong- rd->rdExtraAfterMaxUsed;
	}

    x= rd->rdSizeAcross/8+ rd->rdVerticalFont.vxfFontStruct->ascent;

    tedFormatPageNumber( tlr, &len, &wide, tlr->tlrTopPage );

    appRulerDrawVerticalString( display, win, rd,
				    x, 1+ wide, tlr->tlrPageScratch, len );

    if  ( tlr->tlrBottomPage != tlr->tlrTopPage			&&
	  rd->rdSizeAlong > 4* wide				)
	{
	tedFormatPageNumber( tlr, &len, &wide, tlr->tlrBottomPage );

	appRulerDrawVerticalString( display, win, rd,
			    x, rd->rdSizeAlong- 1, tlr->tlrPageScratch, len );
	}
    }

void tedExposeVerticalRuler(	Widget			w,
				void *			voidtlr,
				XEvent *		event )
    {
    TedLeftRuler *	tlr= (TedLeftRuler *)voidtlr;
    RulerData *		rd= &(tlr->tlrRulerData);

    Display *		display= XtDisplay( w );
    Window		win= XtWindow( w );

    int			ox= 0;
    int			oy= rd->rdVisibleC0- rd->rdMinUsed;

    DocumentRectangle	drClip;

    if  ( ! rd->rdGc )
	{
	const int	vertical= 1;

	appOpenRulerGc( &(tlr->tlrRulerData),
				vertical, 2* rd->rdSizeAcross/ 3- 2, w );
	}

    appCollectExposures( &drClip, ox, oy, display, win,
					    rd->rdGc, event );

    tedDrawVerticalRuler( w, tlr, &drClip, oy, win, display );
    }
