/************************************************************************/
/*									*/
/*  The Row page of the Format Tool.					*/
/*									*/
/************************************************************************/

#   include	"config.h"

#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<debugon.h>

#   include	<Xm/Text.h>
#   include	<Xm/ToggleB.h>

#   include	<appUtil.h>
#   include	<appGeoString.h>
#   include	<appUnit.h>

#   include	"tedApp.h"
#   include	"tedFormatTool.h"

/************************************************************************/
/*									*/
/*  Refresh the row page.						*/
/*									*/
/************************************************************************/

static void tedFormatToolRefreshSectionPage(	SectionTool *	st )
    {
    EditApplication *		ea= st->stApplication;
    SectionProperties *		sp= &(st->stPropertiesChosen);

    char		scratch[50];

    appIntegerToTextWidget( st->stNumberText, st->stSectionNumber+ 1 );

    appGeoLengthToString( scratch, sp->spPageWideTwips, ea->eaUnitInt );
    XmTextSetString( st->stPageWideText, scratch );

    appGeoLengthToString( scratch, sp->spPageHighTwips, ea->eaUnitInt );
    XmTextSetString( st->stPageHighText, scratch );

    appGeoLengthToString( scratch, sp->spLeftMarginTwips, ea->eaUnitInt );
    XmTextSetString( st->stLeftMarginText, scratch );

    appGeoLengthToString( scratch, sp->spRightMarginTwips, ea->eaUnitInt );
    XmTextSetString( st->stRightMarginText, scratch );

    appGeoLengthToString( scratch, sp->spTopMarginTwips, ea->eaUnitInt );
    XmTextSetString( st->stTopMarginText, scratch );

    appGeoLengthToString( scratch, sp->spBottomMarginTwips, ea->eaUnitInt );
    XmTextSetString( st->stBottomMarginText, scratch );

    appGeoLengthToString( scratch, sp->spHeaderYTwips, ea->eaUnitInt );
    XmTextSetString( st->stHeaderYText, scratch );

    appGeoLengthToString( scratch, sp->spFooterYTwips, ea->eaUnitInt );
    XmTextSetString( st->stFooterYText, scratch );

    XtVaSetValues( st->stNumberStyleMenu,
		XmNmenuHistory,	st->stNumberStyleItems[sp->spPageNumberStyle],
		NULL );

    XtVaSetValues( st->stBreakKindMenu,
		XmNmenuHistory,	st->stBreakKindItems[sp->spBreakKind],
		NULL );

    return;
    }

void tedFormatToolRefreshSectionTool(	SectionTool *		st,
					InspectorSubject *	is,
					const BufferSelection *	bs )
    {
    BufferItem *			sectBi;

    sectBi= bs->bsBegin.bpBi;
    while( sectBi && sectBi->biLevel > DOClevSECT )
	{ sectBi= sectBi->biParent;	}

    if  ( ! sectBi )
	{ XDEB(sectBi); return;	}

    st->stSectionNumber= sectBi->biNumberInParent;

    if  ( docCopySectionProperties( &(st->stPropertiesChosen),
					    &(sectBi->biSectProperties) ))
	{ LDEB(1); return ;	}
    if  ( docCopySectionProperties( &(st->stPropertiesSet),
					    &(sectBi->biSectProperties) ) )
	{ LDEB(1); return ;	}

    XtSetSensitive( is->isPrevButton,
			docPrevSection( sectBi ) != (BufferItem *)0 );
    XtSetSensitive( is->isNextButton,
			docNextSection( sectBi ) != (BufferItem *)0 );

    tedFormatToolRefreshSectionPage( st );

    return;
    }

/************************************************************************/
/*									*/
/*  The 'Change' button of the row tool has been pushed.		*/
/*									*/
/************************************************************************/

static void tedFormatChangeSectPushed(	Widget		w,
					XtPointer	voidst,
					XtPointer	voidpbcs )
    {
    SectionTool *		st= (SectionTool *)voidst;
    EditApplication *		ea= st->stApplication;
    EditDocument *		ed= ea->eaCurrentDocument;
    SectionProperties *		spNew= &(st->stPropertiesChosen);

    int				changed;

    unsigned int		taUpdMask= 0;
    unsigned int		ppUpdMask= 0;
    unsigned int		spUpdMask= SPupdALL;

    TextAttribute		taNew;

    docInitTextAttribute ( &taNew );

    if  ( ! ed )
	{ XDEB(ed); return;	}

    if  ( appGetLengthFromTextWidget( st->stPageWideText,
		    &spNew->spPageWideTwips, &changed, ea->eaUnitInt, 1 ) )
	{ return;	}

    if  ( appGetLengthFromTextWidget( st->stPageHighText,
		    &spNew->spPageHighTwips, &changed, ea->eaUnitInt, 1 ) )
	{ return;	}

    if  ( appGetLengthFromTextWidget( st->stLeftMarginText,
		    &spNew->spLeftMarginTwips, &changed, ea->eaUnitInt, 1 ) )
	{ return;	}

    if  ( appGetLengthFromTextWidget( st->stRightMarginText,
		    &spNew->spRightMarginTwips, &changed, ea->eaUnitInt, 1 ) )
	{ return;	}

    if  ( appGetLengthFromTextWidget( st->stTopMarginText,
		    &spNew->spTopMarginTwips, &changed, ea->eaUnitInt, 1 ) )
	{ return;	}

    if  ( appGetLengthFromTextWidget( st->stBottomMarginText,
		    &spNew->spBottomMarginTwips, &changed, ea->eaUnitInt, 1 ) )
	{ return;	}

    if  ( appGetLengthFromTextWidget( st->stHeaderYText,
		    &spNew->spHeaderYTwips, &changed, ea->eaUnitInt, 1 ) )
	{ return;	}

    if  ( appGetLengthFromTextWidget( st->stFooterYText,
		    &spNew->spFooterYTwips, &changed, ea->eaUnitInt, 1 ) )
	{ return;	}

    if  ( tedChangeSelectionProperties( ed,
				    taUpdMask, taNew,
				    ppUpdMask, (ParagraphProperties *)0,
				    spUpdMask, spNew ) )
	{ LDEB(1);	}

    return;
    }

static void tedSectPageLayoutChanged(	Widget		w,
					XtPointer	voidst,
					XtPointer	voidpbcs )
    {
    SectionTool *		st= (SectionTool *)voidst;
    EditApplication *		ea= st->stApplication;
    SectionProperties *		sp= &(st->stPropertiesChosen);

    int				width;
    int				changed;

    if  ( w == st->stPageWideText )
	{ width= sp->spPageWideTwips;	}
    if  ( w == st->stPageHighText )
	{ width= sp->spPageHighTwips;	}

    if  ( w == st->stLeftMarginText )
	{ width= sp->spLeftMarginTwips;	}
    if  ( w == st->stRightMarginText )
	{ width= sp->spRightMarginTwips;	}
    if  ( w == st->stTopMarginText )
	{ width= sp->spTopMarginTwips;	}
    if  ( w == st->stBottomMarginText )
	{ width= sp->spBottomMarginTwips;	}
    if  ( w == st->stHeaderYText )
	{ width= sp->spHeaderYTwips;	}
    if  ( w == st->stFooterYText )
	{ width= sp->spFooterYTwips;	}

    if  ( ! appGetLengthFromTextWidget( w, &width, &changed,
							ea->eaUnitInt, 1 ) )
	{
	char	scratch[50];

	appGeoLengthToString( scratch, width, ea->eaUnitInt );
	XmTextSetString( w, scratch );
	}
    }

static void tedSectNumberStyleChosen(	Widget		w,
					XtPointer	voidst,
					XtPointer	voidtbcs )
    {
    SectionTool *			st= (SectionTool *)voidst;
    SectionProperties *			sp= &(st->stPropertiesChosen);

    short				alignment= -1;

    XtVaGetValues( w,	XmNpositionIndex,	&alignment,
			NULL );

    if  ( alignment >= 0 && alignment < DOCpgn_COUNT )
	{ sp->spPageNumberStyle= alignment;	}
    else{ LLDEB(alignment,DOCpgn_COUNT);	}

    return;
    }

static void tedSectBreakKindChosen(	Widget		w,
					XtPointer	voidst,
					XtPointer	voidtbcs )
    {
    SectionTool *			st= (SectionTool *)voidst;
    SectionProperties *			sp= &(st->stPropertiesChosen);

    short				breakKind= -1;

    XtVaGetValues( w,	XmNpositionIndex,	&breakKind,
			NULL );

    if  ( breakKind >= 0 && breakKind < DOCsbk_COUNT )
	{ sp->spBreakKind= breakKind;		}
    else{ LLDEB(breakKind,DOCsbk_COUNT);	}

    return;
    }

/************************************************************************/
/*									*/
/*  The 'Revert' button of the row tool has been pushed.		*/
/*									*/
/************************************************************************/

static void tedFormatRevertSectPushed(	Widget		w,
					XtPointer	voidst,
					XtPointer	voidpbcs	 )
    {
    SectionTool *	st= (SectionTool *)voidst;

    docCopySectionProperties( &(st->stPropertiesChosen),
						    &(st->stPropertiesSet) );

    tedFormatToolRefreshSectionPage( st );

    return;
    }

/************************************************************************/
/*									*/
/*  Previous/Next Section buttons have been pushed.			*/
/*									*/
/************************************************************************/

static void tedFormatPrevSect(		Widget		w,
					XtPointer	voidst,
					XtPointer	voidpbcs )
    {
    SectionTool *		st= (SectionTool *)voidst;
    EditApplication *		ea= st->stApplication;

    tedSelectWholeSection( ea, -1 );

    return;
    }

static void tedFormatNextSect(		Widget		w,
					XtPointer	voidst,
					XtPointer	voidpbcs )
    {
    SectionTool *		st= (SectionTool *)voidst;
    EditApplication *		ea= st->stApplication;

    tedSelectWholeSection( ea, 1 );

    return;
    }

static void tedFormatSelectSect(	Widget		w,
					XtPointer	voidst,
					XtPointer	voidpbcs )
    {
    SectionTool *		st= (SectionTool *)voidst;
    EditApplication *		ea= st->stApplication;

    tedSelectWholeSection( ea, 0 );

    return;
    }

static void tedFormatDeleteSect(	Widget		w,
					XtPointer	voidst,
					XtPointer	voidpbcs )
    {
    SectionTool *		st= (SectionTool *)voidst;
    EditApplication *		ea= st->stApplication;

    tedDeleteCurrentSection( ea );
    }

static void tedFormatInsertSect(	Widget		w,
					XtPointer	voidst,
					XtPointer	voidpbcs )
    {
    SectionTool *		st= (SectionTool *)voidst;
    EditApplication *		ea= st->stApplication;

    tedInsertSection( ea, 0 );

    return;
    }

static void tedFormatAppendSect(	Widget		w,
					XtPointer	voidst,
					XtPointer	voidpbcs )
    {
    SectionTool *		st= (SectionTool *)voidst;
    EditApplication *		ea= st->stApplication;

    tedInsertSection( ea, 1 );

    return;
    }

/************************************************************************/
/*									*/
/*  Make a section tool. I.E. the 'Section' page of the format tool.	*/
/*									*/
/************************************************************************/

void tedFormatFillSectionPage(	SectionTool *			st,
				const SectionPageResources *	spr,
				InspectorSubject *		is,
				Widget				pageWidget,
				const InspectorSubjectResources * isr )
    {
    Widget	rowLabel;

    Widget	row;

    const int	textColumns= 10;

    /**************/
    st->stSectionNumber= -1;

    docInitSectionProperties( &(st->stPropertiesSet) );
    docInitSectionProperties( &(st->stPropertiesChosen) );

    /**************/
    appMakeLabelAndTextRow( &row, &rowLabel, &(st->stNumberText),
			    pageWidget, isr->isrSubjectName, textColumns, 0 );
    /**************/
    appMakeLabelAndTextRow( &row, &rowLabel, &(st->stPageWideText),
			    pageWidget, spr->sprPageWide, textColumns, 1 );
    XtAddCallback( st->stPageWideText, XmNactivateCallback,
				tedSectPageLayoutChanged, (void *)st );
    /**************/
    appMakeLabelAndTextRow( &row, &rowLabel, &(st->stPageHighText),
			    pageWidget, spr->sprPageHigh, textColumns, 1 );
    XtAddCallback( st->stPageHighText, XmNactivateCallback,
				tedSectPageLayoutChanged, (void *)st );
    /**************/
    appMakeLabelAndTextRow( &row, &rowLabel, &(st->stLeftMarginText),
			    pageWidget, spr->sprLeftMargin, textColumns, 1 );
    XtAddCallback( st->stLeftMarginText, XmNactivateCallback,
				tedSectPageLayoutChanged, (void *)st );
    /**************/
    appMakeLabelAndTextRow( &row, &rowLabel, &(st->stRightMarginText),
			    pageWidget, spr->sprRightMargin, textColumns, 1 );
    XtAddCallback( st->stRightMarginText, XmNactivateCallback,
				tedSectPageLayoutChanged, (void *)st );
    /**************/
    appMakeLabelAndTextRow( &row, &rowLabel, &(st->stTopMarginText),
			    pageWidget, spr->sprTopMargin, textColumns, 1 );
    XtAddCallback( st->stTopMarginText, XmNactivateCallback,
				tedSectPageLayoutChanged, (void *)st );
    /**************/
    appMakeLabelAndTextRow( &row, &rowLabel, &(st->stBottomMarginText),
			    pageWidget, spr->sprBottomMargin, textColumns, 1 );

    XtAddCallback( st->stBottomMarginText, XmNactivateCallback,
				tedSectPageLayoutChanged, (void *)st );
    /**************/
    appMakeLabelAndTextRow( &row, &rowLabel, &(st->stHeaderYText),
			    pageWidget, spr->sprHeaderY, textColumns, 1 );
    XtAddCallback( st->stHeaderYText, XmNactivateCallback,
				tedSectPageLayoutChanged, (void *)st );
    /**************/
    appMakeLabelAndTextRow( &row, &rowLabel, &(st->stFooterYText),
			    pageWidget, spr->sprFooterY, textColumns, 1 );

    XtAddCallback( st->stFooterYText, XmNactivateCallback,
				tedSectPageLayoutChanged, (void *)st );
    /**************/
    appInspectorMakeMenuRow( &row,
			    &(st->stNumberStyleMenu),
			    &(st->stNumberStylePulldown),
			    &rowLabel, pageWidget, spr->sprPageNumberStyle );
    /**************/
    appInspectorMakeMenuRow( &row,
			    &(st->stBreakKindMenu),
			    &(st->stBreakKindPulldown),
			    &rowLabel, pageWidget, spr->sprBreakKind );
    /**************/
    appInspectorMakeButtonRow( &row, pageWidget,
	    &(is->isPrevButton), &(is->isNextButton),
	    isr->isrPrevButtonText, isr->isrNextButtonText,
	    tedFormatPrevSect, tedFormatNextSect, (void *)st );

    appInspectorMakeButtonRow( &row, pageWidget,
	    &(is->isSelectButton), &(is->isDeleteButton),
	    isr->isrSelectButtonText, isr->isrDeleteButtonText,
	    tedFormatSelectSect, tedFormatDeleteSect, (void *)st );

    appInspectorMakeButtonRow( &row, pageWidget,
	    &(is->isInsertButton), &(is->isAppendButton),
	    isr->isrInsertButtonText, isr->isrAppendButtonText,
	    tedFormatInsertSect, tedFormatAppendSect, (void *)st );

    appInspectorMakeButtonRow( &row, pageWidget,
	    &(is->isRevertButton), &(is->isApplyButton),
	    isr->isrRevert, isr->isrApplyToSubject,
	    tedFormatRevertSectPushed, tedFormatChangeSectPushed, (void *)st );

    return;
    }

/************************************************************************/
/*									*/
/*  Finish the section page.						*/
/*									*/
/************************************************************************/

void tedFormatFillSectionChoosers(	SectionTool *			st,
					const SectionPageResources *	spr )
    {
    appFillInspectorMenu( DOCpgn_COUNT, DOCpgnDEC,
			st->stNumberStyleItems, spr->sprNumberStyleMenuTexts,
			st->stNumberStyleMenu, st->stNumberStylePulldown,
			tedSectNumberStyleChosen, (void *)st );

    appFillInspectorMenu( DOCsbk_COUNT, DOCsbkPAGE,
			st->stBreakKindItems, spr->sprBreakKindMenuTexts,
			st->stBreakKindMenu, st->stBreakKindPulldown,
			tedSectBreakKindChosen, (void *)st );

    return;
    }

void tedFormatFinishSectionPage(	SectionTool *			st,
					const SectionPageResources *	spr )
    {
    appPulldownRefreshWidth( st->stNumberStyleMenu );

    appPulldownRefreshWidth( st->stBreakKindMenu );

    return;
    }
