/************************************************************************/
/*  Editor, File actions.						*/
/************************************************************************/

#   include	"config.h"

#   include	<stddef.h>
#   include	<stdio.h>
#   include	<stdlib.h>

#   include	<locale.h>

#   include	<Xm/Xm.h>
#   include	<Xm/MainW.h>
#   include	<Xm/Form.h>
#   include	<Xm/RowColumn.h>
#   include	<Xm/ScrolledW.h>
#   include	<Xm/ScrollBar.h>
#   include	<Xm/ToggleB.h>
#   include	<Xm/DrawingA.h>

#   include	<Xm/Protocols.h>
#   include	<Xm/MwmUtil.h>

#   include	<appSystem.h>
#   include	"appFrame.h"

#   include	<debugon.h>

/************************************************************************/
/*									*/
/*  Callback for the file menu.						*/
/*									*/
/************************************************************************/

void appDocFileSave(	Widget		option,
			XtPointer	voided,
			XtPointer	call_data	 )
    {
    EditDocument *		ed= (EditDocument *)voided;
    EditApplication *		ea= ed->edApplication;

    if  ( ! ea->eaSaveDocument )
	{ XDEB(ea->eaSaveDocument); return;	}

    if  ( ! ed->edFilename		||
	  ed->edFileReadOnly		||
	  ed->edFormat < 0		)
	{ appDocFileSaveAs( option, voided, (void *)0 ); return; }

    if  ( (*ea->eaSaveDocument)( ea, ed, ed->edFormat, ed->edFilename ) )
	{
	appReportSaveFailure( ea, option, ed->edTopWidget, ed->edFilename );
	}
    else{ appDocumentChanged( ed, 0 ); }

    return;
    }

/************************************************************************/
/*									*/
/*  'Close' callback for the file menu and for the window manager menu	*/
/*									*/
/*  1)  As 'mwm' sometimes sends a 'Save yourself' when 'Close' is	*/
/*	selected from the window manager menu, deactivate the protocol.	*/
/*	B.T.W. This does not help.					*/
/*									*/
/*									*/
/************************************************************************/

void appDocFileClose(	Widget		option,
			XtPointer	voided,
			XtPointer	call_data	 )
    {
    EditDocument *	ed= (EditDocument *)voided;
    EditApplication *	ea= ed->edApplication;

    if  ( option == ed->edTopWidget )
	{ option= (Widget)0;	}

    if  ( ed->edHasBeenChanged )
	{ appRunReallyCloseDialog( option, ed ); return; }

    appCloseDocument( ea, ed );
    }

void appDocFileNew(	Widget		option,
			XtPointer	voided,
			XtPointer	call_data	 )
    {
    EditDocument *	ed= (EditDocument *)voided;
    EditApplication *	ea= ed->edApplication;

    appAppFileNew( option, (void *)ea, call_data );
    }

void appDocFileQuit(	Widget		option,
			XtPointer	voided,
			XtPointer	call_data	 )
    {
    EditDocument *	ed= (EditDocument *)voided;
    EditApplication *	ea= ed->edApplication;

    appQuitApplication( option, ed->edTopWidget, ea );

    return;
    }

void appDocFileMini(	Widget		option,
			XtPointer	voided,
			XtPointer	call_data	 )
    {
    EditDocument *		ed= (EditDocument *)voided;
    EditApplication *		ea= ed->edApplication;
    Window			win= XtWindow( ed->edTopWidget );

    XtVaSetValues( ed->edTopWidget, XmNinitialState, IconicState, NULL );
    XIconifyWindow( ea->eaDisplay, win, ea->eaScreen );

    return;
    }

void appDocFileHide(	Widget		option,
			XtPointer	voided,
			XtPointer	call_data	 )
    {
    EditDocument *		ed= (EditDocument *)voided;
    EditApplication *		ea= ed->edApplication;

    if  ( ed->edIsVisible )
	{
	XtUnmapWidget( ed->edTopWidget );

	ed->edIsVisible= 0;

	appDocVisible( ea, ed, ed->edIsVisible );
	}

    return;
    }

void appReportSaveFailure(	EditApplication *	ea,
				Widget			option,
				Widget			relative,
				const char *		filename )
    {
    AppFileMessageResources *	afmr= &(ea->eaFileMessageResources);

    if  ( appTestFileWritable( filename ) )
	{
	appQuestionRunSubjectErrorDialog( ea, relative, option,
					filename, afmr->afmrFileNotWritable );
	}
    else{
	appQuestionRunSubjectErrorDialog( ea, relative, option,
					filename, afmr->afmrFileNotWritable );
	}
    }

/************************************************************************/
/*  The user selected the 'about' option.				*/
/************************************************************************/

void appDocAbout(	Widget		option,
			XtPointer	voided,
			XtPointer	call_data	 )
    {
    EditDocument *		ed= (EditDocument *)voided;
    EditApplication *		ea= ed->edApplication;
    Window			win= XtWindow( ea->eaTopWidget );

    ea->eaMainVisibleAsAbout= 1;

    XtVaSetValues( ea->eaTopWidget, XmNinitialState, NormalState, NULL );
    XMapRaised( ea->eaDisplay, win );

    return;
    }

/************************************************************************/
/*									*/
/*  Print a document.							*/
/*									*/
/*  1)  Avoid user oriented numeric formatting in a PostScript file,	*/
/*	that is meant to be read by computers.				*/
/*									*/
/************************************************************************/

void appCallPrintFunction(	FILE *			f,
				EditDocument *		ed,
				int			nup,
				int			horizontal,
				int			firstPage,
				int			lastPage )
    {
    EditApplication *	ea= ed->edApplication;

    /*  1  */
    setlocale( LC_NUMERIC, "C" );

    (*ea->eaPrintDocument)( f, ed, nup, horizontal, firstPage, lastPage );

    /*  1  */
    setlocale( LC_NUMERIC, "" );

    return;
    }

int appPrintDocument(	EditDocument *		ed,
			int			printer,
			int			nup,
			int			horizontal,
			int			firstPage,
			int			lastPage )
    {
    EditApplication *	ea= ed->edApplication;
    PrintDestination *	pd= ea->eaPrintDestinations+ printer;
    FILE *		f;
    char		scratchName[L_tmpnam+1];
    char		scratchCommand[L_tmpnam+200];

    switch( pd->pdPrintKind )
	{
	case APPprinterPIPE:
	    f= popen( pd->pdCommand, "w" );

	    if  ( ! f )
		{ SXDEB(pd->pdCommand,f); return -1;	}

	    appCallPrintFunction( f, ed,
					nup, horizontal, firstPage, lastPage );

	    pclose( f );
	    return 0;
	case APPprinterTMPFILE:
	    tmpnam( scratchName );
	    f= fopen( scratchName, "w" );
	    if  ( ! f )
		{ SXDEB(scratchName,f); return -1;	}

	    appCallPrintFunction( f, ed,
					nup, horizontal, firstPage, lastPage );

	    fclose( f );

	    sprintf( scratchCommand, pd->pdCommand, scratchName );
	    system( scratchCommand );
	    return 0;
	case APPprinterTOFILE:
	default:
	    LDEB(pd->pdPrintKind);
	    return -1;
	}
    }

static char *	appDocBuildFaxCommand(		const char *	faxCommand,
						int		commandLen,
						const char *	fileName,
						const char *	faxNumber,
						char *		title,
						int		fileCount,
						int		faxCount,
						int		titleCount )
    {
    char *		command;
    const char *	s;
    char *		to;

    int			fileLen;
    int			faxLen;
    int			titleLen;

    if  ( ! title )
	{ title= "";	}

    fileLen= strlen( fileName );
    faxLen= strlen( faxNumber );
    titleLen= strlen( title );

    to= command= (char *)malloc( commandLen+
					fileCount* fileLen+
					faxCount* faxLen+
					titleCount* titleLen+ 1 );
    if  ( ! command )
	{ XDEB(command); return (char *)0;	}

    s= faxCommand;
    while( *s )
	{
	if  ( s[0] == '%' && s[1] == 'f' )
	    {
	    strcpy( to, fileName ); to += fileLen;
	    s += 2; continue;
	    }

	if  ( s[0] == '%' && s[1] == 'n' )
	    {
	    strcpy( to, faxNumber ); to += faxLen;
	    s += 2; continue;
	    }

	if  ( s[0] == '%' && s[1] == 't' )
	    {
	    strcpy( to, title ); to += titleLen;
	    s += 2; continue;
	    }

	*(to++)= *(s++);
	}
    *to= '\0';

    return command;
    }

int appFaxDocument(	EditDocument *		ed,
			const char *		faxNumber,
			int			nup,
			int			horizontal,
			int			firstPage,
			int			lastPage )
    {
    EditApplication *	ea= ed->edApplication;
    int			fileCount= 0;
    int			faxCount= 0;
    int			titleCount= 0;

    int			commandLen;

    char		scratchName[L_tmpnam+1];

    char *		s;
    char *		command;

    FILE *		f;

    if  ( ! ea->eaFaxCommand )
	{ XDEB(ea->eaFaxCommand); return -1;	}

    s= ea->eaFaxCommand;
    while( *s )
	{
	if  ( s[0] == '%' && s[1] == 'f' )
	    { fileCount++; s += 2; continue;	}

	if  ( s[0] == '%' && s[1] == 'n' )
	    { faxCount++; s += 2; continue;	}

	if  ( s[0] == '%' && s[1] == 't' )
	    { titleCount++; s += 2; continue;	}

	s++;
	}

    commandLen= s- ea->eaFaxCommand;

    if  ( faxCount == 0 )
	{ LDEB(faxCount); return -1;	}

    if  ( fileCount > 0 )
	{
	tmpnam( scratchName );
	command= appDocBuildFaxCommand( ea->eaFaxCommand, commandLen,
					scratchName, faxNumber, ed->edTitle,
					fileCount, faxCount, titleCount );

	if  ( ! command )
	    { XDEB(command); return -1;	}

	f= fopen( scratchName, "w" );
	if  ( ! f )
	    { SXDEB(scratchName,f); free( command ); return -1;	}

	appCallPrintFunction( f, ed, nup, horizontal, firstPage, lastPage );

	fclose( f );

	system( command );
	}
    else{
	scratchName[0]= '\0';

	command= appDocBuildFaxCommand( ea->eaFaxCommand, commandLen,
					scratchName, faxNumber, ed->edTitle,
					fileCount, faxCount, titleCount );

	f= popen( command, "w" );
	if  ( ! f )
	    { SXDEB(command,f); free( command ); return -1;	}

	appCallPrintFunction( f, ed, nup, horizontal, firstPage, lastPage );

	pclose( f );
	}

    free( command );

    return 0;
    }

/************************************************************************/
/*									*/
/*  Callback for the item in the files menu.				*/
/*									*/
/************************************************************************/

void appMakeDocVisible(	EditApplication *	ea,
			EditDocument *		ed )
    {
    Window			win= XtWindow( ed->edTopWidget );

    if  ( ! ed->edIsVisible )
	{
	XtVaSetValues( ed->edTopWidget, XmNinitialState, NormalState, NULL );
	XtMapWidget( ed->edTopWidget );

	ed->edIsVisible= 1;

	appDocVisible( ea, ed, ed->edIsVisible );
	}

    XMapRaised( ea->eaDisplay, win );
    }

void appDocToFront(	Widget		option,
			XtPointer	voided,
			XtPointer	call_data	 )
    {
    EditDocument *		ed= (EditDocument *)voided;
    EditApplication *		ea= ed->edApplication;

    Boolean			state;

    appMakeDocVisible( ea, ed );

    state= XmToggleButtonGetState( option );
    XmToggleButtonSetState( option, ! state, False );

    return;
    }

/************************************************************************/
/*									*/
/*  Callbacks for the edit menu.					*/
/*									*/
/************************************************************************/

void appDocEditCopy(	Widget		option,
			XtPointer	voided,
			XtPointer	voidpbcs	 )
    {
    EditDocument *			ed= (EditDocument *)voided;
    EditApplication *			ea= ed->edApplication;
    XmPushButtonCallbackStruct *	pbcs=
					(XmPushButtonCallbackStruct *)voidpbcs;

    if  ( ! ea->eaDocCopy )
	{ SXDEB(ed->edFilename,ea->eaDocCopy); return;	}

    ea->eaDocCopy( ed, pbcs->event ); return;
    }

void appDocEditCut(	Widget		option,
			XtPointer	voided,
			XtPointer	voidpbcs	 )
    {
    EditDocument *			ed= (EditDocument *)voided;
    EditApplication *			ea= ed->edApplication;
    XmPushButtonCallbackStruct *	pbcs=
					(XmPushButtonCallbackStruct *)voidpbcs;

    if  ( ! ea->eaDocCut )
	{ SXDEB(ed->edFilename,ea->eaDocCut); return;	}

    ea->eaDocCut( ed, pbcs->event ); return;
    }

void appDocEditPaste(	Widget		option,
			XtPointer	voided,
			XtPointer	voidpbcs	 )
    {
    EditDocument *			ed= (EditDocument *)voided;
    XmPushButtonCallbackStruct *	pbcs=
					(XmPushButtonCallbackStruct *)voidpbcs;

    appDocAskForPaste( ed, "PRIMARY", pbcs->event->xbutton.time );
    }

void appDocEditSelAll(	Widget		option,
			XtPointer	voided,
			XtPointer	voidpbcs	 )
    {
    EditDocument *			ed= (EditDocument *)voided;
    EditApplication *			ea= ed->edApplication;
    XmPushButtonCallbackStruct *	pbcs=
					(XmPushButtonCallbackStruct *)voidpbcs;

    if  ( ! ea->eaDocSelAll )
	{ SXDEB(ed->edFilename,ea->eaDocSelAll); return;	}

    ea->eaDocSelAll( ed, pbcs->event ); return;
    }

/************************************************************************/
/*									*/
/*  Set the menu of a document window.					*/
/*									*/
/************************************************************************/

static void appDocMakeMenu(	EditDocument *		ed,
				EditApplication *	ea,
				Widget			mainWindow )
    {
    ed->edMenuBar= XmCreateMenuBar( mainWindow, WIDGET_NAME, NULL, 0 );

    ed->edFileMenu= appMakeMenu( &(ed->edFileMenuButton),
			ea, ed->edMenuBar, *(ea->eaDocFileMenuText), 0,
			ea->eaDocFileMenuItems, ea->eaDocFileMenuItemCount,
			(void *)ed );

    ed->edEditMenu= appMakeMenu( &(ed->edEditMenuButton),
			ea, ed->edMenuBar, *(ea->eaDocEditMenuText), 0,
			ea->eaDocEditMenuItems, ea->eaDocEditMenuItemCount,
			(void *)ed );

    if  ( ea->eaMakePrivateDocumentMenus )
	{ (*ea->eaMakePrivateDocumentMenus)( ea, ed, ed->edMenuBar ); }

    ed->edWindowMenu= appMakeMenu( &(ed->edWindowMenuButton),
		    ea, ed->edMenuBar,
		    *(ea->eaDocWindowMenuText), 0,
		    ea->eaDocWindowMenuItems, ea->eaDocWindowMenuItemCount,
		    (void *)ed );

    if  ( ea->eaDocHelpMenuItems )
	{
	ed->edHelpMenu= appMakeMenu( &(ed->edHelpMenuButton),
		    ea, ed->edMenuBar,
		    *(ea->eaDocHelpMenuText), 1,
		    ea->eaDocHelpMenuItems, ea->eaDocHelpMenuItemCount,
		    (void *)ed );
	}

    XtManageChild( ed->edMenuBar );
    }

/************************************************************************/
/*									*/
/*  Close a a document.							*/
/*									*/
/************************************************************************/

void appCloseDocument(	EditApplication *	ea,
			EditDocument *		ed	)
    {
    appRemoveDocument( ea, ed );

    XtDestroyWidget( ed->edTopWidget );

    return;
    }

/************************************************************************/
/*									*/
/*  Callbacks for the scrollbars, and the size admin  of a document.	*/
/*									*/
/************************************************************************/

void appDocVerticalScrollbarCallback(	Widget		w,
					XtPointer	voided,
					XtPointer	voidscbs )
    {
    XmScrollBarCallbackStruct *	scbs= (XmScrollBarCallbackStruct *)voidscbs;
    EditDocument *		ed= (EditDocument *)voided;
    EditApplication *		ea= ed->edApplication;

    Display *			display= XtDisplay( ed->edDocumentWidget );
    Window			win= XtWindow( ed->edDocumentWidget );
    GC				gc= ed->edGc;

    AppDrawingData *		add= &(ed->edDrawingData);
    DocumentRectangle *		drBack= &(add->addBackRect);
    DocumentRectangle		drClip;

    int				ox;
    int				oy;

    int				high;
    int				wide;
    int				d;

    high= ed->edVisibleRect.drY1- ed->edVisibleRect.drY0+ 1;
    wide= ed->edVisibleRect.drX1- ed->edVisibleRect.drX0+ 1;
    d= scbs->value- ed->edVisibleRect.drY0;

    ed->edVisibleRect.drY0 += d;
    ed->edVisibleRect.drY1 += d;

    XSetClipMask( display, gc, None );

    if  ( d == 0 )
	{ return;	}

    drClip= *drBack;

    if  ( d > 0 )
	{
	if  ( high > d )
	    {
	    XCopyArea( display, win, win, gc, 0, d, wide, high- d, 0, 0 );

	    drClip.drY0= ed->edVisibleRect.drY1- d;
	    drClip.drY1= ed->edVisibleRect.drY1;
	    }
	else{
	    drClip.drY0= ed->edVisibleRect.drY0;
	    drClip.drY1= ed->edVisibleRect.drY1;
	    }
	}
    else{
	if  ( high+ d > 0 )
	    {
	    XCopyArea( display, win, win, gc, 0, 0, wide, high+ d, 0, -d );

	    drClip.drY0= ed->edVisibleRect.drY0;
	    drClip.drY1= ed->edVisibleRect.drY0- d;
	    }
	else{
	    drClip.drY0= ed->edVisibleRect.drY0;
	    drClip.drY1= ed->edVisibleRect.drY1;
	    }
	}

    ox= ed->edVisibleRect.drX0;
    oy= ed->edVisibleRect.drY0;

    (*ea->eaDrawRectangle)( ed->edDocumentWidget, ed, &drClip,
							ox, oy, win, gc );

    if  ( ed->edLeftRuler && ea->eaScrollVerticalRuler )
	{
	(*ea->eaScrollVerticalRuler)
				( ed->edLeftRuler, ed->edLeftRulerWidget, d );
	}

    return;
    }


void appDocHorizontalScrollbarCallback(	Widget		w,
					XtPointer	voided,
					XtPointer	voidscbs )
    {
    XmScrollBarCallbackStruct *	scbs= (XmScrollBarCallbackStruct *)voidscbs;
    EditDocument *		ed= (EditDocument *)voided;
    EditApplication *		ea= ed->edApplication;

    Display *			display= XtDisplay( ed->edDocumentWidget );
    Window			win= XtWindow( ed->edDocumentWidget );
    GC				gc= ed->edGc;

    DocumentRectangle		drClip;

    int				ox;
    int				oy;

    int				high;
    int				wide;
    int				d;

    high= ed->edVisibleRect.drY1- ed->edVisibleRect.drY0+ 1;
    wide= ed->edVisibleRect.drX1- ed->edVisibleRect.drX0+ 1;
    d= scbs->value- ed->edVisibleRect.drX0;

    ed->edVisibleRect.drX0 += d;
    ed->edVisibleRect.drX1 += d;

    XSetClipMask( display, gc, None );

    if  ( d == 0 )
	{ return;	}

    drClip= ed->edVisibleRect;

    if  ( d > 0 )
	{
	if  ( wide > d )
	    {
	    XCopyArea( display, win, win, gc, d, 0, wide- d, high, 0, 0 );

	    drClip.drX0= ed->edVisibleRect.drX1- d;
	    drClip.drX1= ed->edVisibleRect.drX1;
	    }
	else{
	    drClip.drX0= ed->edVisibleRect.drX0;
	    drClip.drX1= ed->edVisibleRect.drX1;
	    }
	}
    else{
	if  ( wide+ d > 0 )
	    {
	    XCopyArea( display, win, win, gc, 0, 0, wide+ d, high, -d, 0 );

	    drClip.drX0= ed->edVisibleRect.drX0;
	    drClip.drX1= ed->edVisibleRect.drX0- d;
	    }
	else{
	    drClip.drX0= ed->edVisibleRect.drX0;
	    drClip.drX1= ed->edVisibleRect.drX1;
	    }
	}

    ox= ed->edVisibleRect.drX0;
    oy= ed->edVisibleRect.drY0;

    (*ea->eaDrawRectangle)( ed->edDocumentWidget, ed, &drClip,
							ox, oy, win, gc );

    if  ( ed->edTopRuler && ea->eaScrollHorizontalRuler )
	{
	(*ea->eaScrollHorizontalRuler)
				( ed->edTopRuler, ed->edTopRulerWidget, d );
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Prevent the Shell that contains a document from being resized	*/
/*  beyond normal limits.						*/
/*									*/
/************************************************************************/

void appSetShellConstraints(	EditDocument *		ed )
    {
    Dimension		width;
    Dimension		height;

    Arg			al[20];
    int			ac= 0;

    AppDrawingData *	add= &(ed->edDrawingData);

    XtVaGetValues( ed->edTopWidget,
			XmNwidth,	&width,
			XmNheight,	&height,
			NULL );

    if  ( ed->edShellExtraWidth >= 0 )
	{
	XtWidgetGeometry	preferred;

	int			maxWidth;

	maxWidth= add->addBackRect.drX1- add->addBackRect.drX0;
	maxWidth += ed->edShellExtraWidth;

	XtQueryGeometry( ed->edMenuBar, (XtWidgetGeometry *)0, &preferred );
	if  ( maxWidth < preferred.width+ preferred.border_width )
	    { maxWidth=  preferred.width+ preferred.border_width;	}

	XtSetArg( al[ac],	XmNmaxWidth,	maxWidth ); ac++;

	if  ( width > maxWidth )
	    {
	    XtSetArg( al[ac],	XmNwidth,	maxWidth ); ac++;

	    ed->edVisibleRect.drX1 += maxWidth- width;
	    }
	}

    if  ( ed->edShellExtraHeight >= 0 )
	{
	int	maxHeight;

	maxHeight= add->addBackRect.drY1- add->addBackRect.drY0;

	if  ( maxHeight < add->addPaperRect.drY1- add->addPaperRect.drY0 )
	    { maxHeight=  add->addPaperRect.drY1- add->addPaperRect.drY0; }

	maxHeight += ed->edShellExtraHeight;

	XtSetArg( al[ac],	XmNmaxHeight,	maxHeight ); ac++;

	if  ( height > maxHeight )
	    {
	    XtSetArg( al[ac],	XmNheight,	maxHeight ); ac++;

	    ed->edVisibleRect.drY1 += maxHeight- height;
	    }
	}

    if  ( ac > 0 )
	{ XtSetValues( ed->edTopWidget, al, ac );	}

    return;
    }

static void appFileAdaptHorizontalRulerRange(	EditDocument *	ed,
						int		width )
    {
    EditApplication *		ea= ed->edApplication;

    if  ( ed->edTopRuler && ea->eaSetHorizontalRulerRange )
	{
	Dimension	leftRulerWidth= 0;

	if  ( ed->edLeftRulerWidget )
	    {
	    XtVaGetValues( ed->edLeftRulerWidget,
				XmNwidth,		&leftRulerWidth,
				NULL );
	    }

	(*ea->eaSetHorizontalRulerRange)( ed->edTopRuler,
					ed->edTopRulerWidget,
					ed->edVisibleRect.drX0,
					ed->edVisibleRect.drX1,
					width+ leftRulerWidth );
	}

    return;
    }

static void appFileAdaptVerticalRulerRange(	EditDocument *	ed,
						int		height )
    {
    EditApplication *		ea= ed->edApplication;

    if  ( ed->edLeftRuler && ea->eaSetVerticalRulerRange )
	{
	(*ea->eaSetVerticalRulerRange)( ed->edLeftRuler,
					ed->edLeftRulerWidget,
					ed->edVisibleRect.drY0,
					ed->edVisibleRect.drY1,
					height );
	}

    return;
    }

static void appAdaptToDocumentSize(	EditDocument *	ed,
					int		height,
					int		width )
    {
    Widget		w= ed->edDocumentWidget;
    AppDrawingData *	add= &(ed->edDrawingData);
    int			changed= 0;
    int			d;

    if  ( ed->edVisibleRect.drY1 != ed->edVisibleRect.drY0+ height- 1 )
	{
	ed->edVisibleRect.drY1= ed->edVisibleRect.drY0+ height- 1;
	changed= 1;

	if  ( ed->edVisibleRect.drY1 > add->addBackRect.drY1 )
	    {
	    d= ed->edVisibleRect.drY1- add->addBackRect.drY1;

	    if  ( d > ed->edVisibleRect.drY0 )
		{ d= ed->edVisibleRect.drY0;	}

	    ed->edVisibleRect.drY0 -= d;
	    ed->edVisibleRect.drY1 -= d;
	    }

	appFileAdaptVerticalRulerRange( ed, height );

	XClearArea( XtDisplay( w ), XtWindow( w ), 0, 0, 0, 0, True );
	}

    if  ( ed->edVisibleRect.drX1 != ed->edVisibleRect.drX0+ width+ 1 )
	{
	ed->edVisibleRect.drX1= ed->edVisibleRect.drX0+ width+ 1;
	changed= 1;

	if  ( ed->edVisibleRect.drX1 > add->addBackRect.drX1 )
	    {
	    d= ed->edVisibleRect.drX1- add->addBackRect.drX1;

	    if  ( d > ed->edVisibleRect.drX0 )
		{ d= ed->edVisibleRect.drX0;	}

	    ed->edVisibleRect.drX0 -= d;
	    ed->edVisibleRect.drX1 -= d;
	    }

	appFileAdaptHorizontalRulerRange( ed, width );

	XClearArea( XtDisplay( w ), XtWindow( w ), 0, 0, 0, 0, True );
	}

    if  ( changed )
	{ appDocSetScrollbarValues( ed );	}

    return;
    }

static void appDocTopConfigure(	Widget			w,
				void *			voided,
				XEvent *		event,
				Boolean *		pRefused )
    {
    EditDocument *	ed= (EditDocument *)voided;

    if  ( event->type == MapNotify && ed->edShellExtraWidth < 0 )
	{
	Dimension	shellWidth;
	Dimension	shellHeight;

	Dimension	docWidth;
	Dimension	docHeight;

	XtVaGetValues( ed->edTopWidget,
			    XmNwidth,		&shellWidth,
			    XmNheight,		&shellHeight,
			    NULL );

	XtVaGetValues( ed->edDocumentWidget,
			    XmNwidth,		&docWidth,
			    XmNheight,		&docHeight,
			    NULL );

	ed->edShellExtraWidth= shellWidth- docWidth;
	ed->edShellExtraHeight= shellHeight- docHeight;

	appSetShellConstraints( ed );

	appAdaptToDocumentSize( ed, docHeight, docWidth );

	ed->edMapped= 1;
	}

    if  ( event->type == ConfigureNotify )
	{
	XConfigureEvent *	cevent= (XConfigureEvent *)event;

	if  ( ! ed->edMapped && cevent->y < 0 )
	    {
	    XtVaSetValues( ed->edTopWidget,
				XmNy,		0,
				NULL );
	    }
	}

    *pRefused= 1;
    return;
    }

static void appDocConfigure(	Widget			w,
				void *			voided,
				XEvent *		event,
				Boolean *		pRefused )
    {
    XConfigureEvent *	cevent= &(event->xconfigure);
    EditDocument *	ed= (EditDocument *)voided;

    if  ( event->type != ConfigureNotify )
	{ return;	}

    appAdaptToDocumentSize( ed, cevent->height, cevent->width );

    *pRefused= 1;
    return;
    }

/************************************************************************/
/*									*/
/*  Redraw a rectangle in the document widget. Comes in two variants:	*/
/*									*/
/*  1)  An XmDrawingarea calback.					*/
/*  2)  An Xt Event handler.						*/
/*									*/
/************************************************************************/

void appDrawingExpose(	Widget		w,
			XtPointer	voided,
			XtPointer	voidcbs	)
    {
    XmDrawingAreaCallbackStruct *	cbs;
    Boolean				ignored;

    cbs= (XmDrawingAreaCallbackStruct *)voidcbs;

    appExposeHandler( w, voided, cbs->event, &ignored );

    return;
    }

void appExposeHandler(	Widget			w,
			void *			voided,
			XEvent *		event,
			Boolean *		pRefused )
    {
    EditDocument *		ed= (EditDocument *)voided;
    EditApplication *		ea= ed->edApplication;

    Display *			display= XtDisplay( w );
    Window			win= XtWindow( w );
    GC				gc= ed->edGc;

    int				ox= ed->edVisibleRect.drX0;
    int				oy= ed->edVisibleRect.drY0;

    DocumentRectangle		drClip;

    /*  1  */
    appCollectExposures( &drClip, ox, oy, display, win, gc, event );

    /*  2,3,4  */
    (*ea->eaDrawRectangle)( w, ed, &drClip, ox, oy, win, gc );

    XSetClipMask( display, gc, None );

    return;
    }

/************************************************************************/
/*									*/
/*  Make a document window.						*/
/*									*/
/*  1)  Allocate a document.						*/
/*  3)  Make a shell to contain everything.				*/
/*  4)  Make the window pane.						*/
/*									*/
/************************************************************************/

static int appDocMakeMainWindow(	EditDocument *		ed,
					const char *		title,
					EditApplication *	ea )
    {
    Arg			al[20];
    int			ac= 0;
    Pixmap		pixmap= (Pixmap)0;

    const char *	baseName;

    int			len;
    static char *	fullTitle;
    static char *	fullIconName;

    int			maxWidth= ( 5* ea->eaScreenPixelsWide )/ 6;
    int			maxHeight= ( 5* ea->eaScreenPixelsHigh )/ 6;

    baseName= strrchr( title, '/' );
    if  ( baseName )
	{ baseName++;	}
    else{ baseName= title;	}

    if  ( ea->eaDocumentIcon					&&
	  appGetImagePixmap( ea, ea->eaDocumentIcon, &pixmap )	)
	{ SDEB(ea->eaDocumentIcon); return -1;	}

    len= strlen( ea->eaApplicationName )+ 2+ strlen( title )+ 1;
    fullTitle= (char *)realloc( fullTitle, len );
    if  ( ! fullTitle )
	{ LXDEB(len,fullTitle); return -1;	}

    len= strlen( ea->eaApplicationName )+ 2+ strlen( baseName )+ 1;
    fullIconName= (char *)realloc( fullIconName, len );
    if  ( ! fullIconName )
	{ LXDEB(len,fullIconName); return -1;	}

    sprintf( fullTitle, "%s: %s", ea->eaApplicationName, title );
    sprintf( fullIconName, "%s: %s", ea->eaApplicationName, baseName );

    XtSetArg( al[ac], XmNtitle,			fullTitle ); ac++;
    XtSetArg( al[ac], XmNiconName,		fullIconName ); ac++;
    XtSetArg( al[ac], XmNuserData,		(void *)ed ); ac++;
    XtSetArg( al[ac], XmNdeleteResponse,	XmDO_NOTHING ); ac++;
    XtSetArg( al[ac], XmNallowShellResize,	True ); ac++;
    XtSetArg( al[ac], XmNmappedWhenManaged,	False ); ac++;
    XtSetArg( al[ac], XmNinput,			True ); ac++;
    XtSetArg( al[ac], XmNmaxWidth,		maxWidth ); ac++;
    XtSetArg( al[ac], XmNmaxHeight,		maxHeight ); ac++;

    if  ( pixmap )
	{ XtSetArg( al[ac], XmNiconPixmap,	pixmap ); ac++; }

#   ifdef USE_X11_R5
    ed->edTopWidget= XtAppCreateShell( ea->eaApplicationName, 
					    ea->eaDocumentWidgetName,
					    applicationShellWidgetClass,
					    ea->eaDisplay, al, ac );
#   else
    ed->edTopWidget= XtAppCreateShell( ea->eaApplicationName, 
					    ea->eaDocumentWidgetName,
					    topLevelShellWidgetClass,
					    ea->eaDisplay, al, ac );
#   endif

    if  ( ea->eaCloseAtom != None )
	{
	XmAddWMProtocolCallback( ed->edTopWidget, ea->eaCloseAtom,
					appDocFileClose, (XtPointer)ed );
	}

    ac= 0;
    XtSetArg( al[ac], XmNuserData, (void *)ed ); ac++;
    ed->edMainWindow= XmCreateMainWindow( ed->edTopWidget, WIDGET_NAME, al, ac );

    appDocMakeMenu( ed, ea, ed->edMainWindow );

    ac= 0;
    ed->edMainForm= XmCreateForm( ed->edMainWindow, WIDGET_NAME, al, ac );

    return 0;
    }

int appSetDocumentFilename(	EditDocument *		ed,
				char *			filename )
    {
    char *			s= (char *)0;

    if  ( filename )
	{
	s= strdup( filename );
	if  ( ! s )
	    { XDEB(s); return -1;	}
	}

    if  ( ed->edFilename )
	{ free( ed->edFilename );	}
    ed->edFilename= s;

    return 0;
    }

int appSetDocumentTitle(	EditDocument *		ed,
				char *			title )
    {
    EditApplication *		ea= ed->edApplication;

    char *			s;
    char *			baseName;

    int				len;
    static char *		fullTitle;
    static char *		fullIconName;

    s= strdup( title );
    if  ( ! s )
	{ XDEB(s); return -1;	}

    if  ( ed->edTitle )
	{ free( ed->edTitle );	}
    ed->edTitle= s;

    baseName= strrchr( ed->edTitle, '/' );
    if  ( baseName )
	{ baseName++;			}
    else{ baseName= ed->edTitle;	}

    len= strlen( ea->eaApplicationName )+ 2+ strlen( title )+ 1;
    fullTitle= (char *)realloc( fullTitle, len );
    if  ( ! fullTitle )
	{ LXDEB(len,fullTitle); return -1;	}

    len= strlen( ea->eaApplicationName )+ 2+ strlen( baseName )+ 1;
    fullIconName= (char *)realloc( fullIconName, len );
    if  ( ! fullIconName )
	{ LXDEB(len,fullIconName); return -1;	}

    sprintf( fullTitle, "%s: %s", ea->eaApplicationName, title );
    sprintf( fullIconName, "%s: %s", ea->eaApplicationName, baseName );

    XtVaSetValues( ed->edTopWidget,
			XmNtitle,		fullTitle,
			XmNiconName,		fullIconName,
			NULL );

    appRenameDocumentOptions( ed->edApplication, ed, ed->edTitle );

    return 0;
    }


static int appDocMakeToolbar(	EditDocument *		ed,
				EditApplication *	ea )
    {
#   if 0
    Arg			al[20];
    int			ac= 0;

    Display *		display= ea->eaDisplay;
    int			screen= ea->eaScreen;

    int			verPixPerCM;

    verPixPerCM= ( 10* DisplayHeight( display, screen ) )/
					DisplayHeightMM( display, screen );
    XtSetArg( al[ac], XmNleftAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac], XmNtopAttachment,		XmATTACH_FORM ); ac++;
    XtSetArg( al[ac], XmNrightAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac], XmNheight,		verPixPerCM ); ac++;

    ed->edToolbar= XmCreateForm( ed->edMainForm, WIDGET_NAME, al, ac );

#   endif
    return 0;
    }

static int appDocMakeScrolledWindow(	EditDocument *		ed,
					EditApplication *	ea )
    {
    Arg			al[20];
    int			ac= 0;
    
    ac= 0;
    XtSetArg( al[ac], XmNleftAttachment,	XmATTACH_FORM ); ac++;
    if  ( ed->edToolbar )
	{
	XtSetArg( al[ac], XmNtopAttachment,	XmATTACH_WIDGET ); ac++;
	XtSetArg( al[ac], XmNtopWidget,		ed->edToolbar ); ac++;
	}
    else{
	XtSetArg( al[ac], XmNtopAttachment,	XmATTACH_FORM ); ac++;
	}
    XtSetArg( al[ac], XmNrightAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac], XmNbottomAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac], XmNscrollingPolicy,	XmAPPLICATION_DEFINED ); ac++;
    XtSetArg( al[ac], XmNscrollBarDisplayPolicy, XmSTATIC ); ac++;
    XtSetArg( al[ac], XmNhighlightThickness,	0 ); ac++;
    XtSetArg( al[ac], XmNspacing,		0 ); ac++;

    ed->edScrolledWindow=
	    XmCreateScrolledWindow( ed->edMainForm, WIDGET_NAME, al, ac );

    ac= 0;
    XtSetArg( al[ac], XmNorientation,		XmVERTICAL ); ac++;
    XtSetArg( al[ac], XmNprocessingDirection,	XmMAX_ON_BOTTOM ); ac++;
    XtSetArg( al[ac], XmNtraversalOn,		False ); ac++;
    XtSetArg( al[ac], XmNnavigationType,	XmNONE ); ac++;
    ed->edVerticalScrollbar=
		XmCreateScrollBar( ed->edScrolledWindow, WIDGET_NAME, al, ac );

    XtAddCallback( ed->edVerticalScrollbar, XmNvalueChangedCallback,
			    ea->eaDocVerticalScrollbarCallback, (void *)ed );
    XtAddCallback( ed->edVerticalScrollbar, XmNdragCallback,
			    ea->eaDocVerticalScrollbarCallback, (void *)ed );

    ac= 0;
    XtSetArg( al[ac], XmNorientation,		XmHORIZONTAL ); ac++;
    XtSetArg( al[ac], XmNprocessingDirection,	XmMAX_ON_RIGHT ); ac++;
    XtSetArg( al[ac], XmNtraversalOn,		False ); ac++;
    XtSetArg( al[ac], XmNnavigationType,	XmNONE ); ac++;
    ed->edHorizontalScrollbar=
		XmCreateScrollBar( ed->edScrolledWindow, WIDGET_NAME, al, ac );

    XtAddCallback( ed->edHorizontalScrollbar, XmNvalueChangedCallback,
			    appDocHorizontalScrollbarCallback, (void *)ed );
    XtAddCallback( ed->edHorizontalScrollbar, XmNdragCallback,
			    appDocHorizontalScrollbarCallback, (void *)ed );

    return 0;
    }

int appMakeDocumentWidget(	EditApplication *	ea,
				EditDocument *		ed )
    {
    Arg			al[20];
    int			ac= 0;
    int			mult;

    Widget		form;

    Display *		display= ea->eaDisplay;
    int			screen= ea->eaScreen;
    Pixel		blackPixel= BlackPixel( display, screen );

    double		horPixPerMM;
    double		verPixPerMM;
    double		xfac;
    double		yfac;

    appGetFactors( ea, &horPixPerMM, &verPixPerMM, &xfac, &yfac );

    /*  1  */
    ed->edLeftRulerWidthPixels=
			(int)( ea->eaLeftRulerWidthMM* horPixPerMM );
    ed->edTopRulerHeightPixels=
			(int)( ea->eaTopRulerHeightMM* verPixPerMM );
    ed->edRightRulerWidthPixels=
			(int)( ea->eaRightRulerWidthMM * horPixPerMM );
    ed->edBottomRulerHeightPixels=
			(int)( ea->eaBottomRulerHeightMM* verPixPerMM );

    mult= ea->eaLeftRulerWidthMultiple;
    if  ( mult )
	{
	ed->edLeftRulerWidthPixels=
		    mult* ( ( ed->edLeftRulerWidthPixels+ mult- 1 )/ mult );
	}

    mult= ea->eaTopRulerHeightMultiple;
    if  ( mult )
	{
	ed->edTopRulerHeightPixels=
		    mult* ( ( ed->edTopRulerHeightPixels+ mult- 1 )/ mult );
	}

    mult= ea->eaRightRulerWidthMultiple;
    if  ( mult )
	{
	ed->edRightRulerWidthPixels=
		    mult* ( ( ed->edRightRulerWidthPixels+ mult- 1 )/ mult );
	}

    mult= ea->eaBottomRulerHeightMultiple;
    if  ( mult )
	{
	ed->edBottomRulerHeightPixels=
		    mult* ( ( ed->edBottomRulerHeightPixels+ mult- 1 )/ mult );
	}

    /*  2  */
    form= XmCreateForm( ed->edScrolledWindow, WIDGET_NAME, al, ac );

    /*  3  */
    ac= 0;
    XtSetArg( al[ac],	XmNforeground,		blackPixel ); ac++;
    XtSetArg( al[ac],	XmNtraversalOn,		False ); ac++;
    XtSetArg( al[ac],	XmNleftAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNleftOffset,		0 ); ac++;
    XtSetArg( al[ac],	XmNtopAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNtopOffset,		0 ); ac++;
    XtSetArg( al[ac],	XmNrightAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNrightOffset,		0 ); ac++;
    XtSetArg( al[ac],	XmNbottomAttachment,	XmATTACH_OPPOSITE_FORM ); ac++;
    XtSetArg( al[ac],	XmNbottomOffset,	-ed->edTopRulerHeightPixels ); ac++;

    ed->edTopRulerWidget= XmCreateDrawingArea( form, WIDGET_NAME, al, ac );

    /*  4  */
    ac= 0;
    XtSetArg( al[ac],	XmNforeground,		blackPixel ); ac++;
    XtSetArg( al[ac],	XmNtraversalOn,		False ); ac++;
    XtSetArg( al[ac],	XmNleftAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNleftOffset,		0 ); ac++;
    XtSetArg( al[ac],	XmNtopAttachment,	XmATTACH_OPPOSITE_FORM ); ac++;
    XtSetArg( al[ac],	XmNtopOffset,		-ed->edBottomRulerHeightPixels ); ac++;
    XtSetArg( al[ac],	XmNrightAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNrightOffset,		0 ); ac++;
    XtSetArg( al[ac],	XmNbottomAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNbottomOffset,	0 ); ac++;

    ed->edBottomRulerWidget= XmCreateDrawingArea( form, WIDGET_NAME, al, ac );

    /*  5  */
    ac= 0;
    XtSetArg( al[ac],	XmNforeground,		blackPixel ); ac++;
    XtSetArg( al[ac],	XmNtraversalOn,		False ); ac++;

    XtSetArg( al[ac],	XmNleftAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNleftOffset,		0 ); ac++;

    XtSetArg( al[ac],	XmNtopAttachment,	XmATTACH_WIDGET ); ac++;
    XtSetArg( al[ac],	XmNtopWidget,		ed->edTopRulerWidget ); ac++;
    XtSetArg( al[ac],	XmNtopOffset,		0 ); ac++;

    XtSetArg( al[ac],	XmNrightAttachment,	XmATTACH_OPPOSITE_FORM ); ac++;
    XtSetArg( al[ac],	XmNrightOffset,		-ed->edLeftRulerWidthPixels ); ac++;

    XtSetArg( al[ac],	XmNbottomAttachment,	XmATTACH_WIDGET ); ac++;
    XtSetArg( al[ac],	XmNbottomWidget,	ed->edBottomRulerWidget ); ac++;
    XtSetArg( al[ac],	XmNbottomOffset,	0 ); ac++;

    ed->edLeftRulerWidget= XmCreateDrawingArea( form, WIDGET_NAME, al, ac );

    /*  6  */
    ac= 0;
    XtSetArg( al[ac],	XmNforeground,		blackPixel ); ac++;
    XtSetArg( al[ac],	XmNtraversalOn,		False ); ac++;
    XtSetArg( al[ac],	XmNleftAttachment,	XmATTACH_OPPOSITE_FORM ); ac++;
    XtSetArg( al[ac],	XmNleftOffset,		-ed->edRightRulerWidthPixels ); ac++;
    XtSetArg( al[ac],	XmNtopAttachment,	XmATTACH_WIDGET ); ac++;
    XtSetArg( al[ac],	XmNtopWidget,		ed->edTopRulerWidget ); ac++;
    XtSetArg( al[ac],	XmNtopOffset,		0 ); ac++;
    XtSetArg( al[ac],	XmNrightAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNrightOffset,		0 ); ac++;
    XtSetArg( al[ac],	XmNbottomAttachment,	XmATTACH_WIDGET ); ac++;
    XtSetArg( al[ac],	XmNbottomWidget,	ed->edBottomRulerWidget ); ac++;
    XtSetArg( al[ac],	XmNbottomOffset,	0 ); ac++;

    ed->edRightRulerWidget= XmCreateDrawingArea( form, WIDGET_NAME, al, ac );

    /*  7  */
    ac= 0;
    XtSetArg( al[ac],	XmNtraversalOn,		True ); ac++;

    /* NO: Postpone
    XtSetArg( al[ac],	XmNbackground,		whitePixel ); ac++;
    */
    XtSetArg( al[ac],	XmNleftAttachment,	XmATTACH_WIDGET ); ac++;
    XtSetArg( al[ac],	XmNleftWidget,		ed->edLeftRulerWidget ); ac++;
    XtSetArg( al[ac],	XmNleftOffset,		0 ); ac++;
    XtSetArg( al[ac],	XmNtopAttachment,	XmATTACH_WIDGET ); ac++;
    XtSetArg( al[ac],	XmNtopWidget,		ed->edTopRulerWidget ); ac++;
    XtSetArg( al[ac],	XmNtopOffset,		0 ); ac++;
    XtSetArg( al[ac],	XmNrightAttachment,	XmATTACH_WIDGET ); ac++;
    XtSetArg( al[ac],	XmNrightWidget,		ed->edRightRulerWidget ); ac++;
    XtSetArg( al[ac],	XmNrightOffset,		0 ); ac++;
    XtSetArg( al[ac],	XmNbottomAttachment,	XmATTACH_WIDGET ); ac++;
    XtSetArg( al[ac],	XmNbottomWidget,	ed->edBottomRulerWidget ); ac++;
    XtSetArg( al[ac],	XmNbottomOffset,	0 ); ac++;

    ed->edDocumentWidget= XmCreateDrawingArea( form, WIDGET_NAME, al, ac );

    if  ( ea->eaDocDestroyCallback )
	{
	XtAddCallback( ed->edDocumentWidget,
		XmNdestroyCallback, ea->eaDocDestroyCallback, (void *)ed );
	}

    XtManageChild( ed->edTopRulerWidget );
    XtManageChild( ed->edLeftRulerWidget );
    XtManageChild( ed->edRightRulerWidget );
    XtManageChild( ed->edBottomRulerWidget );
    XtManageChild( ed->edDocumentWidget );
    XtManageChild( form );

    return 0;
    }

/************************************************************************/
/*  Free all resources allocated for a document.			*/
/************************************************************************/

void appFreeDocument(		EditApplication *	ea,
				EditDocument *		ed )
    {
    if  ( ed->edLeftRuler )
	{ (*ea->eaFreeLeftRuler)( ea->eaDisplay, ed->edLeftRuler ); }
    if  ( ed->edTopRuler )
	{ (*ea->eaFreeTopRuler)( ea->eaDisplay, ed->edTopRuler ); }
    if  ( ed->edRightRuler )
	{ (*ea->eaFreeRightRuler)( ea->eaDisplay, ed->edRightRuler ); }
    if  ( ed->edBottomRuler )
	{ (*ea->eaFreeBottomRuler)( ea->eaDisplay, ed->edBottomRuler ); }

    if  ( ed->edFilename )
	{ free( ed->edFilename );	}
    if  ( ed->edTitle )
	{ free( ed->edTitle );	}

    if  ( ed->edCheckpointFilename )
	{
	if  ( appRemoveFile( ed->edCheckpointFilename ) )
	    { SDEB(ed->edCheckpointFilename);	}

	free( ed->edCheckpointFilename );
	}

    if  ( ed->edGc )
	{ XFreeGC( ea->eaDisplay, ed->edGc ); }
    if  ( ed->edInputContext )
	{ XDestroyIC( ed->edInputContext ); }

    appCleanDrawingData( &(ed->edDrawingData) );
    appCleanColors( ea->eaDisplay, &(ed->edColors) );

    free( ed );

    return;
    }

int appMakeDocumentWindow(	EditDocument **		pEd,
				EditApplication *	ea,
				int			readonly,
				const char *		title,
				const char *		filename )
    {
    char			scratch[100];

    EditDocument *		ed;

    if  ( ! title )
	{
	static int			count= 1;

	AppFileMessageResources *	afmr= &(ea->eaFileMessageResources);

	sprintf( scratch, afmr->afmrNamelessTitleFormat, count++ );
	title= scratch;
	}

    /*  1  */
    ed= (EditDocument *)malloc( sizeof(EditDocument) );
    if  ( ! ed )
	{ XDEB(ed); return -1;	}

    ed->edApplication= ea;

    ed->edFilename= (char *)0;
    ed->edTitle= (char *)0;
    ed->edCheckpointFilename= (char *)0;
    ed->edFormat= -1;
    ed->edFileReadOnly= 0;

    ed->edTopWidget= (Widget)0;
    ed->edMenuBar= (Widget)0;
    ed->edMainWindow= (Widget)0;
	ed->edFileMenu= (Widget)0;
	ed->edFileMenuButton= (Widget)0;
	ed->edEditMenu= (Widget)0;
	ed->edEditMenuButton= (Widget)0;
	ed->edWindowMenu= (Widget)0;
	ed->edWindowMenuButton= (Widget)0;
	ed->edHelpMenu= (Widget)0;
	ed->edHelpMenuButton= (Widget)0;

    ed->edMainForm= (Widget)0;
    ed->edToolbar= (Widget)0;
    ed->edScrolledWindow= (Widget)0;
	ed->edVerticalScrollbar= (Widget)0;
	ed->edHorizontalScrollbar= (Widget)0;
	ed->edDocumentWidget= (Widget)0;

	ed->edLeftRulerWidget= (Widget)0;
	ed->edTopRulerWidget= (Widget)0;
	ed->edRightRulerWidget= (Widget)0;
	ed->edBottomRulerWidget= (Widget)0;

	ed->edLeftRuler= (void *)0;
	ed->edTopRuler= (void *)0;
	ed->edRightRuler= (void *)0;
	ed->edBottomRuler= (void *)0;

    ed->edGc= (GC)0;
    ed->edInputContext= (XIC)0;

    ed->edLeftRulerWidthPixels= 0;
    ed->edTopRulerHeightPixels= 0;
    ed->edRightRulerWidthPixels= 0;
    ed->edBottomRulerHeightPixels= 0;

    ed->edHasBeenChanged= 0;
    ed->edIsReadonly= readonly;
    ed->edIsVisible= 0;

    ed->edShellExtraWidth= -1;
    ed->edShellExtraHeight= -1;

    appInitColors( &(ed->edColors) );
    appInitDrawingData( &(ed->edDrawingData) );

    ed->edMapped= 0;

    if  ( filename )
	{ ed->edFilename= strdup( filename );	}
    else{ ed->edFilename= (char *)0;		}

    if  ( title )
	{ ed->edTitle= strdup( title );	}
    else{ ed->edTitle= (char *)0;	}

    if  ( ea->eaSetPrivateData && (*ea->eaSetPrivateData)( ea, ed ) )
	{ appFreeDocument( ea, ed ); return -1; }

    /*  3  */
    if  ( appDocMakeMainWindow( ed, title, ea )		||
	  appDocMakeToolbar( ed, ea )			||
	  appDocMakeScrolledWindow( ed, ea )		||
	  (*ea->eaMakeDocumentWidget)( ea, ed )		)
	{ appFreeDocument( ea, ed ); return -1; }

    XtAddEventHandler( ed->edDocumentWidget, StructureNotifyMask, False,
					appDocConfigure, (void *)ed );
    XtAddEventHandler( ed->edTopWidget, StructureNotifyMask, False,
					appDocTopConfigure, (void *)ed );

    /*  4  */
    XtAddEventHandler( ed->edDocumentWidget,
			PropertyChangeMask, True, appDocSelectionInput, ed );

    XtManageChild( ed->edVerticalScrollbar );
    XtManageChild( ed->edHorizontalScrollbar );

    XtManageChild( ed->edDocumentWidget );
    XtManageChild( ed->edScrolledWindow );

    if  ( ed->edToolbar )
	{ XtManageChild( ed->edToolbar ); }

    XtManageChild( ed->edMainForm );
    XtManageChild( ed->edMainWindow );

    *pEd= ed; return 0;
    }

void appDocSetScrollbarValues(	EditDocument *	ed )
    {
    AppDrawingData *	add= &(ed->edDrawingData);

    int			sliderSize;
    int			maximum;
    int			minimum;
    int			value;

    sliderSize= ed->edVisibleRect.drY1- ed->edVisibleRect.drY0+ 1;
    minimum= add->addBackRect.drY0;
    value= ed->edVisibleRect.drY0;

    if  ( sliderSize >= add->addBackRect.drY1+ 1 )
	{ maximum= ed->edVisibleRect.drY1+ 1;	}
    else{ maximum= add->addBackRect.drY1+ 1;	}

    if  ( value+ sliderSize > maximum )
	{ value= maximum- sliderSize; }

    XtVaSetValues( ed->edVerticalScrollbar,
			XmNminimum,		minimum,
			XmNmaximum,		maximum,
			XmNvalue,		value,
			XmNsliderSize,		sliderSize,
			XmNpageIncrement,	( 9* sliderSize+ 9 )/10,
			XmNincrement,		( sliderSize+ 19 )/ 20,
			NULL );

    sliderSize= ed->edVisibleRect.drX1- ed->edVisibleRect.drX0+ 1;
    minimum= add->addBackRect.drX0;
    value= ed->edVisibleRect.drX0;

    if  ( sliderSize >= add->addBackRect.drX1+ 1 )
	{ maximum= ed->edVisibleRect.drX1+ 1;	}
    else{ maximum= add->addBackRect.drX1+ 1;	}

    XtVaSetValues( ed->edHorizontalScrollbar,
			XmNminimum,	minimum,
			XmNmaximum,	maximum,
			XmNvalue,	value,
			XmNsliderSize,	sliderSize,
			NULL );

    return;
    }

int appSetupDocument(	EditApplication *	ea,
			EditDocument *		ed	)
    {
    AppDrawingData *		add= &(ed->edDrawingData);

    double			horPixPerMM;
    double			verPixPerMM;
    double			xfac;
    double			yfac;

    appGetFactors( ea, &horPixPerMM, &verPixPerMM, &xfac, &yfac );

    ed->edPixelsPerMM= horPixPerMM;

    appSetDrawingEnvironment( add, ea->eaMagnification, xfac,
			    ea->eaAfmDirectory, ea->eaDisplay, ea->eaScreen );

    if  ( (*ea->eaSetupDocument)( ea, ed ) )
	{ SDEB(ed->edFilename); return -1; }

    appDocSetScrollbarValues( ed );

    XtRealizeWidget( ed->edTopWidget );

    add->addDrawable= XtWindow( ed->edDocumentWidget );

    if  ( ea->eaFinishDocumentSetup && (*ea->eaFinishDocumentSetup)( ea, ed ) )
	{ SDEB(ed->edFilename); return -1; }

    XtMapWidget( ed->edTopWidget );

    ed->edIsVisible= 1;
    ed->edHasBeenChanged= 0;

    if  ( ! ea->eaCurrentDocMenuPixel )
	{
	Display *	display= ea->eaDisplay;
	int		screen= ea->eaScreen;
	Colormap	colorMap= DefaultColormap( display, screen );
	const char *	focusColor= "CadetBlue";

	XColor		color;
	XColor		ignored;

	XtVaGetValues( ed->edMenuBar,
			    XmNbackground,	&ea->eaOtherDocMenuPixel,
			    NULL );

	/* mwm:		CadetBlue:	r=96,g=160,b=160 */
	/* fvwm:	#c06077		r=192,g=96,b=119 */

	if  ( ! XAllocNamedColor( display, colorMap, focusColor,
						    &color, &ignored ) )
	    {
	    SDEB(focusColor);
	    ea->eaCurrentDocMenuPixel= ea->eaOtherDocMenuPixel;
	    }
	else{ ea->eaCurrentDocMenuPixel= color.pixel; }
	}

    if  ( ea->eaDocumentCursor )
	{
	XDefineCursor( ea->eaDisplay, XtWindow( ed->edDocumentWidget ),
						    ea->eaDocumentCursor );
	}

    appDocVisible( ea, ed, ed->edIsVisible );

    return 0;
    }

/************************************************************************/
/*									*/
/*  Open a a document.							*/
/*									*/
/*  1)  Allocate a document.						*/
/*  3)  Make a shell to contain everything.				*/
/*  4)  Make the window pane.						*/
/*									*/
/************************************************************************/
EditDocument * appOpenDocument(	EditApplication *	ea,
				Widget			relative,
				Widget			option,
				int			readonly,
				const char *		filename	)
    {
    EditDocument *		ed;
    int				fileReadonly= 0;
    AppFileMessageResources *	afmr= &(ea->eaFileMessageResources);

    if  ( readonly )
	{ fileReadonly= 1;	}
    else{
	if  ( appTestFileReadable( filename ) )
	    {
	    int	resp;

	    resp= appQuestionRunSubjectOkCancelDialog( ea,
					relative, option,
					filename, afmr->afmrFileReadOnly,
					(char *)0, (char *)0 );

	    if  ( resp != AQDrespOK )
		{ return (EditDocument *)0;	}

	    fileReadonly= 1;
	    }
	}

    /*  1,3,4  */
    if  ( appMakeDocumentWindow( &ed, ea, readonly, filename, filename ) )
	{ SDEB(filename); return (EditDocument *)0;	}

    ed->edFileReadOnly= fileReadonly;

    if  ( (*ea->eaOpenDocument)( ea, ed, relative, option, filename ) )
	{ return (EditDocument *)0; }

    if  ( appSetupDocument( ea, ed ) )
	{ SDEB(filename); return (EditDocument *)0; }

    appSetDocument( ea, ed, filename );

    return ed;
    }

int appNewDocument(	EditApplication *	ea,
			const char *		filename )
    {
    const int			readonly= 0;
    const char *		title= filename;

    EditDocument *		ed;

    /*  1,3,4  */
    if  ( appMakeDocumentWindow( &ed, ea, readonly, title, filename ) )
	{ SDEB(filename); return -1;	}

    if  ( (*ea->eaNewDocument)( ea, ed, filename ) )
	{ SDEB(title); return -1; }

    if  ( appSetupDocument( ea, ed ) )
	{ SDEB(title); return -1; }

    appSetDocument( ea, ed, title );

    return 0;
    }

/************************************************************************/
/*									*/
/*  Calculate magnification factors.					*/
/*									*/
/************************************************************************/
void appGetFactors(	EditApplication *	ea,
			double *		pHorPixPerMM,
			double *		pVerPixPerMM,
			double *		pXfac,
			double *		pYfac )
    {
    Display *		display= ea->eaDisplay;
    int			screen= ea->eaScreen;

    double		horPixPerMM;
    double		verPixPerMM;
    double		xfac;
    double		yfac;

    horPixPerMM= ( (double)DisplayWidth( display, screen ) )/
					DisplayWidthMM( display, screen );
    verPixPerMM= ( (double)DisplayHeight( display, screen ) )/
					DisplayHeightMM( display, screen );

    xfac=  ( 25.4/ ( 20.0* 72.0 ) )* horPixPerMM;
    yfac=  ( 25.4/ ( 20.0* 72.0 ) )* verPixPerMM;

    *pHorPixPerMM= horPixPerMM;
    *pVerPixPerMM= verPixPerMM;

    *pXfac= ea->eaMagnification* xfac;
    *pYfac= ea->eaMagnification* yfac;

    return;
    }


/************************************************************************/
/*									*/
/*  Finish the setup of the document after it is realised.		*/
/*									*/
/************************************************************************/

static XIMStyle	appXimPreferences[]=
    {
    XIMPreeditCallbacks,
    XIMPreeditPosition,
    XIMPreeditArea,
    XIMPreeditNothing,
    XIMStatusCallbacks,
    XIMStatusArea,
    XIMStatusNothing,
    };

int appFinishDocumentSetup(	EditApplication *	ea,
				EditDocument *		ed )
    {
    ed->edGc= XCreateGC( ea->eaDisplay, XtWindow( ed->edDocumentWidget ),
								    0, NULL );

    if  ( ea->eaInputMethod )
	{
	XIMStyles *	availableStyles;
	XIMStyle	supportedStyles;
	XIMStyle	likedStyle;
	int		style;

	XGetIMValues( ea->eaInputMethod,
		XNQueryInputStyle, &availableStyles,
		NULL );

	supportedStyles=	XIMPreeditArea		|
				XIMPreeditNothing	|
				XIMPreeditNone		|
				XIMStatusArea		|
				XIMStatusNothing	|
				XIMStatusNone		;

	likedStyle= 0;
	for ( style= 0; style < availableStyles->count_styles; style++ )
	    {
	    XIMStyle	foundStyle= availableStyles->supported_styles[style];
	    unsigned	j;

	    if  ( ( supportedStyles & foundStyle ) != foundStyle )
		{ continue;	}

	    if  ( ! likedStyle )
		{ likedStyle= foundStyle; continue;	}

	    for ( j= 0; j < sizeof(appXimPreferences)/sizeof(XIMStyle); j++ )
		{
		if  (   ( foundStyle & appXimPreferences[j] )	&&
		      ! ( likedStyle & appXimPreferences[j] )	)
		    { likedStyle= foundStyle; break;	}
		}
	    }

	if  ( likedStyle )
	    {
	    ed->edInputContext= XCreateIC( ea->eaInputMethod,
		XNClientWindow,		XtWindow( ed->edDocumentWidget ),
		XNInputStyle,		likedStyle,
		/*
		XNPeeditAttributes,	attributeList,
		XNStatusAttributes,	attributeList,
		*/
		NULL );
	    }
	else{ ed->edInputContext= (XIC)0;	}
	}
    else{ ed->edInputContext= (XIC)0;	}

    return 0;
    }
