#   ifndef	APP_DRAW_H
#   define	APP_DRAW_H

#   include	<X11/Intrinsic.h>
#   include	<appGeo.h>
#   include	<docFont.h>
#   include	<psFont.h>

/************************************************************************/
/*									*/
/*  An attempt to consistently use the same formula to convert from	*/
/*  twips to pixels.							*/
/*									*/
/************************************************************************/

#   define	TWIPStoPIXELS(xmag,t) ((xmag)*(t))
#   define	PIXELStoTWIPS(p,xmag) ((p)/(xmag))

/************************************************************************/
/*									*/
/*  For the administration of the relation between PostScript fonts and	*/
/*  X11 fonts.								*/
/*									*/
/************************************************************************/

typedef struct AppPhysicalFont
    {
    TextAttribute	apfAttribute;
    int			apfDocFamilyNumber;

    int			apfPsFamilyNumber;
    int			apfPsFaceNumber;

    XFontStruct *	apfFontStruct;
    XFontSet		apfFontSet;
    unsigned long	apfUnderLinePositionPixels;
    unsigned long	apfUnderlineThicknessPixels;
    int			apfAscentPixels;
    int			apfDescentPixels;
    int			apfFullSizePixels;
    } AppPhysicalFont;

typedef struct AppPhysicalFontList
    {
    const char *	apflAfmDirectory;
    int			apflCount;
    AppPhysicalFont *	apflFonts;
    } AppPhysicalFontList;

typedef struct AppDrawingData
    {
    DocumentRectangle	addBackRect;
    DocumentRectangle	addPaperRect;
    DocumentRectangle	addDocRect;

    int			addBottomMarginPixels;
    int			addPageGapPixels;
    int			addPageStepPixels;

    double		addMagnification;
    double		addMagnifiedPixelsPerTwip;

    Display *		addDisplay;
    Drawable		addDrawable;
    int			addScreen;
    AppPhysicalFontList	addPhysicalFontList;
    } AppDrawingData;

typedef struct VerticalCharacter
    {
    Pixmap		vcPixmap;
    XCharStruct		vcCharStruct;
    } VerticalCharacter;

typedef struct VerticalXFont
    {
    XVisualInfo		vxfVisualInfo;
    GC			vxfScratchBitmapGc;
    GC			vxfPixmapGc;
    XFontStruct *	vxfFontStruct;
    Pixmap		vxfScratchBitmap;
    Pixmap		vxfStampPixmap;
    int			vxfDirection;
    VerticalCharacter	vxfCharacters[256];
    } VerticalXFont;

/************************************************************************/
/*									*/
/*  Procedure declarations.						*/
/*									*/
/************************************************************************/

extern void appCollectExposures(	DocumentRectangle *	drClip,
					int			ox,
					int			oy,
					Display *		display,
					Window			win,
					GC			gc,
					XEvent *		event );

extern void appInitDrawingData(		AppDrawingData *	add );
extern void appCleanDrawingData(	AppDrawingData *	add );


extern void appInitFontList(	AppPhysicalFontList *	apfl );
extern void appCleanFontList(	Display *		display,
				AppPhysicalFontList *	apfl );

extern void appEnableFontsTool(	void *	voidafc,
				int	enabled );

extern void appShowFontsTool(		void *	voidefc	);

extern int appFontCatalog(	Display *		display,
				AppFontFamily **	pFamilies,
				int *			pCount		);

extern void appFontFormatCurrent(	char *			target,
					AppFontFamily *		aff,
					AppFontTypeface *	aft,
					int			size	);

extern int appOpenDocumentFont(	AppDrawingData *		add,
				const DocumentFontList *	dfl,
				TextAttribute			ta );

extern int appFontSetCurrentFont(	void *		voidefc,
					double		xfac,
					int		fontFamilyNumber,
					unsigned int	updMask,
					TextAttribute	taNew );

extern void appFontAttributeString(	char *			target,
					const char *		familyName,
					unsigned int		updMask,
					TextAttribute		taNew );

extern int appGetPsFont(	int *			pFamily,
				int *			pFace,
				AppFontFamily **	pAff,
				AppFontTypeface **	pAft,
				const char *		afmDirectory,
				const DocumentFont *	df,
				TextAttribute		ta );

extern int appFontXFont(	char *			target,
				Display *		display,
				double			magnifiedPixelsPerTwip,
				const AppFontFamily *	aff,
				const AppFontTypeface *	aft,
				int			twipsSize,
				int			variant );

extern int appCharExistsInFont(	const XFontStruct *	xfs,
				int			ch );

extern int appFontGetAttributes(	char *		attributeString,
					char **		pFamilyName,
					unsigned int *	pUpdMask,
					TextAttribute *	pTaNew );

extern void appInitVerticalXFont(	VerticalXFont *	vxf );

extern void appCleanVerticalXFont(	VerticalXFont *	vxf,
					Display *	display );

extern int appOpenVerticalXFont(	VerticalXFont *	vxf,
					const char *	name,
					int		direction,
					Display *	display,
					Drawable	win,
					Pixel		background );

extern int appDrawVerticalString(	VerticalXFont *	vxf,
					Display *	display,
					Window		win,
					GC		gc,
					int		x,
					int		y,
					char *		string,
					int		length );

extern void appSetDrawingEnvironment(	AppDrawingData *	add,
					double			magnification,
					double			xfac,
					const char *		afmDirectory,
					Display *		display,
					int			screen );

#   endif							/*  }}  */
