/************************************************************************/
/*  A Page Layout tool.							*/
/************************************************************************/

#   include	"config.h"

#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<stdlib.h>
#   include	<debugon.h>

#   include	<X11/Xatom.h>
#   include	<X11/IntrinsicP.h>

#   include	<appFrame.h>
#   include	"appUnit.h"
#   include	<appGeoString.h>
#   include	<appPs.h>

#   include	"appUtil.h"

/************************************************************************/
/*									*/
/*  Draw a text like pattern between some margins.			*/
/*									*/
/************************************************************************/

static void appDrawTextPattern(	Display *			display,
				Window				win,
				GC				gc,
				int				numX1,
				int				numY1,
				int				textX0,
				int				textX1,
				int				textY0,
				int				textY1 )
    {
    int			y;

    y= textY0+ 2;
    while( y < textY1- 1 )
	{
	int	x;
	int	c;

	if  ( y <= numY1 )
	    { x= numX1+ 1;	}
	else{ x= textX0+ 1;	}

	c= (int)( 5+ (75.0* rand() )/ (RAND_MAX+ 1.0 ) );

	while( c > 0 )
	    {
	    int	l;

	    l= (int) ( (1.0* (textX1- textX0)* rand() )/ (RAND_MAX+ 1.0 ) );
	    l= 2+ l/10;

	    if  ( x+ l >= textX1- 1 )
		{
		y += 2;

		if  ( y <= numY1 )
		    { x= numX1+ 1;	}
		else{ x= textX0+ 1;	}

		if  ( y >= textY1- 1 )
		    { break;	}
		}

	    XDrawRectangle( display, win, gc, x, y, l, 0 );

	    x += l+ 2; c--;
	    }

	y += 4;
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Draw the margins of a page.						*/
/*									*/
/************************************************************************/

static void appDrawTextPage(	Display *			display,
				Window				win,
				GC				gc,
				Pixel				marginPixel,
				Pixel				textPixel,
				int				textX0,
				int				textX1,
				int				textY0,
				int				textY1 )
    {
    XSetForeground( display, gc, marginPixel );
    XDrawRectangle( display, win, gc,
	    textX0- 1, textY0- 1, textX1- textX0+ 1, textY1- textY0+ 1 );

    XSetForeground( display, gc, textPixel );
    appDrawTextPattern( display, win, gc,
				    textX0, textY0,
				    textX0, textX1, textY0, textY1 );

    return;
    }

static void appDrawNupTextPage(	Display *			display,
				Window				win,
				GC				gc,
				XFontStruct *			xfs,
				Pixel				marginPixel,
				Pixel				textPixel,
				const char *			label,
				int				textX0,
				int				textX1,
				int				textY0,
				int				textY1 )
    {
    if  ( xfs )
	{
	int				numWide;
	int				numHigh;

	int				l= strlen( label );

	numWide= XTextWidth( xfs, label, l );
	numHigh= xfs->ascent+ xfs->descent;

	XSetForeground( display, gc, marginPixel );
	XDrawRectangle( display, win, gc,
		textX0- 1, textY0- 1, textX1- textX0+ 1, textY1- textY0+ 1 );

	XSetForeground( display, gc, textPixel );

	XDrawString( display, win, gc, textX0+ 2, textY0+ xfs->ascent+ 3,
								label, l );

	appDrawTextPattern( display, win, gc,
				    textX0+ numWide+ 2, textY0+ numHigh+ 2,
				    textX0, textX1, textY0, textY1 );
	}
    else{
	appDrawTextPage( display, win, gc, marginPixel, textPixel,
					    textX0, textX1, textY0, textY1 );

	}
    }

/************************************************************************/
/*									*/
/*  Draw the plane of the sheet of paper.				*/
/*									*/
/************************************************************************/

static void appDrawPageSheet(	int				sheetX0,
				int				sheetX1,
				int				sheetY0,
				int				sheetY1,
				Display *			display,
				Window				win,
				GC				gc )
    {
    int				screen= DefaultScreen( display );

    XSetForeground( display, gc, WhitePixel( display, screen ) );
    XFillRectangle( display, win, gc, sheetX0, sheetY0,
					sheetX1- sheetX0, sheetY1- sheetY0 );

    XSetForeground( display, gc, BlackPixel( display, screen ) );
    XDrawRectangle( display, win, gc, sheetX0- 1, sheetY0- 1,
					sheetX1- sheetX0, sheetY1- sheetY0 );

    return;
    }

/************************************************************************/
/*									*/
/*  Draw a schematic view of a page.					*/
/*									*/
/*  1)  Perform calculations relative to the centre, to avoid assymetry	*/
/*	from rounding when margins are identical.			*/
/*									*/
/************************************************************************/

void appDrawPageDiagram(	Widget				w,
				GC				gc,
				double				widgetHighCm,
				const DocumentGeometry *	dg )
    {
    Display *			display= XtDisplay( w );
    Window			win= XtWindow( w );

    Dimension			wide;
    Dimension			high;
    double			fac;

    int				sheetX0;
    int				sheetX1;
    int				sheetY0;
    int				sheetY1;

    int				pageWidePixels;
    int				pageHighPixels;

    int				textX0;
    int				textX1;
    int				textY0;
    int				textY1;

    Pixel			backPixel;

    XtVaGetValues( w,
			XmNwidth,	&wide,
			XmNheight,	&high,
			XmNbackground,	&backPixel,
			NULL );

    XSetForeground( display, gc, backPixel );
    XFillRectangle( display, win, gc, 0, 0, wide, high );

    fac= ( 2.54* high )/( widgetHighCm* 10.0 * 20.0* 72.27 );

    pageWidePixels= fac* dg->dgPaperWideTwips;
    pageHighPixels= fac* dg->dgPaperHighTwips;

    sheetX0= wide/2- ( pageWidePixels/ 2 );
    sheetY0= high/2- ( pageHighPixels/ 2 );
    sheetX1= wide/2+ ( pageWidePixels/ 2 );
    sheetY1= high/2+ ( pageHighPixels/ 2 );

    appDrawPageSheet( sheetX0, sheetX1, sheetY0, sheetY1, display, win, gc );

    /*  1  */
    textX0= wide/2- fac* ( dg->dgPaperWideTwips/2- dg->dgLeftMarginTwips );
    textX1= wide/2+ fac* ( dg->dgPaperWideTwips/2- dg->dgRightMarginTwips );
    textY0= high/2- fac* ( dg->dgPaperHighTwips/2- dg->dgTopMarginTwips );
    textY1= high/2+ fac* ( dg->dgPaperHighTwips/2- dg->dgBottomMarginTwips );

    appDrawTextPage( display, win, gc,
		backPixel, BlackPixel( display, DefaultScreen( display ) ),
		textX0, textX1, textY0, textY1 );

    return;
    }

/************************************************************************/
/*									*/
/*  Draw a schematic view of the distribution of pages on a sheet of	*/
/*  paper.								*/
/*									*/
/*  1)  When no scaling or moving is required: Use appDrawPageDiagram()	*/
/*	saves work, and is more focused at looking nice.		*/
/*									*/
/************************************************************************/

void appDrawNupDiagram(		Widget				w,
				GC				gc,
				XFontStruct *			xfs,
				double				widgetHighCm,
				const DocumentGeometry *	dgSheet,
				const DocumentGeometry *	dgPage,
				int				nup,
				int				horizontal )
    {
    Display *			display= XtDisplay( w );
    Window			win= XtWindow( w );

    Dimension			wide;
    Dimension			high;
    double			fac;

    int				sheetX0;
    int				sheetX1;
    int				sheetY0;
    int				sheetY1;

    int				pageWidePixels;
    int				pageHighPixels;

    int				textX0;
    int				textX1;
    int				textY0;
    int				textY1;

    Pixel			backPixel;

    double			nupFac;
    int				rotate;
    int				yshift;

    int				textWideTwips;
    int				textHighTwips;

    /*  1  */
    if  ( nup == 1						&&
	  dgSheet->dgPaperWideTwips == dgPage->dgPaperWideTwips	&&
	  dgSheet->dgPaperHighTwips == dgPage->dgPaperHighTwips	)
	{ appDrawPageDiagram( w, gc, widgetHighCm, dgPage ); return;	}

    XtVaGetValues( w,
			XmNwidth,	&wide,
			XmNheight,	&high,
			XmNbackground,	&backPixel,
			NULL );

    XSetForeground( display, gc, backPixel );
    XFillRectangle( display, win, gc, 0, 0, wide, high );

    fac= ( 2.54* high )/( widgetHighCm* 10.0 * 20.0* 72.27 );

    pageWidePixels= fac* dgSheet->dgPaperWideTwips;
    pageHighPixels= fac* dgSheet->dgPaperHighTwips;

    if  ( appPsGetNupFactor( &nupFac, &rotate, &yshift,
					dgPage, dgSheet, nup, horizontal ) )
	{ LDEB(nup); return;	}

    if  ( rotate )
	{
	sheetX0= wide/2- ( pageHighPixels/ 2 );
	sheetY0= high/2- ( pageWidePixels/ 2 );
	sheetX1= wide/2+ ( pageHighPixels/ 2 );
	sheetY1= high/2+ ( pageWidePixels/ 2 );
	}
    else{
	sheetX0= wide/2- ( pageWidePixels/ 2 );
	sheetY0= high/2- ( pageHighPixels/ 2 );
	sheetX1= wide/2+ ( pageWidePixels/ 2 );
	sheetY1= high/2+ ( pageHighPixels/ 2 );
	}

    appDrawPageSheet( sheetX0, sheetX1, sheetY0, sheetY1, display, win, gc );

    textWideTwips= dgPage->dgPaperWideTwips-
				dgPage->dgLeftMarginTwips-
				dgPage->dgRightMarginTwips;
    textHighTwips= dgPage->dgPaperHighTwips-
				dgPage->dgTopMarginTwips-
				dgPage->dgBottomMarginTwips;

    if  ( nup == 1 )
	{
	textX0= sheetX0+ fac* nupFac* dgPage->dgLeftMarginTwips;
	textY0= sheetY0+ fac* nupFac* dgPage->dgTopMarginTwips;

	textX1= textX0+ fac* nupFac* textWideTwips;
	textY1= textY0+ fac* nupFac* textHighTwips;

	appDrawTextPage( display, win, gc,
		backPixel, BlackPixel( display, DefaultScreen( display ) ),
		textX0, textX1, textY0, textY1 );

	return;
	}

    if  ( nup == 2 )
	{
	/*  p 1 */
	textX0= sheetX0+ fac* nupFac* dgPage->dgLeftMarginTwips;
	textY0= sheetY0+ fac* nupFac* dgPage->dgTopMarginTwips;

	textX1= textX0+ fac* nupFac* textWideTwips;
	textY1= textY0+ fac* nupFac* textHighTwips;

	appDrawNupTextPage( display, win, gc, xfs,
		backPixel, BlackPixel( display, DefaultScreen( display ) ),
		"1", textX0, textX1, textY0, textY1 );

	/*  p 2 */
	textX0= sheetX0+ fac* nupFac* dgPage->dgLeftMarginTwips;
	textY0= sheetY0+ fac* nupFac* dgPage->dgTopMarginTwips;

	if  ( rotate )
	    { textX0 += pageHighPixels/ 2;	}
	else{ textY0 += pageHighPixels/ 2;	}

	textX1= textX0+ fac* nupFac* textWideTwips;
	textY1= textY0+ fac* nupFac* textHighTwips;

	appDrawNupTextPage( display, win, gc, xfs,
		backPixel, BlackPixel( display, DefaultScreen( display ) ),
		"2", textX0, textX1, textY0, textY1 );

	return;
	}

    if  ( nup == 4 )
	{
	/*  p 1 */
	textX0= sheetX0+ fac* nupFac* dgPage->dgLeftMarginTwips;
	textY0= sheetY0+ fac* nupFac* dgPage->dgTopMarginTwips;

	textX1= textX0+ fac* nupFac* textWideTwips;
	textY1= textY0+ fac* nupFac* textHighTwips;

	appDrawNupTextPage( display, win, gc, xfs,
		backPixel, BlackPixel( display, DefaultScreen( display ) ),
		"1", textX0, textX1, textY0, textY1 );

	/*  p 2 */
	textX0= sheetX0+ fac* nupFac* dgPage->dgLeftMarginTwips;
	textY0= sheetY0+ fac* nupFac* dgPage->dgTopMarginTwips;

	if  ( rotate )
	    {
	    if  ( horizontal )
		{ textX0 += pageHighPixels/ 2;	}
	    else{ textY0 += pageWidePixels/ 2;	}
	    }
	else{
	    if  ( horizontal )
		{ textX0 += pageWidePixels/ 2;	}
	    else{ textY0 += pageHighPixels/ 2;	}
	    }

	textX1= textX0+ fac* nupFac* textWideTwips;
	textY1= textY0+ fac* nupFac* textHighTwips;

	appDrawNupTextPage( display, win, gc, xfs,
		backPixel, BlackPixel( display, DefaultScreen( display ) ),
		"2", textX0, textX1, textY0, textY1 );

	/*  p 3 */
	textX0= sheetX0+ fac* nupFac* dgPage->dgLeftMarginTwips;
	textY0= sheetY0+ fac* nupFac* dgPage->dgTopMarginTwips;

	if  ( rotate )
	    {
	    if  ( horizontal )
		{ textY0 += pageWidePixels/ 2;	}
	    else{ textX0 += pageHighPixels/ 2;	}
	    }
	else{
	    if  ( horizontal )
		{ textY0 += pageHighPixels/ 2;	}
	    else{ textX0 += pageWidePixels/ 2;	}
	    }

	textX1= textX0+ fac* nupFac* textWideTwips;
	textY1= textY0+ fac* nupFac* textHighTwips;

	appDrawNupTextPage( display, win, gc, xfs,
		backPixel, BlackPixel( display, DefaultScreen( display ) ),
		"3", textX0, textX1, textY0, textY1 );

	/*  p 4 */
	textX0= sheetX0+ fac* nupFac* dgPage->dgLeftMarginTwips;
	textY0= sheetY0+ fac* nupFac* dgPage->dgTopMarginTwips;

	if  ( rotate )
	    {
	    textX0 += pageHighPixels/ 2;
	    textY0 += pageWidePixels/ 2;
	    }
	else{
	    textX0 += pageWidePixels/ 2;
	    textY0 += pageHighPixels/ 2;
	    }

	textX1= textX0+ fac* nupFac* textWideTwips;
	textY1= textY0+ fac* nupFac* textHighTwips;

	appDrawNupTextPage( display, win, gc, xfs,
		backPixel, BlackPixel( display, DefaultScreen( display ) ),
		"4", textX0, textX1, textY0, textY1 );

	return;
	}

    LDEB(nup); return;
    }

