/************************************************************************/
/*									*/
/*  A Format Tool.							*/
/*									*/
/************************************************************************/

#   include	"config.h"

#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<debugon.h>

#   include	<X11/Xatom.h>
#   include	<X11/IntrinsicP.h>
#   include	<X11/keysym.h>

#   include	<Xm/Frame.h>
#   include	<Xm/PanedW.h>
#   include	<Xm/Form.h>

#   include	<Xm/MwmUtil.h>
#   include	<Xm/Protocols.h>

#   include	<appFrame.h>
#   include	<appUtil.h>

/************************************************************************/
/*									*/
/*  Pull a particular subject of the inspector to front.		*/
/*									*/
/************************************************************************/

void appInspectorSelectSubject(	AppInspector *	ai,
				int			andMenu,
				int			subject )
    {
    InspectorSubject *	is;

    if  ( ai->aiCurrentSubject == subject )
	{ return;	}

    if  ( ai->aiCurrentSubject >= 0 )
	{
	is= ai->aiSubjects+ ai->aiCurrentSubject;

	XtVaSetValues( is->isPageFrame,
				XmNmappedWhenManaged,	False,
				NULL );
	}

    if  ( subject >= 0 )
	{
	is= ai->aiSubjects+ subject;

	XtVaSetValues( is->isPageFrame,
				XmNmappedWhenManaged,	True,
				NULL );

	if  ( andMenu )
	    {
	    XtVaSetValues( ai->aiSubjectMenu,
			    XmNmenuHistory,		is->isItem,
			    NULL );
	    }
	}
    else{
	if  ( andMenu )
	    {
	    XtVaSetValues( ai->aiSubjectMenu,
			    XmNmenuHistory,		(Widget)0,
			    NULL );
	    }
	}

    ai->aiCurrentSubject= subject;

    return;
    }

static void appInspectorSubjectChosen(	Widget		w,
					XtPointer	voidaiw,
					XtPointer	voidtbcs )
    {
    AppInspector *		ai= (AppInspector *)voidaiw;

    InspectorSubject *			is;
    int					subject;
    const int				andMenu= 0;

    is= ai->aiSubjects;
    for ( subject= 0; subject < ai->aiSubjectCount; is++, subject++ )
	{
	if  ( is->isItem == w )
	    { appInspectorSelectSubject( ai, andMenu, subject ); return; }
	}

    XDEB(w); return;
    }

/************************************************************************/
/*									*/
/*  An inspector tool must be destroyed.				*/
/*									*/
/************************************************************************/

static void appDestroyInspector(	AppInspector *	ai )
    {
    if  ( ai->aiDestroy )
	{ (*ai->aiDestroy)( ai->aiTarget );	}

    XtDestroyWidget( ai->aiTopWidget );

    free( ai );

    return;
    }

static void appCloseInspector(		Widget		w,
					XtPointer	voidast,
					XtPointer	voidlcs	 )
    {
    AppInspector *	ast= (AppInspector *)voidast;

    appDestroyInspector( ast );

    return;
    }

void appInspectorCancel(		Widget			w,
					void *			voidast,
					XEvent *		event,
					Boolean *		pRefused )
    {
    if  ( event->type == KeyRelease )
	{
	XKeyEvent *		kevent= &(event->xkey);
	AppInspector *		ast= (AppInspector *)voidast;

	if  ( XKeysymToKeycode( XtDisplay(w), XK_Escape ) == kevent->keycode )
	    { appDestroyInspector( ast ); *pRefused= 0; return; }
	}

    *pRefused= 1; return;
    }

/************************************************************************/
/*									*/
/*  Make the form with the two buttons.					*/
/*									*/
/************************************************************************/

void appInspectorMakeButtonRow(	Widget *		pRow,
				Widget			parent,
				Widget *		pLeftButton,
				Widget *		pRightButton,
				char *			leftLabel,
				char *			rightLabel,
				XtCallbackProc		leftCallback,
				XtCallbackProc		rightCallback,
				void *			through )
    {
    Widget	row= appMakeButtonRow( parent, 2 );

    Widget	leftButton;
    Widget	rightButton;

    leftButton= appMakeRowButton( row, leftLabel, leftCallback, through,
								    0, False );
    rightButton= appMakeRowButton( row, rightLabel,rightCallback, through,
								    1, False );

    XtManageChild( leftButton );
    XtManageChild( rightButton );
    XtManageChild( row );

    *pRow= row;
    *pLeftButton= leftButton;
    *pRightButton= rightButton;

    return;
    }

void appInspectorMakeToggleRow(		Widget *		pRow,
					Widget			parent,
					Widget *		pLeftToggle,
					Widget *		pRightToggle,
					char *			leftText,
					char *			rightText,
					XtCallbackProc		leftCallback,
					XtCallbackProc		rightCallback,
					void *			through )
    {
    Widget	row= appMakeButtonRow( parent, 2 );

    Widget	leftToggle;
    Widget	rightToggle;


    leftToggle= appMakeRowToggle( row, leftText, leftCallback, through, 0 );
    rightToggle= appMakeRowToggle( row, rightText, rightCallback, through, 1 );

    XtManageChild( leftToggle );
    XtManageChild( rightToggle );
    XtManageChild( row );

    *pRow= row;
    *pLeftToggle= leftToggle;
    *pRightToggle= rightToggle;

    return;
    }

void appInspectorMakeMenuRow(	Widget *		pRow,
				Widget *		pMenu,
				Widget *		pPulldown,
				Widget *		pLabel,
				Widget			parent,
				const char *		labelText )
    {
    Arg		al[20];
    int		ac= 0;

    Widget	row;
    Widget	menuForm;
    Widget	pulldown;
    Widget	menu;
    Widget	label;

    row= appMakeButtonRow( parent, 2 );

    /**************/
    appMakeRowLabel( &label, row, labelText );

    /**************/
    ac= 0;
    XtSetArg( al[ac], XmNtopAttachment,		XmATTACH_FORM ); ac++;

    XtSetArg( al[ac], XmNleftAttachment,	XmATTACH_POSITION ); ac++;
    XtSetArg( al[ac], XmNleftPosition,		1 ); ac++;

    XtSetArg( al[ac], XmNrightAttachment,	XmATTACH_POSITION ); ac++;
    XtSetArg( al[ac], XmNrightPosition,		2 ); ac++;

    menuForm= XmCreateForm( row, WIDGET_NAME, al, ac );

    appMakePulldownList( &pulldown, &menu, menuForm );

    XtVaSetValues( pulldown,
			XmNresizeWidth,		True,
			NULL );

    XtVaSetValues( menu,
			XmNleftOffset,		0,
			XmNrightOffset,		3,
			NULL );

    XtManageChild( menu );
    XtManageChild( menuForm );
    XtManageChild( row );

    *pRow= row;
    *pMenu= menu;
    *pPulldown= pulldown;
    *pLabel= label;

    return;
    }

/************************************************************************/
/*									*/
/*  Make the menu for the pages.					*/
/*									*/
/************************************************************************/

static Widget appInspectorMakePageMenu(	Widget		parent,
					AppInspector *	ai )
    {
    Widget	menuForm;

    Arg		al[20];
    int		ac;

    ac= 0;
    XtSetArg( al[ac],	XmNallowResize,		True ); ac++;
    XtSetArg( al[ac],	XmNskipAdjust,		True ); ac++;
    menuForm= XmCreateForm( parent, WIDGET_NAME, al, ac );

    appMakePulldownList( &ai->aiSubjectPulldown, &ai->aiSubjectMenu,
								menuForm );

    XtManageChild( ai->aiSubjectMenu );
    XtManageChild( menuForm );

    return menuForm;
    }

/************************************************************************/
/*									*/
/*  Make an inspector subject page.					*/
/*									*/
/************************************************************************/

static void appInspectorMakePage(	InspectorSubject *	is,
					Widget			pageParent )
    {
    Arg		al[20];
    int		ac= 0;

    XtSetArg( al[ac],	XmNleftAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNrightAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNtopAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNbottomAttachment,	XmATTACH_FORM ); ac++;

    XtSetArg( al[ac],	XmNleftOffset,		5 ); ac++;
    XtSetArg( al[ac],	XmNrightOffset,		5 ); ac++;
    XtSetArg( al[ac],	XmNtopOffset,		5 ); ac++;
    XtSetArg( al[ac],	XmNbottomOffset,	5 ); ac++;

    XtSetArg( al[ac],	XmNmappedWhenManaged,	False ); ac++;

    is->isPageFrame= XmCreateFrame( pageParent, WIDGET_NAME, al, ac );

    /**********************/

    ac= 0;
    XtSetArg( al[ac],	XmNsashWidth,		1 ); ac++;
    XtSetArg( al[ac],	XmNsashHeight,		1 ); ac++;
    XtSetArg( al[ac],	XmNseparatorOn,		False ); ac++;
    XtSetArg( al[ac],	XmNmarginWidth,		5 ); ac++;
    XtSetArg( al[ac],	XmNmarginHeight,	5 ); ac++;
    XtSetArg( al[ac],	XmNspacing,		5 ); ac++;

    is->isPagePaned= XmCreatePanedWindow( is->isPageFrame,
						    WIDGET_NAME, al, ac );
    return;
    }

/************************************************************************/
/*									*/
/*  Make the pages,							*/
/*  Fill the menu for the pages.					*/
/*									*/
/************************************************************************/

static void appInspectorFillSubjects(	Widget				parent,
					InspectorSubjectResources *	isr,
					AppInspector *			ai )
    {
    Dimension			width;
    Widget			first= (Widget)0;

    int				subject;
    InspectorSubject *		is;

    XtVaGetValues( ai->aiSubjectMenu,
			XmNwidth,		&width,
			NULL );

    appEmptyPulldownList( ai->aiSubjectPulldown );

    is= ai->aiSubjects;
    for ( subject= 0; subject < ai->aiSubjectCount; is++, isr++, subject++ )
	{
	appInspectorMakePage( is, parent );

	is->isItem= appPulldownMakeOption( ai->aiSubjectPulldown,
				    isr->isrSubjectName, width,
				    appInspectorSubjectChosen, (void *)ai );
	if  ( ! first )
	    { first= is->isItem;	}
	}

    if 	( first )
	{
	XtVaSetValues( ai->aiSubjectMenu,
			    XmNmenuHistory,		first,
			    NULL );
	}

    appPulldownSetWidth( ai->aiSubjectMenu, width );
    }

AppInspector * appMakeInspector(    EditApplication *		ea,
				    Widget			option,
				    const char *		pixmapName,
				    const char *		widgetName,
				    InspectorSubjectResources * isr,
				    int				subjectCount,
				    AppToolDestroy		destroy,
				    void *			through )
    {
    AppInspector *	ai;

    Pixmap			iconPixmap;
    MwmHints			hints;

    Arg				al[20];
    int				ac= 0;

    Widget			menuForm;

    if  ( appGetImagePixmap( ea, pixmapName, &iconPixmap )  )
	{ SDEB(pixmapName); return (AppInspector *)0; }

    ai= (AppInspector *)malloc( sizeof(AppInspector)+ 
				subjectCount* sizeof(InspectorSubject) );
    if  ( ! ai )
	{ LXDEB(subjectCount,ai); return ai;	}

    ai->aiCurrentSubject= -1;
    ai->aiSubjectCount= subjectCount;
    ai->aiDestroy= destroy;
    ai->aiTarget= through;

    hints.flags= MWM_HINTS_FUNCTIONS|MWM_HINTS_DECORATIONS;
    hints.functions=	MWM_FUNC_MOVE		|
			MWM_FUNC_MINIMIZE	|
			MWM_FUNC_CLOSE		;
    hints.decorations=	MWM_DECOR_BORDER	|
			MWM_DECOR_TITLE		|
			MWM_DECOR_MENU		|
			MWM_DECOR_MINIMIZE	;

    XtSetArg( al[ac], XmNuserData,		(void *)ai ); ac++;
    XtSetArg( al[ac], XmNdeleteResponse,	 XmDO_NOTHING ); ac++;
    XtSetArg( al[ac], XmNinput,			True ); ac++;
    XtSetArg( al[ac], XmNallowShellResize,	True );
    XtSetArg( al[ac], XmNmwmDecorations,	hints.decorations ); ac++;
    XtSetArg( al[ac], XmNmwmFunctions,		hints.functions ); ac++;

    if  ( iconPixmap )
	{ XtSetArg( al[ac], XmNiconPixmap,	iconPixmap ); ac++; }

#   ifdef USE_X11_R5
    ai->aiTopWidget= XtAppCreateShell( ea->eaApplicationName,
					    widgetName,
					    applicationShellWidgetClass,
					    ea->eaDisplay, al, ac );
#   else
    ai->aiTopWidget= XtAppCreateShell( ea->eaApplicationName,
					    widgetName,
					    topLevelShellWidgetClass,
					    ea->eaDisplay, al, ac );
#   endif

    appSetShellTitle( ai->aiTopWidget, option, ea->eaApplicationName );

    if  ( ea->eaCloseAtom != None )
	{
	XmAddWMProtocolCallback( ai->aiTopWidget, ea->eaCloseAtom,
					appCloseInspector, (XtPointer)ai );
	}

    ac= 0;
    XtSetArg( al[ac], XmNuserData,	(void *)ai ); ac++;
    XtSetArg( al[ac], XmNallowResize,	True ); ac++;
    ai->aiMainWidget= XmCreateForm( ai->aiTopWidget, WIDGET_NAME, al, ac );

    ac= 0;
    XtSetArg( al[ac],	XmNleftAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNrightAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNtopAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNbottomAttachment,	XmATTACH_FORM ); ac++;

    XtSetArg( al[ac],	XmNsashWidth,		1 ); ac++;
    XtSetArg( al[ac],	XmNsashHeight,		1 ); ac++;
    XtSetArg( al[ac],	XmNseparatorOn,		False ); ac++;
    XtSetArg( al[ac],	XmNmarginWidth,		5 ); ac++;
    XtSetArg( al[ac],	XmNmarginHeight,	5 ); ac++;
    XtSetArg( al[ac],	XmNspacing,		5 ); ac++;
    ai->aiPaned= XmCreatePanedWindow( ai->aiMainWidget, WIDGET_NAME, al, ac );

    menuForm= appInspectorMakePageMenu( ai->aiPaned, ai );

    ac= 0;
    XtSetArg( al[ac],	XmNallowResize,		True ); ac++;
    XtSetArg( al[ac],	XmNskipAdjust,		True ); ac++;
    ai->aiPageParent= XmCreateForm( ai->aiPaned, WIDGET_NAME, al, ac );

    appInspectorFillSubjects( ai->aiPageParent, isr, ai );

    return ai;
    }

void appFinishInspector(	AppInspector *		ai )
    {
    int			subject;
    InspectorSubject *	is;

    is= ai->aiSubjects;
    for ( subject= 0; subject < ai->aiSubjectCount; is++, subject++ )
	{
	XtManageChild( is->isPagePaned );
	XtManageChild( is->isPageFrame );
	}

    XtManageChild( ai->aiPageParent );

    XtManageChild( ai->aiPaned );
    XtManageChild( ai->aiMainWidget );

    XtRealizeWidget( ai->aiTopWidget );
    XtMapWidget( ai->aiTopWidget );

    appPulldownRefreshWidth( ai->aiSubjectMenu );

    return;
    }

/************************************************************************/
/*									*/
/*  Turn an inspector or one of its subjects on or off.			*/
/*									*/
/************************************************************************/

void appEnableInspector(	void *	voidai,
				int	enabled )
    {
    AppInspector *	ai= (AppInspector *)voidai;

    XtSetSensitive( ai->aiMainWidget, enabled != 0 );

    return;
    }

void appEnableInspectorSubject(		AppInspector *		ai,
					int			subject ,
					int			enabled )
    {
    InspectorSubject *	is= ai->aiSubjects+ subject;

    XtSetSensitive( is->isPageFrame, enabled );
    XtSetSensitive( is->isItem, enabled );

    return;
    }

/************************************************************************/
/*									*/
/*  Fill an inspector menu.						*/
/*									*/
/************************************************************************/

void appFillInspectorMenu(	int			count,
				int			current,
				Widget *		items,
				char * const *		texts,
				Widget			menu,
				Widget			pulldown,
				XtCallbackProc		callBack,
				void *			target )
    {
    int			i;
    Dimension		width;

    XtVaGetValues( menu,
			XmNwidth,		&width,
			NULL );

    appEmptyPulldownList( pulldown );

    for ( i= 0; i < count; i++ )
	{
	items[i]= appPulldownMakeOption( pulldown,
				    texts[i], width, callBack, target );
	}

    XtVaSetValues( menu,
			XmNmenuHistory,	items[current],
			NULL );

    appPulldownSetWidth( menu, width );

    return;
    }
