/************************************************************************/
/*									*/
/*  Application, Main module.						*/
/*									*/
/************************************************************************/

#   include	"config.h"

#   include	<stddef.h>
#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<locale.h>

#   include	<Xm/Xm.h>
#   include	<Xm/PanedW.h>
#   include	<Xm/RowColumn.h>
#   include	<Xm/AtomMgr.h>
#   include	<Xm/ToggleB.h>
#   include	<Xm/Form.h>
#   include	<Xm/Label.h>

#   include	<Xm/Protocols.h>
#   include	<Xm/MwmUtil.h>

#   include	<appSystem.h>

#   include	<appFrame.h>
#   include	"appUnit.h"
#   include	"appPaper.h"
#   include	"appGeoString.h"
#   include	<debugon.h>

/************************************************************************/
/*									*/
/*  Hide the application window if shown as an 'about' window.		*/
/*									*/
/************************************************************************/

void appAppFileHide(	Widget		option,
			XtPointer	voidea,
			XtPointer	call_data	 )
    {
    EditApplication *		ea= (EditApplication *)voidea;

    ea->eaMainVisibleAsAbout= 0;

    if  ( ea->eaVisibleDocumentCount > 0 )
	{ XtUnmapWidget( ea->eaTopWidget ); }

    return;
    }

/************************************************************************/
/*									*/
/*  Minimise the application.						*/
/*									*/
/************************************************************************/

void appAppFileMini(	Widget		option,
			XtPointer	voidea,
			XtPointer	call_data	 )
    {
    EditApplication *		ea= (EditApplication *)voidea;
    Window			win= XtWindow( ea->eaTopWidget );

    XtVaSetValues( ea->eaTopWidget, XmNinitialState, IconicState, NULL );
    XIconifyWindow( ea->eaDisplay, win, ea->eaScreen );

    return;
    }

static void appAppWmClose(	Widget		option,
				XtPointer	voidea,
				XtPointer	call_data	 )
    {
    EditApplication *		ea= (EditApplication *)voidea;

    if  ( ea->eaMainVisibleAsAbout )
	{ appAppFileHide( option, voidea, call_data );		}
    else{ appQuitApplication( option, ea->eaTopWidget, ea );	}
    }

/************************************************************************/
/*									*/
/*  Quit the application.						*/
/*									*/
/*  1)  Close the documents in reverse order to avoid the subscript of	*/
/*	unhandeled documents to change.					*/
/*									*/
/************************************************************************/

static int appCountChangedDocuments(	EditApplication *	ea )
    {
    int		i;
    int		changedCount= 0;

    for ( i= 0; i < ea->eaOpenCount; i++ )
	{
	EditDocument *	ed= ea->eaOpenDocuments[i];

	if  ( ed->edHasBeenChanged )
	    { changedCount++;	}
	}

    return changedCount;
    }

static void appExitApplication(	EditApplication *	ea )
    {
    if  ( ea->eaDisplay )
	{ appDiscardImagePixmaps( ea );	}

    XtDestroyWidget( ea->eaTopWidget );

    exit( 0 ); LDEB(1); return;
    }

static void appAskCloseDocuments(	Widget			option,
					EditApplication *	ea )
    {
    int		i;

    /*  1  */
    for ( i= ea->eaOpenCount- 1; i >= 0; i-- )
	{
	EditDocument *	ed= ea->eaOpenDocuments[i];

	if  ( ed->edHasBeenChanged )
	    {
	    appMakeDocVisible( ea, ed );
	    appRunReallyCloseDialog( option, ed );
	    }
	}

    return;
    }

void appQuitApplication(	Widget			option,
				Widget			relative,
				EditApplication *	ea )
    {
    int				rcc;
    int				changedCount= 0;

    changedCount= appCountChangedDocuments( ea );

    if  ( changedCount == 0 )
	{ appExitApplication( ea ); LDEB(1); return;	}

    if  ( changedCount > 1 )
	{
	rcc= appRunReallyQuitDialog( option, relative, ea );

	switch( rcc )
	    {
	    case AQDrespCANCEL:
		return;
	    case AQDrespNO:
		appExitApplication( ea );
		LDEB(1); return;

	    default:
		LDEB(rcc); return;
	    case AQDrespYES:
		break;
	    }
	}

    appAskCloseDocuments( option, ea );

    changedCount= appCountChangedDocuments( ea );
    if  ( changedCount == 0 )
	{ appExitApplication( ea ); LDEB(1); return;	}

    return;
    }

void appAppFileQuit(	Widget		option,
			XtPointer	voidea,
			XtPointer	call_data	 )
    {
    EditApplication *	ea= (EditApplication *)voidea;

    appQuitApplication( option, ea->eaTopWidget, ea );
    }

/************************************************************************/
/*									*/
/*  Open a new document.						*/
/*									*/
/************************************************************************/

void appAppFileNew(	Widget		option,
			XtPointer	voidea,
			XtPointer	call_data	 )
    {
    EditApplication *		ea= (EditApplication *)voidea;
    const char *		filename= (const char *)0;

    if  ( appNewDocument( ea, filename ) )
	{ SDEB(filename); }

    return;
    }

/************************************************************************/
/*									*/
/*  Make a document the current Document.				*/
/*									*/
/************************************************************************/

static void appHighlightCurrentDocument( EditApplication *	ea,
					EditDocument *		ed,
					int			current )
    {
    WidgetList		children;
    Cardinal		childCount;
    int			j;

    Pixel		pix;

    if  ( current )
	{ pix= ea->eaCurrentDocMenuPixel;	}
    else{ pix= ea->eaOtherDocMenuPixel;		}

    XtVaSetValues( ed->edMenuBar,
			XmNbackground,	pix,
			NULL );

    XtVaGetValues( ed->edMenuBar,
			XmNchildren,		&children,
			XmNnumChildren,		&childCount,
			NULL );

    for ( j= childCount- 1; j >= 0; j-- )
	{
	XtVaSetValues( children[j],
			XmNbackground,	pix,
			NULL );
	}

    return;
    }

void appSetCurrentDocument(	EditApplication *	ea,
				EditDocument *		ed	)
    {
    if  ( ea->eaCurrentDocument )
	{ appHighlightCurrentDocument( ea, ea->eaCurrentDocument, 0 ); }

    ea->eaCurrentDocument= ed;

    appHighlightCurrentDocument( ea, ea->eaCurrentDocument, 1 );
    }

/************************************************************************/
/*  Remember whether a document has been changed.			*/
/************************************************************************/
static void appSetWindowsItemState(	Widget		menu,
					EditDocument *	ed,
					int		changed )
    {
    WidgetList		children;
    Cardinal		childCount;

    int			j;

    XtVaGetValues( menu,
			XmNchildren,		&children,
			XmNnumChildren,		&childCount,
			NULL );

    for ( j= childCount- 1; j >= 0; j-- )
	{
	void *	voided= (void *)0;

	XtVaGetValues( children[j],
			XmNuserData,	&voided,
			NULL );

	if  ( voided == (void *)ed )
	    { XmToggleButtonSetState( children[j], changed, False ); }
	}

    return;
    }

void appDocumentChanged(	EditDocument *		ed,
				int			changed	)
    {
    EditApplication *	ea= ed->edApplication;

    changed= changed != 0;

    if  ( ed->edHasBeenChanged != changed )
	{
	int		i;

	for ( i= 0; i < ea->eaOpenCount; i++ )
	    {
	    if  ( ea->eaOpenDocuments[i] == ed )
		{ continue;	}

	    appSetWindowsItemState( ea->eaOpenDocuments[i]->edWindowMenu,
								ed, changed );
	    }

	appSetWindowsItemState( ea->eaWinMenu, ed, changed );

	ed->edHasBeenChanged= changed;
	}
    }

/************************************************************************/
/*  Count the number of visible documents.				*/
/************************************************************************/
void appDocVisible(	EditApplication *	ea,
			EditDocument *		ed,
			int			visible	)
    {
    int	from;

    if  ( visible )
	{ from= ea->eaVisibleDocumentCount++;	}
    else{ from= ea->eaVisibleDocumentCount--;	}

    if  ( ea->eaVisibleDocumentCount == 0 )
	{ XtMapWidget( ea->eaTopWidget ); }
    else{
	if  ( ! ea->eaMainVisibleAsAbout )
	    { XtUnmapWidget( ea->eaTopWidget );	}
	}

    if  ( ea->eaVisibleDocumentCountChanged )
	{
	(*ea->eaVisibleDocumentCountChanged)
				    ( ea, from, ea->eaVisibleDocumentCount );
	}
    }

/************************************************************************/
/*  Remove a reference to a document.					*/
/************************************************************************/
static void appRemoveWindowsOption(	Widget			menu,
					EditDocument *		oldEd )
    {
    WidgetList		children;
    Cardinal		childCount;

    int			i;

    XtVaGetValues( menu,
			XmNchildren,		&children,
			XmNnumChildren,		&childCount,
			NULL );

    for ( i= childCount- 1; i >= 0; i-- )
	{
	void *	voided= (void *)0;

	XtVaGetValues( children[i],
			XmNuserData,	&voided,
			NULL );

	if  ( voided == (void *)oldEd )
	    { XtDestroyWidget( children[i] );	}
	}
    }

static void appRenameWindowsOption(	Widget			menu,
					EditDocument *		ed,
					XmString		titleString )
    {
    WidgetList		children;
    Cardinal		childCount;

    int			i;

    XtVaGetValues( menu,
			XmNchildren,		&children,
			XmNnumChildren,		&childCount,
			NULL );

    for ( i= childCount- 1; i >= 0; i-- )
	{
	void *	voided= (void *)0;

	XtVaGetValues( children[i],
			XmNuserData,	&voided,
			NULL );

	if  ( voided == (void *)ed )
	    {
	    XtVaSetValues( children[i],
				XmNlabelString,	titleString,
				NULL );
	    }
	}
    }

void appRemoveDocument(	EditApplication *	ea,
			EditDocument *		ed	)
    {
    int		i;
    int		count= 0;

    for ( i= 0; i < ea->eaOpenCount; i++ )
	{
	if  ( ea->eaOpenDocuments[i] == ed )
	    { continue;	}

	appRemoveWindowsOption( ea->eaOpenDocuments[i]->edWindowMenu, ed );
	ea->eaOpenDocuments[count++]= ea->eaOpenDocuments[i];
	}

    appRemoveWindowsOption( ea->eaWinMenu, ed );

    ea->eaOpenCount= count;

    if  ( ea->eaCurrentDocument == ed )
	{ ea->eaCurrentDocument= (EditDocument *)0;	}

    if  ( ed->edIsVisible )
	{
	ed->edIsVisible= 0;
	appDocVisible( ea, ed, ed->edIsVisible );
	}

    return;
    }

void appRenameDocumentOptions(	EditApplication *	ea,
				EditDocument *		ed,
				char *			title )
    {
    int		i;
    XmString	titleString;

    titleString= XmStringCreateLocalized( title );

    for ( i= 0; i < ea->eaOpenCount; i++ )
	{
	if  ( ea->eaOpenDocuments[i] == ed )
	    { continue;	}

	appRenameWindowsOption( ea->eaOpenDocuments[i]->edWindowMenu,
							    ed, titleString );
	}

    appRenameWindowsOption( ea->eaWinMenu, ed, titleString );

    XmStringFree( titleString );

    return;
    }

/************************************************************************/
/*  Remember a reference to a document.					*/
/************************************************************************/
static void appAppSetWindowsOption(	Widget			menu,
					EditDocument *		ed,
					XmString		labelString )
    {
    Widget		windowsOption;

    Arg			al[20];
    int			ac= 0;

    XtSetArg( al[ac], XmNlabelString,		labelString ); ac++;
    XtSetArg( al[ac], XmNvisibleWhenOff,	True ); ac++;
    XtSetArg( al[ac], XmNuserData,		ed ); ac++;

    windowsOption= XmCreateToggleButton( menu, WIDGET_NAME, al, ac );

    XtAddCallback( windowsOption, XmNvalueChangedCallback,
						    appDocToFront, ed );
    XtManageChild( windowsOption );

    XmToggleButtonSetState( windowsOption, ed->edHasBeenChanged, False );
    }

void appSetDocument(	EditApplication *	ea,
			EditDocument *		newEd,
			const char *		title	)
    {
    EditDocument **	fresh;
    XmString		labelString;

    int			i;

    Arg			al[20];
    int			ac= 0;

    fresh= (EditDocument **)realloc( ea->eaOpenDocuments,
			( ea->eaOpenCount+ 1 )* sizeof(EditDocument *) );
    if  ( ! fresh )
	{ XDEB(fresh); return;	}

    fresh[ea->eaOpenCount]= newEd;
    ea->eaOpenDocuments= fresh;

    labelString= XmStringCreateLocalized( (char *)title );

    XtSetArg( al[ac], XmNlabelString, labelString ); ac++;

    for ( i= 0; i < ea->eaOpenCount; i++ )
	{
	XmString	otherLabel;

	appAppSetWindowsOption( ea->eaOpenDocuments[i]->edWindowMenu,
							newEd, labelString );

	otherLabel= XmStringCreateLocalized(
				    ea->eaOpenDocuments[i]->edTitle );

	appAppSetWindowsOption( newEd->edWindowMenu,
					ea->eaOpenDocuments[i], otherLabel );

	XmStringFree( otherLabel );
	}

    appAppSetWindowsOption( ea->eaWinMenu, newEd, labelString );

    XmStringFree( labelString );

    ea->eaOpenCount++;

    return;
    }

#   ifndef USE_X11_R5

/************************************************************************/
/*									*/
/*  Session management support.						*/
/*									*/
/*  0)  Choose a short ID, otherwise xsm crashes at starup.		*/
/*  00) Do not save a discard command: Setting it here crashes xsm.	*/
/*  A)  When the shutdown was canceled, or we happen not to be allowed	*/
/*	to interact with the user, tell the session manager that we	*/
/*	have not saved our state and return the token to tell the	*/
/*	session manager that we are done.				*/
/*  1)  When the session is shut down, or if user interaction is	*/
/*	allowed later on, turn off the whole application for the	*/
/*	moment.								*/
/*	NOTE that the comment in the example on p 35 of the X11R6	*/
/*	X11R6 programmers supplement is misleading.			*/
/*  2)  When no interaction is allowed, we can only just save our	*/
/*	state.								*/
/*  3)  If there are unsaved documents tell the session manager that	*/
/*	we want to give the user her word. The callback is removed by	*/
/*	Xt.								*/
/*  z)  For one reason or another, we survived session management. Turn	*/
/*	on the application widgets again.				*/
/*									*/
/************************************************************************/

static void appSessionFreeCommand(	char **		command )
    {
    char **	cmd= command;

    while( *cmd )
	{ free( *cmd ); cmd++; }

    free( command );

    return;
    }

static char ** appSessionAllocateCommand(	int		argc,
						const char *	argv0 )
    {
    char **	rval= malloc( ( argc+ 1 )* sizeof(char *) );

    if  ( ! rval )
	{ XDEB(rval); return rval;	}

    rval[0]= strdup( argv0 );
    if  ( ! rval[0] )
	{ free( rval ); return (char **)0;	}

    rval[1]= (char *)0;

    return rval;
    }

static int appSessionSaveState(	EditApplication *	ea,
				int			saveChanged,
				int			interactive )
    {
    char **	restartCommand= (char **)0;
    char **	discardCommand= (char **)0;

    char **	restartTarget= (char **)0;
    char **	discardTarget= (char **)0;

    char	currentDirectory[1000+1];
    char	saveDirectory[1000+1];
    int		dirlen;

    char	scratch[15];
    char *	id;
    int		idlen;

    char *	sessionID= (char *)0;
    char *	smSaveDir= getenv( "SM_SAVE_DIR" );

    int		rval= 0;
    int		changedCount;
    int		restartCount;

    int		i;

    if  ( appCurrentDirectory( currentDirectory,
					sizeof(currentDirectory)- 1 ) < 0 )
	{ LDEB(1); rval= -1; goto ready;	}

    if  ( smSaveDir )
	{
	dirlen= strlen( smSaveDir );
	if  ( dirlen < 1 || dirlen >= sizeof(saveDirectory) )
	    { LDEB(dirlen); rval= -1; goto ready;	}

	strcpy( saveDirectory, smSaveDir );
	}
    else{
	dirlen= appHomeDirectory( saveDirectory,
		sizeof(saveDirectory)- 3- strlen( ea->eaApplicationName ) );
	if  ( dirlen < 0 )
	    { LDEB(dirlen); rval= -1; goto ready;	}

	strcat( saveDirectory, "/." );
	strcat( saveDirectory, ea->eaApplicationName );

	dirlen= strlen( saveDirectory );
	}

    if  ( appTestDirectory( saveDirectory )	&&
	  appMakeDirectory( saveDirectory )	)
	{ SDEB(saveDirectory); rval= -1; goto ready;	}

    XtVaGetValues( ea->eaTopWidget,
			XtNsessionID,	&sessionID,
			NULL );

    restartCommand= appSessionAllocateCommand(
				    3+ 4* ea->eaOpenCount+ 1, ea->eaArgv[0] );
    if  ( ! restartCommand )
	{ XDEB(restartCommand); rval= -1; goto ready;	}
    restartTarget= restartCommand+ 1;

    discardCommand= appSessionAllocateCommand(
				    3+ 2* ea->eaOpenCount+ 2, ea->eaArgv[0] );
    if  ( ! discardCommand )
	{ XDEB(discardCommand); rval= -1; goto ready;	}
    discardTarget= discardCommand+ 1;

    if  ( sessionID )
	{
	*restartTarget= strdup( "-xtsessionID" );
	if  ( ! *restartTarget )
	    { XDEB(*restartTarget); rval= -1; goto ready;	}
	*(++restartTarget)= (char *)0;

	*restartTarget= strdup( sessionID );
	if  ( ! *restartTarget )
	    { XDEB(*restartTarget); rval= -1; goto ready;	}
	*(++restartTarget)= (char *)0;

	id= sessionID; idlen= strlen( id );
	}

    /*  0 !! */
    sprintf( scratch, "%08lx", appGetTimestamp() );
    id= scratch; idlen= strlen( id );

    changedCount= 0;
    restartCount= 0;
    for ( i= 0; i < ea->eaOpenCount; i++ )
	{
	EditDocument *	ed= ea->eaOpenDocuments[i];

	if  ( ed->edCheckpointFilename )
	    {
	    if  ( appRemoveFile( ed->edCheckpointFilename ) )
		{ SDEB(ed->edCheckpointFilename);	}

	    free( ed->edCheckpointFilename );
	    ed->edCheckpointFilename= (char *)0;
	    }

	if  ( saveChanged					&&
	      ( ed->edHasBeenChanged || ! ed->edFilename )	)
	    {
	    int			size;
	    const char *	ext;
	    char *		name;

	    changedCount++;
	    restartCount++;

	    if  ( ed->edFilename )
		{
		*restartTarget= strdup( "++Upd" );
		if  ( ! *restartTarget )
		    { XDEB(*restartTarget); rval= -1; goto ready;	}
		*(++restartTarget)= (char *)0;
		}
	    else{
		*restartTarget= strdup( "++New" );
		if  ( ! *restartTarget )
		    { XDEB(*restartTarget); rval= -1; goto ready;	}
		*(++restartTarget)= (char *)0;
		}

	    *discardTarget= strdup( "++Old" );
	    if  ( ! *discardTarget )
		{ XDEB(*discardTarget); rval= -1; goto ready;	}
	    *(++discardTarget)= (char *)0;

	    size= dirlen+ 1+ idlen+ 20;
	    ext= ea->eaFileExtensions[ed->edFormat].afeExtension;
	    if  ( ext )
		{
		name= malloc( size+ 1+ strlen( ext ) );
		if  ( ! name )
		    { XDEB(name); rval= -1; goto ready;	}
		sprintf( name, "%s/%s-%d.%s", saveDirectory, id, i, ext );

		}
	    else{
		name= malloc( size+ 1 );
		if  ( ! name )
		    { XDEB(name); rval= -1; goto ready;	}
		sprintf( name, "%s/%s-%d", saveDirectory, id, i );
		}
	    ed->edCheckpointFilename= name;

	    if  ( (*ea->eaSaveDocument)( ea, ed, ed->edFormat, name ) )
		{
		SDEB(name);
		if  ( interactive )
		    {
		    appReportSaveFailure( ea, (Widget)0,
						    ed->edTopWidget, name );
		    }
		}

	    *discardTarget= strdup( name );
	    if  ( ! *discardTarget )
		{ XDEB(*discardTarget); rval= -1; goto ready;	}
	    *(++discardTarget)= (char *)0;

	    *restartTarget= strdup( name );
	    if  ( ! *restartTarget )
		{ XDEB(*restartTarget); rval= -1; goto ready;	}
	    *(++restartTarget)= (char *)0;

	    *restartTarget= strdup( ed->edTitle?ed->edTitle:"?" );
	    if  ( ! *restartTarget )
		{ XDEB(*restartTarget); rval= -1; goto ready;	}
	    *(++restartTarget)= (char *)0;

	    if  ( ed->edFilename )
		{
		*restartTarget= strdup( ed->edFilename );
		if  ( ! *restartTarget )
		    { XDEB(*restartTarget); rval= -1; goto ready;	}
		*(++restartTarget)= (char *)0;
		}
	    }
	else{
	    if  ( ed->edFilename )
		{
		restartCount++;

		*restartTarget= strdup( ed->edFilename );
		if  ( ! *restartTarget )
		    { XDEB(*restartTarget); rval= -1; goto ready;	}
		*(++restartTarget)= (char *)0;
		}
	    }
	}

    *discardTarget= strdup( "++Exit" );
    if  ( ! *discardTarget )
	{ XDEB(*discardTarget); rval= -1; goto ready;	}
    *(++discardTarget)= (char *)0;

    XtVaSetValues( ea->eaTopWidget,
		    XtNcurrentDirectory,	currentDirectory,
		    XtNrestartCommand,		restartCommand,
		    /* 00
		    XtNdiscardCommand,		discardCommand,
		    */
		    NULL );

  ready:
    if  ( restartCommand )
	{ appSessionFreeCommand( restartCommand );	}
    if  ( discardCommand )
	{ appSessionFreeCommand( discardCommand );	}

    return rval;
    }

static void appSessionInteractCallBack(	Widget			w,
					void *			voidea,
					void *			voidxct )
    {
    EditApplication *		ea= (EditApplication *)voidea;
    XtCheckpointToken		xct= (XtCheckpointToken)voidxct;
    Widget			relative= ea->eaTopWidget;

    int				rcc;
    const int			interactive= 1;

    /*  A  */
    if  ( xct->cancel_shutdown				||
	  xct->interact_style == SmInteractStyleNone	)
	{
	xct->save_success= False;
	XtSessionReturnToken( xct );
	return;
	}

    if  ( xct->interact_style == SmInteractStyleAny	&&
	  xct->shutdown					)
	{
	int	changedCount;

	changedCount= appCountChangedDocuments( ea );

	if  ( changedCount > 1 )
	    {
	    rcc= appRunReallyQuitDialog( (Widget)0, relative, ea );

	    switch( rcc )
		{
		case AQDrespCANCEL:
		    xct->request_cancel= True;
		    xct->save_success= False;

		    XtSessionReturnToken( xct );
		    return;

		case AQDrespNO:
		    {
		    const int	saveChanged= 0;

		    if  ( appSessionSaveState( ea, saveChanged, interactive ) )
			{ LDEB(1); xct->save_success= False;	}
		    else{ xct->save_success= True;		}

		    xct->request_cancel= False;

		    XtSessionReturnToken( xct );
		    if  ( xct->shutdown )
			{ appExitApplication( ea ); LDEB(1);	}
		    return;
		    }

		default:
		    LDEB(rcc); return;

		case AQDrespYES:
		    break;
		}
	    }

	appAskCloseDocuments( (Widget)0, ea );

	changedCount= appCountChangedDocuments( ea );
	if  ( changedCount > 0 )
	    {
	    xct->request_cancel= True;
	    xct->save_success= False;
	    XtSessionReturnToken( xct );
	    return;
	    }
	else{
	    const int	saveChanged= 0;

	    if  ( appSessionSaveState( ea, saveChanged, interactive ) )
		{ LDEB(1); xct->save_success= False;	}
	    else{ xct->save_success= True;		}

	    xct->request_cancel= False;

	    XtSessionReturnToken( xct );
	    if  ( xct->shutdown )
		{ appExitApplication( ea ); LDEB(1);	}
	    return;
	    }
	}

    {
    const int	saveChanged= 1;

    if  ( appSessionSaveState( ea, saveChanged, interactive ) )
	{ LDEB(1); xct->save_success= False;	}
    else{ xct->save_success= True;		}

    xct->request_cancel= False;

    XtSessionReturnToken( xct );

    /* NO!
    if  ( xct->shutdown )
	{ appExitApplication( ea ); LDEB(1);	}
    */
    }
    }


static void appSessionMakeSensitive(	EditApplication *	ea,
					int			sensitive )
    {
    int		i;

    XtSetSensitive( ea->eaTopWidget, sensitive );

    for ( i= 0; i < ea->eaOpenCount; i++ )
	{
	EditDocument *	ed= ea->eaOpenDocuments[i];

	XtSetSensitive( ed->edTopWidget, sensitive );
	}

    return;
    }

static void appSessionSaveCallBack(	Widget			w,
					void *			voidea,
					void *			voidxct )
    {
    EditApplication *		ea= (EditApplication *)voidea;
    XtCheckpointToken		xct= (XtCheckpointToken)voidxct;

    /*  1  */
    if  ( xct->shutdown					||
	  xct->interact_style != SmInteractStyleNone	)
	{ appSessionMakeSensitive( ea, False );	}

    /*  2  */
    if  ( xct->interact_style == SmInteractStyleNone )
	{
	const int	saveChanged= 1;
	const int	interactive= 0;

	if  ( appSessionSaveState( ea, saveChanged, interactive ) )
	    { LDEB(1); xct->save_success= False;	}
	else{ xct->save_success= True;			}

	return;
	}

    /*  3  */
    if  ( xct->interact_style == SmInteractStyleAny )
	{ xct->interact_dialog_type= SmDialogNormal;	}
    else{ xct->interact_dialog_type= SmDialogError;		}

    XtAddCallback( ea->eaTopWidget, XtNinteractCallback,
					appSessionInteractCallBack, voidea );

    return;
    }

static void appSessionDieCallBack(	Widget			w,
					void *			voidea,
					void *			vnull )
    {
    EditApplication *		ea= (EditApplication *)voidea;

    appExitApplication( ea );

    return;
    }

static void appSessionCancelCallBack(	Widget			w,
					void *			voidea,
					void *			vnull )
    {
    EditApplication *		ea= (EditApplication *)voidea;

    /*  z  */
    appSessionMakeSensitive( ea, True );

    return;
    }

static void appSessionSaveCompleteCallBack( Widget		w,
					void *			voidea,
					void *			vnull )
    {
    EditApplication *		ea= (EditApplication *)voidea;

    /*  z  */
    appSessionMakeSensitive( ea, True );

    return;
    }

#   endif

/************************************************************************/
/*									*/
/*  Generic editor type application main().				*/
/*									*/
/*  1)  Make an application shell.					*/
/*  1b) Obtain resource values.						*/
/*  2)  Remember X11 properties.					*/
/*  3)  Remember Window manager properties.				*/
/*									*/
/************************************************************************/
# define xx(x)	x,x

static float OneDotTwo= 1.2;

static XtResource	APP_ApplicationResourceTable[]=
    {
	{ xx("unit"), XtRString, sizeof(char *),
		    offsetof(EditApplication,eaUnitString),
		    XtRString, "inch" },
	{ xx("paper"), XtRString, sizeof(char *),
		    offsetof(EditApplication,eaPaperString),
		    XtRString, "letter" },
	{ xx("leftMargin"), XtRString, sizeof(char *),
		    offsetof(EditApplication,eaLeftMarginString),
		    XtRString, "90pt" },
	{ xx("rightMargin"), XtRString, sizeof(char *),
		    offsetof(EditApplication,eaRightMarginString),
		    XtRString, "90pt" },
	{ xx("topMargin"), XtRString, sizeof(char *),
		    offsetof(EditApplication,eaTopMarginString),
		    XtRString, "72pt" },
	{ xx("bottomMargin"), XtRString, sizeof(char *),
		    offsetof(EditApplication,eaBottomMarginString),
		    XtRString, "72pt" },
	{ xx("magnification"), XtRFloat, sizeof(float),
		    offsetof(EditApplication,eaMagnification),
		    XtRFloat, (void *)&OneDotTwo },

	{ xx("author"), XtRString, sizeof(char *),
		    offsetof(EditApplication,eaAuthor),
		    XtRString, (char *)0 },

	{ xx("pageNumberFormat"), XtRString, sizeof(char *),
		    offsetof(EditApplication,eaPageNumberFormat),
		    XtRString, "Page %d" },

	/*
	{ xx("sessionManagement"), XtRBool, sizeof(Bool),
		    offsetof(EditApplication,eaSessionManagement),
		    XtRString, "False" },
	*/

	{ xx("afmDirectory"), XtRString, sizeof(char *),
		    offsetof(EditApplication,eaAfmDirectory),
		    XtRString, AFMDIR },
	{ xx("defaultFont"), XtRString, sizeof(char *),
		    offsetof(EditApplication,eaDefaultFont),
		    XtRString, "Helvetica,,,10" },
	{ xx("faxCommand"), XtRString, sizeof(char *),
		    offsetof(EditApplication,eaFaxCommand),
		    XtRString, (char *)0 },
    };

int appMain(	EditApplication *	ea,
		int			argc,
		char *			argv[]	)
    {
    int				arg;

    AppDocSelectionType *	adst;
    AppAppSelectionType *	aast;
    int				i;

    Arg				al[20];
    int				ac= 0;

    Pixmap			pixmap= (Pixmap)0;

    int				paperFormat;
    int				width;
    int				height;

    Widget			pictureLabel= (Widget)0;
    Widget			label;
    XmString			labelString;

    MwmHints			hints;
    char *			cloneCommand[2];

    setlocale( LC_ALL, "" );

    if  ( ! XSupportsLocale() )
	{ SLDEB(getenv("LANG"),XSupportsLocale()); }

#   if 0
    This is actually counter productive.
    XtSetLanguageProc( NULL, NULL, NULL );
#   endif

    arg= 1;
    while( arg < argc && ! strcmp( argv[arg], "++Old" ) )
	{
	if  ( arg+ 1 >= argc )
	    { LLDEB(arg,argc); arg= argc- 1; continue;	}

	if  ( appRemoveFile( argv[arg+ 1] ) )
	    { SDEB(argv[arg+ 1]);	}

	arg += 2; continue;
	}

    if  ( arg < argc && ! strcmp( argv[arg], "++Exit" ) )
	{ return 0;	}

    /*  1  */
    hints.flags= MWM_HINTS_FUNCTIONS|MWM_HINTS_DECORATIONS;
    hints.functions=	MWM_FUNC_MOVE		|
			MWM_FUNC_MINIMIZE	|
			MWM_FUNC_CLOSE		;
    hints.decorations=	MWM_DECOR_BORDER	|
			MWM_DECOR_TITLE		|
			MWM_DECOR_MENU		|
			MWM_DECOR_MINIMIZE	;
    ac= 0;
    XtSetArg( al[ac], XmNinput,			True ); ac++;
    XtSetArg( al[ac], XmNallowShellResize,	True ); ac++;
    XtSetArg( al[ac], XmNmwmDecorations,	hints.decorations ); ac++;
    XtSetArg( al[ac], XmNmwmFunctions,		hints.functions ); ac++;

#   ifdef USE_X11_R5
    /* applicationShellWidgetClass, */
    ea->eaTopWidget= XtAppInitialize( &(ea->eaContext),
					ea->eaApplicationName,
					NULL, 0, &argc, argv, NULL, al, ac );
#   else
    cloneCommand[0]= argv[0];
    cloneCommand[1]= (char *)0;
    XtSetArg( al[ac], XtNcloneCommand,		cloneCommand ); ac++;

    ea->eaTopWidget= XtOpenApplication( &(ea->eaContext),
					ea->eaApplicationName,
					NULL, 0, &argc, argv, NULL,
					sessionShellWidgetClass,
					al, ac );
#   endif

    /*  1b  */
    XtGetApplicationResources( ea->eaTopWidget, ea->eaResourceData,
			ea->eaResourceTable, ea->eaResourceCount,
			NULL, 0 );

    XtGetApplicationResources( ea->eaTopWidget, &(ea->eaFileMessageResources),
			ea->eaFileMessageResourceTable,
			ea->eaFileMessageResourceCount,
			NULL, 0 );

#   ifndef USE_X11_R5
    XtAddCallback( ea->eaTopWidget, XtNsaveCallback,
				appSessionSaveCallBack, (void *)ea );
    XtAddCallback( ea->eaTopWidget, XtNcancelCallback,
				appSessionCancelCallBack, (void *)ea );
    XtAddCallback( ea->eaTopWidget, XtNsaveCompleteCallback,
				appSessionSaveCompleteCallBack, (void *)ea );
    XtAddCallback( ea->eaTopWidget, XtNdieCallback,
				appSessionDieCallBack, (void *)ea );
#   endif

    appInitDocumentGeometry( &(ea->eaDefaultDocumentGeometry) );
    XtGetApplicationResources( ea->eaTopWidget, (void *)ea,
			APP_ApplicationResourceTable,
			XtNumber(APP_ApplicationResourceTable),
			NULL, 0 );

    ea->eaUnitInt= appUnitTypeInt( ea->eaUnitString );
    if  ( ea->eaUnitInt < 0 )
	{ SDEB(ea->eaUnitString); ea->eaUnitInt= UNITtyINCH; }

    paperFormat= appPaperGetByMnemonic( ea->eaPaperString );
    if  ( paperFormat >= 0 )
	{
	if  ( ! appPaperGetInformation( paperFormat,
					&width, &height, (const char **)0 ) )
	    {
	    ea->eaDefaultDocumentGeometry.dgPaperWideTwips= width;
	    ea->eaDefaultDocumentGeometry.dgPaperHighTwips= height;
	    }
	else{ SDEB(ea->eaPaperString);	}
	}
    else{
	if  ( ! appGeoRectangleFromString( ea->eaPaperString, ea->eaUnitInt,
							    &width, &height ) )
	    {
	    ea->eaDefaultDocumentGeometry.dgPaperWideTwips= width;
	    ea->eaDefaultDocumentGeometry.dgPaperHighTwips= height;
	    }
	else{ SDEB(ea->eaPaperString);	}
	}

    if  ( ea->eaLeftMarginString )
	{
	if  ( appGeoLengthFromString( ea->eaLeftMarginString, ea->eaUnitInt,
			&ea->eaDefaultDocumentGeometry.dgLeftMarginTwips ) )
	    { SDEB(ea->eaLeftMarginString);	}
	}

    if  ( ea->eaRightMarginString )
	{
	if  ( appGeoLengthFromString( ea->eaRightMarginString, ea->eaUnitInt,
			&ea->eaDefaultDocumentGeometry.dgRightMarginTwips ) )
	    { SDEB(ea->eaRightMarginString);	}
	}

    if  ( ea->eaTopMarginString )
	{
	if  ( appGeoLengthFromString( ea->eaTopMarginString, ea->eaUnitInt,
			&ea->eaDefaultDocumentGeometry.dgTopMarginTwips ) )
	    { SDEB(ea->eaTopMarginString);	}
	}

    if  ( ea->eaBottomMarginString )
	{
	if  ( appGeoLengthFromString( ea->eaBottomMarginString, ea->eaUnitInt,
			&ea->eaDefaultDocumentGeometry.dgBottomMarginTwips ) )
	    { SDEB(ea->eaBottomMarginString);	}
	}

    /*  2  */
    ea->eaDisplay= XtDisplay( ea->eaTopWidget );
    ea->eaScreen= DefaultScreen( ea->eaDisplay );
    ea->eaScreenPixelsWide= DisplayWidth( ea->eaDisplay, ea->eaScreen );
    ea->eaScreenPixelsHigh= DisplayHeight( ea->eaDisplay, ea->eaScreen );
    ea->eaInputMethod= XOpenIM( ea->eaDisplay, NULL, NULL, NULL );
    ea->eaDocumentCursor= (Cursor)0;

    /*  3  */
    ea->eaCloseAtom= XmInternAtom( ea->eaDisplay, "WM_DELETE_WINDOW", False );

    ea->eaArgc= argc;
    ea->eaArgv= argv;

    if  ( ea->eaCloseAtom )
	{
	XtVaSetValues( ea->eaTopWidget,
			    XmNdeleteResponse,		XmDO_NOTHING,
			    NULL );

	XmAddWMProtocolCallback( ea->eaTopWidget, ea->eaCloseAtom,
						appAppWmClose, (void *)ea );
	}

    ac= 0;
    XtSetArg( al[ac],	XmNsashWidth,		1 ); ac++;
    XtSetArg( al[ac],	XmNsashHeight,		1 ); ac++;
    XtSetArg( al[ac],	XmNseparatorOn,		False ); ac++;
    XtSetArg( al[ac],	XmNmarginWidth,		0 ); ac++;
    XtSetArg( al[ac],	XmNmarginHeight,	0 ); ac++;
    XtSetArg( al[ac],	XmNspacing,		0 ); ac++;
    ea->eaMainWindow= XmCreatePanedWindow( ea->eaTopWidget,
						    WIDGET_NAME, al, ac );

    ac= 0;
    ea->eaMenuBar= XmCreateMenuBar( ea->eaMainWindow, WIDGET_NAME, al, ac );

    ea->eaFileMenu= appMakeMenu( &(ea->eaFileMenuButton), ea, ea->eaMenuBar,
			*(ea->eaAppFileMenuText), 0,
			ea->eaAppFileMenuItems, ea->eaAppFileMenuItemCount,
			(void *)ea );

    ea->eaWinMenu= appMakeMenu( &(ea->eaWinMenuButton), ea, ea->eaMenuBar,
			*(ea->eaAppWinMenuText), 0,
			ea->eaAppWinMenuItems, ea->eaAppWinMenuItemCount,
			(void *)ea );

    if  ( ea->eaMakePrivateApplicationMenus )
	{ (*ea->eaMakePrivateApplicationMenus)( ea, ea->eaMenuBar ); }

    if  ( ea->eaAppHelpMenuItems )
	{
	ea->eaHelpMenu= appMakeMenu( &(ea->eaHelpMenuButton),
		    ea, ea->eaMenuBar,
		    *(ea->eaAppHelpMenuText), 1,
		    ea->eaAppHelpMenuItems, ea->eaAppHelpMenuItemCount,
		    (void *)ea );
	}

    if  ( ea->eaMainPicture )
	{
	ac= 0;
	XtSetArg( al[ac], XmNlabelType,		XmPIXMAP ); ac++;
	XtSetArg( al[ac], XmNrecomputeSize,	True ); ac++;
	XtSetArg( al[ac], XmNallowResize,	True ); ac++;

	pictureLabel= XmCreateLabel( ea->eaMainWindow, WIDGET_NAME, al, ac );
	XtManageChild( pictureLabel );
	}

    if  ( ea->eaNameAndVersion )
	{
	labelString= XmStringCreateLocalized( ea->eaNameAndVersion );

	ac= 0;
	XtSetArg( al[ac], XmNlabelString,	labelString ); ac++;
	XtSetArg( al[ac], XmNmarginWidth,	5 ); ac++;
	XtSetArg( al[ac], XmNmarginHeight,	0 ); ac++;
	XtSetArg( al[ac], XmNmarginTop,		4 ); ac++;
	XtSetArg( al[ac], XmNmarginBottom,	2 ); ac++;
	XtSetArg( al[ac], XmNalignment,		XmALIGNMENT_BEGINNING ); ac++;
	label= XmCreateLabel( ea->eaMainWindow, WIDGET_NAME, al, ac );

	XmStringFree( labelString );

	XtManageChild( label );
	}

    if  ( ea->eaReference )
	{
	labelString= XmStringCreateLocalized( ea->eaReference );

	ac= 0;
	XtSetArg( al[ac], XmNlabelString,	labelString ); ac++;
	XtSetArg( al[ac], XmNmarginWidth,	5 ); ac++;
	XtSetArg( al[ac], XmNmarginHeight,	0 ); ac++;
	XtSetArg( al[ac], XmNmarginTop,		2 ); ac++;
	XtSetArg( al[ac], XmNmarginBottom,	4 ); ac++;
	XtSetArg( al[ac], XmNalignment,		XmALIGNMENT_BEGINNING ); ac++;
	label= XmCreateLabel( ea->eaMainWindow, WIDGET_NAME, al, ac );

	XmStringFree( labelString );

	XtManageChild( label );
	}

    XtManageChild( ea->eaMenuBar );
    XtManageChild( ea->eaMainWindow );
    XtRealizeWidget( ea->eaTopWidget );

    if  ( pictureLabel )
	{
	Pixmap			labelPixmap;

	if  ( ! appGetImagePixmap( ea, ea->eaMainPicture, &labelPixmap ) )
	    {
	    XtVaSetValues( pictureLabel,
				XmNborderWidth,	0,
				XmNlabelPixmap,	labelPixmap,
				NULL );
	    }
	}

    if  ( ea->eaMainIcon					&&
	  appGetImagePixmap( ea, ea->eaMainIcon, &pixmap )	)
	{ SDEB(ea->eaMainIcon); return -1;	}

    if  ( pixmap )
	{
	XtVaSetValues( ea->eaTopWidget,
			    XmNiconPixmap,	pixmap,
			    NULL );
	}

    if  ( ea->eaPrintDestinationCount == 0				&&
	  utilPrinterGetPrinters( &(ea->eaPrintDestinationCount),
				    &(ea->eaDefaultPrintDestination),
				    &(ea->eaPrintDestinations) )	)
	{ LDEB(1); 	}

    adst= ea->eaDocSelectionTypes;
    for ( i= 0; i < ea->eaDocSelectionTypeCount; adst++, i++ )
	{
	int				j;
	AppDocSelectionTargetType *	adstt;

	if  ( adst->adstSelectionAtom == None )
	    {
	    adst->adstSelectionAtom= XInternAtom( ea->eaDisplay,
					    adst->adstSelectionString, 0 );
	    if  ( adst->adstSelectionAtom == None )
		{ SLDEB(adst->adstSelectionString, adst->adstSelectionAtom); }
	    }
	adstt= adst->adstTargetTypes;
	for ( j= 0; j < adst->adstTargetTypeCount; adstt++, j++ )
	    {
	    if  ( adstt->adsttTargetAtom == None )
		{
		adstt->adsttTargetAtom= XInternAtom( ea->eaDisplay,
					    adstt->adsttTargetString, 0 );
		if  ( adstt->adsttTargetAtom == None )
		    { SLDEB(adstt->adsttTargetString, adstt->adsttTargetAtom); }
		}
	    }
	}

    aast= ea->eaAppSelectionTypes;
    for ( i= 0; i < ea->eaAppSelectionTypeCount; aast++, i++ )
	{
	int				j;
	AppAppSelectionTargetType *	aastt;

	if  ( aast->aastSelectionAtom == None )
	    {
	    aast->aastSelectionAtom= XInternAtom( ea->eaDisplay,
					    aast->aastSelectionString, 0 );
	    if  ( aast->aastSelectionAtom == None )
		{ SLDEB(aast->aastSelectionString, aast->aastSelectionAtom); }
	    }
	aastt= aast->aastTargetTypes;
	for ( j= 0; j < aast->aastTargetTypeCount; aastt++, j++ )
	    {
	    if  ( aastt->aasttTargetAtom == None )
		{
		aastt->aasttTargetAtom= XInternAtom( ea->eaDisplay,
					    aastt->aasttTargetString, 0 );
		if  ( aastt->aasttTargetAtom == None )
		    { SLDEB(aastt->aasttTargetString, aastt->aasttTargetAtom); }
		}
	    }
	}

    if  ( ea->eaAppSelectionTypeCount > 0 )
	{
	XtAddEventHandler( ea->eaTopWidget,
			PropertyChangeMask, True, appAppSelectionInput, ea );
	}

    for ( arg= 1; arg < argc; arg++ )
	{
	const int			readonly= 0;
	AppFileMessageResources *	afmr= &(ea->eaFileMessageResources);
	static char			absolute[1000+1];

	EditDocument *		ed;

	if  ( ! strcmp( argv[arg], "++Upd" ) )
	    {
	    if  ( arg+ 3 >= argc )
		{ LLDEB(arg,argc); arg= argc- 1; continue;	}

	    ed= appOpenDocument( ea, ea->eaTopWidget, (Widget)0,
						readonly, argv[arg+ 1] );

	    if  ( ed )
		{
		appSetDocumentTitle( ed, argv[arg+ 2] );
		appSetDocumentFilename( ed, argv[arg+ 3] );

		appDocumentChanged( ed, 1 );
		}

	    if  ( appRemoveFile( argv[arg+ 1] ) )
		{ SDEB(argv[arg+ 1]);	}

	    arg += 3; continue;
	    }

	if  ( ! strcmp( argv[arg], "++New" ) )
	    {
	    if  ( arg+ 2 >= argc )
		{ LLDEB(arg,argc); arg= argc- 1; continue;	}

	    ed= appOpenDocument( ea, ea->eaTopWidget, (Widget)0,
						readonly, argv[arg+ 1] );

	    if  ( ed )
		{
		appSetDocumentTitle( ed, argv[arg+ 2] );
		appSetDocumentFilename( ed, (char *)0 );

		appDocumentChanged( ed, 1 );
		}

	    if  ( appRemoveFile( argv[arg+ 1] ) )
		{ SDEB(argv[arg+ 1]);	}

	    arg += 2; continue;
	    }

	if  ( ! strcmp( argv[arg], "++Old" ) )
	    {
	    if  ( arg+ 1 >= argc )
		{ LLDEB(arg,argc); arg= argc- 1; continue;	}

	    if  ( appRemoveFile( argv[arg+ 1] ) )
		{ SDEB(argv[arg+ 1]);	}

	    arg += 1; continue;
	    }

	if  ( ! strcmp( argv[arg], "++Exit" ) )
	    { continue;	}

	if  ( appTestFileExists( argv[arg] ) )
	    {
	    if  ( ea->eaCreateNewFromCommand )
		{
		int	res;

		res= appQuestionRunSubjectYesNoDialog( ea, ea->eaTopWidget,
		    (Widget)0, argv[arg], afmr->afmrMakeItQuestion,
		    (char *)0, (char *)0 );

		if  ( res == AQDrespYES )
		    {
		    if  ( appNewDocument( ea, argv[arg] ) )
			{ SDEB(argv[arg]);	}
		    }
		}
	    else{
		appQuestionRunSubjectErrorDialog( ea, ea->eaTopWidget,
		    (Widget)0, argv[arg], afmr->afmrNoSuchFileMessage );
		}
	    }
	else{
	    if  ( appAbsoluteName( absolute, 1000, argv[arg] ) < 0 )
		{
		SDEB(argv[arg]);
		ed= appOpenDocument( ea, ea->eaTopWidget, (Widget)0,
						    readonly, argv[arg] );
		}
	    else{
		ed= appOpenDocument( ea, ea->eaTopWidget, (Widget)0,
						    readonly, absolute );
		}
	    }
	}

    if  ( ea->eaVisibleDocumentCount > 0 )
	{ XtUnmapWidget( ea->eaTopWidget ); }

    {
    double			horPixPerMM;
    double			verPixPerMM;
    double			xfac;
    double			yfac;

    appGetFactors( ea, &horPixPerMM, &verPixPerMM, &xfac, &yfac );
    }

    XtAppMainLoop( ea->eaContext );

    return 0;
    }
