#   include	<stdlib.h>
#   include	<stdio.h>

#   include	<appMetricRuler.h>

#   include	<appDraw.h>

#   include	<debugon.h>

#   define	BACK_MARG(s)	((s)/2)

#   define	TAG_TOP(s)	((s)/3)
#   define	TAG_BOTTOM(s)	(s)
#   define	TAG_HEIGHT(s)	(TAG_BOTTOM(s)-TAG_TOP(s))

#   define	FONT_HEIGHT(s)	(5*(s)/9)
#   define	FONT_BOTTOM(s)	(5*(s)/9)

#   define	TICK_TOP(s)	((2*(s))/3)
#   define	TICK_BOTTOM(s)	(s)
#   define	TICK_HEIGHT(s)	(TICK_BOTTOM(s)-TICK_TOP(s))

#   define	HAIR_THICK(s)	(2)

typedef struct MetricRuler
    {
    RulerData		mrRulerData;

    int			mrV0Pixels;		/*  hairs lower value	*/
    int			mrV1Pixels;		/*  hairs upper value	*/
    } MetricRuler;

void * appMakeMetricRuler(	int			sizeAcross,
				double			pixPerMM,
				double			magnification,

				Pixel			forePixel,
				Pixel			backPixel,
				Pixel			topPixel,
				Pixel			bottomPixel,

				int			minUnused,
				int			maxUnused,

				int			documentC0,
				int			documentC1,
				int			visibleC0,
				int			visibleC1,
				int			rulerC1,

				int			whatUnit )
    {
    MetricRuler *		mr;
    RulerData *		rd;

    mr= malloc( sizeof( MetricRuler ) );
    if  ( ! mr )
	{ XDEB(mr); return (void *)0;	}
    rd= &(mr->mrRulerData);

    appInitRulerData( &(mr->mrRulerData) );

    rd->rdUnit= whatUnit;
    appRulerCalculateIncrements( rd, pixPerMM, magnification );

    rd->rdMinUsed= minUnused;
    rd->rdMaxUsed= 0;
    rd->rdExtraAfterMaxUsed= maxUnused;

    rd->rdDocumentC0= documentC0;
    rd->rdDocumentC1= documentC1;
    rd->rdVisibleC0= visibleC0;
    rd->rdVisibleC1= visibleC1;
    rd->rdRulerC1= rulerC1;

    rd->rdSizeAcross= sizeAcross;
    rd->rdSizeAlong= 0;

    rd->rdForePixel= forePixel;
    rd->rdBackPixel= backPixel;
    rd->rdTopPixel= topPixel;
    rd->rdBottomPixel= bottomPixel;

    mr->mrV0Pixels= -1;
    mr->mrV1Pixels= -1;

    return (void *)mr;
    }

void appFreeMetricRuler(	Display *	display,
			void *		voidmr )
    {
    MetricRuler *		mr= (MetricRuler *)voidmr;

    appCleanRulerData( display, &(mr->mrRulerData) );

    free( mr );

    return;
    }

/************************************************************************/
/*									*/
/*  Draw a horizontal ruler.						*/
/*									*/
/************************************************************************/

static void appDrawHorizontalRuler(	Widget			w,
					MetricRuler *		mr,
					DocumentRectangle *	drClip,
					int			ox,
					Window			win,
					Display *		display )
    {
    RulerData *		rd= &(mr->mrRulerData);
    int			sizeAcross= rd->rdSizeAcross;

    double		item;
    int			tick;
    double		units;

    int			tagY= TAG_TOP(sizeAcross);
    int			tagH= TAG_HEIGHT(sizeAcross);

    int			tickY= TICK_TOP(sizeAcross);
    int			tickH= TICK_HEIGHT(sizeAcross);

    int			fontBottom= FONT_BOTTOM(sizeAcross);

    int			x0;
    int			x1;

    if  ( rd->rdSizeAlong == 0 )
	{
	XtVaGetValues( w, XtNwidth, &rd->rdSizeAlong, NULL );
	rd->rdMaxUsed= rd->rdSizeAlong- rd->rdExtraAfterMaxUsed;
	}

    appRulerDrawBackground( display, win, rd,
		    drClip->drX0- ox, drClip->drY0,
		    drClip->drX1- drClip->drX0, drClip->drY1- drClip->drY0 );

    if  ( mr->mrV0Pixels >= 0 || mr->mrV1Pixels >= 0 )
	{
	DocumentRectangle	drSel;

	drSel.drX0= 0;
	drSel.drX1= rd->rdMaxUsed;
	if  ( mr->mrV0Pixels >= 0 )
	    { drSel.drX0= mr->mrV0Pixels;	}
	if  ( mr->mrV1Pixels >= 0 )
	    { drSel.drX1= mr->mrV1Pixels;	}
	drSel.drY0= 0;
	drSel.drY1= sizeAcross;

	if  ( docIntersectRectangle( &drSel, &drSel, drClip ) )
	    {
	    XSetForeground( display, rd->rdGc, rd->rdBottomPixel );

	    XFillRectangle( display, win, rd->rdGc,
		    drSel.drX0- ox, drSel.drY0,
		    drSel.drX1- drSel.drX0, drSel.drY1- drSel.drY0 );
	    }
	}

    XSetForeground( display, rd->rdGc, rd->rdForePixel );

    x0= 0- ox;
    x1= rd->rdRulerC1- ox;

    if  ( x0 < rd->rdMinUsed )
	{ x0=  rd->rdMinUsed;	}
    if  ( x1 > rd->rdMaxUsed )
	{ x1=  rd->rdMaxUsed;	}

    if  ( x1 >= x0 )
	{
	XFillRectangle( display, win, rd->rdGc, x0, sizeAcross- 1, x1- x0, 1 );
	}

    item= 0; units= 0; tick= 0;
    while( item < drClip->drX1 )
	{
	if  ( item- ox <  rd->rdMinUsed	||
	      item- ox >= rd->rdMaxUsed	)
	    {
	    tick++;
	    units += rd->rdUnitsPerTick;
	    item += rd->rdTickDistancePixels;
	    continue;
	    }

	if  ( item != 0 && tick % rd->rdTicksPerText == 0 )
	    {
	    char	scratch[30];
	    int		len;
	    int		wide;

	    appRulerTagText( scratch, &wide, &len, rd, units );

	    if  ( item- ox- wide/2 >= rd->rdMinUsed	&&
		  item- ox+ wide/2 <  rd->rdMaxUsed	)
		{
		XDrawString( display, win, rd->rdGc, 
			    item- ox- wide/2, fontBottom, scratch, len );

		XFillRectangle( display, win, rd->rdGc,
			item- ox, fontBottom+ 2, 1, sizeAcross- fontBottom- 2 );
		}
	    else{
		XFillRectangle( display, win, rd->rdGc,
					    item- ox, tagY, 1, tagH );
		}
	    }
	else{
	    if  ( tick % rd->rdTicksPerTag == 0 )
		{
		XFillRectangle( display, win, rd->rdGc,
						item- ox, tagY, 1, tagH );
		}
	    else{
		XFillRectangle( display, win, rd->rdGc,
						item- ox, tickY, 1, tickH );
		}
	    }

	tick++;
	units += rd->rdUnitsPerTick;
	item += rd->rdTickDistancePixels;
	continue;
	}

    /* Already done:
    XSetForeground( display, rd->rdGc, mr->trForePixel );
    */

    if  ( mr->mrV0Pixels >= 0				&&
	  mr->mrV0Pixels- ox- 2 >= rd->rdMinUsed		&&
	  mr->mrV0Pixels- ox <  rd->rdMaxUsed		)
	{
	XFillRectangle( display, win, rd->rdGc,
			    mr->mrV0Pixels- ox- 1, 0, 1, sizeAcross );
	}

    if  ( mr->mrV1Pixels >= 0				&&
	  mr->mrV1Pixels- ox >= rd->rdMinUsed		&&
	  mr->mrV1Pixels- ox+ 2 <  rd->rdMaxUsed	)
	{
	XFillRectangle( display, win, rd->rdGc,
				mr->mrV1Pixels- ox, 0, 1, sizeAcross );
	}

    if  ( mr->mrV0Pixels >= 0 || mr->mrV1Pixels >= 0 )
	{
	int	screen= DefaultScreen( display );
	Pixel	whitePixel= WhitePixel( display, screen );

	XSetForeground( display, rd->rdGc, whitePixel );

	if  ( mr->mrV0Pixels >= 0				&&
	      mr->mrV0Pixels- ox- 2 >= rd->rdMinUsed	&&
	      mr->mrV0Pixels- ox <  rd->rdMaxUsed		)
	    {
	    XFillRectangle( display, win, rd->rdGc,
			    mr->mrV0Pixels- ox- 2, 0, 1, sizeAcross );
	    }

	if  ( mr->mrV1Pixels >= 0				&&
	      mr->mrV1Pixels- ox >= rd->rdMinUsed		&&
	      mr->mrV1Pixels- ox+ 2 <  rd->rdMaxUsed	)
	    {
	    XFillRectangle( display, win, rd->rdGc,
			    mr->mrV1Pixels- ox+ 1, 0, 1, sizeAcross );
	    }
	}

    XSetClipMask( display, rd->rdGc, None );

    return;
    }

/************************************************************************/
/*									*/
/*  Draw a vertical ruler.						*/
/*									*/
/************************************************************************/
static void appDrawVerticalRuler(	Widget			w,
					MetricRuler *		mr,
					DocumentRectangle *	drClip,
					int			oy,
					Window			win,
					Display *		display )
    {
    RulerData *		rd= &(mr->mrRulerData);
    int			sizeAcross= rd->rdSizeAcross;

    double		item;
    int			tick;
    double		units;

    int			tagX= TAG_TOP(sizeAcross);
    int			tagW= TAG_HEIGHT(sizeAcross);

    int			tickX= TICK_TOP(sizeAcross);
    int			tickW= TICK_HEIGHT(sizeAcross);

    int			fontBottom= FONT_BOTTOM(sizeAcross);

    int			y0;
    int			y1;

    /*  2  */
    XSetForeground( display, rd->rdGc, rd->rdBackPixel );

    if  ( rd->rdSizeAlong == 0 )
	{
	XtVaGetValues( w, XtNheight, &rd->rdSizeAlong, NULL );
	rd->rdMaxUsed= rd->rdSizeAlong- rd->rdExtraAfterMaxUsed;
	}

    XFillRectangle( display, win, rd->rdGc,
		    drClip->drX0, drClip->drY0- oy,
		    drClip->drX1- drClip->drX0, drClip->drY1- drClip->drY0 );

    if  ( mr->mrV0Pixels >= 0 || mr->mrV1Pixels >= 0 )
	{
	DocumentRectangle	drSel;

	drSel.drX0= 0;
	drSel.drX1= sizeAcross;
	drSel.drY0= 0;
	drSel.drY1= rd->rdMaxUsed;
	if  ( mr->mrV0Pixels >= 0 )
	    { drSel.drY0= mr->mrV0Pixels;	}
	if  ( mr->mrV1Pixels >= 0 )
	    { drSel.drY1= mr->mrV1Pixels;	}

	if  ( docIntersectRectangle( &drSel, &drSel, drClip ) )
	    {
	    XSetForeground( display, rd->rdGc, rd->rdBottomPixel );

	    XFillRectangle( display, win, rd->rdGc,
		    drSel.drX0, drSel.drY0- oy,
		    drSel.drX1- drSel.drX0, drSel.drY1- drSel.drY0 );
	    }
	}

    XSetForeground( display, rd->rdGc, rd->rdForePixel );

    y0= 0- oy;
    y1= rd->rdRulerC1- oy;

    if  ( y0 < rd->rdMinUsed )
	{ y0=  rd->rdMinUsed;	}
    if  ( y1 > rd->rdMaxUsed )
	{ y1=  rd->rdMaxUsed;	}

    if  ( y1 >= y0 )
	{
	XFillRectangle( display, win, rd->rdGc, sizeAcross- 1, y0, 1, y1- y0 );
	}

    item= 0; units= 0; tick= 0;
    while( item < drClip->drY1 )
	{
	if  ( item- oy <  rd->rdMinUsed	||
	      item- oy >= rd->rdMaxUsed	)
	    {
	    tick++;
	    units += rd->rdUnitsPerTick;
	    item += rd->rdTickDistancePixels;
	    continue;
	    }

	if  ( item != 0 && tick % rd->rdTicksPerText == 0 )
	    {
	    char	scratch[30];
	    int		len;
	    int		wide;

	    int		ascent;
	    int		descent;

	    appRulerTagText( scratch, &wide, &len, rd, units );

	    ascent= FONT_HEIGHT(sizeAcross)/ 2;
	    descent= FONT_HEIGHT(sizeAcross)/ 2;

	    if  ( item- oy- descent >= rd->rdMinUsed	&&
		  item- oy+ ascent  <  rd->rdMaxUsed	)
		{
		appRulerDrawVerticalString( display, win, rd,
			    fontBottom, item+ wide/2- oy, scratch, len );

		XFillRectangle( display, win, rd->rdGc,
				fontBottom+ 2, item- oy,
				sizeAcross- fontBottom- 2, 1 );
		}
	    else{
		XFillRectangle( display, win, rd->rdGc,
					    tagX, item- oy, tagW, 1 );
		}
	    }
	else{
	    if  ( tick % rd->rdTicksPerTag == 0 )
		{
		XFillRectangle( display, win, rd->rdGc,
						tagX, item- oy, tagW, 1 );
		}
	    else{
		XFillRectangle( display, win, rd->rdGc,
						tickX, item- oy, tickW, 1 );
		}
	    }

	tick++;
	units += rd->rdUnitsPerTick;
	item += rd->rdTickDistancePixels;
	continue;
	}

    /* Already done:
    XSetForeground( display, rd->rdGc, mr->trForePixel );
    */

    if  ( mr->mrV0Pixels >= 0				&&
	  mr->mrV0Pixels- oy- 2 >= rd->rdMinUsed		&&
	  mr->mrV0Pixels- oy <  rd->rdMaxUsed		)
	{
	XFillRectangle( display, win, rd->rdGc,
				    0, mr->mrV0Pixels- oy- 1, sizeAcross, 1 );
	}

    if  ( mr->mrV1Pixels >= 0				&&
	  mr->mrV1Pixels- oy >= rd->rdMinUsed		&&
	  mr->mrV1Pixels- oy+ 2 <  rd->rdMaxUsed	)
	{
	XFillRectangle( display, win, rd->rdGc,
				    0, mr->mrV1Pixels- oy, sizeAcross, 1 );
	}

    if  ( mr->mrV0Pixels >= 0 || mr->mrV1Pixels >= 0 )
	{
	int	screen= DefaultScreen( display );
	Pixel	whitePixel= WhitePixel( display, screen );

	XSetForeground( display, rd->rdGc, whitePixel );

	if  ( mr->mrV0Pixels >= 0				&&
	      mr->mrV0Pixels- oy- 2 >= rd->rdMinUsed	&&
	      mr->mrV0Pixels- oy <  rd->rdMaxUsed		)
	    {
	    XFillRectangle( display, win, rd->rdGc,
				    0, mr->mrV0Pixels- oy- 2, sizeAcross, 1 );
	    }

	if  ( mr->mrV1Pixels >= 0				&&
	      mr->mrV1Pixels- oy >= rd->rdMinUsed		&&
	      mr->mrV1Pixels- oy+ 2 <  rd->rdMaxUsed	)
	    {
	    XFillRectangle( display, win, rd->rdGc,
				    0, mr->mrV1Pixels- oy+ 1, sizeAcross, 1 );
	    }
	}

    XSetClipMask( display, rd->rdGc, None );

    return;
    }

void appExposeHorizontalRuler(	Widget			w,
				void *			voidmr,
				XEvent *		event )
    {
    MetricRuler *		mr= (MetricRuler *)voidmr;
    RulerData *		rd= &(mr->mrRulerData);

    Display *		display= XtDisplay( w );
    Window		win= XtWindow( w );

    int			ox= rd->rdVisibleC0- rd->rdMinUsed;
    int			oy= 0;

    DocumentRectangle	drClip;

    if  ( ! rd->rdGc )
	{
	const int	vertical= 0;

	appOpenRulerGc( rd, vertical, FONT_HEIGHT(rd->rdSizeAcross), w );
	}

    appCollectExposures( &drClip, ox, oy, display, win, rd->rdGc, event );

    appDrawHorizontalRuler( w, mr, &drClip, ox, win, display );
    }

void appExposeVerticalRuler(	Widget			w,
				void *			voidmr,
				XEvent *		event )
    {
    MetricRuler *		mr= (MetricRuler *)voidmr;
    RulerData *		rd= &(mr->mrRulerData);

    Display *		display= XtDisplay( w );
    Window		win= XtWindow( w );

    int			ox= 0;
    int			oy= rd->rdVisibleC0- rd->rdMinUsed;

    DocumentRectangle	drClip;

    if  ( ! rd->rdGc )
	{
	const int	vertical= 1;

	appOpenRulerGc( rd, vertical, FONT_HEIGHT(rd->rdSizeAcross), w );
	}

    appCollectExposures( &drClip, ox, oy, display, win, rd->rdGc, event );

    appDrawVerticalRuler( w, mr, &drClip, oy, win, display );
    }

void appHorizontalRulerConfigure(	Widget			w,
					void *			voidmr,
					XEvent *		event,
					Boolean *		pRefused )
    {
    XConfigureEvent *	cevent= &(event->xconfigure);
    MetricRuler *		mr= (MetricRuler *)voidmr;
    RulerData *		rd= &(mr->mrRulerData);

    if  ( cevent->type != ConfigureNotify )
	{ return;	}

    rd->rdSizeAlong= cevent->width;
    rd->rdMaxUsed= cevent->width- rd->rdExtraAfterMaxUsed;

    return;
    }

void appVerticalRulerConfigure(	Widget			w,
					void *			voidmr,
					XEvent *		event,
					Boolean *		pRefused )
    {
    XConfigureEvent *	cevent= &(event->xconfigure);
    MetricRuler *		mr= (MetricRuler *)voidmr;
    RulerData *		rd= &(mr->mrRulerData);

    if  ( cevent->type != ConfigureNotify )
	{ return;	}

    rd->rdSizeAlong= cevent->height;
    rd->rdMaxUsed= cevent->height- rd->rdExtraAfterMaxUsed;

    return;
    }

/************************************************************************/
/*									*/
/*  Scroll the ruler.							*/
/*									*/
/************************************************************************/

void appScrollHorMetricRuler(	void *			voidmr,
				Widget			w,
				int			d	)
    {
    MetricRuler *			mr= (MetricRuler *)voidmr;
    RulerData *			rd= &(mr->mrRulerData);
    GC				gc= rd->rdGc;
    Display *			display= XtDisplay( w );
    Window			win= XtWindow( w );

    DocumentRectangle		drClip;

    int				ox;

    appScrollHorizontalRuler( rd, &drClip, d, display, win, gc );

    ox= rd->rdVisibleC0- rd->rdMinUsed;

    appDrawHorizontalRuler( w, mr, &drClip, ox, win, display );
    }

void appScrollVertMetricRuler(	void *			voidmr,
				Widget			w,
				int			d	)
    {
    MetricRuler *			mr= (MetricRuler *)voidmr;
    RulerData *			rd= &(mr->mrRulerData);
    GC				gc= rd->rdGc;
    Display *			display= XtDisplay( w );
    Window			win= XtWindow( w );

    DocumentRectangle		drClip;

    int				oy;

    appScrollVerticalRuler( rd, &drClip, d, display, win, gc );

    oy= rd->rdVisibleC0- rd->rdMinUsed;

    appDrawVerticalRuler( w, mr, &drClip, oy, win, display );
    }

void appSetMetricRulerRange(	void *			voidmr,
			Widget			w,
			int			docVisV0,
			int			docVisV1,
			int			docBackV1 )
    {
    MetricRuler *			mr= (MetricRuler *)voidmr;
    RulerData *			rd= &(mr->mrRulerData);

    Display *			display= XtDisplay( w );
    Window			win= XtWindow( w );

    rd->rdVisibleC0= docVisV0;
    rd->rdVisibleC1= docVisV1;
    rd->rdMaxUsed= docBackV1;

    XClearArea( display, win, 0, 0, 0, 0, True );

    return;
    }

/************************************************************************/
/*									*/
/*  Let the user drag one of the controls on the top ruler.		*/
/*									*/
/*  1)  Draw a 'hair' over the document that corresponds to the		*/
/*	position of the control.					*/
/*  2)  Until the mouse button is released, observe events.		*/
/*  3)  Merge as many mouse motion events as possible.			*/
/*  4)  If the event is a mouse motion or a button release, redraw the	*/
/*	old an new position of the control on the ruler. (Assume that	*/
/*	the buttons are wider than the tab marks.			*/
/*  5)  Let the document move the hair.					*/
/*  6)  When the mouse button is released, the loop ends.		*/
/*  7)  Let the toolkit dispatch all events that are not relevant for	*/
/*	this routine.							*/
/*									*/
/************************************************************************/

static void appHorizontalRulerDragItem(		int *		pValue,
						int *		pChanged,
						MetricRuler *	mr,
						int		ox,
						void *		voided,
						DragHair	dragHair,
						int		asUpper,
						XEvent *	event,
						Widget		w,
						Window		win,
						Display *	display )
    {
    RulerData *			rd= &(mr->mrRulerData);
    int				value= *pValue;
    DocumentRectangle		drClip;

    /*  1  */
    (*dragHair)( voided, -1, value, asUpper );

    /*  2  */
    for (;;)
	{
	/*  3  */
	if  ( QLength( display ) > 0 )
	    {
	    while( QLength( display ) > 0 )
		{
		XNextEvent( display, event );

		if  ( event->type != MotionNotify )
		    { break;	}
		}
	    }
	else{ XNextEvent( display, event );	}

	if  ( event->type == MotionNotify )
	    { *pChanged= 1;	}

	/*  4  */
	if  ( event->type == ButtonRelease	||
	      event->type == MotionNotify	)
	    {
	    int		newValue= event->xbutton.x+ ox;

	    if  ( newValue < rd->rdDocumentC0 )
		{ newValue= rd->rdDocumentC0;	}
	    if  ( newValue > rd->rdDocumentC1 )
		{ newValue= rd->rdDocumentC1;	}

	    if  ( *pValue == newValue && event->type == MotionNotify )
		{ continue;	}
		
	    *pValue= newValue;

	    if  ( value < newValue )
		{
		drClip.drX0= value- rd->rdSizeAcross;
		drClip.drX1= newValue+ rd->rdSizeAcross;
		drClip.drY0= 0;
		drClip.drY1= rd->rdSizeAcross;
		}
	    else{
		drClip.drX0= newValue- rd->rdSizeAcross;
		drClip.drX1= value+ rd->rdSizeAcross;
		drClip.drY0= 0;
		drClip.drY1= rd->rdSizeAcross;
		}

	    appDrawHorizontalRuler( w, mr, &drClip, ox, win, display );

	    /*  5  */
	    (*dragHair)( voided, value, newValue, asUpper );

	    value= newValue;
	    }

	if  ( event->type == MotionNotify )
	    { continue;	}

	/*  6  */
	if  ( event->type == ButtonRelease )
	    { (*dragHair)( voided, *pValue, -1, asUpper ); return; }

	/*  7  */
	XtDispatchEvent( event );
	}
    }

/************************************************************************/
/*									*/
/*  The user clicked on the top ruler, make her happy.			*/
/*									*/
/*  1)  Only accept 'left mouse down' events.				*/
/*									*/
/************************************************************************/

void     appHorizontalRulerTrackMouse(	Widget		w,
					void *		voidmr,
					int *		pChanged,
					XEvent *	pEvent,
					int *		pX0pixels,
					int *		pX1pixels,
					void *		voided,
					DragHair	dragHair )
    {
    Display *		display= XtDisplay( w );
    Window		win= XtWindow( w );

    MetricRuler  *	mr= (MetricRuler *)voidmr;
    RulerData *		rd= &(mr->mrRulerData);

    int			ox= rd->rdVisibleC0- rd->rdMinUsed;

    XEvent		event= *pEvent;
    XButtonEvent *	bevent= &(event.xbutton);

    int			changed= 0;
    int			asUpper;

    int			xp= 2;

    /*  1  */
    if  ( event.type != ButtonPress	||
	  bevent->button != Button1	)
	{ return; }


    if  ( mr->mrV0Pixels >= 0					&&
	  bevent->x+ ox <  mr->mrV0Pixels + xp			&&
	  bevent->x+ ox >= mr->mrV0Pixels- HAIR_THICK( sizeAcross ) -xp	)
	{
	asUpper= 0;

	appHorizontalRulerDragItem( &mr->mrV0Pixels, &changed,
					    mr, ox, voided, dragHair, asUpper,
					    &event, w, win, display );
	goto ready;
	}

    if  ( mr->mrV1Pixels >= 0					&&
	  bevent->x+ ox >= mr->mrV1Pixels- xp			&&
	  bevent->x+ ox <  mr->mrV1Pixels+ HAIR_THICK( sizeAcross )+ xp	)
	{
	asUpper= 1;

	appHorizontalRulerDragItem( &mr->mrV1Pixels, &changed,
					    mr, ox, voided, dragHair, asUpper,
					    &event, w, win, display );
	goto ready;
	}

    if  ( mr->mrV0Pixels >= 0 )
	{
	if  ( mr->mrV1Pixels >= 0 )
	    { return;	}

	if  ( bevent->x+ ox < mr->mrV1Pixels )
	    {
	    asUpper= 0;
	    mr->mrV1Pixels= mr->mrV0Pixels;

	    mr->mrV0Pixels= bevent->x+ ox;

	    changed= 1;
	    (*dragHair)( voided, -1, bevent->x+ ox, asUpper );

	    appHorizontalRulerDragItem( &mr->mrV0Pixels, &changed,
					    mr, ox, voided, dragHair, asUpper,
					    &event, w, win, display );
	    goto ready;
	    }
	else{
	    asUpper= 1;
	    mr->mrV1Pixels= bevent->x+ ox;

	    changed= 1;
	    (*dragHair)( voided, -1, bevent->x+ ox, asUpper );

	    appHorizontalRulerDragItem( &mr->mrV1Pixels, &changed,
					    mr, ox, voided, dragHair, asUpper,
					    &event, w, win, display );
	    goto ready;
	    }
	}

    if  ( mr->mrV1Pixels >= 0 )
	{
	if  ( bevent->x+ ox > mr->mrV1Pixels )
	    {
	    asUpper= 1;
	    mr->mrV0Pixels= mr->mrV1Pixels;

	    mr->mrV1Pixels= bevent->x+ ox;

	    changed= 1;
	    (*dragHair)( voided, -1, bevent->x+ ox, asUpper );

	    appHorizontalRulerDragItem( &mr->mrV1Pixels, &changed,
					    mr, ox, voided, dragHair, asUpper,
					    &event, w, win, display );
	    goto ready;
	    }
	else{
	    asUpper= 0;
	    mr->mrV0Pixels= bevent->x+ ox;

	    changed= 1;
	    (*dragHair)( voided, -1, bevent->x+ ox, asUpper );

	    appHorizontalRulerDragItem( &mr->mrV0Pixels, &changed,
					    mr, ox, voided, dragHair, asUpper,
					    &event, w, win, display );
	    goto ready;
	    }
	}

    if  ( bevent->x+ ox <= ( rd->rdMinUsed+ rd->rdMaxUsed )/ 2 )
	{
	asUpper= 0;
	mr->mrV0Pixels= bevent->x+ ox;

	changed= 1;
	(*dragHair)( voided, -1, bevent->x+ ox, asUpper );

	appHorizontalRulerDragItem( &mr->mrV0Pixels, &changed,
					    mr, ox, voided, dragHair, asUpper,
					    &event, w, win, display );
	}
    else{
	asUpper= 1;
	mr->mrV1Pixels= bevent->x+ ox;

	changed= 1;
	(*dragHair)( voided, -1, bevent->x+ ox, asUpper );

	appHorizontalRulerDragItem( &mr->mrV1Pixels, &changed,
					    mr, ox, voided, dragHair, asUpper,
					    &event, w, win, display );
	}

  ready:
    if  ( mr->mrV0Pixels >= 0			&&
	  mr->mrV1Pixels >= 0			&&
	  mr->mrV0Pixels > mr->mrV1Pixels	)
	{
	int		swap;

	changed= 1;

	(*dragHair)( voided, mr->mrV1Pixels, -1, 1 );
	(*dragHair)( voided, mr->mrV0Pixels, -1, 0 );

	swap= mr->mrV1Pixels;
	mr->mrV1Pixels= mr->mrV0Pixels;
	mr->mrV0Pixels= swap;

	(*dragHair)( voided, -1, mr->mrV1Pixels, 1 );
	(*dragHair)( voided, -1, mr->mrV0Pixels, 0 );
	}

    if  ( changed )
	{
	*pChanged= 1;
	XClearArea( display, win, 0, 0, 0, 0, True );
	}

    *pX0pixels= mr->mrV0Pixels;
    *pX1pixels= mr->mrV1Pixels;

    return;
    }

/************************************************************************/
/*									*/
/*  Let the user drag one of the controls on the left ruler.		*/
/*									*/
/*  1)  Draw a 'hair' over the document that corresponds to the		*/
/*	position of the control.					*/
/*  2)  Until the mouse button is released, observe events.		*/
/*  3)  Merge as many mouse motion events as possible.			*/
/*  4)  If the event is a mouse motion or a button release, redraw the	*/
/*	old an new position of the control on the ruler. (Assume that	*/
/*	the buttons are wider than the tab marks.			*/
/*  5)  Let the document move the hair.					*/
/*  6)  When the mouse button is released, the loop ends.		*/
/*  7)  Let the toolkit dispatch all events that are not relevant for	*/
/*	this routine.							*/
/*									*/
/************************************************************************/

static void appVerticalRulerDragItem(	int *		pValue,
					int *		pChanged,
					MetricRuler *	mr,
					int		oy,
					void *		voided,
					DragHair	dragHair,
					int		asUpper,
					XEvent *	event,
					Widget		w,
					Window		win,
					Display *	display )
    {
    RulerData *			rd= &(mr->mrRulerData);
    int				value= *pValue;
    DocumentRectangle		drClip;

    /*  1  */
    (*dragHair)( voided, -1, value, asUpper );

    /*  2  */
    for (;;)
	{
	/*  3  */
	if  ( QLength( display ) > 0 )
	    {
	    while( QLength( display ) > 0 )
		{
		XNextEvent( display, event );

		if  ( event->type != MotionNotify )
		    { break;	}
		}
	    }
	else{ XNextEvent( display, event );	}

	if  ( event->type == MotionNotify )
	    { *pChanged= 1;	}

	/*  4  */
	if  ( event->type == ButtonRelease	||
	      event->type == MotionNotify	)
	    {
	    int		newValue= event->xbutton.y+ oy;

	    if  ( newValue < rd->rdDocumentC0 )
		{ newValue= rd->rdDocumentC0;	}
	    if  ( newValue > rd->rdDocumentC1 )
		{ newValue= rd->rdDocumentC1;	}

	    if  ( *pValue == newValue && event->type == MotionNotify )
		{ continue;	}
		
	    *pValue= newValue;

	    if  ( value < newValue )
		{
		drClip.drX0= 0;
		drClip.drX1= rd->rdSizeAcross;
		drClip.drY0= value- rd->rdSizeAcross;
		drClip.drY1= newValue+ rd->rdSizeAcross;
		}
	    else{
		drClip.drX0= 0;
		drClip.drX1= rd->rdSizeAcross;
		drClip.drY0= newValue- rd->rdSizeAcross;
		drClip.drY1= value+ rd->rdSizeAcross;
		}

	    appDrawVerticalRuler( w, mr, &drClip, oy, win, display );

	    /*  5  */
	    (*dragHair)( voided, value, newValue, asUpper );

	    value= newValue;
	    }

	if  ( event->type == MotionNotify )
	    { continue;	}

	/*  6  */
	if  ( event->type == ButtonRelease )
	    { (*dragHair)( voided, *pValue, -1, asUpper ); return; }

	/*  7  */
	XtDispatchEvent( event );
	}
    }

/************************************************************************/
/*									*/
/*  The user clicked on the left ruler, make her happy.			*/
/*									*/
/*  1)  Only accept 'left mouse down' events.				*/
/*									*/
/************************************************************************/

void     appVerticalRulerTrackMouse(	Widget		w,
					void *		voidmr,
					int *		pChanged,
					XEvent *	pEvent,
					int *		pY0pixels,
					int *		pY1pixels,
					void *		voided,
					DragHair	dragHair )
    {
    Display *		display= XtDisplay( w );
    Window		win= XtWindow( w );

    MetricRuler  *	mr= (MetricRuler *)voidmr;
    RulerData *		rd= &(mr->mrRulerData);

    int			oy= rd->rdVisibleC0- rd->rdMinUsed;

    XEvent		event= *pEvent;
    XButtonEvent *	bevent= &(event.xbutton);

    int			changed= 0;
    int			asUpper;

    int			xp= 2;

    /*  1  */
    if  ( event.type != ButtonPress	||
	  bevent->button != Button1	)
	{ return; }


    if  ( mr->mrV0Pixels >= 0					&&
	  bevent->y+ oy <  mr->mrV0Pixels + xp			&&
	  bevent->y+ oy >= mr->mrV0Pixels- HAIR_THICK( sizeAcross ) -xp	)
	{
	asUpper= 0;

	appVerticalRulerDragItem( &mr->mrV0Pixels, &changed,
					    mr, oy, voided, dragHair, asUpper,
					    &event, w, win, display );
	goto ready;
	}

    if  ( mr->mrV1Pixels >= 0					&&
	  bevent->y+ oy >= mr->mrV1Pixels- xp			&&
	  bevent->y+ oy <  mr->mrV1Pixels+ HAIR_THICK( sizeAcross )+ xp	)
	{
	asUpper= 1;

	appVerticalRulerDragItem( &mr->mrV1Pixels, &changed,
					    mr, oy, voided, dragHair, asUpper,
					    &event, w, win, display );
	goto ready;
	}

    if  ( mr->mrV0Pixels >= 0 )
	{
	if  ( mr->mrV1Pixels >= 0 )
	    { return;	}

	if  ( bevent->y+ oy < mr->mrV1Pixels )
	    {
	    asUpper= 0;
	    mr->mrV1Pixels= mr->mrV0Pixels;

	    mr->mrV0Pixels= bevent->y+ oy;

	    changed= 1;
	    (*dragHair)( voided, -1, bevent->y+ oy, asUpper );

	    appVerticalRulerDragItem( &mr->mrV0Pixels, &changed,
					    mr, oy, voided, dragHair, asUpper,
					    &event, w, win, display );
	    goto ready;
	    }
	else{
	    asUpper= 1;
	    mr->mrV1Pixels= bevent->y+ oy;

	    changed= 1;
	    (*dragHair)( voided, -1, bevent->y+ oy, asUpper );

	    appVerticalRulerDragItem( &mr->mrV1Pixels, &changed,
					    mr, oy, voided, dragHair, asUpper,
					    &event, w, win, display );
	    goto ready;
	    }
	}

    if  ( mr->mrV1Pixels >= 0 )
	{
	if  ( bevent->y+ oy > mr->mrV1Pixels )
	    {
	    asUpper= 1;
	    mr->mrV0Pixels= mr->mrV1Pixels;

	    mr->mrV1Pixels= bevent->y+ oy;

	    changed= 1;
	    (*dragHair)( voided, -1, bevent->y+ oy, asUpper );

	    appVerticalRulerDragItem( &mr->mrV1Pixels, &changed,
					    mr, oy, voided, dragHair, asUpper,
					    &event, w, win, display );
	    goto ready;
	    }
	else{
	    asUpper= 0;
	    mr->mrV0Pixels= bevent->y+ oy;

	    changed= 1;
	    (*dragHair)( voided, -1, bevent->y+ oy, asUpper );

	    appVerticalRulerDragItem( &mr->mrV0Pixels, &changed,
					    mr, oy, voided, dragHair, asUpper,
					    &event, w, win, display );
	    goto ready;
	    }
	}

    if  ( bevent->x+ oy <= ( rd->rdMinUsed+ rd->rdMaxUsed )/ 2 )
	{
	asUpper= 0;
	mr->mrV0Pixels= bevent->y+ oy;

	changed= 1;
	(*dragHair)( voided, -1, bevent->y+ oy, asUpper );

	appVerticalRulerDragItem( &mr->mrV0Pixels, &changed,
					    mr, oy, voided, dragHair, asUpper,
					    &event, w, win, display );
	}
    else{
	asUpper= 1;
	mr->mrV1Pixels= bevent->y+ oy;

	changed= 1;
	(*dragHair)( voided, -1, bevent->y+ oy, asUpper );

	appVerticalRulerDragItem( &mr->mrV1Pixels, &changed,
					    mr, oy, voided, dragHair, asUpper,
					    &event, w, win, display );
	}

  ready:
    if  ( mr->mrV0Pixels >= 0			&&
	  mr->mrV1Pixels >= 0			&&
	  mr->mrV0Pixels > mr->mrV1Pixels	)
	{
	int		swap;

	changed= 1;

	(*dragHair)( voided, mr->mrV1Pixels, -1, 1 );
	(*dragHair)( voided, mr->mrV0Pixels, -1, 0 );

	swap= mr->mrV1Pixels;
	mr->mrV1Pixels= mr->mrV0Pixels;
	mr->mrV0Pixels= swap;

	(*dragHair)( voided, -1, mr->mrV1Pixels, 1 );
	(*dragHair)( voided, -1, mr->mrV0Pixels, 0 );
	}

    if  ( changed )
	{
	*pChanged= 1;
	XClearArea( display, win, 0, 0, 0, 0, True );
	}

    *pY0pixels= mr->mrV0Pixels;
    *pY1pixels= mr->mrV1Pixels;

    return;
    }

/************************************************************************/
/*									*/
/*  Set the selection of a horizontal/vertical ruler.			*/
/*									*/
/************************************************************************/

void appSetMetricRulerValues(	Widget		w,
				void *		voidmr,
				int		v0,
				int		v1 )
    {
    Display *		display= XtDisplay( w );
    Window		win= XtWindow( w );

    MetricRuler  *	mr= (MetricRuler *)voidmr;

    int			changed= 0;

    if  ( v0 != mr->mrV0Pixels )
	{ mr->mrV0Pixels= v0; changed= 1; }

    if  ( v1 != mr->mrV1Pixels )
	{ mr->mrV1Pixels= v1; changed= 1; }

    if  ( changed )
	{ XClearArea( display, win, 0, 0, 0, 0, True );	}
    }

void appSetMetricRulerMagnification(	Widget		w,
				void *		voidmr,
				double		pixPerMM,
				double		magnification )
    {
    Display *			display= XtDisplay( w );
    Window			win= XtWindow( w );

    MetricRuler  *		mr= (MetricRuler *)voidmr;
    RulerData *			rd= &(mr->mrRulerData);

    appRulerCalculateIncrements( rd, pixPerMM, magnification );

    XClearArea( display, win, 0, 0, 0, 0, True );

    return;
    }

void appAdaptMetricRuler(	Widget		w,
			void *		voidmr,
			int		documentC1,
			int		rulerC1 )
    {
    MetricRuler  *	mr= (MetricRuler *)voidmr;
    RulerData *		rd= &(mr->mrRulerData);

    rd->rdDocumentC1= documentC1;

    if  ( rd->rdRulerC1 != rulerC1 )
	{
	Display *	display= XtDisplay( w );
	Window		win= XtWindow( w );

	rd->rdRulerC1= rulerC1;

	XClearArea( display, win, 0, 0, 0, 0, True );
	}
    }

