/************************************************************************/
/*									*/
/*  A Page Layout tool.							*/
/*									*/
/************************************************************************/

#   include	"config.h"

#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<debugon.h>

#   include	<X11/Xatom.h>
#   include	<X11/IntrinsicP.h>

#   include	<Xm/Text.h>
#   include	<Xm/DrawingA.h>
#   include	<Xm/Form.h>
#   include	<Xm/PanedW.h>

#   include	<Xm/MwmUtil.h>
#   include	<Xm/Protocols.h>


#   include	<appFrame.h>
#   include	"appUnit.h"
#   include	<appGeoString.h>
#   include	"appPaper.h"
#   include	"appPaperChooser.h"

#   include	"appUtil.h"

#   define	DRH_CM	4.5

/************************************************************************/
/*									*/
/*  Represents A Page Layout tool:					*/
/*									*/
/************************************************************************/
#   define	FILEL	400

typedef struct AppPageToolResources
    {
    char *	aptrUnitType;

    char *	aptrPaper;
    char *	aptrCustom;
    char *	aptrPortrait;
    char *	aptrLandscape;
    char *	aptrMargins;
    char *	aptrLeftMargin;
    char *	aptrTopMargin;
    char *	aptrRightMargin;
    char *	aptrBottomMargin;
    char *	aptrRevert;
    char *	aptrSet;
#   if 0
    For use as a dialog.
    char *	aptrOk;
    char *	aptrCancel;
#   endif
    } AppPageToolResources;

typedef struct AppPageTool
    {
    GC				aptGc;

    Widget			aptTopWidget;
    Widget			aptMainWidget;

    Widget			aptPageDrawing;

    PaperChooser		aptPaperChooser;

    Widget			aptLeftMarginText;
    Widget			aptTopMarginText;
    Widget			aptRightMarginText;
    Widget			aptBottomMarginText;

    AppToolDestroy		aptDestroy;
    PageToolSetProperties	aptSetProperties;

    DocumentGeometry		aptGeometryChosen;
    DocumentGeometry		aptGeometrySet;

    int				aptUnitType;
    int				aptCustomPaperSize;
    void *			aptTarget;
    } AppPageTool;


# define xx(x)	x,x

static XtResource APP_PageToolresourceTable[]=
    {
	{ xx("unit"), XtRString, sizeof(char *),
		    offsetof(AppPageToolResources,aptrUnitType),
		    XtRString, "inch" },
	{ xx("pageToolPaper"), XtRString, sizeof(char *),
		    offsetof(AppPageToolResources,aptrPaper),
		    XtRString, "Paper" },
	{ xx("pageToolCustom"), XtRString, sizeof(char *),
		    offsetof(AppPageToolResources,aptrCustom),
		    XtRString, "Custom" },
	{ xx("pageToolPortrait"), XtRString, sizeof(char *),
		    offsetof(AppPageToolResources,aptrPortrait),
		    XtRString, "Portrait" },
	{ xx("pageToolLandscape"), XtRString, sizeof(char *),
		    offsetof(AppPageToolResources,aptrLandscape),
		    XtRString, "Landscape" },
	{ xx("pageToolMargins"), XtRString, sizeof(char *),
		    offsetof(AppPageToolResources,aptrMargins),
		    XtRString, "Margins" },
	{ xx("pageToolLeftMargin"), XtRString, sizeof(char *),
		    offsetof(AppPageToolResources,aptrLeftMargin),
		    XtRString, "Left" },
	{ xx("pageToolTopMargin"), XtRString, sizeof(char *),
		    offsetof(AppPageToolResources,aptrTopMargin),
		    XtRString, "Top" },
	{ xx("pageToolRightMargin"), XtRString, sizeof(char *),
		    offsetof(AppPageToolResources,aptrRightMargin),
		    XtRString, "Right" },
	{ xx("pageToolBottomMargin"), XtRString, sizeof(char *),
		    offsetof(AppPageToolResources,aptrBottomMargin),
		    XtRString, "Bottom" },
	{ xx("pageToolRevert"), XtRString, sizeof(char *),
		    offsetof(AppPageToolResources,aptrRevert),
		    XtRString, "Revert" },
	{ xx("pageToolSet"), XtRString, sizeof(char *),
		    offsetof(AppPageToolResources,aptrSet),
		    XtRString, "Set" },
#	if 0
	For use as a Dialog.
	{ xx("pageToolOk"), XtRString, sizeof(char *),
		    offsetof(AppPageToolResources,aptrOk),
		    XtRString, "Ok" },
	{ xx("pageToolCancel"), XtRString, sizeof(char *),
		    offsetof(AppPageToolResources,aptrCancel),
		    XtRString, "Cancel" },
#	endif
    };

/************************************************************************/
/*									*/
/*  Show the size of the paper in the Text.				*/
/*									*/
/************************************************************************/

static void appPageToolShowMargins(	AppPageTool *	apt )
    {
    char			scratch[50];
    int				unitInt= apt->aptUnitType;
    const DocumentGeometry *	dg= &(apt->aptGeometryChosen);

    appGeoLengthToString( scratch, dg->dgLeftMarginTwips, unitInt );
    XmTextSetString( apt->aptLeftMarginText, scratch );

    appGeoLengthToString( scratch, dg->dgTopMarginTwips, unitInt );
    XmTextSetString( apt->aptTopMarginText, scratch );

    appGeoLengthToString( scratch, dg->dgRightMarginTwips, unitInt );
    XmTextSetString( apt->aptRightMarginText, scratch );

    appGeoLengthToString( scratch, dg->dgBottomMarginTwips, unitInt );
    XmTextSetString( apt->aptBottomMarginText, scratch );

    return;
    }

/************************************************************************/
/*									*/
/*  Draw a schematic view of the page.					*/
/*									*/
/************************************************************************/
static void appPageToolDrawPage(	Widget		w,
					XtPointer	voidapt,
					XtPointer	voidcbs	)
    {
    AppPageTool *		apt= (AppPageTool *)voidapt;

    appDrawPageDiagram( w, apt->aptGc, DRH_CM, &(apt->aptGeometryChosen) );
    }

/************************************************************************/
/*									*/
/*  A paper size was chosen.						*/
/*									*/
/************************************************************************/

static void appPageToolSizeChosen(	Widget		w,
					XtPointer	voidapt,
					XtPointer	voidpbcs	 )
    {
    AppPageTool *		apt= (AppPageTool *)voidapt;

    appPaperSizeChosen( w, &(apt->aptPaperChooser),
						&(apt->aptGeometryChosen),
						apt->aptUnitType );

    XClearArea( XtDisplay( apt->aptPageDrawing ),
				    XtWindow( apt->aptPageDrawing ),
				    0, 0, 0, 0, True );

    return;
    }

static void appPageToolPaperSizeChanged(	Widget		w,
						XtPointer	voidapt,
						XtPointer	voidcbs	 )
    {
    AppPageTool *		apt= (AppPageTool *)voidapt;
    const int			set= 0;

    appPaperSizeChanged( &(apt->aptPaperChooser),
						&(apt->aptGeometryChosen),
						apt->aptUnitType );

    appPaperChooserAdaptToGeometry( &(apt->aptPaperChooser),
			apt->aptUnitType, &(apt->aptGeometryChosen), set );

    XClearArea( XtDisplay( apt->aptPageDrawing ),
				    XtWindow( apt->aptPageDrawing ),
				    0, 0, 0, 0, True );

    return;
    }

/************************************************************************/
/*									*/
/*  The buttons have been pushed.					*/
/*									*/
/************************************************************************/

static void appPageToolRevertPushed(	Widget		w,
					XtPointer	voidapt,
					XtPointer	voidpbcs	 )
    {
    AppPageTool *		apt= (AppPageTool *)voidapt;
    const int			set= 0;

    apt->aptGeometryChosen.dgPaperWideTwips=
					apt->aptGeometrySet.dgPaperWideTwips;
    apt->aptGeometryChosen.dgPaperHighTwips=
					apt->aptGeometrySet.dgPaperHighTwips;

    appPaperChooserAdaptToGeometry( &(apt->aptPaperChooser),
			    apt->aptUnitType, &(apt->aptGeometryChosen), set );

    apt->aptGeometryChosen.dgLeftMarginTwips=
					apt->aptGeometrySet.dgLeftMarginTwips;
    apt->aptGeometryChosen.dgTopMarginTwips=
					apt->aptGeometrySet.dgTopMarginTwips;
    apt->aptGeometryChosen.dgRightMarginTwips=
					apt->aptGeometrySet.dgRightMarginTwips;
    apt->aptGeometryChosen.dgBottomMarginTwips=
					apt->aptGeometrySet.dgBottomMarginTwips;
    appPageToolShowMargins( apt );

    XClearArea( XtDisplay( apt->aptPageDrawing ),
				    XtWindow( apt->aptPageDrawing ),
				    0, 0, 0, 0, True );
    return;
    }

/************************************************************************/
/*									*/
/*  Get one/all margins from the text widgets.				*/
/*									*/
/************************************************************************/

static int appPageToolGetMargins(	AppPageTool *		apt,
					DocumentGeometry *	dg )
    {
    int		changed;
    const int	requirePositive= 1;

    if  ( appGetLengthFromTextWidget( apt->aptLeftMarginText,
				    &dg->dgLeftMarginTwips, &changed,
				    apt->aptUnitType, requirePositive ) )
	{ return -1;	}

    if  ( appGetLengthFromTextWidget( apt->aptTopMarginText,
				    &dg->dgTopMarginTwips, &changed,
				    apt->aptUnitType, requirePositive ) )
	{ return -1;	}

    if  ( appGetLengthFromTextWidget( apt->aptRightMarginText,
				    &dg->dgRightMarginTwips, &changed,
				    apt->aptUnitType, requirePositive ) )
	{ return -1;	}

    if  ( appGetLengthFromTextWidget( apt->aptBottomMarginText,
				    &dg->dgBottomMarginTwips, &changed,
				    apt->aptUnitType, requirePositive ) )
	{ return -1;	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Check margins.							*/
/*									*/
/************************************************************************/

static int appPageToolCheckMargins(	DocumentGeometry *	dg )
    {
    if  ( dg->dgLeftMarginTwips+ dg->dgRightMarginTwips >= 
						    dg->dgPaperWideTwips )
	{ return -1;	}

    if  ( dg->dgTopMarginTwips+ dg->dgBottomMarginTwips >= 
						    dg->dgPaperWideTwips )
	{ return -1;	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  The 'Set' button has been pushed:					*/
/*									*/
/*	Retrieve the margins and the paper size.			*/
/*	Check the values						*/
/*	If things are ok, report the values back to the application.	*/
/*									*/
/************************************************************************/

static void appPageToolSetPushed(	Widget		w,
					XtPointer	voidapt,
					XtPointer	voidpbcs	 )
    {
    AppPageTool *	apt= (AppPageTool *)voidapt;

    DocumentGeometry	dg;

    if  ( appPageToolGetMargins( apt, &dg ) )
	{ return;	}

    if  ( appPaperChooserGetSize( &(apt->aptPaperChooser),
						    apt->aptUnitType, &dg ) )
	{ LDEB(1); return;	}

    if  ( appPageToolCheckMargins( &dg ) )
	{ return;	}

    if  ( ! apt->aptSetProperties )
	{ XDEB(apt->aptSetProperties); return;	}

    appPageToolSetProperties( (void *)apt, &dg );

    (*apt->aptSetProperties)( apt->aptTarget, &dg );

    return;
    }

/************************************************************************/
/*									*/
/*  The page layout tool is closed by the user.				*/
/*									*/
/************************************************************************/

static void appClosePageTool(		Widget		w,
					XtPointer	voidapt,
					XtPointer	voidlcs	 )
    {
    AppPageTool *	apt= (AppPageTool *)voidapt;

    if  ( apt->aptDestroy )
	{ (*apt->aptDestroy)( apt->aptTarget );	}

    if  ( apt->aptGc )
	{ XFreeGC( XtDisplay( w ), apt->aptGc );	}

    XtDestroyWidget( w );

    free( apt );

    return;
    }

/************************************************************************/
/*									*/
/*  One of the margins has been changed by the user.			*/
/*									*/
/************************************************************************/
static void appPageToolMarginChanged(	Widget		w,
					XtPointer	voidapt,
					XtPointer	voidcbs	 )
    {
    AppPageTool *			apt= (AppPageTool *)voidapt;

    DocumentGeometry			dg;

    dg= apt->aptGeometryChosen;

    if  ( appPageToolGetMargins( apt, &dg ) )
	{ return;	}

    if  ( appPageToolCheckMargins( &dg ) )
	{ return;	}

    apt->aptGeometryChosen.dgLeftMarginTwips= dg.dgLeftMarginTwips;
    apt->aptGeometryChosen.dgTopMarginTwips= dg.dgTopMarginTwips;
    apt->aptGeometryChosen.dgRightMarginTwips= dg.dgRightMarginTwips;
    apt->aptGeometryChosen.dgBottomMarginTwips= dg.dgBottomMarginTwips;

    XClearArea( XtDisplay( apt->aptPageDrawing ),
				    XtWindow( apt->aptPageDrawing ),
				    0, 0, 0, 0, True );

    return;
    }

/************************************************************************/
/*									*/
/*  One of the orientation toggle buttons has been pushed.		*/
/*									*/
/************************************************************************/

static void appPageToolOrientationChosen(	Widget		w,
						XtPointer	voidapt,
						XtPointer	voidtbcs  )
    {
    AppPageTool *			apt= (AppPageTool *)voidapt;
    int					changed;

    appPaperOrientationChosen( w, voidtbcs, &changed,
						&(apt->aptPaperChooser),
						apt->aptUnitType,
						&(apt->aptGeometryChosen) );

    if  ( changed )
	{
	XClearArea( XtDisplay( apt->aptPageDrawing ),
				    XtWindow( apt->aptPageDrawing ),
				    0, 0, 0, 0, True );
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Make the frame for the margins.					*/
/*									*/
/************************************************************************/

static Widget appPageToolMakeMarginFrame( XtAppContext		appContext,
					Widget			parent,
					Display *		display,
					AppPageToolResources *	aptr,
					AppPageTool *		apt )
    {
    Widget	marginFrame;

    Widget	insideWidget;
    Widget	leftColumn;
    Widget	rightColumn;

    Widget	label;
    Widget	row;

    Arg		al[20];
    int		ac;

    appMakeFrame( &marginFrame, parent, aptr->aptrMargins );

    ac= 0;
    XtSetArg( al[ac],	XmNfractionBase,	2 ); ac++;
    XtSetArg( al[ac],	XmNallowResize,		True ); ac++;
    XtSetArg( al[ac],	XmNskipAdjust,		True ); ac++;
    insideWidget= XmCreateForm( marginFrame, WIDGET_NAME, al, ac );

    ac= 0;
    XtSetArg( al[ac],	XmNtopAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNbottomAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNleftAttachment,	XmATTACH_POSITION ); ac++;
    XtSetArg( al[ac],	XmNrightAttachment,	XmATTACH_POSITION ); ac++;
    XtSetArg( al[ac],	XmNleftPosition,	0 ); ac++;
    XtSetArg( al[ac],	XmNrightPosition,	1 ); ac++;

    XtSetArg( al[ac],	XmNsashWidth,		1 ); ac++;
    XtSetArg( al[ac],	XmNsashHeight,		1 ); ac++;
    XtSetArg( al[ac],	XmNseparatorOn,		False ); ac++;
    XtSetArg( al[ac],	XmNmarginWidth,		5 ); ac++;
    XtSetArg( al[ac],	XmNmarginHeight,	5 ); ac++;
    XtSetArg( al[ac],	XmNspacing,		5 ); ac++;
    XtSetArg( al[ac],	XmNallowResize,		True ); ac++;
    leftColumn= XmCreatePanedWindow( insideWidget, WIDGET_NAME, al, ac );

    ac= 0;
    XtSetArg( al[ac],	XmNtopAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNbottomAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNleftAttachment,	XmATTACH_POSITION ); ac++;
    XtSetArg( al[ac],	XmNrightAttachment,	XmATTACH_POSITION ); ac++;
    XtSetArg( al[ac],	XmNleftPosition,	1 ); ac++;
    XtSetArg( al[ac],	XmNrightPosition,	2 ); ac++;

    XtSetArg( al[ac],	XmNsashWidth,		1 ); ac++;
    XtSetArg( al[ac],	XmNsashHeight,		1 ); ac++;
    XtSetArg( al[ac],	XmNseparatorOn,		False ); ac++;
    XtSetArg( al[ac],	XmNmarginWidth,		5 ); ac++;
    XtSetArg( al[ac],	XmNmarginHeight,	5 ); ac++;
    XtSetArg( al[ac],	XmNspacing,		5 ); ac++;
    XtSetArg( al[ac],	XmNallowResize,		True ); ac++;
    rightColumn= XmCreatePanedWindow( insideWidget, WIDGET_NAME, al, ac );

    /**********************/

    appMakeLabelAndTextRow( &row, &label, &(apt->aptLeftMarginText),
			    leftColumn, aptr->aptrLeftMargin, 8, True );
    appMakeLabelAndTextRow( &row, &label, &(apt->aptRightMarginText),
			    leftColumn, aptr->aptrRightMargin, 8, True );

    appMakeLabelAndTextRow( &row, &label, &(apt->aptTopMarginText),
			    rightColumn, aptr->aptrTopMargin, 8, True );
    appMakeLabelAndTextRow( &row, &label, &(apt->aptBottomMarginText),
			    rightColumn, aptr->aptrBottomMargin, 8, True );

    XtAddCallback( apt->aptLeftMarginText, XmNactivateCallback,
				    appPageToolMarginChanged, (void *)apt );

    XtAddCallback( apt->aptTopMarginText, XmNactivateCallback,
				    appPageToolMarginChanged, (void *)apt );

    XtAddCallback( apt->aptRightMarginText, XmNactivateCallback,
				    appPageToolMarginChanged, (void *)apt );

    XtAddCallback( apt->aptBottomMarginText, XmNactivateCallback,
				    appPageToolMarginChanged, (void *)apt );

    XtManageChild( leftColumn );
    XtManageChild( rightColumn );
    XtManageChild( insideWidget );
    XtManageChild( marginFrame );

    return marginFrame;
    }

/************************************************************************/
/*									*/
/*  Make the frame for paper size and orientation.			*/
/*									*/
/************************************************************************/

static void appPageToolMakePaperFrame(	Widget			parent,
					AppPageToolResources *	aptr,
					AppPageTool *		apt )
    {
    appMakePaperChooserWidgets( parent, aptr->aptrPaper,
						&(apt->aptPaperChooser),
						appPageToolPaperSizeChanged,
						(void *)apt );

    appPaperChooserAddOrientationToggles( &(apt->aptPaperChooser),
						aptr->aptrPortrait,
						aptr->aptrLandscape,
						appPageToolOrientationChosen,
						(void *)apt );

    appPaperChooserManageWidgets( &(apt->aptPaperChooser) );

    return;
    }

/************************************************************************/
/*									*/
/*  Make the location for the schematic drawing of the page.		*/
/*									*/
/************************************************************************/

static Widget appPageToolMakeDrawing(	Widget		parent,
					Display *	display,
					AppPageTool *	apt )
    {
    Widget	drawing;

    Arg		al[20];
    int		ac;

    int		screen= DefaultScreen( display );
    int		pixPerCM;

    pixPerCM= ( 10* DisplayHeight( display, screen ) )/
					DisplayHeightMM( display, screen );

    ac= 0;
    XtSetArg( al[ac],	XmNheight,		(int)(DRH_CM* pixPerCM) ); ac++;
    XtSetArg( al[ac],	XmNtraversalOn,		False ); ac++;
    XtSetArg( al[ac],	XmNsensitive,		False ); ac++;
    XtSetArg( al[ac],	XmNnavigationType,	XmNONE ); ac++;

    drawing= XmCreateDrawingArea( parent, WIDGET_NAME, al, ac );

    XtAddCallback( drawing, XmNexposeCallback,
					appPageToolDrawPage, (void *)apt );

    XtManageChild( drawing );

    return drawing;
    }

/************************************************************************/
/*									*/
/*  Make the form with the two buttons.					*/
/*									*/
/************************************************************************/
static Widget appPageToolMakeButtonRow( Widget			parent,
					AppPageToolResources *	aptr,
					AppPageTool *		apt )
    {
    Widget	row;
    Widget	revertButton;
    Widget	setButton;

    row= appMakeButtonRow( parent, 2 );

    /**********************/

    revertButton= appMakeRowButton( row, aptr->aptrRevert,
			    appPageToolRevertPushed, (void *)apt, 0, False );

    /**********************/

    setButton= appMakeRowButton( row, aptr->aptrSet,
			    appPageToolSetPushed, (void *)apt, 1, False );

    /**********************/

    XtManageChild( row );

    return row;
    }

/************************************************************************/
/*  The page tool has been closed, remember so.				*/
/************************************************************************/
static void appPageToolClosed(		void *		voidea	)
    {
    EditApplication *	ea=	(EditApplication *)voidea;

    ea->eaPageTool= (void *)0;
    }

/************************************************************************/
/*									*/
/*  make a page tool.							*/
/*									*/
/************************************************************************/

void * appMakePageTool(	EditApplication *	ea,
			Widget			pageOption,
			const char *		pixmapName,
			PageToolSetProperties	setProperties )
    {
    Display *		display= XtDisplay( pageOption );
    AppPageTool *	apt;
    
    Arg			al[20];
    int			ac= 0;

    Widget		marginFrame;
    Widget		buttonForm;

    MwmHints		hints;

    Pixmap		iconPixmap= (Pixmap)0;

    static int			gotResources= 0;
    static AppPageToolResources	aptr;

    if  ( ! gotResources )
	{
	XtGetApplicationResources( ea->eaTopWidget, &aptr,
				    APP_PageToolresourceTable,
				    XtNumber(APP_PageToolresourceTable),
				    NULL, 0 );

	gotResources= 1;
	}

    if  ( appGetImagePixmap( ea, pixmapName, &iconPixmap )  )
	{ SDEB(pixmapName); return (void *)0;	}

    hints.flags= MWM_HINTS_FUNCTIONS|MWM_HINTS_DECORATIONS;
    hints.functions=	MWM_FUNC_MOVE		|
			MWM_FUNC_MINIMIZE	|
			MWM_FUNC_CLOSE		;
    hints.decorations=	MWM_DECOR_BORDER	|
			MWM_DECOR_TITLE		|
			MWM_DECOR_MENU		|
			MWM_DECOR_MINIMIZE	;

    apt= (AppPageTool *)malloc( sizeof(AppPageTool) );
    if  ( ! apt )
	{ XDEB(apt); return (void *)0;	}

    apt->aptUnitType= appUnitTypeInt( aptr.aptrUnitType );
    if  ( apt->aptUnitType < 0 )
	{ SDEB(aptr.aptrUnitType); apt->aptUnitType= UNITtyINCH;	}

    apt->aptDestroy= appPageToolClosed;
    apt->aptSetProperties= setProperties;
    apt->aptTarget= (void *)ea;
    apt->aptCustomPaperSize= -1;

    appInitPaperChooser( &(apt->aptPaperChooser) );
    apt->aptGc= (GC)0;
    
    XtSetArg( al[ac], XmNuserData,	(void *)apt ); ac++;
    XtSetArg( al[ac], XmNdeleteResponse, XmDO_NOTHING ); ac++;
    XtSetArg( al[ac], XmNmwmDecorations, hints.decorations ); ac++;
    XtSetArg( al[ac], XmNmwmFunctions,	hints.functions ); ac++;
    XtSetArg( al[ac], XmNinput,		True ); ac++;

    if  ( iconPixmap )
	{ XtSetArg( al[ac], XmNiconPixmap,	iconPixmap ); ac++; }

    XtSetArg( al[ac], XmNwidth, 275 ); ac++;

#   ifdef USE_X11_R5
    apt->aptTopWidget= XtAppCreateShell( ea->eaApplicationName,
					    ea->eaPageToolName,
					    applicationShellWidgetClass,
					    ea->eaDisplay, al, ac );
#   else
    apt->aptTopWidget= XtAppCreateShell( ea->eaApplicationName,
					    ea->eaPageToolName,
					    topLevelShellWidgetClass,
					    ea->eaDisplay, al, ac );
#   endif


    appSetShellTitle( apt->aptTopWidget, pageOption, ea->eaApplicationName );

    XtAddEventHandler( apt->aptTopWidget, StructureNotifyMask, False,
					appFixSize, (void *)apt );


    if  ( ea->eaCloseAtom > 0 )
	{
	XmAddWMProtocolCallback( apt->aptTopWidget, ea->eaCloseAtom,
					appClosePageTool, (XtPointer)apt );
	}

    ac= 0;
    XtSetArg( al[ac], XmNuserData, (void *)apt ); ac++;

    ac= 0;
    XtSetArg( al[ac],	XmNuserData,		(void *)apt ); ac++;
    XtSetArg( al[ac],	XmNsashWidth,		1 ); ac++;
    XtSetArg( al[ac],	XmNsashHeight,		1 ); ac++;
    XtSetArg( al[ac],	XmNseparatorOn,		False ); ac++;
    XtSetArg( al[ac],	XmNmarginWidth,		5 ); ac++;
    XtSetArg( al[ac],	XmNmarginHeight,	5 ); ac++;
    XtSetArg( al[ac],	XmNspacing,		5 ); ac++;
    apt->aptMainWidget= XmCreatePanedWindow( apt->aptTopWidget, WIDGET_NAME, al, ac );

    apt->aptPageDrawing= appPageToolMakeDrawing( apt->aptMainWidget,
								display, apt );

    appPageToolMakePaperFrame( apt->aptMainWidget, &aptr, apt );

    marginFrame= appPageToolMakeMarginFrame( ea->eaContext, apt->aptMainWidget,
							display, &aptr, apt );
    buttonForm= appPageToolMakeButtonRow( apt->aptMainWidget, &aptr, apt );

    XtManageChild( apt->aptMainWidget );

    appPaperChooserFillMenu( &(apt->aptPaperChooser), aptr.aptrCustom,
					appPageToolSizeChosen, (void *)apt );

    XtRealizeWidget( apt->aptTopWidget );
    XtMapWidget( apt->aptTopWidget );

    appPaperChooserRetreshMenuWidth( &(apt->aptPaperChooser) );

    apt->aptGc= XCreateGC( XtDisplay( apt->aptPageDrawing ),
				    XtWindow( apt->aptPageDrawing ), 0, NULL );

    return (void *)apt;
    }

/************************************************************************/
/*									*/
/*  Draw a find tool to front.						*/
/*									*/
/************************************************************************/
void appShowPageTool(		void *	voidapt	)
    {
    AppPageTool *		apt= (AppPageTool *)voidapt;

    XtVaSetValues( apt->aptTopWidget, XmNinitialState, NormalState, NULL );
    XtMapWidget( apt->aptTopWidget );
    XRaiseWindow( XtDisplay( apt->aptTopWidget ),
					    XtWindow( apt->aptTopWidget ) );
    }

/************************************************************************/
/*									*/
/*  Set the properties in a page tool.					*/
/*									*/
/************************************************************************/


void appPageToolSetProperties(	void *				voidapt,
				const DocumentGeometry *	dg )
    {
    AppPageTool *	apt= (AppPageTool *)voidapt;
    const int		set= 1;

    apt->aptGeometryChosen= apt->aptGeometrySet= *dg;

    appPaperChooserAdaptToGeometry( &(apt->aptPaperChooser),
			    apt->aptUnitType, &(apt->aptGeometryChosen), set );

    appPageToolShowMargins( apt );

    XClearArea( XtDisplay( apt->aptPageDrawing ),
				    XtWindow( apt->aptPageDrawing ),
				    0, 0, 0, 0, True );

    return;
    }

void appEnablePageTool(		void *	voidapt,
				int	enabled )
    {
    AppPageTool *	apt= (AppPageTool *)voidapt;

    XtSetSensitive( apt->aptMainWidget, enabled != 0 );

    return;
    }
