/************************************************************************/
/*									*/
/*  Choose a paper format common functionality.				*/
/*									*/
/************************************************************************/

#   include	<Xm/Text.h>
#   include	<Xm/ToggleB.h>

#   include	<appGeoString.h>
#   include	<appPaper.h>

#   include	"appPaperChooser.h"
#   include	<appUtil.h>

#   include	<debugon.h>

/************************************************************************/
/*									*/
/*  Update the text widget.						*/
/*									*/
/************************************************************************/

void appPaperChooserShowWidthHeight(	PaperChooser *		pc,
				const DocumentGeometry *	dg,
				int				unitInt )
    {
    char			scratch[50];

    appGeoRectangleToString( scratch,
		    dg->dgPaperWideTwips, dg->dgPaperHighTwips, unitInt );

    XmTextSetString( pc->pcSizeText, scratch );

    return;
    }

/************************************************************************/
/*									*/
/*  Update the menu.							*/
/*									*/
/************************************************************************/

void appPaperChooserShowSize(	PaperChooser *			pc,
				int				sizeChosen )
    {
    WidgetList		children;
    Cardinal		childCount= 0;

    XtVaGetValues( pc->pcPulldown,
		    XmNchildren,	&children,
		    XmNnumChildren,	&childCount,
		    NULL );

    if  ( sizeChosen >= 0 && sizeChosen < (int)childCount )
	{
	XtVaSetValues( pc->pcMenu,
			    XmNmenuHistory,	children[sizeChosen],
			    NULL );
	}
    else{
	XtVaSetValues( pc->pcMenu,
			    XmNmenuHistory,	(Widget)0,
			    NULL );
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Extract a paper size from the Text widget.				*/
/*									*/
/************************************************************************/
int appPaperChooserGetSize(	PaperChooser *		pc,
				int			unitType,
				DocumentGeometry *	dg )
    {
    char *	s;

    s= XmTextGetString( pc->pcSizeText );

    if  ( appGeoRectangleFromString( s, unitType,
		    &(dg->dgPaperWideTwips), &(dg->dgPaperHighTwips) )	||
	  dg->dgPaperWideTwips <= 0					||
	  dg->dgPaperHighTwips <= 0					)
	{
	XmTextSetSelection( pc->pcSizeText, 0, strlen( s ),
		    XtLastTimestampProcessed( XtDisplay( pc->pcSizeText ) ) );

	XtFree( s );
	XmProcessTraversal( pc->pcSizeText, XmTRAVERSE_CURRENT );
	return -1;
	}

    XtFree( s );

    return 0;
    }


/************************************************************************/
/*									*/
/*  Fill the list of paper sizes.					*/
/*									*/
/************************************************************************/

void appPaperChooserFillMenu(	PaperChooser *		pc,
				const char *		customLabel,
				XtCallbackProc		callback,
				void *			through )
    {
    Dimension		width;
    const char *	label;
    int			i;
    Widget		child0= (Widget)0;
    Widget		fresh;

    XtVaGetValues( pc->pcMenu,
			XmNwidth,		&width,
			NULL );

    appEmptyPulldownList( pc->pcPulldown );

    i= 0;
    while( ! appPaperGetInformation( i, (int *)0, (int *)0, &label ) )
	{
	fresh= appPulldownMakeOption( pc->pcPulldown,
					label, width, callback, through );

	if  ( i == 0 )
	    { child0= fresh;	}

	i++;
	}

    pc->pcCustomPaperSize= i;
    fresh= appPulldownMakeOption( pc->pcPulldown,
				customLabel, width, callback, through );
    if  ( i == 0 )
	{ child0= fresh;	}

    if  ( child0 )
	{
	XtVaSetValues( pc->pcMenu,
			    XmNmenuHistory,		child0,
			    NULL );

	pc->pcSizeChosen= 0;
	}

    appPulldownSetWidth( pc->pcMenu, width );

    return;
    }

/************************************************************************/
/*									*/
/*  Adapt to geometry.							*/
/*									*/
/************************************************************************/

void appPaperChooserAdaptToGeometry(	PaperChooser *		pc,
					int			unitType,
					const DocumentGeometry * dg,
					int			set )
    {
    int		i;

    appPaperChooserShowWidthHeight( pc, dg, unitType );

    if  ( dg->dgPaperWideTwips > dg->dgPaperHighTwips	)
	{
	i= appPaperGetBySize( dg->dgPaperHighTwips, dg->dgPaperWideTwips );

	pc->pcLandscapeChosen= 1;
	}
    else{
	i= appPaperGetBySize( dg->dgPaperWideTwips, dg->dgPaperHighTwips );

	pc->pcLandscapeChosen= 0;
	}

    if  ( i < 0 )
	{ i= pc->pcCustomPaperSize;	}

    pc->pcSizeChosen= i;

    appPaperChooserShowSize( pc, pc->pcSizeChosen );

    if  ( pc->pcPortraitRadio && pc->pcLandscapeRadio )
	{ appPageToolShowOrientation( pc );	}

    if  ( set )
	{
	pc->pcLandscapeChosen= pc->pcLandscapeSet;
	pc->pcSizeChosen= pc->pcSizeSet;
	}

    return;
    }

/************************************************************************/
/*									*/
/*  A paper size was chosen in the menu. As far as the paper chooser	*/
/*  is concerned no more needs to be done.				*/
/*									*/
/************************************************************************/

void appPaperSizeChosen(		Widget			w,
					PaperChooser *		pc,
					DocumentGeometry *	dg,
					int			unitType )
    {
    short		sizeChosen= -1;

    int			width;
    int			height;

    XtVaGetValues( w,
			XmNpositionIndex,	&sizeChosen,
			NULL );

    pc->pcSizeChosen= sizeChosen;

    if  ( appPaperGetInformation( sizeChosen,
					&width, &height, (const char **)0 ) )
	{ return;	}

    if  ( pc->pcLandscapeChosen )
	{
	dg->dgPaperWideTwips= height;
	dg->dgPaperHighTwips= width;
	}
    else{
	dg->dgPaperWideTwips= width;
	dg->dgPaperHighTwips= height;
	}

    appPaperChooserShowWidthHeight( pc, dg, unitType );

    return;
    }

/************************************************************************/
/*									*/
/*  The paper size in the text field was activated by the user. As far	*/
/*  as the paper chooser is concerned no more needs to be done.		*/
/*									*/
/************************************************************************/

void appPaperSizeChanged(		PaperChooser *		pc,
					DocumentGeometry *	dg,
					int			unitType )
    {
    DocumentGeometry	dgScratch;

    dgScratch= *dg;

    if  ( appPaperChooserGetSize( pc, unitType, &dgScratch ) )
	{ LDEB(1); return;	}

    if  ( pc->pcLandscapeChosen )
	{
	dg->dgPaperWideTwips= dgScratch.dgPaperHighTwips;
	dg->dgPaperHighTwips= dgScratch.dgPaperWideTwips;
	}
    else{
	dg->dgPaperWideTwips= dgScratch.dgPaperWideTwips;
	dg->dgPaperHighTwips= dgScratch.dgPaperHighTwips;
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Show the orientation in the radiogroup.				*/
/*									*/
/************************************************************************/

void appPageToolShowOrientation(	PaperChooser *	pc )
    {
    if  ( ! pc->pcPortraitRadio || ! pc->pcLandscapeRadio )
	{ XXDEB(pc->pcPortraitRadio,pc->pcLandscapeRadio); return;	}

    if  ( pc->pcLandscapeChosen )
	{
	XmToggleButtonSetState( pc->pcPortraitRadio,  0, False );
	XmToggleButtonSetState( pc->pcLandscapeRadio, 1, False );
	}
    else{
	XmToggleButtonSetState( pc->pcPortraitRadio,  1, False );
	XmToggleButtonSetState( pc->pcLandscapeRadio, 0, False );
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Make the widgets for a paper chooser.				*/
/*									*/
/************************************************************************/

void appMakePaperChooserWidgets(	Widget			parent,
					const char *		title,
					PaperChooser *		pc,
					XtCallbackProc		changedCallback,
					void *			through )
    {
    appMakeVerticalFrame( &(pc->pcFrame), &(pc->pcBulletinBoard),
				    &(pc->pcVerticalColumn), parent, title );


    appMakePulldownList( &(pc->pcPulldown), &(pc->pcMenu),
						    pc->pcVerticalColumn );

    appMakeColumnText( &(pc->pcSizeText), pc->pcVerticalColumn, 0, True );

    XtAddCallback( pc->pcSizeText, XmNactivateCallback,
						changedCallback, through );
    
    return;
    }

/************************************************************************/
/*									*/
/*  Add Portrait/Landscape toggles to a paper chooser.			*/
/*									*/
/************************************************************************/

void appPaperChooserAddOrientationToggles( PaperChooser *	pc,
					const char *		portrait,
					const char *		landscape,
					XtCallbackProc		callback,
					void *			through )
    {
    pc->pcOrientationRow= appMakeButtonRow( pc->pcVerticalColumn, 2 );

    pc->pcPortraitRadio= appMakeRowToggle( pc->pcOrientationRow,
					    portrait, callback, through, 0 );

    pc->pcLandscapeRadio= appMakeRowToggle( pc->pcOrientationRow,
					    landscape, callback, through, 1 );

    return;
    }

/************************************************************************/
/*									*/
/*  One of the orientation toggles was activated. As far as the paper	*/
/*  chooser is concerned no more needs to be done.			*/
/*									*/
/************************************************************************/

void appPaperOrientationChosen(		Widget			w,
					XtPointer		voidtbcs,
					int *			pChanged,
					PaperChooser *		pc,
					int			unitType,
					DocumentGeometry *	dg )
    {
    XmToggleButtonCallbackStruct *	tbcs;
    int					change= 0;

    tbcs= (XmToggleButtonCallbackStruct *)voidtbcs;

    if  ( w == pc->pcPortraitRadio )
	{
	if  ( pc->pcLandscapeChosen )
	    {
	    XmToggleButtonSetState( pc->pcLandscapeRadio, 0, False );
	    pc->pcLandscapeChosen= 0;
	    change= 1;
	    }
	else{
	    XmToggleButtonSetState( pc->pcPortraitRadio, 1, False );
	    }
	}

    if  ( w == pc->pcLandscapeRadio )
	{
	if  ( pc->pcLandscapeChosen )
	    {
	    XmToggleButtonSetState( pc->pcLandscapeRadio, 1, False );
	    }
	else{
	    XmToggleButtonSetState( pc->pcPortraitRadio, 0, False );
	    pc->pcLandscapeChosen= 1;
	    change= 1;
	    }
	}

    if  ( change )
	{
	int	swap;

	swap= dg->dgPaperWideTwips;
	dg->dgPaperWideTwips= dg->dgPaperHighTwips;
	dg->dgPaperHighTwips= swap;

	appPaperChooserShowWidthHeight( pc, dg, unitType );
	}

    *pChanged= change; return;
    }

/************************************************************************/
/*									*/
/*  Manage the widgets in a paperchooser.				*/
/*									*/
/************************************************************************/

void appPaperChooserManageWidgets(	PaperChooser *		pc )
    {
    XtManageChild( pc->pcMenu );

    if  ( pc->pcOrientationRow )
	{ XtManageChild( pc->pcOrientationRow );	}

    XtManageChild( pc->pcVerticalColumn );
    XtManageChild( pc->pcBulletinBoard );
    XtManageChild( pc->pcFrame );

    return;
    }

/************************************************************************/
/*									*/
/*  Initialize a fresh PaperChooser.					*/
/*									*/
/************************************************************************/

void appInitPaperChooser(	PaperChooser *	pc )
    {
    pc->pcFrame= (Widget)0;
    pc->pcBulletinBoard= (Widget)0;
    pc->pcVerticalColumn= (Widget)0;

    pc->pcMenu= (Widget)0;
    pc->pcPulldown= (Widget)0;
    pc->pcSizeText= (Widget)0;

    pc->pcOrientationRow= (Widget)0;
    pc->pcPortraitRadio= (Widget)0;
    pc->pcLandscapeRadio= (Widget)0;

    pc->pcCustomPaperSize= -1;
    pc->pcSizeChosen= -1;
    pc->pcSizeSet= -1;
    pc->pcLandscapeChosen= 0;
    pc->pcLandscapeSet= 0;

    return;
    }
