/************************************************************************/
/*									*/
/*  Print Dialog.							*/
/*									*/
/************************************************************************/

#   include	"config.h"

#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<debugon.h>

#   include	<X11/Xatom.h>
#   include	<X11/IntrinsicP.h>

#   include	<Xm/Text.h>
#   include	<Xm/Form.h>
#   include	<Xm/DrawingA.h>
#   include	<Xm/PanedW.h>

#   include	<appFrame.h>
#   include	"appUnit.h"
#   include	"appGeoString.h"
#   include	"appPaperChooser.h"
#   include	"appPaper.h"

#   include	"appUtil.h"

#   define	DRH_CM	4.5

/************************************************************************/
/*									*/
/*  Represents a print dialog.						*/
/*									*/
/************************************************************************/

#   define	FILEL	400

#   define	PRINTselAll_PAGES	0
#   define	PRINTselPAGE_RANGE	1
#   define	PRINTselCURRENT_PAGE	2

#   define	PRINTselCOUNT		3

typedef struct PlacementSpecification
    {
    int		psNup;
    int		psHorizontal;
    char *	psOptionText;
    Widget	psOptionWidget;
    } PlacementSpecification;

#   define	PRINTps_1_UP	0
#   define	PRINTps_2_UP	1
#   define	PRINTps_4H_UP	2
#   define	PRINTps_4V_UP	3

#   define	PRINTpsCOUNT	4

static PlacementSpecification	APP_Placements[]=
    {
	{ 1, 0, "1 Up"		},
	{ 2, 1, "2 Up"		},
	{ 4, 1, "4 Up Hor"	},
	{ 4, 0, "4 Up Ver"	},
    };

typedef struct AppPrintDialog
    {
    Widget			apdTopWidget;
    Widget			apdDialog;

    Widget			apdPrintButton;
    Widget			apdCancelButton;

    Widget			apdPrinterPulldown;
    Widget			apdPrinterMenu;

    int				apdUnitType;
    PaperChooser		apdPaperChooser;

    Widget			apdPlacementPulldown;
    Widget			apdPlacementMenu;

    Widget			apdSelectionPulldown;
    Widget			apdSelectionMenu;

    Widget			apdFaxNumberText;
    Widget			apdPrintToFileOption;

    Widget			apdPageDrawing;
    GC				apdGc;
    XFontStruct *		apdFontStruct;

    Widget			apdSelectionOptions[PRINTselCOUNT];
    Widget			apdPageFromTextWidget;
    Widget			apdPageToTextWidget;

    int				apdPrinterCount;
    int				apdDestinationChosen;

    int				apdActivateFaxOptions;
    int				apdFileDestination;
    int				apdFaxDestination;

    DocumentGeometry		apdDocumentGeometry;
    DocumentGeometry		apdPrinterGeometry;
    PlacementSpecification	apdPlacements[PRINTpsCOUNT];
    int				apdPlacementChosen;

    int				apdPageCount;
    int				apdFirstSelectedPage;
    int				apdLastSelectedPage;
    int				apdSelectionChosen;
    int				apdFirstPageChosen;
    int				apdLastPageChosen;

    void *			apdTarget;
    int				apdResponse;

    char *			apdPrinterText;
    char *			apdNoneText;
    char *			apdCancelText;
    char *			apdPrintText;
    char *			apdPrintToFileText;

    char *			apdFaxText;

    char *			apdPaperSizeText;
    char *			apdCustomPaperSizeText;
    char *			apdPlacementText;

    char *			apdSelectionText;
    char *			apdAllPagesText;
    char *			apdPageRangeText;
    char *			apdCurrentPageText;
    char *			apdFromPageText;
    char *			apdToPageText;

    char *			apdDefaultPrinter;
    } AppPrintDialog;

static AppPrintDialog *	APP_PrintDialog;

# define xx(x)	x,x

static XtResource APP_PrintDialogresourceTable[]=
    {
	{ xx("printDialogPrinter"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdPrinterText),
		    XtRString, "Printer" },
	{ xx("printDialogNone"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdNoneText),
		    XtRString, "None" },
	{ xx("printDialogCancel"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdCancelText),
		    XtRString, "Cancel" },
	{ xx("printDialogPrint"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdPrintText),
		    XtRString, "Print" },
	{ xx("printDialogPrintToFile"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdPrintToFileText),
		    XtRString, "Print to File" },

	{ xx("printDialogFax"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdFaxText),
		    XtRString, "Fax" },

	{ xx("printDialogPaperSize"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdPaperSizeText),
		    XtRString, "Paper Size" },
	{ xx("printDialogCustomPaperSize"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdCustomPaperSizeText),
		    XtRString, "Custom" },

	{ xx("printDialogPlacement"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdPlacementText),
		    XtRString, "Placement" },

	{ xx("printDialog_1_UpText"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,
				apdPlacements[PRINTps_1_UP].psOptionText),
		    XtRString, "1 Page/Sheet" },
	{ xx("printDialog_2_UpText"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,
				apdPlacements[PRINTps_2_UP].psOptionText),
		    XtRString, "2 Pages/Sheet" },
	{ xx("printDialog_4H_UpText"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,
				apdPlacements[PRINTps_4H_UP].psOptionText),
		    XtRString, "4 Pages/Sheet Horizontal" },
	{ xx("printDialog_4V_UpText"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,
				apdPlacements[PRINTps_4V_UP].psOptionText),
		    XtRString, "4 Pages/Sheet Vertical" },

	{ xx("printDialogSelection"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdSelectionText),
		    XtRString, "Selection" },

	{ xx("printDialogAllPages"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdAllPagesText),
		    XtRString, "All Pages" },
	{ xx("printDialogPageRange"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdPageRangeText),
		    XtRString, "Page Range" },
	{ xx("printDialogCurrentPage"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdCurrentPageText),
		    XtRString, "Current Page" },

	{ xx("printDialogFromPage"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdFromPageText),
		    XtRString, "From" },
	{ xx("printDialogToPage"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdToPageText),
		    XtRString, "To" },

	{ xx("defaultPrinter"), XtRString, sizeof(char *),
		    offsetof(AppPrintDialog,apdDefaultPrinter),
		    XtRString, "" },
    };

static void appPrintDialogShowPrinter(	AppPrintDialog *	apd )
    {
    WidgetList		children;
    Cardinal		childCount= 0;

    XtVaGetValues( apd->apdPrinterPulldown,
		    XmNchildren,	&children,
		    XmNnumChildren,	&childCount,
		    NULL );

    if  ( apd->apdDestinationChosen >= 0		&&
	  apd->apdDestinationChosen < (int)childCount	)
	{
	XtVaSetValues( apd->apdPrinterMenu,
			XmNmenuHistory,	children[apd->apdDestinationChosen],
			NULL );
	}
    else{
	XtVaSetValues( apd->apdPrinterMenu,
			    XmNmenuHistory,	(Widget)0,
			    NULL );
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Show a range of pages in the Print Dialog.				*/
/*									*/
/************************************************************************/

static void appPrintDialogShowFromTo(	AppPrintDialog *	apd,
					int			fromPage,
					int			toPage,
					int			enabled )
    {
    appIntegerToTextWidget( apd->apdPageFromTextWidget, fromPage );
    appIntegerToTextWidget( apd->apdPageToTextWidget, toPage );

    appEnableText( apd->apdPageFromTextWidget, enabled );
    appEnableText( apd->apdPageToTextWidget, enabled );

    return;
    }

/************************************************************************/
/*									*/
/*  Fax number changed.							*/
/*									*/
/************************************************************************/

static void appPrintCheckFaxNumber(	AppPrintDialog *	apd )
    {
    char *	faxNumber;
    char *	s;

    s= faxNumber= XmTextGetString( apd->apdFaxNumberText );

    while( *s == ' ' )
	{ s++;	}

    XtSetSensitive( apd->apdPrintButton, *s != '\0' );

    XtFree( faxNumber );

    return;
    }

static void appPrintFaxChanged(	Widget		w,
				XtPointer	voidapd,
				XtPointer	call_data )
    {
    AppPrintDialog *	apd= (AppPrintDialog *)voidapd;

    appPrintCheckFaxNumber( apd );

    return;
    }

/************************************************************************/
/*									*/
/*  A printer was chosen.						*/
/*									*/
/************************************************************************/

static void appPrinterAdaptToDestination(	AppPrintDialog *	apd )
    {
    if  ( apd->apdDestinationChosen == apd->apdFaxDestination )
	{
	if  ( apd->apdFaxNumberText )
	    {
	    appEnableText( apd->apdFaxNumberText, 1 );

	    appPrintCheckFaxNumber( apd );
	    }
	}
    else{
	if  ( apd->apdFaxNumberText )
	    {
	    appEnableText( apd->apdFaxNumberText, 0 );
	    }

	XtSetSensitive( apd->apdPrintButton, True );

	XtVaSetValues( apd->apdDialog,
			XmNdefaultButton,	apd->apdPrintButton,
			NULL );
	}
    }

static void appDestinationChosen(	Widget		w,
					XtPointer	voidapd,
					XtPointer	voidpbcs	 )
    {
    AppPrintDialog *	apd= (AppPrintDialog *)voidapd;
    short		printerChosen= -1;

    XtVaGetValues( w,
			XmNpositionIndex,	&printerChosen,
			NULL );

    apd->apdDestinationChosen= printerChosen;

    appPrinterAdaptToDestination( apd );

    return;
    }

/************************************************************************/
/*									*/
/*  A placement was chosen.						*/
/*									*/
/************************************************************************/

static void appPlacementChosen(		Widget		w,
					XtPointer	voidapd,
					XtPointer	voidpbcs	 )
    {
    AppPrintDialog *	apd= (AppPrintDialog *)voidapd;
    short		placementChosen= -1;

    XtVaGetValues( w,
			XmNpositionIndex,	&placementChosen,
			NULL );

    apd->apdPlacementChosen= placementChosen;

    XClearArea( XtDisplay( apd->apdPageDrawing ),
				    XtWindow( apd->apdPageDrawing ),
				    0, 0, 0, 0, True );

    return;
    }

/************************************************************************/
/*									*/
/*  A selection mode was chosen.					*/
/*									*/
/************************************************************************/

static void appSelectionChosen(		Widget		w,
					XtPointer	voidapd,
					XtPointer	voidpbcs	 )
    {
    AppPrintDialog *	apd= (AppPrintDialog *)voidapd;
    short		selectionChosen= -1;

    XSetInputFocus( XtDisplay( apd->apdTopWidget ),
		XtWindow( apd->apdTopWidget ), RevertToNone, CurrentTime );

    XtVaGetValues( w,
			XmNpositionIndex,	&selectionChosen,
			NULL );

    switch( selectionChosen )
	{
	case PRINTselAll_PAGES:
	    apd->apdFirstPageChosen= 0;
	    apd->apdLastPageChosen= apd->apdPageCount- 1;

	    appPrintDialogShowFromTo( apd, 1, apd->apdPageCount, 0 );
	    apd->apdSelectionChosen= selectionChosen;
	    return;

	case PRINTselPAGE_RANGE:
	    appEnableText( apd->apdPageFromTextWidget, 1 );
	    appEnableText( apd->apdPageToTextWidget, 1 );
	    apd->apdSelectionChosen= selectionChosen;

	    XmProcessTraversal( apd->apdPageFromTextWidget,
							XmTRAVERSE_CURRENT );
	    return;

	case PRINTselCURRENT_PAGE:
	    apd->apdFirstPageChosen= apd->apdFirstSelectedPage;
	    apd->apdLastPageChosen= apd->apdLastSelectedPage;

	    appPrintDialogShowFromTo( apd, apd->apdFirstSelectedPage+ 1,
					    apd->apdLastSelectedPage+ 1, 0 );
	    apd->apdSelectionChosen= selectionChosen;
	    return;

	default:
	    LDEB(selectionChosen); return;
	}
    }

/************************************************************************/
/*									*/
/*  The buttons have been pushed.					*/
/*									*/
/************************************************************************/

static void appPrintDialogCancelPushed(	Widget		w,
					XtPointer	voidapd,
					XtPointer	voidpbcs	 )
    {
    AppPrintDialog *	apd= (AppPrintDialog *)voidapd;

    apd->apdResponse= AQDrespCANCEL;

    return;
    }

static void appPrintDialogPrintPushed(	Widget		w,
					XtPointer	voidapd,
					XtPointer	voidpbcs	 )
    {
    AppPrintDialog *	apd= (AppPrintDialog *)voidapd;

    if  ( apd->apdDestinationChosen < 0	)
	{ LDEB(apd->apdDestinationChosen); return;	}

    if  ( apd->apdSelectionChosen == PRINTselPAGE_RANGE )
	{
	int		from;
	int		to;

	if  ( appGetIntegerFromTextWidget( apd->apdPageFromTextWidget,
					&from, 1, 0, apd->apdPageCount, 0 ) )
	    { return;	}

	if  ( appGetIntegerFromTextWidget( apd->apdPageToTextWidget,
					&to, from, 0, apd->apdPageCount, 1 ) )
	    { return;	}

	apd->apdFirstPageChosen= from- 1;
	apd->apdLastPageChosen= to- 1;
	}

    apd->apdResponse= AQDrespOK;

    return;
    }

/************************************************************************/
/*									*/
/*  A paper size was chosen.						*/
/*									*/
/************************************************************************/

static void appPrintDialogSizeChosen(	Widget		w,
					XtPointer	voidapd,
					XtPointer	voidpbcs	 )
    {
    AppPrintDialog *		apd= (AppPrintDialog *)voidapd;

    appPaperSizeChosen( w, &(apd->apdPaperChooser),
						&(apd->apdPrinterGeometry),
						apd->apdUnitType );

    XClearArea( XtDisplay( apd->apdPageDrawing ),
				    XtWindow( apd->apdPageDrawing ),
				    0, 0, 0, 0, True );

    return;
    }

static void appPrintDialogPaperSizeChanged(	Widget		w,
						XtPointer	voidapd,
						XtPointer	voidcbs	 )
    {
    AppPrintDialog *		apd= (AppPrintDialog *)voidapd;
    const int			set= 0;

    appPaperSizeChanged( &(apd->apdPaperChooser),
						&(apd->apdPrinterGeometry),
						apd->apdUnitType );

    appPaperChooserAdaptToGeometry( &(apd->apdPaperChooser),
			apd->apdUnitType, &(apd->apdPrinterGeometry), set );

    XClearArea( XtDisplay( apd->apdPageDrawing ),
				    XtWindow( apd->apdPageDrawing ),
				    0, 0, 0, 0, True );

    return;
    }

/************************************************************************/
/*									*/
/*  Draw a schematic view of the page placement.			*/
/*									*/
/************************************************************************/

static void appPrintDialogDrawPage(	Widget		w,
					XtPointer	voidapd,
					XtPointer	voidcbs	)
    {
    AppPrintDialog *		apd= (AppPrintDialog *)voidapd;
    PlacementSpecification *	ps;

    if  ( apd->apdPlacementChosen < 0			||
	  apd->apdPlacementChosen >= PRINTpsCOUNT	)
	{ LLDEB(apd->apdPlacementChosen,PRINTpsCOUNT); return;	}

    ps= apd->apdPlacements+ apd->apdPlacementChosen;

    appDrawNupDiagram( w, apd->apdGc, apd->apdFontStruct, DRH_CM,
					    &(apd->apdPrinterGeometry),
					    &(apd->apdDocumentGeometry),
					    ps->psNup, ps->psHorizontal );
    }

/************************************************************************/
/*									*/
/*  Fill the list of printers.						*/
/*									*/
/************************************************************************/

static void appPrintDialogFillPrinterMenu( AppPrintDialog *	apd,
					int			printerCount,
					int			defaultPrinter,
					PrintDestination *	pd )
    {
    Dimension		width;
    int			i;
    Widget		defaultChild= (Widget)0;
    Widget		fresh;
    int			destinationChosen= 0;

    XtVaGetValues( apd->apdPrinterMenu,
			XmNwidth,		&width,
			NULL );

    appEmptyPulldownList( apd->apdPrinterPulldown );

    for ( i= 0; i < printerCount; pd++, i++ )
	{
	fresh= appPulldownMakeOption( apd->apdPrinterPulldown,
			pd->pdPrinterName, width, appDestinationChosen,
			(void *)apd );

	if  ( i == defaultPrinter )
	    { destinationChosen= i; defaultChild= fresh;	}
	if  ( i == 0 )
	    { destinationChosen= i; defaultChild= fresh;	}
	}

    fresh= appPulldownMakeOption( apd->apdPrinterPulldown,
		    apd->apdFaxText, width, appDestinationChosen, (void *)apd );

    if  ( ! apd->apdActivateFaxOptions )
	{ XtSetSensitive( fresh, False ); }

    apd->apdFaxDestination= i++;
    
    fresh= apd->apdPrintToFileOption=
	appPulldownMakeOption( apd->apdPrinterPulldown,
	    apd->apdPrintToFileText, width, appDestinationChosen, (void *)apd );

    if  ( ! defaultChild )
	{ destinationChosen= printerCount+ 1; defaultChild= fresh;	}

    apd->apdFileDestination= i++;
    
    if  ( defaultChild )
	{
	XtVaSetValues( apd->apdPrinterMenu,
			    XmNmenuHistory,		defaultChild,
			    NULL );
	}

    apd->apdPrinterCount= printerCount;

    if  ( defaultPrinter >= 0 )
	{ apd->apdDestinationChosen= defaultPrinter;	}
    else{ apd->apdDestinationChosen= destinationChosen; }

    appPulldownSetWidth( apd->apdPrinterMenu, width );
    }

/************************************************************************/
/*									*/
/*  Fill the list of possible page placements.				*/
/*									*/
/************************************************************************/

static void appPrintDialogFillPlacementMenu( AppPrintDialog *	apd )
    {
    Dimension			width;
    int				i;

    PlacementSpecification *	ps;

    XtVaGetValues( apd->apdPlacementMenu,
			XmNwidth,		&width,
			NULL );

    appEmptyPulldownList( apd->apdPlacementPulldown );

    ps= apd->apdPlacements;
    for ( i= 0; i < PRINTpsCOUNT; ps++, i++ )
	{
	ps->psOptionWidget= appPulldownMakeOption( apd->apdPlacementPulldown,
				ps->psOptionText, width,
				appPlacementChosen, (void *)apd );
	}

    apd->apdPlacementChosen= 0;

    XtVaSetValues( apd->apdPlacementMenu,
		XmNmenuHistory,
		    apd->apdPlacements[apd->apdPlacementChosen].psOptionWidget,
		NULL );

    appPulldownSetWidth( apd->apdPlacementMenu, width );
    }

/************************************************************************/
/*									*/
/*  Fill the list of selection options.					*/
/*									*/
/************************************************************************/

#   define	PRINTselAll_PAGES	0
#   define	PRINTselPAGE_RANGE	1
#   define	PRINTselCURRENT_PAGE	2

static void appPrintDialogFillSelectionMenu( AppPrintDialog *	apd )
    {
    Dimension			width;

    XtVaGetValues( apd->apdSelectionMenu,
			XmNwidth,		&width,
			NULL );

    appEmptyPulldownList( apd->apdSelectionPulldown );

    apd->apdSelectionOptions[PRINTselAll_PAGES]= appPulldownMakeOption(
					    apd->apdSelectionPulldown,
					    apd->apdAllPagesText, width,
					    appSelectionChosen, (void *)apd );
    apd->apdSelectionOptions[PRINTselPAGE_RANGE]= appPulldownMakeOption(
					    apd->apdSelectionPulldown,
					    apd->apdPageRangeText, width,
					    appSelectionChosen, (void *)apd );
    apd->apdSelectionOptions[PRINTselCURRENT_PAGE]= appPulldownMakeOption(
					    apd->apdSelectionPulldown,
					    apd->apdCurrentPageText, width,
					    appSelectionChosen, (void *)apd );

    apd->apdSelectionChosen= PRINTselAll_PAGES;
    XtVaSetValues( apd->apdSelectionMenu,
		XmNmenuHistory,
		    apd->apdSelectionOptions[apd->apdSelectionChosen],
		NULL );

    appPulldownSetWidth( apd->apdSelectionMenu, width );
    }

/************************************************************************/
/*									*/
/*  Protocol callback: The print dialog is closed through the window	*/
/*  manager. Interpret this as 'Cancel'					*/
/*									*/
/************************************************************************/

static void editClosePrintDialog(		Widget		w,
						XtPointer	voidapd,
						XtPointer	voidlcs	 )
    {
    AppPrintDialog *	apd= (AppPrintDialog *)voidapd;

    apd->apdResponse= AQDrespCANCEL;

    return;
    }

/************************************************************************/
/*									*/
/*  Make the frame for selecting a printer.				*/
/*									*/
/************************************************************************/

static Widget appPrintDialogMakePrinterFrame( Widget		parent,
					AppPrintDialog *	apd )
    {
    Widget	printerFrame;

    Widget	row;
    Widget	menuForm;

    Arg		al[20];
    int		ac;

    /**********************/

    appMakeFrame( &printerFrame, parent, apd->apdPrinterText );

    XtVaSetValues( printerFrame,
			XmNmarginWidth,		3,
			XmNmarginHeight,	2,
			NULL );

    row= appMakeButtonRow( printerFrame, 2 );

    /**********************/

    ac= 0;
    XtSetArg( al[ac], XmNtopAttachment,		XmATTACH_FORM ); ac++;

    XtSetArg( al[ac], XmNleftAttachment,	XmATTACH_POSITION ); ac++;
    XtSetArg( al[ac], XmNleftPosition,		0 ); ac++;

    XtSetArg( al[ac], XmNrightAttachment,	XmATTACH_POSITION ); ac++;
    XtSetArg( al[ac], XmNrightPosition,		1 ); ac++;

    menuForm= XmCreateForm( row, WIDGET_NAME, al, ac );

    appMakePulldownList( &apd->apdPrinterPulldown,
					    &apd->apdPrinterMenu, menuForm );

    XtVaSetValues( apd->apdPrinterPulldown,
			XmNresizeWidth,		True,
			NULL );

    XtVaSetValues( apd->apdPrinterMenu,
			XmNleftOffset,		0,
			XmNrightOffset,		3,
			NULL );

    /**********************/

    appMakeRowText( &apd->apdFaxNumberText, row, 15, False );

    XtAddCallback( apd->apdFaxNumberText, XmNvalueChangedCallback,
					    appPrintFaxChanged, (void *)apd );

    /**********************/

    XtManageChild( apd->apdPrinterMenu );

    XtManageChild( menuForm );
    XtManageChild( row );
    XtManageChild( printerFrame );

    return printerFrame;
    }

/************************************************************************/
/*									*/
/*  Make the frame for paper size and orientation.			*/
/*									*/
/************************************************************************/

static void appPrintDialogMakePaperFrame( Widget		parent,
					AppPrintDialog *	apd )
    {
    appMakePaperChooserWidgets( parent, apd->apdPaperSizeText,
						&(apd->apdPaperChooser),
						appPrintDialogPaperSizeChanged,
						(void *)apd );

    appPaperChooserManageWidgets( &(apd->apdPaperChooser) );

    return;
    }

/************************************************************************/
/*									*/
/*  Make the frame for giving page placement.				*/
/*									*/
/************************************************************************/

static Widget appPrintDialogMakePlacementFrame( Widget		parent,
					int *			pHigh,
					AppPrintDialog *	apd )
    {
    Widget		placementFrame;
    Widget		bboard;
    Widget		paned;

    Arg			al[20];
    int			ac;

    Display *		display= XtDisplay( parent );
    int			screen= DefaultScreen( display );
    int			pixPerCM;
    int			high;

    appMakeVerticalFrame( &placementFrame, &bboard, &paned, parent,
						    apd->apdPlacementText );

    /**********************/

    appMakePulldownList( &(apd->apdPlacementPulldown),
					    &(apd->apdPlacementMenu), paned );

    /**********************/

    pixPerCM= ( 10* DisplayHeight( display, screen ) )/
					DisplayHeightMM( display, screen );
    high= DRH_CM* pixPerCM;

    ac= 0;
    XtSetArg( al[ac],	XmNheight,		high ); ac++;
    XtSetArg( al[ac],	XmNtraversalOn,		False ); ac++;
    XtSetArg( al[ac],	XmNsensitive,		False ); ac++;
    XtSetArg( al[ac],	XmNnavigationType,	XmNONE ); ac++;

    apd->apdPageDrawing= XmCreateDrawingArea( paned, WIDGET_NAME, al, ac );

    XtAddCallback( apd->apdPageDrawing, XmNexposeCallback,
					appPrintDialogDrawPage, (void *)apd );

    /**********************/

    XtManageChild( apd->apdPlacementMenu );

    XtManageChild( apd->apdPageDrawing );

    XtManageChild( paned );
    XtManageChild( bboard );
    XtManageChild( placementFrame );

    *pHigh= high;
    return placementFrame;
    }

/************************************************************************/
/*									*/
/*  Make the frame to select what pages are to be printed.		*/
/*									*/
/************************************************************************/

static Widget appPrintDialogMakeSelectionFrame( Widget		parent,
					AppPrintDialog *	apd )
    {
    Widget		selectionFrame;
    Widget		bboard;
    Widget		paned;

    Widget		form;
    Widget		row;
    Widget		leftColumn;
    Widget		rightColumn;
    Widget		fromLabel;
    Widget		toLabel;

    Arg			al[20];
    int			ac;

    appMakeVerticalFrame( &selectionFrame, &bboard, &paned, parent,
						    apd->apdSelectionText );

    /**********************/

    appMakePulldownList( &(apd->apdSelectionPulldown),
					    &(apd->apdSelectionMenu), paned );

    /**********************/

    ac= 0;
    XtSetArg( al[ac],	XmNfractionBase,	2 ); ac++;
    XtSetArg( al[ac],	XmNallowResize,		True ); ac++;
    XtSetArg( al[ac],	XmNskipAdjust,		True ); ac++;
    form= XmCreateForm( paned, WIDGET_NAME, al, ac );

    ac= 0;
    XtSetArg( al[ac],	XmNtopAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNbottomAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNleftAttachment,	XmATTACH_POSITION ); ac++;
    XtSetArg( al[ac],	XmNrightAttachment,	XmATTACH_POSITION ); ac++;
    XtSetArg( al[ac],	XmNleftPosition,	0 ); ac++;
    XtSetArg( al[ac],	XmNrightPosition,	1 ); ac++;

    XtSetArg( al[ac],	XmNsashWidth,		1 ); ac++;
    XtSetArg( al[ac],	XmNsashHeight,		1 ); ac++;
    XtSetArg( al[ac],	XmNseparatorOn,		False ); ac++;
    XtSetArg( al[ac],	XmNmarginWidth,		5 ); ac++;
    XtSetArg( al[ac],	XmNmarginHeight,	5 ); ac++;
    XtSetArg( al[ac],	XmNspacing,		5 ); ac++;
    XtSetArg( al[ac],	XmNallowResize,		True ); ac++;
    leftColumn= XmCreatePanedWindow( form, WIDGET_NAME, al, ac );

    ac= 0;
    XtSetArg( al[ac],	XmNtopAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNbottomAttachment,	XmATTACH_FORM ); ac++;
    XtSetArg( al[ac],	XmNleftAttachment,	XmATTACH_POSITION ); ac++;
    XtSetArg( al[ac],	XmNrightAttachment,	XmATTACH_POSITION ); ac++;
    XtSetArg( al[ac],	XmNleftPosition,	1 ); ac++;
    XtSetArg( al[ac],	XmNrightPosition,	2 ); ac++;

    XtSetArg( al[ac],	XmNsashWidth,		1 ); ac++;
    XtSetArg( al[ac],	XmNsashHeight,		1 ); ac++;
    XtSetArg( al[ac],	XmNseparatorOn,		False ); ac++;
    XtSetArg( al[ac],	XmNmarginWidth,		5 ); ac++;
    XtSetArg( al[ac],	XmNmarginHeight,	5 ); ac++;
    XtSetArg( al[ac],	XmNspacing,		5 ); ac++;
    XtSetArg( al[ac],	XmNallowResize,		True ); ac++;
    rightColumn= XmCreatePanedWindow( form, WIDGET_NAME, al, ac );

    /**********************/

    appMakeLabelAndTextRow( &row, &fromLabel, &(apd->apdPageFromTextWidget),
			    leftColumn, apd->apdFromPageText, 8, False );

    appMakeLabelAndTextRow( &row, &toLabel, &(apd->apdPageToTextWidget),
			    rightColumn, apd->apdToPageText, 8, False );

    /**********************/

    XtManageChild( apd->apdSelectionMenu );

    XtManageChild( leftColumn );
    XtManageChild( rightColumn );
    XtManageChild( form );

    XtManageChild( paned );
    XtManageChild( bboard );
    XtManageChild( selectionFrame );

    return selectionFrame;
    }

/************************************************************************/
/*									*/
/*  Make the form with the two buttons.					*/
/*									*/
/************************************************************************/

static Widget appPrintDialogMakeButtonRow(	Widget			parent,
						AppPrintDialog *	apd )
    {
    Widget	row;

    row= appMakeButtonRow( parent, 2 );

    /**********************/

    apd->apdPrintButton= appMakeRowButton( row, apd->apdPrintText,
			appPrintDialogPrintPushed, (void *)apd, 0, True );

    /**********************/

    apd->apdCancelButton= appMakeRowButton( row, apd->apdCancelText,
			appPrintDialogCancelPushed, (void *)apd, 1, False );

    XtVaSetValues( apd->apdDialog,
			XmNcancelButton,	apd->apdCancelButton,
			NULL );

    /**********************/

    XtManageChild( row );

    return row;
    }

/************************************************************************/
/*									*/
/*  make a page tool.							*/
/*									*/
/************************************************************************/

static AppPrintDialog * appMakePrintDialog( EditApplication *	ea,
					EditDocument *		ed,
					Widget			printOption,
					const char *		pixmapName )
    {
    AppPrintDialog *	apd;
    
    Widget		printerFrame;
    Widget		placementFrame;
    Widget		selectionFrame;
    Widget		paned;
    Widget		buttonRow;

    int			i;
    int			placeHigh;

    const int		withSeparator= 0;
    Pixmap		iconPixmap= (Pixmap)0;
    
    char		name[128];

    if  ( appGetImagePixmap( ea, pixmapName, &iconPixmap )  )
	{ SDEB(pixmapName); return (AppPrintDialog *)0;	}

    if  ( ea->eaPrintDestinationCount == 0				&&
	  utilPrinterGetPrinters( &(ea->eaPrintDestinationCount),
				    &(ea->eaDefaultPrintDestination),
				    &(ea->eaPrintDestinations) )	)
	{ LDEB(1); return (AppPrintDialog *)0;	}

    apd= (AppPrintDialog *)malloc( sizeof(AppPrintDialog) );
    if  ( ! apd )
	{ XDEB(apd); return (AppPrintDialog *)0;	}

    for ( i= 0; i < PRINTpsCOUNT; i++ )
	{ apd->apdPlacements[i]= APP_Placements[i];	}

    XtGetApplicationResources( ea->eaTopWidget, apd,
				    APP_PrintDialogresourceTable,
				    XtNumber(APP_PrintDialogresourceTable),
				    NULL, 0 );


    if  ( apd->apdDefaultPrinter && apd->apdDefaultPrinter[0] )
	{
	PrintDestination *	pd= ea->eaPrintDestinations;

	for ( i= 0; i < ea->eaPrintDestinationCount; pd++, i++ )
	    {
	    if  ( ! strcmp( pd->pdPrinterName, apd->apdDefaultPrinter ) )
		{ ea->eaDefaultPrintDestination= i; break;	}
	    }
	}

    if  ( ea->eaFaxCommand )
	{ apd->apdActivateFaxOptions= 1;	}
    else{ apd->apdActivateFaxOptions= 0;	}

    apd->apdTarget= (void *)ea;
    apd->apdPrinterCount= 0;
    apd->apdFileDestination= -1;
    apd->apdFaxDestination= -1;
    apd->apdUnitType= ea->eaUnitInt;

    appInitPaperChooser( &(apd->apdPaperChooser) );

    appMakeVerticalDialog( &(apd->apdTopWidget), &(apd->apdDialog),
			    &paned, ea, editClosePrintDialog, (void *)apd,
			    iconPixmap, withSeparator, ea->eaPrintDialogName );

    appSetShellTitle( apd->apdTopWidget, printOption, ea->eaApplicationName );

    XtAddEventHandler( apd->apdTopWidget, StructureNotifyMask, False,
						    appFixSize, (void *)apd );

    printerFrame= appPrintDialogMakePrinterFrame( paned, apd );

    appPrintDialogMakePaperFrame( paned, apd );

    placementFrame= appPrintDialogMakePlacementFrame( paned, &placeHigh, apd );
    selectionFrame= appPrintDialogMakeSelectionFrame( paned, apd );
    buttonRow= appPrintDialogMakeButtonRow( paned, apd );

    appSetRelativeCallback( ed->edTopWidget, apd->apdDialog );
    appSetFocusCallback( apd->apdDialog );

    XtManageChild( paned );

    appPrintDialogFillPrinterMenu( apd,
				    ea->eaPrintDestinationCount,
				    ea->eaDefaultPrintDestination,
				    ea->eaPrintDestinations );

    appPrintDialogFillPlacementMenu( apd );
    appPrintDialogFillSelectionMenu( apd );


    appPaperChooserFillMenu( &(apd->apdPaperChooser),
				apd->apdCustomPaperSizeText,
				appPrintDialogSizeChosen, (void *)apd );

    XtManageChild( apd->apdDialog );

    appPaperChooserRetreshMenuWidth( &(apd->apdPaperChooser) );

    appDialogRelative( ed->edTopWidget, apd->apdDialog );

    appPulldownRefreshWidth( apd->apdPrinterMenu );
    appPulldownRefreshWidth( apd->apdPlacementMenu );
    appPulldownRefreshWidth( apd->apdSelectionMenu );

    apd->apdGc= XCreateGC( XtDisplay( apd->apdPageDrawing ),
				    XtWindow( apd->apdPageDrawing ), 0, NULL );

    sprintf( name,
	    "-*-lucidatypewriter-bold-r-*-*-%d-*-*-*-m-*-iso8859-1",
	    placeHigh/ 12 );

    apd->apdFontStruct= XLoadQueryFont(
				    XtDisplay( apd->apdPageDrawing ), name );

    sprintf( name,
	    "-*-*-bold-r-*-*-%d-*-*-*-m-*-iso8859-1", placeHigh/ 12 );

    apd->apdFontStruct= XLoadQueryFont(
				    XtDisplay( apd->apdPageDrawing ), name );

    return apd;
    }

/************************************************************************/
/*									*/
/*  Show the 'Print...' dialog.						*/
/*									*/
/*  1)	Make or just show it.						*/
/*									*/
/************************************************************************/

void appRunPrintDialog(			EditDocument *		ed,
					const DocumentGeometry * dgDocument,
					int			pageCount,
					int			firstSelected,
					int			lastSelected,
					Widget			printOption,
					const char *		pixmapName )
    {
    EditApplication *		ea= ed->edApplication;
    AppPrintDialog *		apd= APP_PrintDialog;
    PlacementSpecification *	ps;

    /*  1  */
    if  ( ! apd )
	{
	apd= APP_PrintDialog=
			appMakePrintDialog( ea, ed, printOption, pixmapName );

	if  ( ! apd )
	    { XDEB(apd); return;	}

	/*  2  */
	if  ( ea->eaDefaultPrintDestination >= 0			  &&
	      ea->eaDefaultPrintDestination < ea->eaPrintDestinationCount )
	    { apd->apdDestinationChosen= ea->eaDefaultPrintDestination; }
	else{
	    if  ( apd->apdPrinterCount > 0 )
		{ apd->apdDestinationChosen= 0;	}
	    else{ apd->apdDestinationChosen= 1; }
	    }
	}
    else{
	appSetShellTitle( apd->apdTopWidget, printOption,
						    ea->eaApplicationName );

	appSetRelativeCallback( ed->edTopWidget, apd->apdDialog );
	appDialogRelative( ed->edTopWidget, apd->apdDialog );

	appSetFocusCallback( apd->apdDialog );

	XtManageChild( apd->apdDialog );
	}

    apd->apdPrinterGeometry= ea->eaDefaultDocumentGeometry;
    apd->apdDocumentGeometry= *dgDocument;

    apd->apdPageCount= pageCount;
    apd->apdFirstSelectedPage= firstSelected;
    apd->apdLastSelectedPage= lastSelected;
    apd->apdFirstPageChosen= 0;
    apd->apdLastPageChosen= apd->apdPageCount- 1;

    apd->apdSelectionChosen= PRINTselAll_PAGES;
    XtVaSetValues( apd->apdSelectionMenu,
		XmNmenuHistory,
		    apd->apdSelectionOptions[apd->apdSelectionChosen],
		NULL );
    appPrintDialogShowFromTo( apd, 1, apd->apdPageCount, 0 );

    XtSetSensitive( apd->apdSelectionOptions[PRINTselCURRENT_PAGE],
			    ( apd->apdFirstSelectedPage >= 0	&&
			      apd->apdLastSelectedPage >= 0	)	);

    appPrintDialogShowPrinter( apd );
    appPrinterAdaptToDestination( apd );

    appPaperChooserAdaptToGeometry( &(apd->apdPaperChooser),
			apd->apdUnitType, &(apd->apdPrinterGeometry), 1 );

    apd->apdResponse= AQDrespNONE;

    while( apd->apdResponse == AQDrespNONE )
	{ XtAppProcessEvent( ea->eaContext, XtIMAll ); }

    if  ( apd->apdPlacementChosen < 0			||
	  apd->apdPlacementChosen >= PRINTpsCOUNT	)
	{
	LLDEB(apd->apdPlacementChosen,PRINTpsCOUNT);
	XtUnmanageChild( apd->apdDialog );
	return;
	}

    ps= apd->apdPlacements+ apd->apdPlacementChosen;

    switch( apd->apdResponse )
	{
	case AQDrespOK:
	    if  ( apd->apdDestinationChosen == apd->apdFileDestination )
		{
		appDocPrintToFile( printOption, apd->apdTopWidget, ed,
						ps->psNup, ps->psHorizontal,
						apd->apdFirstPageChosen,
						apd->apdLastPageChosen );
		}
	    else{
		if  ( apd->apdDestinationChosen == apd->apdFaxDestination )
		    {
		    char *	faxNumber;
		    char *	s;

		    s= faxNumber= XmTextGetString( apd->apdFaxNumberText );

		    while( *s == ' ' )
			{ s++;	}

		    if  ( ! *s )
			{ SDEB(faxNumber);	}
		    else{
			if  ( appFaxDocument( ed, s,
						ps->psNup, ps->psHorizontal,
						apd->apdFirstPageChosen,
						apd->apdLastPageChosen ) )
			    { SDEB(faxNumber);	}
			}

		    XtFree( faxNumber );
		    }
		else{
		    if  ( appPrintDocument( ed, apd->apdDestinationChosen,
			    ps->psNup, ps->psHorizontal,
			    apd->apdFirstPageChosen, apd->apdLastPageChosen ) )
			{ LDEB(apd->apdDestinationChosen); }
		    }
		}

	    XtUnmanageChild( apd->apdDialog );
	    return;

	default:
	    LDEB(apd->apdResponse);
	    /*FALLTHROUGH*/
	case AQDrespCANCEL:
	    XtUnmanageChild( apd->apdDialog );
	    return;
	}

    }
