/************************************************************************/
/*									*/
/*  Common resources used by Rulers.					*/
/*									*/
/************************************************************************/

#   include	<stdio.h>

#   define	y0	math_y0
#   define	y1	math_y1
#   include	<math.h>
#   undef	y0
#   undef	y1

#   include	<appUnit.h>
#   include	<appRuler.h>

#   include	<debugon.h>

#   ifndef	M_LN10
#	define	M_LN10	2.30258509299404568402
#   endif

void appInitRulerData(	RulerData *	rd )
    {
    rd->rdSizeAcross= 0;
    rd->rdSizeAlong= 0;
    rd->rdFontHeight= 0;

    rd->rdMinUsed= 0;
    rd->rdMaxUsed= 0;
    rd->rdExtraAfterMaxUsed= 0;

    rd->rdDocumentC0= 0;
    rd->rdDocumentC1= 0;
    rd->rdVisibleC0= 0;
    rd->rdVisibleC1= 0;
    rd->rdRulerC1= 0;

    rd->rdFontStruct= (XFontStruct *)0;
    appInitVerticalXFont( &(rd->rdVerticalFont) );
    rd->rdGc= None;
    rd->rdBackStipple= None;

    return;
    }

void appCleanRulerData(	Display *	display,
			RulerData *	rd )
    {
    if  ( rd->rdFontStruct )
	{ XFreeFont( display, rd->rdFontStruct );	}

    appCleanVerticalXFont( &(rd->rdVerticalFont), display );

    if  ( rd->rdGc )
	{ XFreeGC( display, rd->rdGc );	}

    if  ( rd->rdBackStipple )
	{ XFreePixmap( display, rd->rdBackStipple );	}

    return;
    }

void appRulerDrawBackground(	Display *	display,
				Window		win,
				RulerData *	rd,
				int		x,
				int		y,
				int		wide,
				int		high )
    {
    XSetForeground( display, rd->rdGc, rd->rdBackPixel );
    if  ( rd->rdBackStipple )
	{ XSetFillStyle( display, rd->rdGc, FillStippled ); }

    XFillRectangle( display, win, rd->rdGc, x, y, wide, high );

    if  ( rd->rdBackStipple )
	{ XSetFillStyle( display, rd->rdGc, FillSolid ); }

    return;
    }

void appOpenRulerGc(		RulerData *	rd,
				int		vertical,
				int		fontHeight,
				Widget		w )
    {
    Display *		display= XtDisplay( w );
    Window		win= XtWindow( w );
    int			screen= DefaultScreen( display );

    char		name[128];

    rd->rdFontHeight= fontHeight;

    rd->rdGc= XCreateGC( display, win, 0, NULL );
    XSetLineAttributes( display, rd->rdGc,
			    1, LineSolid, CapProjecting, JoinMiter );

    /*
    sprintf( name, "-*-lucidatypewriter-medium-r-*-*-%d-*-*-*-m-*-iso8859-1",
								fontHeight );
    */

    sprintf( name, "-*-helvetica-medium-r-*-*-%d-*-*-*-*-*-iso8859-1",
								fontHeight );

    rd->rdFontStruct= XLoadQueryFont( display, name );

    if  ( ! rd->rdFontStruct )
	{
	SXDEB(name,rd->rdFontStruct);
	sprintf( name, "-*-*-medium-r-*-*-%d-*-*-*-m-*-iso8859-1",
							    fontHeight );

	rd->rdFontStruct= XLoadQueryFont( display, name );
	}

    if  ( vertical							&&
	  rd->rdFontStruct						&&
	  appOpenVerticalXFont( &(rd->rdVerticalFont), name, -1,
			    display, win, rd->rdBackPixel )	)
	{ SDEB(name);	}

    if  ( rd->rdFontStruct )
	{ XSetFont( display, rd->rdGc, rd->rdFontStruct->fid );	}
    else{ SXDEB(name,rd->rdFontStruct);				}

    if  ( DefaultDepth( display, screen ) <= 2 )
	{
	int		i;

	for ( i= 0; i < 16; i++ )
	    {
	    memset( name+ 8* i+ 0, 0xaa, 4 );
	    memset( name+ 8* i+ 4, 0x55, 4 );
	    }

	rd->rdBackStipple= XCreateBitmapFromData( display, win,
							name, 32, 32 );
	if  ( ! rd->rdBackStipple )
	    { XDEB(rd->rdBackStipple);	}

	rd->rdBackPixel= BlackPixel( display, screen );

	XSetStipple( display, rd->rdGc, rd->rdBackStipple );
	}

    return;
    }

int appRulerTextWidth(	RulerData *	rd,
			const char *	s,
			int		len )
    {
    if  ( rd->rdFontStruct )
	{ return  XTextWidth( rd->rdFontStruct, s, len );	}
    else{ return  (6*rd->rdFontHeight*len)/10;		}
    }

void appRulerDrawVerticalString(	Display *	display,
					Window		win,
					RulerData *	rd,
					int		x,
					int		y,
					char *		s,
					int		len )
    {
    appDrawVerticalString( &(rd->rdVerticalFont),
				    display, win, rd->rdGc, x, y, s, len );

    return;
    }

/************************************************************************/
/*									*/
/*  Calculate suitable distances for Texts, tags, ticks			*/
/*									*/
/*  unitsPerTick= unitsPerTag/ticksPerTag;				*/
/*  screenUnitsPerTick= unitsPerTick* magnification.			*/
/*  pixelsPerTick= FAC* screenUnitsPerTick.				*/
/*									*/
/************************************************************************/

static double appRulerPerItem(	double		unitsPerItem )
    {
    double aa;
    double bb;

    aa= log( unitsPerItem )/ M_LN10;
    bb= (int)( aa+ 20 )- 20;

    if  ( aa- bb > log( 5.0 )/ M_LN10	)
	{ aa= bb+ 1;	}
    else{
	if  ( aa- bb > log( 2.0 )/ M_LN10	)
	    { aa= bb+ log( 5.0 )/ M_LN10;	}
	else{
	    if  ( aa- bb > 0.0			)
		{ aa= bb+ log( 2.0 )/ M_LN10;	}
	    }
	}

    return exp( M_LN10* aa );
    }

void appRulerCalculateIncrements(	RulerData *	rd,
					double		pixPerMM,
					double		magnification )
    {
    double		unitsPerMM;

    double		unitsPerText;
    double		unitsPerTag;
    double		unitsPerTick;

    int			tagsPerText;

    switch( rd->rdUnit )
	{
	case UNITtyMM:		unitsPerMM= 1.0;	break;
	case UNITtyCM:		unitsPerMM= 10.0;	break;
	case UNITtyINCH:	unitsPerMM= 24.4;	break;
	case UNITtyPOINTS:	unitsPerMM= 25.4/ 72.0;	break;
	case UNITtyPICAS:	unitsPerMM= 25.4/ 6.0;	break;
	default:
	    LDEB(rd->rdUnit);	unitsPerMM= 1.0;	break;
	}

    unitsPerText= appRulerPerItem( ( 30.0/ unitsPerMM )/ magnification );
    unitsPerTag=  appRulerPerItem( (  8.0/ unitsPerMM )/ magnification );
    unitsPerTick= appRulerPerItem( (  3.0/ unitsPerMM )/ magnification );

    tagsPerText= unitsPerText/ unitsPerTag+ 0.49;
    rd->rdTicksPerTag= unitsPerTag/ unitsPerTick+ 0.49;

    rd->rdTicksPerText= tagsPerText* rd->rdTicksPerTag;
    rd->rdUnitsPerTick= unitsPerText/ rd->rdTicksPerText;

    rd->rdTickDistancePixels=
		    unitsPerMM* pixPerMM* magnification* rd->rdUnitsPerTick;

    return;
    }

void appRulerTagText(	char *		to,
			int *		pWide,
			int *		pLen,
			RulerData *	rd,
			double		units )
    {
    char	scratch[30];
    char *	s;

    int		wide;
    int		len;

    sprintf ( scratch, "%12.3g", units );
    s= scratch; while( *s == ' ' ) { s++;	}
    len= strlen( s );

    wide= appRulerTextWidth( rd, s, len );

    strcpy( to, s ); *pWide= wide; *pLen= len;

    return;
    }

void appScrollHorizontalRuler(		RulerData *		rd,
					DocumentRectangle *	drClip,
					int			d,
					Display *		display,
					Window			win,
					GC			gc )
    {
    int				wide= rd->rdMaxUsed+ 1;
    int				high= rd->rdSizeAcross;

    int				x0;

    rd->rdVisibleC0 += d;
    rd->rdVisibleC1 += d;

    x0= rd->rdMinUsed;

    drClip->drY0= 0;
    drClip->drY1= high;

    if  ( d > 0 )
	{
	if  ( wide > d )
	    {
	    XCopyArea( display, win, win, gc,
					d+ x0, 0, wide- d- x0, high, x0, 0 );

	    drClip->drX0= rd->rdVisibleC1- d;
	    drClip->drX1= rd->rdVisibleC1;
	    }
	else{
	    drClip->drX0= rd->rdVisibleC0;
	    drClip->drX1= rd->rdVisibleC1;
	    }
	}
    else{
	if  ( high+ d > 0 )
	    {
	    XCopyArea( display, win, win, gc,
					x0, 0, wide+ d- x0, high, x0- d, 0 );

	    drClip->drX0= rd->rdVisibleC0;
	    drClip->drX1= rd->rdVisibleC0- d;
	    }
	else{
	    drClip->drX0= rd->rdVisibleC0;
	    drClip->drX1= rd->rdVisibleC1;
	    }
	}

    return;
    }

void appScrollVerticalRuler(		RulerData *		rd,
					DocumentRectangle *	drClip,
					int			d,
					Display *		display,
					Window			win,
					GC			gc )
    {
    int				high= rd->rdMaxUsed+ 1;
    int				wide= rd->rdSizeAcross;

    int				y0;

    rd->rdVisibleC0 += d;
    rd->rdVisibleC1 += d;

    y0= rd->rdMinUsed;

    drClip->drX0= 0;
    drClip->drX1= wide;

    if  ( d > 0 )
	{
	if  ( high > d )
	    {
	    XCopyArea( display, win, win, gc,
					0, d+ y0, wide, high- d- y0, 0, y0 );

	    drClip->drY0= rd->rdVisibleC1- d;
	    drClip->drY1= rd->rdVisibleC1;
	    }
	else{
	    drClip->drY0= rd->rdVisibleC0;
	    drClip->drY1= rd->rdVisibleC1;
	    }
	}
    else{
	if  ( high+ d > 0 )
	    {
	    XCopyArea( display, win, win, gc,
					0, y0, wide, high+ d- y0, 0, y0- d );

	    drClip->drY0= rd->rdVisibleC0;
	    drClip->drY1= rd->rdVisibleC0- d;
	    }
	else{
	    drClip->drY0= rd->rdVisibleC0;
	    drClip->drY1= rd->rdVisibleC1;
	    }
	}

    return;
    }
