/************************************************************************/
/*									*/
/*  Infrastructure for building and managing hierarchy trees.		*/
/*									*/
/************************************************************************/

#   include	<X11/Xlib.h>
#   include	<appFrame.h>

typedef struct TreeNode
    {
    /*  1  */
    int			tnChildCount;
    int			tnNumberInParent;
    struct TreeNode **	tnChildren;
    struct TreeNode *	tnParent;

    /*  2  */
    unsigned long	tnId;
    char *		tnLabel;
    char *		tnIconName;
    Pixmap		tnPixmap;
    int			tnLabelStrlen;
    int			tnLabelWidth;

    /*  3  */
    int			tnIsOpen;
    int			tnX0;
    int			tnY0;
    int			tnY1;
    } TreeNode;

typedef struct TreeDocument
    {
    TreeNode *		tdRootNode;
    TreeNode *		tdCurrentNode;

    Pixel		tdBlackPixel;
    Pixel		tdWhitePixel;
    Pixel		tdLinePixel;
    Pixel		tdCurrentPixel;
    XFontStruct *	tdFontStruct;

    int			tdGridSize;
    } TreeDocument;

typedef enum ClickPosition
    {
    CLICKposSTRIP,
    CLICKposOPEN,
    CLICKposICON,
    CLICKposLABEL
    } ClickPosition;

#   define	appTreeMakeCurrent( td, tn ) (td)->tdCurrentNode= (tn)

/************************************************************************/
/*									*/
/*  Routine declarations.						*/
/*									*/
/************************************************************************/

extern TreeNode * appTreeAddNode(	TreeDocument *		td );

extern void appInitTreeDocument(	TreeDocument *		td );

extern void appCleanTreeDocument(	TreeDocument *		td );

extern void appDeleteNode(		TreeDocument *		td,
					TreeNode *		tn );

extern int appTreeSetLabel(		TreeNode *		tn,
					const char *		label );

extern int appTreeSetIconName(		TreeNode *		tn,
					const char *		iconName );

extern int appLayoutTree(		TreeDocument *		td,
					EditApplication *	ea );

extern int appDrawTree(			TreeDocument *			td,
					AppDrawingData *		add,
					const DocumentRectangle *	drClip,
					int				ox,
					int				oy,
					Window				win,
					GC				gc );

extern TreeNode * appTreeFindNode(	TreeDocument *		td,
					int *			pPosition,
					int			x,
					int			y );

extern int appTreeProcessMouseClick( TreeDocument *		td,
				EditApplication *		ea,
				Display *			display,
				Window				win,
				XButtonEvent *			bevent,
				int				ox,
				int				oy );

extern int appTreeAllocateResources(	TreeDocument *		td,
					AppDrawingData *	add,
					AppColors *		ac,
					const char *		font );

void appDrawTreeRectangle(	TreeDocument *		td,
				AppDrawingData *	add,
				DocumentRectangle *	drVisible,
				DocumentRectangle *	drClip,
				int			ox,
				int			oy,
				Window			win,
				GC			gc );
