/************************************************************************/
/*									*/
/*  Manage a zoom menu.							*/
/*									*/
/************************************************************************/

#   include	<stddef.h>
#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<ctype.h>

#   define	y0	math_y0
#   define	y1	math_y1
#   include	<math.h>
#   undef	y0
#   undef	y1

#   include	<Xm/ToggleB.h>

#   include	<appFrame.h>
#   include	<appZoomMenu.h>

#   include	<debugon.h>

/************************************************************************/
/*									*/
/*  Normalize a zoom factor.						*/
/*									*/
/************************************************************************/

static double appZoomFactor(	int	logs2mag	)
    {
    if  ( logs2mag == 0 )
	{ return 1.0;	}

    if  ( logs2mag > 0 )
	{ return	pow( sqrt( 2.0 ),  (double)logs2mag );	}
    else{ return 1.0/	pow( sqrt( 2.0 ), -(double)logs2mag );	}
    }

/************************************************************************/
/*									*/
/*  Adapt the zoom menu to the current factor.				*/
/*									*/
/************************************************************************/

void appUpdateZoomMenu(	Widget		menu,
			int		logs2mag )
    {
    WidgetList			children;
    Cardinal			childCount= 0;

    int				i;

    XtVaGetValues( menu,
			XmNchildren,		&children,
			XmNnumChildren,		&childCount,
			NULL );

    for ( i= 0; i < childCount; i++ )
	{
	char		scratch[40];
	XmString	labelString;

	sprintf( scratch, "%6.0f %%",
		    100.0 * appZoomFactor( logs2mag+ i - childCount/ 2 ) );

	labelString= XmStringCreateSimple( scratch );
	XtVaSetValues( children[i],
			    XmNlabelString,	labelString,
			    NULL );

	XmStringFree( labelString );

	XmToggleButtonSetState( children[i], i == childCount/ 2, False );
	}
    }

/************************************************************************/
/*									*/
/*  The user has chosen another magnification:				*/
/*									*/
/************************************************************************/

void appZoomChangeFactor(	Widget			option,
				int *			pLogSqrt2Magnification,
				double *		pMagnification,
				DocumentRectangle *	drSelected,
				DocumentRectangle *	drCut )
    {
    short		num= -1;
    Cardinal		zoomCount= 0;
    double		magStep;

    XtVaGetValues( XtParent( option ),
			XmNnumChildren,		&zoomCount,
			NULL );

    XtVaGetValues( option,
			XmNpositionIndex,	&num,
			NULL );

    if  ( num < 0 || num >= zoomCount )
	{ LDEB(num); return;	}

    if  ( num == zoomCount/2 )
	{ return;	}

    *pLogSqrt2Magnification += num- zoomCount/2;
    magStep= appZoomFactor( *pLogSqrt2Magnification )/ *pMagnification;
    *pMagnification= appZoomFactor( *pLogSqrt2Magnification );

    appUpdateZoomMenu( XtParent( option ), *pLogSqrt2Magnification );

    if  ( drSelected )
	{
	if  ( drSelected->drX0 >= 0 )
	    { drSelected->drX0 *= magStep;	}
	if  ( drSelected->drX1 >= 0 )
	    { drSelected->drX1 *= magStep;	}
	if  ( drSelected->drY0 >= 0 )
	    { drSelected->drY0 *= magStep;	}
	if  ( drSelected->drY1 >= 0 )
	    { drSelected->drY1 *= magStep;	}
	}

    if  ( drCut )
	{
	if  ( drCut->drX0 >= 0 )
	    { drCut->drX0 *= magStep;	}
	if  ( drCut->drX1 >= 0 )
	    { drCut->drX1 *= magStep;	}
	if  ( drCut->drY0 >= 0 )
	    { drCut->drY0 *= magStep;	}
	if  ( drCut->drY1 >= 0 )
	    { drCut->drY1 *= magStep;	}
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Make a zoom menu.							*/
/*									*/
/************************************************************************/

void appMakeZoomMenu(		EditApplication *	ea,
				EditDocument *		ed,
				Widget			menubar,
				int			logSqrt2Magnification,
				const char *		zoomText,
				AppMenuItem *		zoomItems,
				int			zoomCount )
    {
    int			i;
    Widget		zoomMenu;

    Widget		button;

    for ( i= 1; i < zoomCount; i++ )
	{ zoomItems[i]= zoomItems[0];	}

    zoomMenu= appMakeMenu( &button, ea, menubar,
			    zoomText, 0, zoomItems, zoomCount, (void *)ed );

    appUpdateZoomMenu( zoomMenu, logSqrt2Magnification );

    return;
    }
